function OT(e,t){for(var n=0;n<t.length;n++){const r=t[n];if(typeof r!="string"&&!Array.isArray(r)){for(const a in r)if(a!=="default"&&!(a in e)){const i=Object.getOwnPropertyDescriptor(r,a);i&&Object.defineProperty(e,a,i.get?i:{enumerable:!0,get:()=>r[a]})}}}return Object.freeze(Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}))}(function(){const t=document.createElement("link").relList;if(t&&t.supports&&t.supports("modulepreload"))return;for(const a of document.querySelectorAll('link[rel="modulepreload"]'))r(a);new MutationObserver(a=>{for(const i of a)if(i.type==="childList")for(const l of i.addedNodes)l.tagName==="LINK"&&l.rel==="modulepreload"&&r(l)}).observe(document,{childList:!0,subtree:!0});function n(a){const i={};return a.integrity&&(i.integrity=a.integrity),a.referrerPolicy&&(i.referrerPolicy=a.referrerPolicy),a.crossOrigin==="use-credentials"?i.credentials="include":a.crossOrigin==="anonymous"?i.credentials="omit":i.credentials="same-origin",i}function r(a){if(a.ep)return;a.ep=!0;const i=n(a);fetch(a.href,i)}})();var xu=typeof globalThis<"u"?globalThis:typeof window<"u"?window:typeof global<"u"?global:typeof self<"u"?self:{};function Ra(e){return e&&e.__esModule&&Object.prototype.hasOwnProperty.call(e,"default")?e.default:e}function Er(e){if(e.__esModule)return e;var t=e.default;if(typeof t=="function"){var n=function r(){return this instanceof r?Reflect.construct(t,arguments,this.constructor):t.apply(this,arguments)};n.prototype=t.prototype}else n={};return Object.defineProperty(n,"__esModule",{value:!0}),Object.keys(e).forEach(function(r){var a=Object.getOwnPropertyDescriptor(e,r);Object.defineProperty(n,r,a.get?a:{enumerable:!0,get:function(){return e[r]}})}),n}var O1={exports:{}},Ec={};/**
 * @license React
 * react-jsx-runtime.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var _T=Symbol.for("react.transitional.element"),TT=Symbol.for("react.fragment");function _1(e,t,n){var r=null;if(n!==void 0&&(r=""+n),t.key!==void 0&&(r=""+t.key),"key"in t){n={};for(var a in t)a!=="key"&&(n[a]=t[a])}else n=t;return t=n.ref,{$$typeof:_T,type:e,key:r,ref:t!==void 0?t:null,props:n}}Ec.Fragment=TT;Ec.jsx=_1;Ec.jsxs=_1;O1.exports=Ec;var g=O1.exports,T1={exports:{}},F={};/**
 * @license React
 * react.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Jh=Symbol.for("react.transitional.element"),wT=Symbol.for("react.portal"),AT=Symbol.for("react.fragment"),RT=Symbol.for("react.strict_mode"),CT=Symbol.for("react.profiler"),$T=Symbol.for("react.consumer"),NT=Symbol.for("react.context"),MT=Symbol.for("react.forward_ref"),jT=Symbol.for("react.suspense"),DT=Symbol.for("react.memo"),w1=Symbol.for("react.lazy"),yv=Symbol.iterator;function BT(e){return e===null||typeof e!="object"?null:(e=yv&&e[yv]||e["@@iterator"],typeof e=="function"?e:null)}var A1={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},R1=Object.assign,C1={};function Ji(e,t,n){this.props=e,this.context=t,this.refs=C1,this.updater=n||A1}Ji.prototype.isReactComponent={};Ji.prototype.setState=function(e,t){if(typeof e!="object"&&typeof e!="function"&&e!=null)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,e,t,"setState")};Ji.prototype.forceUpdate=function(e){this.updater.enqueueForceUpdate(this,e,"forceUpdate")};function $1(){}$1.prototype=Ji.prototype;function Ih(e,t,n){this.props=e,this.context=t,this.refs=C1,this.updater=n||A1}var ey=Ih.prototype=new $1;ey.constructor=Ih;R1(ey,Ji.prototype);ey.isPureReactComponent=!0;var mv=Array.isArray,Ae={H:null,A:null,T:null,S:null},N1=Object.prototype.hasOwnProperty;function ty(e,t,n,r,a,i){return n=i.ref,{$$typeof:Jh,type:e,key:t,ref:n!==void 0?n:null,props:i}}function kT(e,t){return ty(e.type,t,void 0,void 0,void 0,e.props)}function ny(e){return typeof e=="object"&&e!==null&&e.$$typeof===Jh}function zT(e){var t={"=":"=0",":":"=2"};return"$"+e.replace(/[=:]/g,function(n){return t[n]})}var vv=/\/+/g;function Vf(e,t){return typeof e=="object"&&e!==null&&e.key!=null?zT(""+e.key):t.toString(36)}function gv(){}function LT(e){switch(e.status){case"fulfilled":return e.value;case"rejected":throw e.reason;default:switch(typeof e.status=="string"?e.then(gv,gv):(e.status="pending",e.then(function(t){e.status==="pending"&&(e.status="fulfilled",e.value=t)},function(t){e.status==="pending"&&(e.status="rejected",e.reason=t)})),e.status){case"fulfilled":return e.value;case"rejected":throw e.reason}}throw e}function ri(e,t,n,r,a){var i=typeof e;(i==="undefined"||i==="boolean")&&(e=null);var l=!1;if(e===null)l=!0;else switch(i){case"bigint":case"string":case"number":l=!0;break;case"object":switch(e.$$typeof){case Jh:case wT:l=!0;break;case w1:return l=e._init,ri(l(e._payload),t,n,r,a)}}if(l)return a=a(e),l=r===""?"."+Vf(e,0):r,mv(a)?(n="",l!=null&&(n=l.replace(vv,"$&/")+"/"),ri(a,t,n,"",function(u){return u})):a!=null&&(ny(a)&&(a=kT(a,n+(a.key==null||e&&e.key===a.key?"":(""+a.key).replace(vv,"$&/")+"/")+l)),t.push(a)),1;l=0;var o=r===""?".":r+":";if(mv(e))for(var s=0;s<e.length;s++)r=e[s],i=o+Vf(r,s),l+=ri(r,t,n,i,a);else if(s=BT(e),typeof s=="function")for(e=s.call(e),s=0;!(r=e.next()).done;)r=r.value,i=o+Vf(r,s++),l+=ri(r,t,n,i,a);else if(i==="object"){if(typeof e.then=="function")return ri(LT(e),t,n,r,a);throw t=String(e),Error("Objects are not valid as a React child (found: "+(t==="[object Object]"?"object with keys {"+Object.keys(e).join(", ")+"}":t)+"). If you meant to render a collection of children, use an array instead.")}return l}function fs(e,t,n){if(e==null)return e;var r=[],a=0;return ri(e,r,"","",function(i){return t.call(n,i,a++)}),r}function UT(e){if(e._status===-1){var t=e._result;t=t(),t.then(function(n){(e._status===0||e._status===-1)&&(e._status=1,e._result=n)},function(n){(e._status===0||e._status===-1)&&(e._status=2,e._result=n)}),e._status===-1&&(e._status=0,e._result=t)}if(e._status===1)return e._result.default;throw e._result}var bv=typeof reportError=="function"?reportError:function(e){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof e=="object"&&e!==null&&typeof e.message=="string"?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",e);return}};function HT(){}F.Children={map:fs,forEach:function(e,t,n){fs(e,function(){t.apply(this,arguments)},n)},count:function(e){var t=0;return fs(e,function(){t++}),t},toArray:function(e){return fs(e,function(t){return t})||[]},only:function(e){if(!ny(e))throw Error("React.Children.only expected to receive a single React element child.");return e}};F.Component=Ji;F.Fragment=AT;F.Profiler=CT;F.PureComponent=Ih;F.StrictMode=RT;F.Suspense=jT;F.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=Ae;F.act=function(){throw Error("act(...) is not supported in production builds of React.")};F.cache=function(e){return function(){return e.apply(null,arguments)}};F.cloneElement=function(e,t,n){if(e==null)throw Error("The argument must be a React element, but you passed "+e+".");var r=R1({},e.props),a=e.key,i=void 0;if(t!=null)for(l in t.ref!==void 0&&(i=void 0),t.key!==void 0&&(a=""+t.key),t)!N1.call(t,l)||l==="key"||l==="__self"||l==="__source"||l==="ref"&&t.ref===void 0||(r[l]=t[l]);var l=arguments.length-2;if(l===1)r.children=n;else if(1<l){for(var o=Array(l),s=0;s<l;s++)o[s]=arguments[s+2];r.children=o}return ty(e.type,a,void 0,void 0,i,r)};F.createContext=function(e){return e={$$typeof:NT,_currentValue:e,_currentValue2:e,_threadCount:0,Provider:null,Consumer:null},e.Provider=e,e.Consumer={$$typeof:$T,_context:e},e};F.createElement=function(e,t,n){var r,a={},i=null;if(t!=null)for(r in t.key!==void 0&&(i=""+t.key),t)N1.call(t,r)&&r!=="key"&&r!=="__self"&&r!=="__source"&&(a[r]=t[r]);var l=arguments.length-2;if(l===1)a.children=n;else if(1<l){for(var o=Array(l),s=0;s<l;s++)o[s]=arguments[s+2];a.children=o}if(e&&e.defaultProps)for(r in l=e.defaultProps,l)a[r]===void 0&&(a[r]=l[r]);return ty(e,i,void 0,void 0,null,a)};F.createRef=function(){return{current:null}};F.forwardRef=function(e){return{$$typeof:MT,render:e}};F.isValidElement=ny;F.lazy=function(e){return{$$typeof:w1,_payload:{_status:-1,_result:e},_init:UT}};F.memo=function(e,t){return{$$typeof:DT,type:e,compare:t===void 0?null:t}};F.startTransition=function(e){var t=Ae.T,n={};Ae.T=n;try{var r=e(),a=Ae.S;a!==null&&a(n,r),typeof r=="object"&&r!==null&&typeof r.then=="function"&&r.then(HT,bv)}catch(i){bv(i)}finally{Ae.T=t}};F.unstable_useCacheRefresh=function(){return Ae.H.useCacheRefresh()};F.use=function(e){return Ae.H.use(e)};F.useActionState=function(e,t,n){return Ae.H.useActionState(e,t,n)};F.useCallback=function(e,t){return Ae.H.useCallback(e,t)};F.useContext=function(e){return Ae.H.useContext(e)};F.useDebugValue=function(){};F.useDeferredValue=function(e,t){return Ae.H.useDeferredValue(e,t)};F.useEffect=function(e,t){return Ae.H.useEffect(e,t)};F.useId=function(){return Ae.H.useId()};F.useImperativeHandle=function(e,t,n){return Ae.H.useImperativeHandle(e,t,n)};F.useInsertionEffect=function(e,t){return Ae.H.useInsertionEffect(e,t)};F.useLayoutEffect=function(e,t){return Ae.H.useLayoutEffect(e,t)};F.useMemo=function(e,t){return Ae.H.useMemo(e,t)};F.useOptimistic=function(e,t){return Ae.H.useOptimistic(e,t)};F.useReducer=function(e,t,n){return Ae.H.useReducer(e,t,n)};F.useRef=function(e){return Ae.H.useRef(e)};F.useState=function(e){return Ae.H.useState(e)};F.useSyncExternalStore=function(e,t,n){return Ae.H.useSyncExternalStore(e,t,n)};F.useTransition=function(){return Ae.H.useTransition()};F.version="19.0.0";T1.exports=F;var m=T1.exports;const ne=Ra(m),Sv=OT({__proto__:null,default:ne},[m]);var M1={exports:{}};/*!
	Copyright (c) 2018 Jed Watson.
	Licensed under the MIT License (MIT), see
	http://jedwatson.github.io/classnames
*/(function(e){(function(){var t={}.hasOwnProperty;function n(){for(var i="",l=0;l<arguments.length;l++){var o=arguments[l];o&&(i=a(i,r(o)))}return i}function r(i){if(typeof i=="string"||typeof i=="number")return i;if(typeof i!="object")return"";if(Array.isArray(i))return n.apply(null,i);if(i.toString!==Object.prototype.toString&&!i.toString.toString().includes("[native code]"))return i.toString();var l="";for(var o in i)t.call(i,o)&&i[o]&&(l=a(l,o));return l}function a(i,l){return l?i?i+" "+l:i+l:i}e.exports?(n.default=n,e.exports=n):window.classNames=n})()})(M1);var qT=M1.exports;const z=Ra(qT);function k(){return k=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},k.apply(null,arguments)}function bn(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.includes(r))continue;n[r]=e[r]}return n}function Ev(e){return"default"+e.charAt(0).toUpperCase()+e.substr(1)}function GT(e){var t=YT(e,"string");return typeof t=="symbol"?t:String(t)}function YT(e,t){if(typeof e!="object"||e===null)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(typeof r!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}function j1(e,t,n){var r=m.useRef(e!==void 0),a=m.useState(t),i=a[0],l=a[1],o=e!==void 0,s=r.current;return r.current=o,!o&&s&&i!==t&&l(t),[o?e:i,m.useCallback(function(u){for(var c=arguments.length,d=new Array(c>1?c-1:0),f=1;f<c;f++)d[f-1]=arguments[f];n&&n.apply(void 0,[u].concat(d)),l(u)},[n])]}function VT(e,t){return Object.keys(t).reduce(function(n,r){var a,i=n,l=i[Ev(r)],o=i[r],s=bn(i,[Ev(r),r].map(GT)),u=t[r],c=j1(o,l,e[u]),d=c[0],f=c[1];return k({},s,(a={},a[r]=d,a[u]=f,a))},e)}function cp(e,t){return cp=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(n,r){return n.__proto__=r,n},cp(e,t)}function XT(e,t){e.prototype=Object.create(t.prototype),e.prototype.constructor=e,cp(e,t)}const PT=["xxl","xl","lg","md","sm","xs"],FT="xs",xc=m.createContext({prefixes:{},breakpoints:PT,minBreakpoint:FT});function H(e,t){const{prefixes:n}=m.useContext(xc);return e||n[t]||t}function D1(){const{breakpoints:e}=m.useContext(xc);return e}function B1(){const{minBreakpoint:e}=m.useContext(xc);return e}function k1(){const{dir:e}=m.useContext(xc);return e==="rtl"}function ko(e){return e&&e.ownerDocument||document}function QT(e){var t=ko(e);return t&&t.defaultView||window}function KT(e,t){return QT(e).getComputedStyle(e,t)}var ZT=/([A-Z])/g;function WT(e){return e.replace(ZT,"-$1").toLowerCase()}var JT=/^ms-/;function ds(e){return WT(e).replace(JT,"-ms-")}var IT=/^((translate|rotate|scale)(X|Y|Z|3d)?|matrix(3d)?|perspective|skew(X|Y)?)$/i;function ew(e){return!!(e&&IT.test(e))}function fr(e,t){var n="",r="";if(typeof t=="string")return e.style.getPropertyValue(ds(t))||KT(e).getPropertyValue(ds(t));Object.keys(t).forEach(function(a){var i=t[a];!i&&i!==0?e.style.removeProperty(ds(a)):ew(a)?r+=a+"("+i+") ":n+=ds(a)+": "+i+";"}),r&&(n+="transform: "+r+";"),e.style.cssText+=";"+n}var z1={exports:{}},tw="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED",nw=tw,rw=nw;function L1(){}function U1(){}U1.resetWarningCache=L1;var aw=function(){function e(r,a,i,l,o,s){if(s!==rw){var u=new Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw u.name="Invariant Violation",u}}e.isRequired=e;function t(){return e}var n={array:e,bigint:e,bool:e,func:e,number:e,object:e,string:e,symbol:e,any:e,arrayOf:t,element:e,elementType:e,instanceOf:t,node:e,objectOf:t,oneOf:t,oneOfType:t,shape:t,exact:t,checkPropTypes:U1,resetWarningCache:L1};return n.PropTypes=n,n};z1.exports=aw();var iw=z1.exports;const he=Ra(iw);var H1={exports:{}},vt={};/**
 * @license React
 * react-dom.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var lw=m;function q1(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function wr(){}var yt={d:{f:wr,r:function(){throw Error(q1(522))},D:wr,C:wr,L:wr,m:wr,X:wr,S:wr,M:wr},p:0,findDOMNode:null},ow=Symbol.for("react.portal");function sw(e,t,n){var r=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:ow,key:r==null?null:""+r,children:e,containerInfo:t,implementation:n}}var Ll=lw.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function Oc(e,t){if(e==="font")return"";if(typeof t=="string")return t==="use-credentials"?t:""}vt.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=yt;vt.createPortal=function(e,t){var n=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!t||t.nodeType!==1&&t.nodeType!==9&&t.nodeType!==11)throw Error(q1(299));return sw(e,t,null,n)};vt.flushSync=function(e){var t=Ll.T,n=yt.p;try{if(Ll.T=null,yt.p=2,e)return e()}finally{Ll.T=t,yt.p=n,yt.d.f()}};vt.preconnect=function(e,t){typeof e=="string"&&(t?(t=t.crossOrigin,t=typeof t=="string"?t==="use-credentials"?t:"":void 0):t=null,yt.d.C(e,t))};vt.prefetchDNS=function(e){typeof e=="string"&&yt.d.D(e)};vt.preinit=function(e,t){if(typeof e=="string"&&t&&typeof t.as=="string"){var n=t.as,r=Oc(n,t.crossOrigin),a=typeof t.integrity=="string"?t.integrity:void 0,i=typeof t.fetchPriority=="string"?t.fetchPriority:void 0;n==="style"?yt.d.S(e,typeof t.precedence=="string"?t.precedence:void 0,{crossOrigin:r,integrity:a,fetchPriority:i}):n==="script"&&yt.d.X(e,{crossOrigin:r,integrity:a,fetchPriority:i,nonce:typeof t.nonce=="string"?t.nonce:void 0})}};vt.preinitModule=function(e,t){if(typeof e=="string")if(typeof t=="object"&&t!==null){if(t.as==null||t.as==="script"){var n=Oc(t.as,t.crossOrigin);yt.d.M(e,{crossOrigin:n,integrity:typeof t.integrity=="string"?t.integrity:void 0,nonce:typeof t.nonce=="string"?t.nonce:void 0})}}else t==null&&yt.d.M(e)};vt.preload=function(e,t){if(typeof e=="string"&&typeof t=="object"&&t!==null&&typeof t.as=="string"){var n=t.as,r=Oc(n,t.crossOrigin);yt.d.L(e,n,{crossOrigin:r,integrity:typeof t.integrity=="string"?t.integrity:void 0,nonce:typeof t.nonce=="string"?t.nonce:void 0,type:typeof t.type=="string"?t.type:void 0,fetchPriority:typeof t.fetchPriority=="string"?t.fetchPriority:void 0,referrerPolicy:typeof t.referrerPolicy=="string"?t.referrerPolicy:void 0,imageSrcSet:typeof t.imageSrcSet=="string"?t.imageSrcSet:void 0,imageSizes:typeof t.imageSizes=="string"?t.imageSizes:void 0,media:typeof t.media=="string"?t.media:void 0})}};vt.preloadModule=function(e,t){if(typeof e=="string")if(t){var n=Oc(t.as,t.crossOrigin);yt.d.m(e,{as:typeof t.as=="string"&&t.as!=="script"?t.as:void 0,crossOrigin:n,integrity:typeof t.integrity=="string"?t.integrity:void 0})}else yt.d.m(e)};vt.requestFormReset=function(e){yt.d.r(e)};vt.unstable_batchedUpdates=function(e,t){return e(t)};vt.useFormState=function(e,t,n){return Ll.H.useFormState(e,t,n)};vt.useFormStatus=function(){return Ll.H.useHostTransitionStatus()};vt.version="19.0.0";function G1(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(G1)}catch{}}G1(),H1.exports=vt;var Y1=H1.exports;const fa=Ra(Y1),xv={disabled:!1},V1=ne.createContext(null);var uw=function(t){return t.scrollTop},Cl="unmounted",$r="exited",Vt="entering",tr="entered",ji="exiting",xr=function(e){XT(t,e);function t(r,a){var i;i=e.call(this,r,a)||this;var l=a,o=l&&!l.isMounting?r.enter:r.appear,s;return i.appearStatus=null,r.in?o?(s=$r,i.appearStatus=Vt):s=tr:r.unmountOnExit||r.mountOnEnter?s=Cl:s=$r,i.state={status:s},i.nextCallback=null,i}t.getDerivedStateFromProps=function(a,i){var l=a.in;return l&&i.status===Cl?{status:$r}:null};var n=t.prototype;return n.componentDidMount=function(){this.updateStatus(!0,this.appearStatus)},n.componentDidUpdate=function(a){var i=null;if(a!==this.props){var l=this.state.status;this.props.in?l!==Vt&&l!==tr&&(i=Vt):(l===Vt||l===tr)&&(i=ji)}this.updateStatus(!1,i)},n.componentWillUnmount=function(){this.cancelNextCallback()},n.getTimeouts=function(){var a=this.props.timeout,i,l,o;return i=l=o=a,a!=null&&typeof a!="number"&&(i=a.exit,l=a.enter,o=a.appear!==void 0?a.appear:l),{exit:i,enter:l,appear:o}},n.updateStatus=function(a,i){if(a===void 0&&(a=!1),i!==null)if(this.cancelNextCallback(),i===Vt){if(this.props.unmountOnExit||this.props.mountOnEnter){var l=this.props.nodeRef?this.props.nodeRef.current:fa.findDOMNode(this);l&&uw(l)}this.performEnter(a)}else this.performExit();else this.props.unmountOnExit&&this.state.status===$r&&this.setState({status:Cl})},n.performEnter=function(a){var i=this,l=this.props.enter,o=this.context?this.context.isMounting:a,s=this.props.nodeRef?[o]:[fa.findDOMNode(this),o],u=s[0],c=s[1],d=this.getTimeouts(),f=o?d.appear:d.enter;if(!a&&!l||xv.disabled){this.safeSetState({status:tr},function(){i.props.onEntered(u)});return}this.props.onEnter(u,c),this.safeSetState({status:Vt},function(){i.props.onEntering(u,c),i.onTransitionEnd(f,function(){i.safeSetState({status:tr},function(){i.props.onEntered(u,c)})})})},n.performExit=function(){var a=this,i=this.props.exit,l=this.getTimeouts(),o=this.props.nodeRef?void 0:fa.findDOMNode(this);if(!i||xv.disabled){this.safeSetState({status:$r},function(){a.props.onExited(o)});return}this.props.onExit(o),this.safeSetState({status:ji},function(){a.props.onExiting(o),a.onTransitionEnd(l.exit,function(){a.safeSetState({status:$r},function(){a.props.onExited(o)})})})},n.cancelNextCallback=function(){this.nextCallback!==null&&(this.nextCallback.cancel(),this.nextCallback=null)},n.safeSetState=function(a,i){i=this.setNextCallback(i),this.setState(a,i)},n.setNextCallback=function(a){var i=this,l=!0;return this.nextCallback=function(o){l&&(l=!1,i.nextCallback=null,a(o))},this.nextCallback.cancel=function(){l=!1},this.nextCallback},n.onTransitionEnd=function(a,i){this.setNextCallback(i);var l=this.props.nodeRef?this.props.nodeRef.current:fa.findDOMNode(this),o=a==null&&!this.props.addEndListener;if(!l||o){setTimeout(this.nextCallback,0);return}if(this.props.addEndListener){var s=this.props.nodeRef?[this.nextCallback]:[l,this.nextCallback],u=s[0],c=s[1];this.props.addEndListener(u,c)}a!=null&&setTimeout(this.nextCallback,a)},n.render=function(){var a=this.state.status;if(a===Cl)return null;var i=this.props,l=i.children;i.in,i.mountOnEnter,i.unmountOnExit,i.appear,i.enter,i.exit,i.timeout,i.addEndListener,i.onEnter,i.onEntering,i.onEntered,i.onExit,i.onExiting,i.onExited,i.nodeRef;var o=bn(i,["children","in","mountOnEnter","unmountOnExit","appear","enter","exit","timeout","addEndListener","onEnter","onEntering","onEntered","onExit","onExiting","onExited","nodeRef"]);return ne.createElement(V1.Provider,{value:null},typeof l=="function"?l(a,o):ne.cloneElement(ne.Children.only(l),o))},t}(ne.Component);xr.contextType=V1;xr.propTypes={};function Ya(){}xr.defaultProps={in:!1,mountOnEnter:!1,unmountOnExit:!1,appear:!1,enter:!0,exit:!0,onEnter:Ya,onEntering:Ya,onEntered:Ya,onExit:Ya,onExiting:Ya,onExited:Ya};xr.UNMOUNTED=Cl;xr.EXITED=$r;xr.ENTERING=Vt;xr.ENTERED=tr;xr.EXITING=ji;function X1(e){return e.code==="Escape"||e.keyCode===27}function cw(){const e=m.version.split(".");return{major:+e[0],minor:+e[1],patch:+e[2]}}function zo(e){if(!e||typeof e=="function")return null;const{major:t}=cw();return t>=19?e.props.ref:e.ref}const _c=!!(typeof window<"u"&&window.document&&window.document.createElement);var fp=!1,dp=!1;try{var Xf={get passive(){return fp=!0},get once(){return dp=fp=!0}};_c&&(window.addEventListener("test",Xf,Xf),window.removeEventListener("test",Xf,!0))}catch{}function fw(e,t,n,r){if(r&&typeof r!="boolean"&&!dp){var a=r.once,i=r.capture,l=n;!dp&&a&&(l=n.__once||function o(s){this.removeEventListener(t,o,i),n.call(this,s)},n.__once=l),e.addEventListener(t,l,fp?r:i)}e.addEventListener(t,n,r)}function dw(e,t,n,r){var a=r&&typeof r!="boolean"?r.capture:r;e.removeEventListener(t,n,a),n.__once&&e.removeEventListener(t,n.__once,a)}function ir(e,t,n,r){return fw(e,t,n,r),function(){dw(e,t,n,r)}}function pw(e,t,n,r){if(r===void 0&&(r=!0),e){var a=document.createEvent("HTMLEvents");a.initEvent(t,n,r),e.dispatchEvent(a)}}function hw(e){var t=fr(e,"transitionDuration")||"",n=t.indexOf("ms")===-1?1e3:1;return parseFloat(t)*n}function yw(e,t,n){n===void 0&&(n=5);var r=!1,a=setTimeout(function(){r||pw(e,"transitionend",!0)},t+n),i=ir(e,"transitionend",function(){r=!0},{once:!0});return function(){clearTimeout(a),i()}}function mw(e,t,n,r){n==null&&(n=hw(e)||0);var a=yw(e,n,r),i=ir(e,"transitionend",t);return function(){a(),i()}}function Ov(e,t){const n=fr(e,t)||"",r=n.indexOf("ms")===-1?1e3:1;return parseFloat(n)*r}function ry(e,t){const n=Ov(e,"transitionDuration"),r=Ov(e,"transitionDelay"),a=mw(e,i=>{i.target===e&&(a(),t(i))},n+r)}function pl(...e){return e.filter(t=>t!=null).reduce((t,n)=>{if(typeof n!="function")throw new Error("Invalid Argument Type, must only provide functions, undefined, or null.");return t===null?n:function(...a){t.apply(this,a),n.apply(this,a)}},null)}function P1(e){e.offsetHeight}const _v=e=>!e||typeof e=="function"?e:t=>{e.current=t};function vw(e,t){const n=_v(e),r=_v(t);return a=>{n&&n(a),r&&r(a)}}function ay(e,t){return m.useMemo(()=>vw(e,t),[e,t])}function Ou(e){return e&&"setState"in e?fa.findDOMNode(e):e??null}const iy=ne.forwardRef(({onEnter:e,onEntering:t,onEntered:n,onExit:r,onExiting:a,onExited:i,addEndListener:l,children:o,childRef:s,...u},c)=>{const d=m.useRef(null),f=ay(d,s),h=O=>{f(Ou(O))},b=O=>T=>{O&&d.current&&O(d.current,T)},S=m.useCallback(b(e),[e]),x=m.useCallback(b(t),[t]),y=m.useCallback(b(n),[n]),p=m.useCallback(b(r),[r]),v=m.useCallback(b(a),[a]),E=m.useCallback(b(i),[i]),_=m.useCallback(b(l),[l]);return g.jsx(xr,{ref:c,...u,onEnter:S,onEntered:y,onEntering:x,onExit:p,onExited:E,onExiting:v,addEndListener:_,nodeRef:d,children:typeof o=="function"?(O,T)=>o(O,{...T,ref:h}):ne.cloneElement(o,{ref:h})})}),gw={height:["marginTop","marginBottom"],width:["marginLeft","marginRight"]};function bw(e,t){const n=`offset${e[0].toUpperCase()}${e.slice(1)}`,r=t[n],a=gw[e];return r+parseInt(fr(t,a[0]),10)+parseInt(fr(t,a[1]),10)}const Sw={[$r]:"collapse",[ji]:"collapsing",[Vt]:"collapsing",[tr]:"collapse show"},Tc=ne.forwardRef(({onEnter:e,onEntering:t,onEntered:n,onExit:r,onExiting:a,className:i,children:l,dimension:o="height",in:s=!1,timeout:u=300,mountOnEnter:c=!1,unmountOnExit:d=!1,appear:f=!1,getDimensionValue:h=bw,...b},S)=>{const x=typeof o=="function"?o():o,y=m.useMemo(()=>pl(O=>{O.style[x]="0"},e),[x,e]),p=m.useMemo(()=>pl(O=>{const T=`scroll${x[0].toUpperCase()}${x.slice(1)}`;O.style[x]=`${O[T]}px`},t),[x,t]),v=m.useMemo(()=>pl(O=>{O.style[x]=null},n),[x,n]),E=m.useMemo(()=>pl(O=>{O.style[x]=`${h(x,O)}px`,P1(O)},r),[r,h,x]),_=m.useMemo(()=>pl(O=>{O.style[x]=null},a),[x,a]);return g.jsx(iy,{ref:S,addEndListener:ry,...b,"aria-expanded":b.role?s:null,onEnter:y,onEntering:p,onEntered:v,onExit:E,onExiting:_,childRef:zo(l),in:s,timeout:u,mountOnEnter:c,unmountOnExit:d,appear:f,children:(O,T)=>ne.cloneElement(l,{...T,className:z(i,l.props.className,Sw[O],x==="width"&&"collapse-horizontal")})})});function Ew(e){const t=m.useRef(e);return m.useEffect(()=>{t.current=e},[e]),t}function Ii(e){const t=Ew(e);return m.useCallback(function(...n){return t.current&&t.current(...n)},[t])}const ly=e=>m.forwardRef((t,n)=>g.jsx("div",{...t,ref:n,className:z(t.className,e)}));function Tv(){return m.useState(null)}function xw(e){const t=m.useRef(e);return m.useEffect(()=>{t.current=e},[e]),t}function An(e){const t=xw(e);return m.useCallback(function(...n){return t.current&&t.current(...n)},[t])}function F1(){const e=m.useRef(!0),t=m.useRef(()=>e.current);return m.useEffect(()=>(e.current=!0,()=>{e.current=!1}),[]),t.current}function Ow(e){const t=m.useRef(null);return m.useEffect(()=>{t.current=e}),t.current}const _w=typeof global<"u"&&global.navigator&&global.navigator.product==="ReactNative",Tw=typeof document<"u",wv=Tw||_w?m.useLayoutEffect:m.useEffect,ww=["as","disabled"];function Aw(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function Rw(e){return!e||e.trim()==="#"}function Q1({tagName:e,disabled:t,href:n,target:r,rel:a,role:i,onClick:l,tabIndex:o=0,type:s}){e||(n!=null||r!=null||a!=null?e="a":e="button");const u={tagName:e};if(e==="button")return[{type:s||"button",disabled:t},u];const c=f=>{if((t||e==="a"&&Rw(n))&&f.preventDefault(),t){f.stopPropagation();return}l==null||l(f)},d=f=>{f.key===" "&&(f.preventDefault(),c(f))};return e==="a"&&(n||(n="#"),t&&(n=void 0)),[{role:i??"button",disabled:void 0,tabIndex:t?void 0:o,href:n,target:e==="a"?r:void 0,"aria-disabled":t||void 0,rel:e==="a"?a:void 0,onClick:c,onKeyDown:d},u]}const Cw=m.forwardRef((e,t)=>{let{as:n,disabled:r}=e,a=Aw(e,ww);const[i,{tagName:l}]=Q1(Object.assign({tagName:n,disabled:r},a));return g.jsx(l,Object.assign({},a,i,{ref:t}))});Cw.displayName="Button";const $w={[Vt]:"show",[tr]:"show"},uo=m.forwardRef(({className:e,children:t,transitionClasses:n={},onEnter:r,...a},i)=>{const l={in:!1,timeout:300,mountOnEnter:!1,unmountOnExit:!1,appear:!1,...a},o=m.useCallback((s,u)=>{P1(s),r==null||r(s,u)},[r]);return g.jsx(iy,{ref:i,addEndListener:ry,...l,onEnter:o,childRef:zo(t),children:(s,u)=>m.cloneElement(t,{...u,className:z("fade",e,t.props.className,$w[s],n[s])})})});uo.displayName="Fade";const Nw={"aria-label":he.string,onClick:he.func,variant:he.oneOf(["white"])},wc=m.forwardRef(({className:e,variant:t,"aria-label":n="Close",...r},a)=>g.jsx("button",{ref:a,type:"button",className:z("btn-close",t&&`btn-close-${t}`,e),"aria-label":n,...r}));wc.displayName="CloseButton";wc.propTypes=Nw;const K1=m.forwardRef(({bsPrefix:e,bg:t="primary",pill:n=!1,text:r,className:a,as:i="span",...l},o)=>{const s=H(e,"badge");return g.jsx(i,{ref:o,...l,className:z(a,s,n&&"rounded-pill",r&&`text-${r}`,t&&`bg-${t}`)})});K1.displayName="Badge";const Lo=m.forwardRef(({as:e,bsPrefix:t,variant:n="primary",size:r,active:a=!1,disabled:i=!1,className:l,...o},s)=>{const u=H(t,"btn"),[c,{tagName:d}]=Q1({tagName:e,disabled:i,...o}),f=d;return g.jsx(f,{...c,...o,ref:s,disabled:i,className:z(l,u,a&&"active",n&&`${u}-${n}`,r&&`${u}-${r}`,o.href&&i&&"disabled")})});Lo.displayName="Button";const oy=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"card-body"),g.jsx(n,{ref:a,className:z(e,t),...r})));oy.displayName="CardBody";const Z1=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"card-footer"),g.jsx(n,{ref:a,className:z(e,t),...r})));Z1.displayName="CardFooter";const W1=m.createContext(null);W1.displayName="CardHeaderContext";const J1=m.forwardRef(({bsPrefix:e,className:t,as:n="div",...r},a)=>{const i=H(e,"card-header"),l=m.useMemo(()=>({cardHeaderBsPrefix:i}),[i]);return g.jsx(W1.Provider,{value:l,children:g.jsx(n,{ref:a,...r,className:z(t,i)})})});J1.displayName="CardHeader";const I1=m.forwardRef(({bsPrefix:e,className:t,variant:n,as:r="img",...a},i)=>{const l=H(e,"card-img");return g.jsx(r,{ref:i,className:z(n?`${l}-${n}`:l,t),...a})});I1.displayName="CardImg";const eS=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"card-img-overlay"),g.jsx(n,{ref:a,className:z(e,t),...r})));eS.displayName="CardImgOverlay";const tS=m.forwardRef(({className:e,bsPrefix:t,as:n="a",...r},a)=>(t=H(t,"card-link"),g.jsx(n,{ref:a,className:z(e,t),...r})));tS.displayName="CardLink";const Mw=ly("h6"),nS=m.forwardRef(({className:e,bsPrefix:t,as:n=Mw,...r},a)=>(t=H(t,"card-subtitle"),g.jsx(n,{ref:a,className:z(e,t),...r})));nS.displayName="CardSubtitle";const rS=m.forwardRef(({className:e,bsPrefix:t,as:n="p",...r},a)=>(t=H(t,"card-text"),g.jsx(n,{ref:a,className:z(e,t),...r})));rS.displayName="CardText";const jw=ly("h5"),aS=m.forwardRef(({className:e,bsPrefix:t,as:n=jw,...r},a)=>(t=H(t,"card-title"),g.jsx(n,{ref:a,className:z(e,t),...r})));aS.displayName="CardTitle";const iS=m.forwardRef(({bsPrefix:e,className:t,bg:n,text:r,border:a,body:i=!1,children:l,as:o="div",...s},u)=>{const c=H(e,"card");return g.jsx(o,{ref:u,...s,className:z(t,c,n&&`bg-${n}`,r&&`text-${r}`,a&&`border-${a}`),children:i?g.jsx(oy,{children:l}):l})});iS.displayName="Card";const Ei=Object.assign(iS,{Img:I1,Title:aS,Subtitle:nS,Body:oy,Link:tS,Text:rS,Header:J1,Footer:Z1,ImgOverlay:eS});function Dw(){const e=m.useRef(!0),t=m.useRef(()=>e.current);return m.useEffect(()=>(e.current=!0,()=>{e.current=!1}),[]),t.current}function Bw(e){const t=m.useRef(e);return t.current=e,t}function kw(e){const t=Bw(e);m.useEffect(()=>()=>t.current(),[])}const pp=2**31-1;function lS(e,t,n){const r=n-Date.now();e.current=r<=pp?setTimeout(t,r):setTimeout(()=>lS(e,t,n),pp)}function oS(){const e=Dw(),t=m.useRef();return kw(()=>clearTimeout(t.current)),m.useMemo(()=>{const n=()=>clearTimeout(t.current);function r(a,i=0){e()&&(n(),i<=pp?t.current=setTimeout(a,i):lS(t,a,Date.now()+i))}return{set:r,clear:n,handleRef:t}},[])}function zw(e,t){return m.Children.toArray(e).some(n=>m.isValidElement(n)&&n.type===t)}function Lw({as:e,bsPrefix:t,className:n,...r}){t=H(t,"col");const a=D1(),i=B1(),l=[],o=[];return a.forEach(s=>{const u=r[s];delete r[s];let c,d,f;typeof u=="object"&&u!=null?{span:c,offset:d,order:f}=u:c=u;const h=s!==i?`-${s}`:"";c&&l.push(c===!0?`${t}${h}`:`${t}${h}-${c}`),f!=null&&o.push(`order${h}-${f}`),d!=null&&o.push(`offset${h}-${d}`)}),[{...r,className:z(n,...l,...o)},{as:e,bsPrefix:t,spans:l}]}const On=m.forwardRef((e,t)=>{const[{className:n,...r},{as:a="div",bsPrefix:i,spans:l}]=Lw(e);return g.jsx(a,{...r,ref:t,className:z(n,!l.length&&i)})});On.displayName="Col";const sS=m.forwardRef(({bsPrefix:e,fluid:t=!1,as:n="div",className:r,...a},i)=>{const l=H(e,"container"),o=typeof t=="string"?`-${t}`:"-fluid";return g.jsx(n,{ref:i,...a,className:z(r,t?`${l}${o}`:l)})});sS.displayName="Container";var Uw=Function.prototype.bind.call(Function.prototype.call,[].slice);function Va(e,t){return Uw(e.querySelectorAll(t))}var Av=Object.prototype.hasOwnProperty;function Rv(e,t,n){for(n of e.keys())if(Ul(n,t))return n}function Ul(e,t){var n,r,a;if(e===t)return!0;if(e&&t&&(n=e.constructor)===t.constructor){if(n===Date)return e.getTime()===t.getTime();if(n===RegExp)return e.toString()===t.toString();if(n===Array){if((r=e.length)===t.length)for(;r--&&Ul(e[r],t[r]););return r===-1}if(n===Set){if(e.size!==t.size)return!1;for(r of e)if(a=r,a&&typeof a=="object"&&(a=Rv(t,a),!a)||!t.has(a))return!1;return!0}if(n===Map){if(e.size!==t.size)return!1;for(r of e)if(a=r[0],a&&typeof a=="object"&&(a=Rv(t,a),!a)||!Ul(r[1],t.get(a)))return!1;return!0}if(n===ArrayBuffer)e=new Uint8Array(e),t=new Uint8Array(t);else if(n===DataView){if((r=e.byteLength)===t.byteLength)for(;r--&&e.getInt8(r)===t.getInt8(r););return r===-1}if(ArrayBuffer.isView(e)){if((r=e.byteLength)===t.byteLength)for(;r--&&e[r]===t[r];);return r===-1}if(!n||typeof e=="object"){r=0;for(n in e)if(Av.call(e,n)&&++r&&!Av.call(t,n)||!(n in t)||!Ul(e[n],t[n]))return!1;return Object.keys(t).length===r}}return e!==e&&t!==t}function Hw(e){const t=F1();return[e[0],m.useCallback(n=>{if(t())return e[1](n)},[t,e[1]])]}var bt="top",tn="bottom",nn="right",St="left",sy="auto",Uo=[bt,tn,nn,St],Di="start",co="end",qw="clippingParents",uS="viewport",hl="popper",Gw="reference",Cv=Uo.reduce(function(e,t){return e.concat([t+"-"+Di,t+"-"+co])},[]),cS=[].concat(Uo,[sy]).reduce(function(e,t){return e.concat([t,t+"-"+Di,t+"-"+co])},[]),Yw="beforeRead",Vw="read",Xw="afterRead",Pw="beforeMain",Fw="main",Qw="afterMain",Kw="beforeWrite",Zw="write",Ww="afterWrite",Jw=[Yw,Vw,Xw,Pw,Fw,Qw,Kw,Zw,Ww];function $n(e){return e.split("-")[0]}function Dt(e){if(e==null)return window;if(e.toString()!=="[object Window]"){var t=e.ownerDocument;return t&&t.defaultView||window}return e}function xa(e){var t=Dt(e).Element;return e instanceof t||e instanceof Element}function Nn(e){var t=Dt(e).HTMLElement;return e instanceof t||e instanceof HTMLElement}function uy(e){if(typeof ShadowRoot>"u")return!1;var t=Dt(e).ShadowRoot;return e instanceof t||e instanceof ShadowRoot}var ha=Math.max,_u=Math.min,Bi=Math.round;function hp(){var e=navigator.userAgentData;return e!=null&&e.brands&&Array.isArray(e.brands)?e.brands.map(function(t){return t.brand+"/"+t.version}).join(" "):navigator.userAgent}function fS(){return!/^((?!chrome|android).)*safari/i.test(hp())}function ki(e,t,n){t===void 0&&(t=!1),n===void 0&&(n=!1);var r=e.getBoundingClientRect(),a=1,i=1;t&&Nn(e)&&(a=e.offsetWidth>0&&Bi(r.width)/e.offsetWidth||1,i=e.offsetHeight>0&&Bi(r.height)/e.offsetHeight||1);var l=xa(e)?Dt(e):window,o=l.visualViewport,s=!fS()&&n,u=(r.left+(s&&o?o.offsetLeft:0))/a,c=(r.top+(s&&o?o.offsetTop:0))/i,d=r.width/a,f=r.height/i;return{width:d,height:f,top:c,right:u+d,bottom:c+f,left:u,x:u,y:c}}function cy(e){var t=ki(e),n=e.offsetWidth,r=e.offsetHeight;return Math.abs(t.width-n)<=1&&(n=t.width),Math.abs(t.height-r)<=1&&(r=t.height),{x:e.offsetLeft,y:e.offsetTop,width:n,height:r}}function dS(e,t){var n=t.getRootNode&&t.getRootNode();if(e.contains(t))return!0;if(n&&uy(n)){var r=t;do{if(r&&e.isSameNode(r))return!0;r=r.parentNode||r.host}while(r)}return!1}function Wr(e){return e?(e.nodeName||"").toLowerCase():null}function hr(e){return Dt(e).getComputedStyle(e)}function Iw(e){return["table","td","th"].indexOf(Wr(e))>=0}function na(e){return((xa(e)?e.ownerDocument:e.document)||window.document).documentElement}function Ac(e){return Wr(e)==="html"?e:e.assignedSlot||e.parentNode||(uy(e)?e.host:null)||na(e)}function $v(e){return!Nn(e)||hr(e).position==="fixed"?null:e.offsetParent}function eA(e){var t=/firefox/i.test(hp()),n=/Trident/i.test(hp());if(n&&Nn(e)){var r=hr(e);if(r.position==="fixed")return null}var a=Ac(e);for(uy(a)&&(a=a.host);Nn(a)&&["html","body"].indexOf(Wr(a))<0;){var i=hr(a);if(i.transform!=="none"||i.perspective!=="none"||i.contain==="paint"||["transform","perspective"].indexOf(i.willChange)!==-1||t&&i.willChange==="filter"||t&&i.filter&&i.filter!=="none")return a;a=a.parentNode}return null}function Ho(e){for(var t=Dt(e),n=$v(e);n&&Iw(n)&&hr(n).position==="static";)n=$v(n);return n&&(Wr(n)==="html"||Wr(n)==="body"&&hr(n).position==="static")?t:n||eA(e)||t}function fy(e){return["top","bottom"].indexOf(e)>=0?"x":"y"}function Hl(e,t,n){return ha(e,_u(t,n))}function tA(e,t,n){var r=Hl(e,t,n);return r>n?n:r}function pS(){return{top:0,right:0,bottom:0,left:0}}function hS(e){return Object.assign({},pS(),e)}function yS(e,t){return t.reduce(function(n,r){return n[r]=e,n},{})}var nA=function(t,n){return t=typeof t=="function"?t(Object.assign({},n.rects,{placement:n.placement})):t,hS(typeof t!="number"?t:yS(t,Uo))};function rA(e){var t,n=e.state,r=e.name,a=e.options,i=n.elements.arrow,l=n.modifiersData.popperOffsets,o=$n(n.placement),s=fy(o),u=[St,nn].indexOf(o)>=0,c=u?"height":"width";if(!(!i||!l)){var d=nA(a.padding,n),f=cy(i),h=s==="y"?bt:St,b=s==="y"?tn:nn,S=n.rects.reference[c]+n.rects.reference[s]-l[s]-n.rects.popper[c],x=l[s]-n.rects.reference[s],y=Ho(i),p=y?s==="y"?y.clientHeight||0:y.clientWidth||0:0,v=S/2-x/2,E=d[h],_=p-f[c]-d[b],O=p/2-f[c]/2+v,T=Hl(E,O,_),w=s;n.modifiersData[r]=(t={},t[w]=T,t.centerOffset=T-O,t)}}function aA(e){var t=e.state,n=e.options,r=n.element,a=r===void 0?"[data-popper-arrow]":r;a!=null&&(typeof a=="string"&&(a=t.elements.popper.querySelector(a),!a)||dS(t.elements.popper,a)&&(t.elements.arrow=a))}const iA={name:"arrow",enabled:!0,phase:"main",fn:rA,effect:aA,requires:["popperOffsets"],requiresIfExists:["preventOverflow"]};function zi(e){return e.split("-")[1]}var lA={top:"auto",right:"auto",bottom:"auto",left:"auto"};function oA(e,t){var n=e.x,r=e.y,a=t.devicePixelRatio||1;return{x:Bi(n*a)/a||0,y:Bi(r*a)/a||0}}function Nv(e){var t,n=e.popper,r=e.popperRect,a=e.placement,i=e.variation,l=e.offsets,o=e.position,s=e.gpuAcceleration,u=e.adaptive,c=e.roundOffsets,d=e.isFixed,f=l.x,h=f===void 0?0:f,b=l.y,S=b===void 0?0:b,x=typeof c=="function"?c({x:h,y:S}):{x:h,y:S};h=x.x,S=x.y;var y=l.hasOwnProperty("x"),p=l.hasOwnProperty("y"),v=St,E=bt,_=window;if(u){var O=Ho(n),T="clientHeight",w="clientWidth";if(O===Dt(n)&&(O=na(n),hr(O).position!=="static"&&o==="absolute"&&(T="scrollHeight",w="scrollWidth")),O=O,a===bt||(a===St||a===nn)&&i===co){E=tn;var $=d&&O===_&&_.visualViewport?_.visualViewport.height:O[T];S-=$-r.height,S*=s?1:-1}if(a===St||(a===bt||a===tn)&&i===co){v=nn;var R=d&&O===_&&_.visualViewport?_.visualViewport.width:O[w];h-=R-r.width,h*=s?1:-1}}var N=Object.assign({position:o},u&&lA),G=c===!0?oA({x:h,y:S},Dt(n)):{x:h,y:S};if(h=G.x,S=G.y,s){var U;return Object.assign({},N,(U={},U[E]=p?"0":"",U[v]=y?"0":"",U.transform=(_.devicePixelRatio||1)<=1?"translate("+h+"px, "+S+"px)":"translate3d("+h+"px, "+S+"px, 0)",U))}return Object.assign({},N,(t={},t[E]=p?S+"px":"",t[v]=y?h+"px":"",t.transform="",t))}function sA(e){var t=e.state,n=e.options,r=n.gpuAcceleration,a=r===void 0?!0:r,i=n.adaptive,l=i===void 0?!0:i,o=n.roundOffsets,s=o===void 0?!0:o,u={placement:$n(t.placement),variation:zi(t.placement),popper:t.elements.popper,popperRect:t.rects.popper,gpuAcceleration:a,isFixed:t.options.strategy==="fixed"};t.modifiersData.popperOffsets!=null&&(t.styles.popper=Object.assign({},t.styles.popper,Nv(Object.assign({},u,{offsets:t.modifiersData.popperOffsets,position:t.options.strategy,adaptive:l,roundOffsets:s})))),t.modifiersData.arrow!=null&&(t.styles.arrow=Object.assign({},t.styles.arrow,Nv(Object.assign({},u,{offsets:t.modifiersData.arrow,position:"absolute",adaptive:!1,roundOffsets:s})))),t.attributes.popper=Object.assign({},t.attributes.popper,{"data-popper-placement":t.placement})}const uA={name:"computeStyles",enabled:!0,phase:"beforeWrite",fn:sA,data:{}};var ps={passive:!0};function cA(e){var t=e.state,n=e.instance,r=e.options,a=r.scroll,i=a===void 0?!0:a,l=r.resize,o=l===void 0?!0:l,s=Dt(t.elements.popper),u=[].concat(t.scrollParents.reference,t.scrollParents.popper);return i&&u.forEach(function(c){c.addEventListener("scroll",n.update,ps)}),o&&s.addEventListener("resize",n.update,ps),function(){i&&u.forEach(function(c){c.removeEventListener("scroll",n.update,ps)}),o&&s.removeEventListener("resize",n.update,ps)}}const fA={name:"eventListeners",enabled:!0,phase:"write",fn:function(){},effect:cA,data:{}};var dA={left:"right",right:"left",bottom:"top",top:"bottom"};function qs(e){return e.replace(/left|right|bottom|top/g,function(t){return dA[t]})}var pA={start:"end",end:"start"};function Mv(e){return e.replace(/start|end/g,function(t){return pA[t]})}function dy(e){var t=Dt(e),n=t.pageXOffset,r=t.pageYOffset;return{scrollLeft:n,scrollTop:r}}function py(e){return ki(na(e)).left+dy(e).scrollLeft}function hA(e,t){var n=Dt(e),r=na(e),a=n.visualViewport,i=r.clientWidth,l=r.clientHeight,o=0,s=0;if(a){i=a.width,l=a.height;var u=fS();(u||!u&&t==="fixed")&&(o=a.offsetLeft,s=a.offsetTop)}return{width:i,height:l,x:o+py(e),y:s}}function yA(e){var t,n=na(e),r=dy(e),a=(t=e.ownerDocument)==null?void 0:t.body,i=ha(n.scrollWidth,n.clientWidth,a?a.scrollWidth:0,a?a.clientWidth:0),l=ha(n.scrollHeight,n.clientHeight,a?a.scrollHeight:0,a?a.clientHeight:0),o=-r.scrollLeft+py(e),s=-r.scrollTop;return hr(a||n).direction==="rtl"&&(o+=ha(n.clientWidth,a?a.clientWidth:0)-i),{width:i,height:l,x:o,y:s}}function hy(e){var t=hr(e),n=t.overflow,r=t.overflowX,a=t.overflowY;return/auto|scroll|overlay|hidden/.test(n+a+r)}function mS(e){return["html","body","#document"].indexOf(Wr(e))>=0?e.ownerDocument.body:Nn(e)&&hy(e)?e:mS(Ac(e))}function ql(e,t){var n;t===void 0&&(t=[]);var r=mS(e),a=r===((n=e.ownerDocument)==null?void 0:n.body),i=Dt(r),l=a?[i].concat(i.visualViewport||[],hy(r)?r:[]):r,o=t.concat(l);return a?o:o.concat(ql(Ac(l)))}function yp(e){return Object.assign({},e,{left:e.x,top:e.y,right:e.x+e.width,bottom:e.y+e.height})}function mA(e,t){var n=ki(e,!1,t==="fixed");return n.top=n.top+e.clientTop,n.left=n.left+e.clientLeft,n.bottom=n.top+e.clientHeight,n.right=n.left+e.clientWidth,n.width=e.clientWidth,n.height=e.clientHeight,n.x=n.left,n.y=n.top,n}function jv(e,t,n){return t===uS?yp(hA(e,n)):xa(t)?mA(t,n):yp(yA(na(e)))}function vA(e){var t=ql(Ac(e)),n=["absolute","fixed"].indexOf(hr(e).position)>=0,r=n&&Nn(e)?Ho(e):e;return xa(r)?t.filter(function(a){return xa(a)&&dS(a,r)&&Wr(a)!=="body"}):[]}function gA(e,t,n,r){var a=t==="clippingParents"?vA(e):[].concat(t),i=[].concat(a,[n]),l=i[0],o=i.reduce(function(s,u){var c=jv(e,u,r);return s.top=ha(c.top,s.top),s.right=_u(c.right,s.right),s.bottom=_u(c.bottom,s.bottom),s.left=ha(c.left,s.left),s},jv(e,l,r));return o.width=o.right-o.left,o.height=o.bottom-o.top,o.x=o.left,o.y=o.top,o}function vS(e){var t=e.reference,n=e.element,r=e.placement,a=r?$n(r):null,i=r?zi(r):null,l=t.x+t.width/2-n.width/2,o=t.y+t.height/2-n.height/2,s;switch(a){case bt:s={x:l,y:t.y-n.height};break;case tn:s={x:l,y:t.y+t.height};break;case nn:s={x:t.x+t.width,y:o};break;case St:s={x:t.x-n.width,y:o};break;default:s={x:t.x,y:t.y}}var u=a?fy(a):null;if(u!=null){var c=u==="y"?"height":"width";switch(i){case Di:s[u]=s[u]-(t[c]/2-n[c]/2);break;case co:s[u]=s[u]+(t[c]/2-n[c]/2);break}}return s}function fo(e,t){t===void 0&&(t={});var n=t,r=n.placement,a=r===void 0?e.placement:r,i=n.strategy,l=i===void 0?e.strategy:i,o=n.boundary,s=o===void 0?qw:o,u=n.rootBoundary,c=u===void 0?uS:u,d=n.elementContext,f=d===void 0?hl:d,h=n.altBoundary,b=h===void 0?!1:h,S=n.padding,x=S===void 0?0:S,y=hS(typeof x!="number"?x:yS(x,Uo)),p=f===hl?Gw:hl,v=e.rects.popper,E=e.elements[b?p:f],_=gA(xa(E)?E:E.contextElement||na(e.elements.popper),s,c,l),O=ki(e.elements.reference),T=vS({reference:O,element:v,strategy:"absolute",placement:a}),w=yp(Object.assign({},v,T)),$=f===hl?w:O,R={top:_.top-$.top+y.top,bottom:$.bottom-_.bottom+y.bottom,left:_.left-$.left+y.left,right:$.right-_.right+y.right},N=e.modifiersData.offset;if(f===hl&&N){var G=N[a];Object.keys(R).forEach(function(U){var X=[nn,tn].indexOf(U)>=0?1:-1,q=[bt,tn].indexOf(U)>=0?"y":"x";R[U]+=G[q]*X})}return R}function bA(e,t){t===void 0&&(t={});var n=t,r=n.placement,a=n.boundary,i=n.rootBoundary,l=n.padding,o=n.flipVariations,s=n.allowedAutoPlacements,u=s===void 0?cS:s,c=zi(r),d=c?o?Cv:Cv.filter(function(b){return zi(b)===c}):Uo,f=d.filter(function(b){return u.indexOf(b)>=0});f.length===0&&(f=d);var h=f.reduce(function(b,S){return b[S]=fo(e,{placement:S,boundary:a,rootBoundary:i,padding:l})[$n(S)],b},{});return Object.keys(h).sort(function(b,S){return h[b]-h[S]})}function SA(e){if($n(e)===sy)return[];var t=qs(e);return[Mv(e),t,Mv(t)]}function EA(e){var t=e.state,n=e.options,r=e.name;if(!t.modifiersData[r]._skip){for(var a=n.mainAxis,i=a===void 0?!0:a,l=n.altAxis,o=l===void 0?!0:l,s=n.fallbackPlacements,u=n.padding,c=n.boundary,d=n.rootBoundary,f=n.altBoundary,h=n.flipVariations,b=h===void 0?!0:h,S=n.allowedAutoPlacements,x=t.options.placement,y=$n(x),p=y===x,v=s||(p||!b?[qs(x)]:SA(x)),E=[x].concat(v).reduce(function(V,ve){return V.concat($n(ve)===sy?bA(t,{placement:ve,boundary:c,rootBoundary:d,padding:u,flipVariations:b,allowedAutoPlacements:S}):ve)},[]),_=t.rects.reference,O=t.rects.popper,T=new Map,w=!0,$=E[0],R=0;R<E.length;R++){var N=E[R],G=$n(N),U=zi(N)===Di,X=[bt,tn].indexOf(G)>=0,q=X?"width":"height",C=fo(t,{placement:N,boundary:c,rootBoundary:d,altBoundary:f,padding:u}),j=X?U?nn:St:U?tn:bt;_[q]>O[q]&&(j=qs(j));var D=qs(j),K=[];if(i&&K.push(C[G]<=0),o&&K.push(C[j]<=0,C[D]<=0),K.every(function(V){return V})){$=N,w=!1;break}T.set(N,K)}if(w)for(var ie=b?3:1,Pe=function(ve){var zt=E.find(function(_t){var it=T.get(_t);if(it)return it.slice(0,ve).every(function(Lt){return Lt})});if(zt)return $=zt,"break"},ke=ie;ke>0;ke--){var xe=Pe(ke);if(xe==="break")break}t.placement!==$&&(t.modifiersData[r]._skip=!0,t.placement=$,t.reset=!0)}}const xA={name:"flip",enabled:!0,phase:"main",fn:EA,requiresIfExists:["offset"],data:{_skip:!1}};function Dv(e,t,n){return n===void 0&&(n={x:0,y:0}),{top:e.top-t.height-n.y,right:e.right-t.width+n.x,bottom:e.bottom-t.height+n.y,left:e.left-t.width-n.x}}function Bv(e){return[bt,nn,tn,St].some(function(t){return e[t]>=0})}function OA(e){var t=e.state,n=e.name,r=t.rects.reference,a=t.rects.popper,i=t.modifiersData.preventOverflow,l=fo(t,{elementContext:"reference"}),o=fo(t,{altBoundary:!0}),s=Dv(l,r),u=Dv(o,a,i),c=Bv(s),d=Bv(u);t.modifiersData[n]={referenceClippingOffsets:s,popperEscapeOffsets:u,isReferenceHidden:c,hasPopperEscaped:d},t.attributes.popper=Object.assign({},t.attributes.popper,{"data-popper-reference-hidden":c,"data-popper-escaped":d})}const _A={name:"hide",enabled:!0,phase:"main",requiresIfExists:["preventOverflow"],fn:OA};function TA(e,t,n){var r=$n(e),a=[St,bt].indexOf(r)>=0?-1:1,i=typeof n=="function"?n(Object.assign({},t,{placement:e})):n,l=i[0],o=i[1];return l=l||0,o=(o||0)*a,[St,nn].indexOf(r)>=0?{x:o,y:l}:{x:l,y:o}}function wA(e){var t=e.state,n=e.options,r=e.name,a=n.offset,i=a===void 0?[0,0]:a,l=cS.reduce(function(c,d){return c[d]=TA(d,t.rects,i),c},{}),o=l[t.placement],s=o.x,u=o.y;t.modifiersData.popperOffsets!=null&&(t.modifiersData.popperOffsets.x+=s,t.modifiersData.popperOffsets.y+=u),t.modifiersData[r]=l}const AA={name:"offset",enabled:!0,phase:"main",requires:["popperOffsets"],fn:wA};function RA(e){var t=e.state,n=e.name;t.modifiersData[n]=vS({reference:t.rects.reference,element:t.rects.popper,strategy:"absolute",placement:t.placement})}const CA={name:"popperOffsets",enabled:!0,phase:"read",fn:RA,data:{}};function $A(e){return e==="x"?"y":"x"}function NA(e){var t=e.state,n=e.options,r=e.name,a=n.mainAxis,i=a===void 0?!0:a,l=n.altAxis,o=l===void 0?!1:l,s=n.boundary,u=n.rootBoundary,c=n.altBoundary,d=n.padding,f=n.tether,h=f===void 0?!0:f,b=n.tetherOffset,S=b===void 0?0:b,x=fo(t,{boundary:s,rootBoundary:u,padding:d,altBoundary:c}),y=$n(t.placement),p=zi(t.placement),v=!p,E=fy(y),_=$A(E),O=t.modifiersData.popperOffsets,T=t.rects.reference,w=t.rects.popper,$=typeof S=="function"?S(Object.assign({},t.rects,{placement:t.placement})):S,R=typeof $=="number"?{mainAxis:$,altAxis:$}:Object.assign({mainAxis:0,altAxis:0},$),N=t.modifiersData.offset?t.modifiersData.offset[t.placement]:null,G={x:0,y:0};if(O){if(i){var U,X=E==="y"?bt:St,q=E==="y"?tn:nn,C=E==="y"?"height":"width",j=O[E],D=j+x[X],K=j-x[q],ie=h?-w[C]/2:0,Pe=p===Di?T[C]:w[C],ke=p===Di?-w[C]:-T[C],xe=t.elements.arrow,V=h&&xe?cy(xe):{width:0,height:0},ve=t.modifiersData["arrow#persistent"]?t.modifiersData["arrow#persistent"].padding:pS(),zt=ve[X],_t=ve[q],it=Hl(0,T[C],V[C]),Lt=v?T[C]/2-ie-it-zt-R.mainAxis:Pe-it-zt-R.mainAxis,Vn=v?-T[C]/2+ie+it+_t+R.mainAxis:ke+it+_t+R.mainAxis,Xn=t.elements.arrow&&Ho(t.elements.arrow),Pn=Xn?E==="y"?Xn.clientTop||0:Xn.clientLeft||0:0,lt=(U=N==null?void 0:N[E])!=null?U:0,fe=j+Lt-lt-Pn,Y=j+Vn-lt,Ie=Hl(h?_u(D,fe):D,j,h?ha(K,Y):K);O[E]=Ie,G[E]=Ie-j}if(o){var dt,Tt=E==="x"?bt:St,fl=E==="x"?tn:nn,Oe=O[_],cn=_==="y"?"height":"width",Ua=Oe+x[Tt],Ha=Oe-x[fl],la=[bt,St].indexOf(y)!==-1,dl=(dt=N==null?void 0:N[_])!=null?dt:0,qa=la?Ua:Oe-T[cn]-w[cn]-dl+R.altAxis,Tr=la?Oe+T[cn]+w[cn]-dl-R.altAxis:Ha,M=h&&la?tA(qa,Oe,Tr):Hl(h?qa:Ua,Oe,h?Tr:Ha);O[_]=M,G[_]=M-Oe}t.modifiersData[r]=G}}const MA={name:"preventOverflow",enabled:!0,phase:"main",fn:NA,requiresIfExists:["offset"]};function jA(e){return{scrollLeft:e.scrollLeft,scrollTop:e.scrollTop}}function DA(e){return e===Dt(e)||!Nn(e)?dy(e):jA(e)}function BA(e){var t=e.getBoundingClientRect(),n=Bi(t.width)/e.offsetWidth||1,r=Bi(t.height)/e.offsetHeight||1;return n!==1||r!==1}function kA(e,t,n){n===void 0&&(n=!1);var r=Nn(t),a=Nn(t)&&BA(t),i=na(t),l=ki(e,a,n),o={scrollLeft:0,scrollTop:0},s={x:0,y:0};return(r||!r&&!n)&&((Wr(t)!=="body"||hy(i))&&(o=DA(t)),Nn(t)?(s=ki(t,!0),s.x+=t.clientLeft,s.y+=t.clientTop):i&&(s.x=py(i))),{x:l.left+o.scrollLeft-s.x,y:l.top+o.scrollTop-s.y,width:l.width,height:l.height}}function zA(e){var t=new Map,n=new Set,r=[];e.forEach(function(i){t.set(i.name,i)});function a(i){n.add(i.name);var l=[].concat(i.requires||[],i.requiresIfExists||[]);l.forEach(function(o){if(!n.has(o)){var s=t.get(o);s&&a(s)}}),r.push(i)}return e.forEach(function(i){n.has(i.name)||a(i)}),r}function LA(e){var t=zA(e);return Jw.reduce(function(n,r){return n.concat(t.filter(function(a){return a.phase===r}))},[])}function UA(e){var t;return function(){return t||(t=new Promise(function(n){Promise.resolve().then(function(){t=void 0,n(e())})})),t}}function HA(e){var t=e.reduce(function(n,r){var a=n[r.name];return n[r.name]=a?Object.assign({},a,r,{options:Object.assign({},a.options,r.options),data:Object.assign({},a.data,r.data)}):r,n},{});return Object.keys(t).map(function(n){return t[n]})}var kv={placement:"bottom",modifiers:[],strategy:"absolute"};function zv(){for(var e=arguments.length,t=new Array(e),n=0;n<e;n++)t[n]=arguments[n];return!t.some(function(r){return!(r&&typeof r.getBoundingClientRect=="function")})}function qA(e){e===void 0&&(e={});var t=e,n=t.defaultModifiers,r=n===void 0?[]:n,a=t.defaultOptions,i=a===void 0?kv:a;return function(o,s,u){u===void 0&&(u=i);var c={placement:"bottom",orderedModifiers:[],options:Object.assign({},kv,i),modifiersData:{},elements:{reference:o,popper:s},attributes:{},styles:{}},d=[],f=!1,h={state:c,setOptions:function(y){var p=typeof y=="function"?y(c.options):y;S(),c.options=Object.assign({},i,c.options,p),c.scrollParents={reference:xa(o)?ql(o):o.contextElement?ql(o.contextElement):[],popper:ql(s)};var v=LA(HA([].concat(r,c.options.modifiers)));return c.orderedModifiers=v.filter(function(E){return E.enabled}),b(),h.update()},forceUpdate:function(){if(!f){var y=c.elements,p=y.reference,v=y.popper;if(zv(p,v)){c.rects={reference:kA(p,Ho(v),c.options.strategy==="fixed"),popper:cy(v)},c.reset=!1,c.placement=c.options.placement,c.orderedModifiers.forEach(function(R){return c.modifiersData[R.name]=Object.assign({},R.data)});for(var E=0;E<c.orderedModifiers.length;E++){if(c.reset===!0){c.reset=!1,E=-1;continue}var _=c.orderedModifiers[E],O=_.fn,T=_.options,w=T===void 0?{}:T,$=_.name;typeof O=="function"&&(c=O({state:c,options:w,name:$,instance:h})||c)}}}},update:UA(function(){return new Promise(function(x){h.forceUpdate(),x(c)})}),destroy:function(){S(),f=!0}};if(!zv(o,s))return h;h.setOptions(u).then(function(x){!f&&u.onFirstUpdate&&u.onFirstUpdate(x)});function b(){c.orderedModifiers.forEach(function(x){var y=x.name,p=x.options,v=p===void 0?{}:p,E=x.effect;if(typeof E=="function"){var _=E({state:c,name:y,instance:h,options:v}),O=function(){};d.push(_||O)}})}function S(){d.forEach(function(x){return x()}),d=[]}return h}}const GA=qA({defaultModifiers:[_A,CA,uA,fA,AA,xA,MA,iA]}),YA=["enabled","placement","strategy","modifiers"];function VA(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}const XA={name:"applyStyles",enabled:!1,phase:"afterWrite",fn:()=>{}},PA={name:"ariaDescribedBy",enabled:!0,phase:"afterWrite",effect:({state:e})=>()=>{const{reference:t,popper:n}=e.elements;if("removeAttribute"in t){const r=(t.getAttribute("aria-describedby")||"").split(",").filter(a=>a.trim()!==n.id);r.length?t.setAttribute("aria-describedby",r.join(",")):t.removeAttribute("aria-describedby")}},fn:({state:e})=>{var t;const{popper:n,reference:r}=e.elements,a=(t=n.getAttribute("role"))==null?void 0:t.toLowerCase();if(n.id&&a==="tooltip"&&"setAttribute"in r){const i=r.getAttribute("aria-describedby");if(i&&i.split(",").indexOf(n.id)!==-1)return;r.setAttribute("aria-describedby",i?`${i},${n.id}`:n.id)}}},FA=[];function QA(e,t,n={}){let{enabled:r=!0,placement:a="bottom",strategy:i="absolute",modifiers:l=FA}=n,o=VA(n,YA);const s=m.useRef(l),u=m.useRef(),c=m.useCallback(()=>{var x;(x=u.current)==null||x.update()},[]),d=m.useCallback(()=>{var x;(x=u.current)==null||x.forceUpdate()},[]),[f,h]=Hw(m.useState({placement:a,update:c,forceUpdate:d,attributes:{},styles:{popper:{},arrow:{}}})),b=m.useMemo(()=>({name:"updateStateModifier",enabled:!0,phase:"write",requires:["computeStyles"],fn:({state:x})=>{const y={},p={};Object.keys(x.elements).forEach(v=>{y[v]=x.styles[v],p[v]=x.attributes[v]}),h({state:x,styles:y,attributes:p,update:c,forceUpdate:d,placement:x.placement})}}),[c,d,h]),S=m.useMemo(()=>(Ul(s.current,l)||(s.current=l),s.current),[l]);return m.useEffect(()=>{!u.current||!r||u.current.setOptions({placement:a,strategy:i,modifiers:[...S,b,XA]})},[i,a,b,r,S]),m.useEffect(()=>{if(!(!r||e==null||t==null))return u.current=GA(e,t,Object.assign({},o,{placement:a,strategy:i,modifiers:[...S,PA,b]})),()=>{u.current!=null&&(u.current.destroy(),u.current=void 0,h(x=>Object.assign({},x,{attributes:{},styles:{popper:{}}})))}},[r,e,t]),f}function po(e,t){if(e.contains)return e.contains(t);if(e.compareDocumentPosition)return e===t||!!(e.compareDocumentPosition(t)&16)}var KA=function(){},ZA=KA;const WA=Ra(ZA),Lv=()=>{};function JA(e){return e.button===0}function IA(e){return!!(e.metaKey||e.altKey||e.ctrlKey||e.shiftKey)}const Gs=e=>e&&("current"in e?e.current:e),Uv={click:"mousedown",mouseup:"mousedown",pointerup:"pointerdown"};function eR(e,t=Lv,{disabled:n,clickTrigger:r="click"}={}){const a=m.useRef(!1),i=m.useRef(!1),l=m.useCallback(u=>{const c=Gs(e);WA(!!c,"ClickOutside captured a close event but does not have a ref to compare it to. useClickOutside(), should be passed a ref that resolves to a DOM node"),a.current=!c||IA(u)||!JA(u)||!!po(c,u.target)||i.current,i.current=!1},[e]),o=An(u=>{const c=Gs(e);c&&po(c,u.target)?i.current=!0:i.current=!1}),s=An(u=>{a.current||t(u)});m.useEffect(()=>{var u,c;if(n||e==null)return;const d=ko(Gs(e)),f=d.defaultView||window;let h=(u=f.event)!=null?u:(c=f.parent)==null?void 0:c.event,b=null;Uv[r]&&(b=ir(d,Uv[r],o,!0));const S=ir(d,r,l,!0),x=ir(d,r,p=>{if(p===h){h=void 0;return}s(p)});let y=[];return"ontouchstart"in d.documentElement&&(y=[].slice.call(d.body.children).map(p=>ir(p,"mousemove",Lv))),()=>{b==null||b(),S(),x(),y.forEach(p=>p())}},[e,n,r,l,o,s])}function tR(e){const t={};return Array.isArray(e)?(e==null||e.forEach(n=>{t[n.name]=n}),t):e||t}function nR(e={}){return Array.isArray(e)?e:Object.keys(e).map(t=>(e[t].name=t,e[t]))}function rR({enabled:e,enableEvents:t,placement:n,flip:r,offset:a,fixed:i,containerPadding:l,arrowElement:o,popperConfig:s={}}){var u,c,d,f,h;const b=tR(s.modifiers);return Object.assign({},s,{placement:n,enabled:e,strategy:i?"fixed":s.strategy,modifiers:nR(Object.assign({},b,{eventListeners:{enabled:t,options:(u=b.eventListeners)==null?void 0:u.options},preventOverflow:Object.assign({},b.preventOverflow,{options:l?Object.assign({padding:l},(c=b.preventOverflow)==null?void 0:c.options):(d=b.preventOverflow)==null?void 0:d.options}),offset:{options:Object.assign({offset:a},(f=b.offset)==null?void 0:f.options)},arrow:Object.assign({},b.arrow,{enabled:!!o,options:Object.assign({},(h=b.arrow)==null?void 0:h.options,{element:o})}),flip:Object.assign({enabled:!!r},b.flip)}))})}const aR=m.createContext(null),iR="data-rr-ui-";function lR(e){return`${iR}${e}`}const gS=m.createContext(_c?window:void 0);gS.Provider;function yy(){return m.useContext(gS)}const oR=typeof global<"u"&&global.navigator&&global.navigator.product==="ReactNative",sR=typeof document<"u",bS=sR||oR?m.useLayoutEffect:m.useEffect,SS=m.createContext(null);SS.displayName="InputGroupContext";const qo=m.createContext(null);qo.displayName="NavbarContext";he.string,he.bool,he.bool,he.bool,he.bool;const ES=m.forwardRef(({bsPrefix:e,className:t,fluid:n=!1,rounded:r=!1,roundedCircle:a=!1,thumbnail:i=!1,...l},o)=>(e=H(e,"img"),g.jsx("img",{ref:o,...l,className:z(t,n&&`${e}-fluid`,r&&"rounded",a&&"rounded-circle",i&&`${e}-thumbnail`)})));ES.displayName="Image";const uR={type:he.string,tooltip:he.bool,as:he.elementType},Rc=m.forwardRef(({as:e="div",className:t,type:n="valid",tooltip:r=!1,...a},i)=>g.jsx(e,{...a,ref:i,className:z(t,`${n}-${r?"tooltip":"feedback"}`)}));Rc.displayName="Feedback";Rc.propTypes=uR;const yr=m.createContext({}),Go=m.forwardRef(({id:e,bsPrefix:t,className:n,type:r="checkbox",isValid:a=!1,isInvalid:i=!1,as:l="input",...o},s)=>{const{controlId:u}=m.useContext(yr);return t=H(t,"form-check-input"),g.jsx(l,{...o,ref:s,type:r,id:e||u,className:z(n,t,a&&"is-valid",i&&"is-invalid")})});Go.displayName="FormCheckInput";const Tu=m.forwardRef(({bsPrefix:e,className:t,htmlFor:n,...r},a)=>{const{controlId:i}=m.useContext(yr);return e=H(e,"form-check-label"),g.jsx("label",{...r,ref:a,htmlFor:n||i,className:z(t,e)})});Tu.displayName="FormCheckLabel";const xS=m.forwardRef(({id:e,bsPrefix:t,bsSwitchPrefix:n,inline:r=!1,reverse:a=!1,disabled:i=!1,isValid:l=!1,isInvalid:o=!1,feedbackTooltip:s=!1,feedback:u,feedbackType:c,className:d,style:f,title:h="",type:b="checkbox",label:S,children:x,as:y="input",...p},v)=>{t=H(t,"form-check"),n=H(n,"form-switch");const{controlId:E}=m.useContext(yr),_=m.useMemo(()=>({controlId:e||E}),[E,e]),O=!x&&S!=null&&S!==!1||zw(x,Tu),T=g.jsx(Go,{...p,type:b==="switch"?"checkbox":b,ref:v,isValid:l,isInvalid:o,disabled:i,as:y});return g.jsx(yr.Provider,{value:_,children:g.jsx("div",{style:f,className:z(d,O&&t,r&&`${t}-inline`,a&&`${t}-reverse`,b==="switch"&&n),children:x||g.jsxs(g.Fragment,{children:[T,O&&g.jsx(Tu,{title:h,children:S}),u&&g.jsx(Rc,{type:c,tooltip:s,children:u})]})})})});xS.displayName="FormCheck";const wu=Object.assign(xS,{Input:Go,Label:Tu}),OS=m.forwardRef(({bsPrefix:e,type:t,size:n,htmlSize:r,id:a,className:i,isValid:l=!1,isInvalid:o=!1,plaintext:s,readOnly:u,as:c="input",...d},f)=>{const{controlId:h}=m.useContext(yr);return e=H(e,"form-control"),g.jsx(c,{...d,type:t,size:r,ref:f,readOnly:u,id:a||h,className:z(i,s?`${e}-plaintext`:e,n&&`${e}-${n}`,t==="color"&&`${e}-color`,l&&"is-valid",o&&"is-invalid")})});OS.displayName="FormControl";const cR=Object.assign(OS,{Feedback:Rc}),_S=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"form-floating"),g.jsx(n,{ref:a,className:z(e,t),...r})));_S.displayName="FormFloating";const my=m.forwardRef(({controlId:e,as:t="div",...n},r)=>{const a=m.useMemo(()=>({controlId:e}),[e]);return g.jsx(yr.Provider,{value:a,children:g.jsx(t,{...n,ref:r})})});my.displayName="FormGroup";const TS=m.forwardRef(({as:e="label",bsPrefix:t,column:n=!1,visuallyHidden:r=!1,className:a,htmlFor:i,...l},o)=>{const{controlId:s}=m.useContext(yr);t=H(t,"form-label");let u="col-form-label";typeof n=="string"&&(u=`${u} ${u}-${n}`);const c=z(a,t,r&&"visually-hidden",n&&u);return i=i||s,n?g.jsx(On,{ref:o,as:"label",className:c,htmlFor:i,...l}):g.jsx(e,{ref:o,className:c,htmlFor:i,...l})});TS.displayName="FormLabel";const wS=m.forwardRef(({bsPrefix:e,className:t,id:n,...r},a)=>{const{controlId:i}=m.useContext(yr);return e=H(e,"form-range"),g.jsx("input",{...r,type:"range",ref:a,className:z(t,e),id:n||i})});wS.displayName="FormRange";const AS=m.forwardRef(({bsPrefix:e,size:t,htmlSize:n,className:r,isValid:a=!1,isInvalid:i=!1,id:l,...o},s)=>{const{controlId:u}=m.useContext(yr);return e=H(e,"form-select"),g.jsx("select",{...o,size:n,ref:s,className:z(r,e,t&&`${e}-${t}`,a&&"is-valid",i&&"is-invalid"),id:l||u})});AS.displayName="FormSelect";const RS=m.forwardRef(({bsPrefix:e,className:t,as:n="small",muted:r,...a},i)=>(e=H(e,"form-text"),g.jsx(n,{...a,ref:i,className:z(t,e,r&&"text-muted")})));RS.displayName="FormText";const CS=m.forwardRef((e,t)=>g.jsx(wu,{...e,ref:t,type:"switch"}));CS.displayName="Switch";const fR=Object.assign(CS,{Input:wu.Input,Label:wu.Label}),$S=m.forwardRef(({bsPrefix:e,className:t,children:n,controlId:r,label:a,...i},l)=>(e=H(e,"form-floating"),g.jsxs(my,{ref:l,className:z(t,e),controlId:r,...i,children:[n,g.jsx("label",{htmlFor:r,children:a})]})));$S.displayName="FloatingLabel";const dR={_ref:he.any,validated:he.bool,as:he.elementType},vy=m.forwardRef(({className:e,validated:t,as:n="form",...r},a)=>g.jsx(n,{...r,ref:a,className:z(e,t&&"was-validated")}));vy.displayName="Form";vy.propTypes=dR;const pt=Object.assign(vy,{Group:my,Control:cR,Floating:_S,Check:wu,Switch:fR,Label:TS,Text:RS,Range:wS,Select:AS,FloatingLabel:$S}),Cc=m.forwardRef(({className:e,bsPrefix:t,as:n="span",...r},a)=>(t=H(t,"input-group-text"),g.jsx(n,{ref:a,className:z(e,t),...r})));Cc.displayName="InputGroupText";const pR=e=>g.jsx(Cc,{children:g.jsx(Go,{type:"checkbox",...e})}),hR=e=>g.jsx(Cc,{children:g.jsx(Go,{type:"radio",...e})}),NS=m.forwardRef(({bsPrefix:e,size:t,hasValidation:n,className:r,as:a="div",...i},l)=>{e=H(e,"input-group");const o=m.useMemo(()=>({}),[]);return g.jsx(SS.Provider,{value:o,children:g.jsx(a,{ref:l,...i,className:z(r,e,t&&`${e}-${t}`,n&&"has-validation")})})});NS.displayName="InputGroup";const dr=Object.assign(NS,{Text:Cc,Radio:hR,Checkbox:pR}),Hv=e=>!e||typeof e=="function"?e:t=>{e.current=t};function yR(e,t){const n=Hv(e),r=Hv(t);return a=>{n&&n(a),r&&r(a)}}function $c(e,t){return m.useMemo(()=>yR(e,t),[e,t])}function Pf(e){e===void 0&&(e=ko());try{var t=e.activeElement;return!t||!t.nodeName?null:t}catch{return e.body}}function mR(e){const t=m.useRef(e);return t.current=e,t}function vR(e){const t=mR(e);m.useEffect(()=>()=>t.current(),[])}function gR(e=document){const t=e.defaultView;return Math.abs(t.innerWidth-e.documentElement.clientWidth)}const qv=lR("modal-open");class gy{constructor({ownerDocument:t,handleContainerOverflow:n=!0,isRTL:r=!1}={}){this.handleContainerOverflow=n,this.isRTL=r,this.modals=[],this.ownerDocument=t}getScrollbarWidth(){return gR(this.ownerDocument)}getElement(){return(this.ownerDocument||document).body}setModalAttributes(t){}removeModalAttributes(t){}setContainerStyle(t){const n={overflow:"hidden"},r=this.isRTL?"paddingLeft":"paddingRight",a=this.getElement();t.style={overflow:a.style.overflow,[r]:a.style[r]},t.scrollBarWidth&&(n[r]=`${parseInt(fr(a,r)||"0",10)+t.scrollBarWidth}px`),a.setAttribute(qv,""),fr(a,n)}reset(){[...this.modals].forEach(t=>this.remove(t))}removeContainerStyle(t){const n=this.getElement();n.removeAttribute(qv),Object.assign(n.style,t.style)}add(t){let n=this.modals.indexOf(t);return n!==-1||(n=this.modals.length,this.modals.push(t),this.setModalAttributes(t),n!==0)||(this.state={scrollBarWidth:this.getScrollbarWidth(),style:{}},this.handleContainerOverflow&&this.setContainerStyle(this.state)),n}remove(t){const n=this.modals.indexOf(t);n!==-1&&(this.modals.splice(n,1),!this.modals.length&&this.handleContainerOverflow&&this.removeContainerStyle(this.state),this.removeModalAttributes(t))}isTopModal(t){return!!this.modals.length&&this.modals[this.modals.length-1]===t}}const Ff=(e,t)=>_c?e==null?(t||ko()).body:(typeof e=="function"&&(e=e()),e&&"current"in e&&(e=e.current),e&&("nodeType"in e||e.getBoundingClientRect)?e:null):null;function mp(e,t){const n=yy(),[r,a]=m.useState(()=>Ff(e,n==null?void 0:n.document));if(!r){const i=Ff(e);i&&a(i)}return m.useEffect(()=>{},[t,r]),m.useEffect(()=>{const i=Ff(e);i!==r&&a(i)},[e,r]),r}function bR({children:e,in:t,onExited:n,mountOnEnter:r,unmountOnExit:a}){const i=m.useRef(null),l=m.useRef(t),o=An(n);m.useEffect(()=>{t?l.current=!0:o(i.current)},[t,o]);const s=$c(i,e.ref),u=m.cloneElement(e,{ref:s});return t?u:a||!l.current&&r?null:u}const SR=["onEnter","onEntering","onEntered","onExit","onExiting","onExited","addEndListener","children"];function ER(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}function xR(e){let{onEnter:t,onEntering:n,onEntered:r,onExit:a,onExiting:i,onExited:l,addEndListener:o,children:s}=e,u=ER(e,SR);const c=m.useRef(null),d=$c(c,zo(s)),f=E=>_=>{E&&c.current&&E(c.current,_)},h=m.useCallback(f(t),[t]),b=m.useCallback(f(n),[n]),S=m.useCallback(f(r),[r]),x=m.useCallback(f(a),[a]),y=m.useCallback(f(i),[i]),p=m.useCallback(f(l),[l]),v=m.useCallback(f(o),[o]);return Object.assign({},u,{nodeRef:c},t&&{onEnter:h},n&&{onEntering:b},r&&{onEntered:S},a&&{onExit:x},i&&{onExiting:y},l&&{onExited:p},o&&{addEndListener:v},{children:typeof s=="function"?(E,_)=>s(E,Object.assign({},_,{ref:d})):m.cloneElement(s,{ref:d})})}const OR=["component"];function _R(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}const TR=m.forwardRef((e,t)=>{let{component:n}=e,r=_R(e,OR);const a=xR(r);return g.jsx(n,Object.assign({ref:t},a))});function wR({in:e,onTransition:t}){const n=m.useRef(null),r=m.useRef(!0),a=An(t);return wv(()=>{if(!n.current)return;let i=!1;return a({in:e,element:n.current,initial:r.current,isStale:()=>i}),()=>{i=!0}},[e,a]),wv(()=>(r.current=!1,()=>{r.current=!0}),[]),n}function AR({children:e,in:t,onExited:n,onEntered:r,transition:a}){const[i,l]=m.useState(!t);t&&i&&l(!1);const o=wR({in:!!t,onTransition:u=>{const c=()=>{u.isStale()||(u.in?r==null||r(u.element,u.initial):(l(!0),n==null||n(u.element)))};Promise.resolve(a(u)).then(c,d=>{throw u.in||l(!0),d})}}),s=$c(o,e.ref);return i&&!t?null:m.cloneElement(e,{ref:s})}function vp(e,t,n){return e?g.jsx(TR,Object.assign({},n,{component:e})):t?g.jsx(AR,Object.assign({},n,{transition:t})):g.jsx(bR,Object.assign({},n))}const RR=["show","role","className","style","children","backdrop","keyboard","onBackdropClick","onEscapeKeyDown","transition","runTransition","backdropTransition","runBackdropTransition","autoFocus","enforceFocus","restoreFocus","restoreFocusOptions","renderDialog","renderBackdrop","manager","container","onShow","onHide","onExit","onExited","onExiting","onEnter","onEntering","onEntered"];function CR(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.indexOf(r)>=0)continue;n[r]=e[r]}return n}let Qf;function $R(e){return Qf||(Qf=new gy({ownerDocument:e==null?void 0:e.document})),Qf}function NR(e){const t=yy(),n=e||$R(t),r=m.useRef({dialog:null,backdrop:null});return Object.assign(r.current,{add:()=>n.add(r.current),remove:()=>n.remove(r.current),isTopModal:()=>n.isTopModal(r.current),setDialogRef:m.useCallback(a=>{r.current.dialog=a},[]),setBackdropRef:m.useCallback(a=>{r.current.backdrop=a},[])})}const MS=m.forwardRef((e,t)=>{let{show:n=!1,role:r="dialog",className:a,style:i,children:l,backdrop:o=!0,keyboard:s=!0,onBackdropClick:u,onEscapeKeyDown:c,transition:d,runTransition:f,backdropTransition:h,runBackdropTransition:b,autoFocus:S=!0,enforceFocus:x=!0,restoreFocus:y=!0,restoreFocusOptions:p,renderDialog:v,renderBackdrop:E=Y=>g.jsx("div",Object.assign({},Y)),manager:_,container:O,onShow:T,onHide:w=()=>{},onExit:$,onExited:R,onExiting:N,onEnter:G,onEntering:U,onEntered:X}=e,q=CR(e,RR);const C=yy(),j=mp(O),D=NR(_),K=F1(),ie=Ow(n),[Pe,ke]=m.useState(!n),xe=m.useRef(null);m.useImperativeHandle(t,()=>D,[D]),_c&&!ie&&n&&(xe.current=Pf(C==null?void 0:C.document)),n&&Pe&&ke(!1);const V=An(()=>{if(D.add(),Vn.current=ir(document,"keydown",it),Lt.current=ir(document,"focus",()=>setTimeout(zt),!0),T&&T(),S){var Y,Ie;const dt=Pf((Y=(Ie=D.dialog)==null?void 0:Ie.ownerDocument)!=null?Y:C==null?void 0:C.document);D.dialog&&dt&&!po(D.dialog,dt)&&(xe.current=dt,D.dialog.focus())}}),ve=An(()=>{if(D.remove(),Vn.current==null||Vn.current(),Lt.current==null||Lt.current(),y){var Y;(Y=xe.current)==null||Y.focus==null||Y.focus(p),xe.current=null}});m.useEffect(()=>{!n||!j||V()},[n,j,V]),m.useEffect(()=>{Pe&&ve()},[Pe,ve]),vR(()=>{ve()});const zt=An(()=>{if(!x||!K()||!D.isTopModal())return;const Y=Pf(C==null?void 0:C.document);D.dialog&&Y&&!po(D.dialog,Y)&&D.dialog.focus()}),_t=An(Y=>{Y.target===Y.currentTarget&&(u==null||u(Y),o===!0&&w())}),it=An(Y=>{s&&X1(Y)&&D.isTopModal()&&(c==null||c(Y),Y.defaultPrevented||w())}),Lt=m.useRef(),Vn=m.useRef(),Xn=(...Y)=>{ke(!0),R==null||R(...Y)};if(!j)return null;const Pn=Object.assign({role:r,ref:D.setDialogRef,"aria-modal":r==="dialog"?!0:void 0},q,{style:i,className:a,tabIndex:-1});let lt=v?v(Pn):g.jsx("div",Object.assign({},Pn,{children:m.cloneElement(l,{role:"document"})}));lt=vp(d,f,{unmountOnExit:!0,mountOnEnter:!0,appear:!0,in:!!n,onExit:$,onExiting:N,onExited:Xn,onEnter:G,onEntering:U,onEntered:X,children:lt});let fe=null;return o&&(fe=E({ref:D.setBackdropRef,onClick:_t}),fe=vp(h,b,{in:!!n,appear:!0,mountOnEnter:!0,unmountOnExit:!0,children:fe})),g.jsx(g.Fragment,{children:fa.createPortal(g.jsxs(g.Fragment,{children:[fe,lt]}),j)})});MS.displayName="Modal";const MR=Object.assign(MS,{Manager:gy});function gp(e,t){return e.classList?!!t&&e.classList.contains(t):(" "+(e.className.baseVal||e.className)+" ").indexOf(" "+t+" ")!==-1}function jR(e,t){e.classList?e.classList.add(t):gp(e,t)||(typeof e.className=="string"?e.className=e.className+" "+t:e.setAttribute("class",(e.className&&e.className.baseVal||"")+" "+t))}function Gv(e,t){return e.replace(new RegExp("(^|\\s)"+t+"(?:\\s|$)","g"),"$1").replace(/\s+/g," ").replace(/^\s*|\s*$/g,"")}function DR(e,t){e.classList?e.classList.remove(t):typeof e.className=="string"?e.className=Gv(e.className,t):e.setAttribute("class",Gv(e.className&&e.className.baseVal||"",t))}const Xa={FIXED_CONTENT:".fixed-top, .fixed-bottom, .is-fixed, .sticky-top",STICKY_CONTENT:".sticky-top",NAVBAR_TOGGLER:".navbar-toggler"};class jS extends gy{adjustAndStore(t,n,r){const a=n.style[t];n.dataset[t]=a,fr(n,{[t]:`${parseFloat(fr(n,t))+r}px`})}restore(t,n){const r=n.dataset[t];r!==void 0&&(delete n.dataset[t],fr(n,{[t]:r}))}setContainerStyle(t){super.setContainerStyle(t);const n=this.getElement();if(jR(n,"modal-open"),!t.scrollBarWidth)return;const r=this.isRTL?"paddingLeft":"paddingRight",a=this.isRTL?"marginLeft":"marginRight";Va(n,Xa.FIXED_CONTENT).forEach(i=>this.adjustAndStore(r,i,t.scrollBarWidth)),Va(n,Xa.STICKY_CONTENT).forEach(i=>this.adjustAndStore(a,i,-t.scrollBarWidth)),Va(n,Xa.NAVBAR_TOGGLER).forEach(i=>this.adjustAndStore(a,i,t.scrollBarWidth))}removeContainerStyle(t){super.removeContainerStyle(t);const n=this.getElement();DR(n,"modal-open");const r=this.isRTL?"paddingLeft":"paddingRight",a=this.isRTL?"marginLeft":"marginRight";Va(n,Xa.FIXED_CONTENT).forEach(i=>this.restore(r,i)),Va(n,Xa.STICKY_CONTENT).forEach(i=>this.restore(a,i)),Va(n,Xa.NAVBAR_TOGGLER).forEach(i=>this.restore(a,i))}}let Kf;function BR(e){return Kf||(Kf=new jS(e)),Kf}const DS=m.createContext({onHide(){}}),kR=m.forwardRef(({closeLabel:e="Close",closeVariant:t,closeButton:n=!1,onHide:r,children:a,...i},l)=>{const o=m.useContext(DS),s=Ii(()=>{o==null||o.onHide(),r==null||r()});return g.jsxs("div",{ref:l,...i,children:[a,n&&g.jsx(wc,{"aria-label":e,variant:t,onClick:s})]})}),BS=m.forwardRef(({bsPrefix:e,className:t,as:n,...r},a)=>{e=H(e,"navbar-brand");const i=n||(r.href?"a":"span");return g.jsx(i,{...r,ref:a,className:z(t,e)})});BS.displayName="NavbarBrand";const kS=m.forwardRef(({children:e,bsPrefix:t,...n},r)=>{t=H(t,"navbar-collapse");const a=m.useContext(qo);return g.jsx(Tc,{in:!!(a&&a.expanded),...n,children:g.jsx("div",{ref:r,className:t,children:e})})});kS.displayName="NavbarCollapse";const zS=m.forwardRef(({bsPrefix:e,className:t,children:n,label:r="Toggle navigation",as:a="button",onClick:i,...l},o)=>{e=H(e,"navbar-toggler");const{onToggle:s,expanded:u}=m.useContext(qo)||{},c=Ii(d=>{i&&i(d),s&&s()});return a==="button"&&(l.type="button"),g.jsx(a,{...l,ref:o,onClick:c,"aria-label":r,className:z(t,e,!u&&"collapsed"),children:n||g.jsx("span",{className:`${e}-icon`})})});zS.displayName="NavbarToggle";const bp=new WeakMap,Yv=(e,t)=>{if(!e||!t)return;const n=bp.get(t)||new Map;bp.set(t,n);let r=n.get(e);return r||(r=t.matchMedia(e),r.refCount=0,n.set(r.media,r)),r};function zR(e,t=typeof window>"u"?void 0:window){const n=Yv(e,t),[r,a]=m.useState(()=>n?n.matches:!1);return bS(()=>{let i=Yv(e,t);if(!i)return a(!1);let l=bp.get(t);const o=()=>{a(i.matches)};return i.refCount++,i.addListener(o),o(),()=>{i.removeListener(o),i.refCount--,i.refCount<=0&&(l==null||l.delete(i.media)),i=void 0}},[e]),r}function LR(e){const t=Object.keys(e);function n(o,s){return o===s?s:o?`${o} and ${s}`:s}function r(o){return t[Math.min(t.indexOf(o)+1,t.length-1)]}function a(o){const s=r(o);let u=e[s];return typeof u=="number"?u=`${u-.2}px`:u=`calc(${u} - 0.2px)`,`(max-width: ${u})`}function i(o){let s=e[o];return typeof s=="number"&&(s=`${s}px`),`(min-width: ${s})`}function l(o,s,u){let c;typeof o=="object"?(c=o,u=s,s=!0):(s=s||!0,c={[o]:s});let d=m.useMemo(()=>Object.entries(c).reduce((f,[h,b])=>((b==="up"||b===!0)&&(f=n(f,i(h))),(b==="down"||b===!0)&&(f=n(f,a(h))),f),""),[JSON.stringify(c)]);return zR(d,u)}return l}const UR=LR({xs:0,sm:576,md:768,lg:992,xl:1200,xxl:1400}),LS=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"offcanvas-body"),g.jsx(n,{ref:a,className:z(e,t),...r})));LS.displayName="OffcanvasBody";const HR={[Vt]:"show",[tr]:"show"},US=m.forwardRef(({bsPrefix:e,className:t,children:n,in:r=!1,mountOnEnter:a=!1,unmountOnExit:i=!1,appear:l=!1,...o},s)=>(e=H(e,"offcanvas"),g.jsx(iy,{ref:s,addEndListener:ry,in:r,mountOnEnter:a,unmountOnExit:i,appear:l,...o,childRef:zo(n),children:(u,c)=>m.cloneElement(n,{...c,className:z(t,n.props.className,(u===Vt||u===ji)&&`${e}-toggling`,HR[u])})})));US.displayName="OffcanvasToggling";const HS=m.forwardRef(({bsPrefix:e,className:t,closeLabel:n="Close",closeButton:r=!1,...a},i)=>(e=H(e,"offcanvas-header"),g.jsx(kR,{ref:i,...a,className:z(t,e),closeLabel:n,closeButton:r})));HS.displayName="OffcanvasHeader";const qR=ly("h5"),qS=m.forwardRef(({className:e,bsPrefix:t,as:n=qR,...r},a)=>(t=H(t,"offcanvas-title"),g.jsx(n,{ref:a,className:z(e,t),...r})));qS.displayName="OffcanvasTitle";function GR(e){return g.jsx(US,{...e})}function YR(e){return g.jsx(uo,{...e})}const GS=m.forwardRef(({bsPrefix:e,className:t,children:n,"aria-labelledby":r,placement:a="start",responsive:i,show:l=!1,backdrop:o=!0,keyboard:s=!0,scroll:u=!1,onEscapeKeyDown:c,onShow:d,onHide:f,container:h,autoFocus:b=!0,enforceFocus:S=!0,restoreFocus:x=!0,restoreFocusOptions:y,onEntered:p,onExit:v,onExiting:E,onEnter:_,onEntering:O,onExited:T,backdropClassName:w,manager:$,renderStaticNode:R=!1,...N},G)=>{const U=m.useRef();e=H(e,"offcanvas");const[X,q]=m.useState(!1),C=Ii(f),j=UR(i||"xs","up");m.useEffect(()=>{q(i?l&&!j:l)},[l,i,j]);const D=m.useMemo(()=>({onHide:C}),[C]);function K(){return $||(u?(U.current||(U.current=new jS({handleContainerOverflow:!1})),U.current):BR())}const ie=(V,...ve)=>{V&&(V.style.visibility="visible"),_==null||_(V,...ve)},Pe=(V,...ve)=>{V&&(V.style.visibility=""),T==null||T(...ve)},ke=m.useCallback(V=>g.jsx("div",{...V,className:z(`${e}-backdrop`,w)}),[w,e]),xe=V=>g.jsx("div",{...V,...N,className:z(t,i?`${e}-${i}`:e,`${e}-${a}`),"aria-labelledby":r,children:n});return g.jsxs(g.Fragment,{children:[!X&&(i||R)&&xe({}),g.jsx(DS.Provider,{value:D,children:g.jsx(MR,{show:X,ref:G,backdrop:o,container:h,keyboard:s,autoFocus:b,enforceFocus:S&&!u,restoreFocus:x,restoreFocusOptions:y,onEscapeKeyDown:c,onShow:d,onHide:C,onEnter:ie,onEntering:O,onEntered:p,onExit:v,onExiting:E,onExited:Pe,manager:K(),transition:GR,backdropTransition:YR,renderBackdrop:ke,renderDialog:xe})})]})});GS.displayName="Offcanvas";const $l=Object.assign(GS,{Body:LS,Header:HS,Title:qS}),YS=m.forwardRef(({onHide:e,...t},n)=>{const r=m.useContext(qo),a=Ii(()=>{r==null||r.onToggle==null||r.onToggle(),e==null||e()});return g.jsx($l,{ref:n,show:!!(r!=null&&r.expanded),...t,renderStaticNode:!0,onHide:a})});YS.displayName="NavbarOffcanvas";const VS=m.forwardRef(({className:e,bsPrefix:t,as:n="span",...r},a)=>(t=H(t,"navbar-text"),g.jsx(n,{ref:a,className:z(e,t),...r})));VS.displayName="NavbarText";const XS=m.forwardRef((e,t)=>{const{bsPrefix:n,expand:r=!0,variant:a="light",bg:i,fixed:l,sticky:o,className:s,as:u="nav",expanded:c,onToggle:d,onSelect:f,collapseOnSelect:h=!1,...b}=VT(e,{expanded:"onToggle"}),S=H(n,"navbar"),x=m.useCallback((...v)=>{f==null||f(...v),h&&c&&(d==null||d(!1))},[f,h,c,d]);b.role===void 0&&u!=="nav"&&(b.role="navigation");let y=`${S}-expand`;typeof r=="string"&&(y=`${y}-${r}`);const p=m.useMemo(()=>({onToggle:()=>d==null?void 0:d(!c),bsPrefix:S,expanded:!!c,expand:r}),[S,c,r,d]);return g.jsx(qo.Provider,{value:p,children:g.jsx(aR.Provider,{value:x,children:g.jsx(u,{ref:t,...b,className:z(s,S,r&&y,a&&`${S}-${a}`,i&&`bg-${i}`,o&&`sticky-${o}`,l&&`fixed-${l}`)})})})});XS.displayName="Navbar";const Zf=Object.assign(XS,{Brand:BS,Collapse:kS,Offcanvas:YS,Text:VS,Toggle:zS}),VR=()=>{};function XR(e,t,{disabled:n,clickTrigger:r}={}){const a=t||VR;eR(e,a,{disabled:n,clickTrigger:r});const i=An(l=>{X1(l)&&a(l)});m.useEffect(()=>{if(n||e==null)return;const l=ko(Gs(e));let o=(l.defaultView||window).event;const s=ir(l,"keyup",u=>{if(u===o){o=void 0;return}i(u)});return()=>{s()}},[e,n,i])}const PS=m.forwardRef((e,t)=>{const{flip:n,offset:r,placement:a,containerPadding:i,popperConfig:l={},transition:o,runTransition:s}=e,[u,c]=Tv(),[d,f]=Tv(),h=$c(c,t),b=mp(e.container),S=mp(e.target),[x,y]=m.useState(!e.show),p=QA(S,u,rR({placement:a,enableEvents:!!e.show,containerPadding:i||5,flip:n,offset:r,arrowElement:d,popperConfig:l}));e.show&&x&&y(!1);const v=(...N)=>{y(!0),e.onExited&&e.onExited(...N)},E=e.show||!x;if(XR(u,e.onHide,{disabled:!e.rootClose||e.rootCloseDisabled,clickTrigger:e.rootCloseEvent}),!E)return null;const{onExit:_,onExiting:O,onEnter:T,onEntering:w,onEntered:$}=e;let R=e.children(Object.assign({},p.attributes.popper,{style:p.styles.popper,ref:h}),{popper:p,placement:a,show:!!e.show,arrowProps:Object.assign({},p.attributes.arrow,{style:p.styles.arrow,ref:f})});return R=vp(o,s,{in:!!e.show,appear:!0,mountOnEnter:!0,unmountOnExit:!0,children:R,onExit:_,onExiting:O,onExited:v,onEnter:T,onEntering:w,onEntered:$}),b?fa.createPortal(R,b):null});PS.displayName="Overlay";const FS=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"popover-header"),g.jsx(n,{ref:a,className:z(e,t),...r})));FS.displayName="PopoverHeader";const by=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"popover-body"),g.jsx(n,{ref:a,className:z(e,t),...r})));by.displayName="PopoverBody";function QS(e,t){let n=e;return e==="left"?n=t?"end":"start":e==="right"&&(n=t?"start":"end"),n}function KS(e="absolute"){return{position:e,top:"0",left:"0",opacity:"0",pointerEvents:"none"}}const PR=m.forwardRef(({bsPrefix:e,placement:t="right",className:n,style:r,children:a,body:i,arrowProps:l,hasDoneInitialMeasure:o,popper:s,show:u,...c},d)=>{const f=H(e,"popover"),h=k1(),[b]=(t==null?void 0:t.split("-"))||[],S=QS(b,h);let x=r;return u&&!o&&(x={...r,...KS(s==null?void 0:s.strategy)}),g.jsxs("div",{ref:d,role:"tooltip",style:x,"x-placement":b,className:z(n,f,b&&`bs-popover-${S}`),...c,children:[g.jsx("div",{className:"popover-arrow",...l}),i?g.jsx(by,{children:a}):a]})}),FR=Object.assign(PR,{Header:FS,Body:by,POPPER_OFFSET:[0,8]}),ZS=m.forwardRef(({bsPrefix:e,placement:t="right",className:n,style:r,children:a,arrowProps:i,hasDoneInitialMeasure:l,popper:o,show:s,...u},c)=>{e=H(e,"tooltip");const d=k1(),[f]=(t==null?void 0:t.split("-"))||[],h=QS(f,d);let b=r;return s&&!l&&(b={...r,...KS(o==null?void 0:o.strategy)}),g.jsxs("div",{ref:c,style:b,role:"tooltip","x-placement":f,className:z(n,e,`bs-tooltip-${h}`),...u,children:[g.jsx("div",{className:"tooltip-arrow",...i}),g.jsx("div",{className:`${e}-inner`,children:a})]})});ZS.displayName="Tooltip";const WS=Object.assign(ZS,{TOOLTIP_OFFSET:[0,6]});function QR(e){const t=m.useRef(null),n=H(void 0,"popover"),r=H(void 0,"tooltip"),a=m.useMemo(()=>({name:"offset",options:{offset:()=>{if(e)return e;if(t.current){if(gp(t.current,n))return FR.POPPER_OFFSET;if(gp(t.current,r))return WS.TOOLTIP_OFFSET}return[0,0]}}}),[e,n,r]);return[t,[a]]}function KR(e,t){const{ref:n}=e,{ref:r}=t;e.ref=n.__wrapped||(n.__wrapped=a=>n(Ou(a))),t.ref=r.__wrapped||(r.__wrapped=a=>r(Ou(a)))}const JS=m.forwardRef(({children:e,transition:t=uo,popperConfig:n={},rootClose:r=!1,placement:a="top",show:i=!1,...l},o)=>{const s=m.useRef({}),[u,c]=m.useState(null),[d,f]=QR(l.offset),h=ay(o,d),b=t===!0?uo:t||void 0,S=Ii(x=>{c(x),n==null||n.onFirstUpdate==null||n.onFirstUpdate(x)});return bS(()=>{u&&l.target&&(s.current.scheduleUpdate==null||s.current.scheduleUpdate())},[u,l.target]),m.useEffect(()=>{i||c(null)},[i]),g.jsx(PS,{...l,ref:h,popperConfig:{...n,modifiers:f.concat(n.modifiers||[]),onFirstUpdate:S},transition:b,rootClose:r,placement:a,show:i,children:(x,{arrowProps:y,popper:p,show:v})=>{var E;KR(x,y);const _=p==null?void 0:p.placement,O=Object.assign(s.current,{state:p==null?void 0:p.state,scheduleUpdate:p==null?void 0:p.update,placement:_,outOfBoundaries:(p==null||(E=p.state)==null||(E=E.modifiersData.hide)==null?void 0:E.isReferenceHidden)||!1,strategy:n.strategy}),T=!!u;return typeof e=="function"?e({...x,placement:_,show:v,...!t&&v&&{className:"show"},popper:O,arrowProps:y,hasDoneInitialMeasure:T}):m.cloneElement(e,{...x,placement:_,arrowProps:y,popper:O,hasDoneInitialMeasure:T,className:z(e.props.className,!t&&v&&"show"),style:{...e.props.style,...x.style}})}})});JS.displayName="Overlay";function ZR(e){return e&&typeof e=="object"?e:{show:e,hide:e}}function Vv(e,t,n){const[r]=t,a=r.currentTarget,i=r.relatedTarget||r.nativeEvent[n];(!i||i!==a)&&!po(a,i)&&e(...t)}he.oneOf(["click","hover","focus"]);const WR=({trigger:e=["hover","focus"],overlay:t,children:n,popperConfig:r={},show:a,defaultShow:i=!1,onToggle:l,delay:o,placement:s,flip:u=s&&s.indexOf("auto")!==-1,...c})=>{const d=m.useRef(null),f=ay(d,zo(n)),h=oS(),b=m.useRef(""),[S,x]=j1(a,i,l),y=ZR(o),{onFocus:p,onBlur:v,onClick:E}=typeof n!="function"?m.Children.only(n).props:{},_=q=>{f(Ou(q))},O=m.useCallback(()=>{if(h.clear(),b.current="show",!y.show){x(!0);return}h.set(()=>{b.current==="show"&&x(!0)},y.show)},[y.show,x,h]),T=m.useCallback(()=>{if(h.clear(),b.current="hide",!y.hide){x(!1);return}h.set(()=>{b.current==="hide"&&x(!1)},y.hide)},[y.hide,x,h]),w=m.useCallback((...q)=>{O(),p==null||p(...q)},[O,p]),$=m.useCallback((...q)=>{T(),v==null||v(...q)},[T,v]),R=m.useCallback((...q)=>{x(!S),E==null||E(...q)},[E,x,S]),N=m.useCallback((...q)=>{Vv(O,q,"fromElement")},[O]),G=m.useCallback((...q)=>{Vv(T,q,"toElement")},[T]),U=e==null?[]:[].concat(e),X={ref:_};return U.indexOf("click")!==-1&&(X.onClick=R),U.indexOf("focus")!==-1&&(X.onFocus=w,X.onBlur=$),U.indexOf("hover")!==-1&&(X.onMouseOver=N,X.onMouseOut=G),g.jsxs(g.Fragment,{children:[typeof n=="function"?n(X):m.cloneElement(n,X),g.jsx(JS,{...c,show:S,onHide:T,flip:u,placement:s,popperConfig:r,target:d.current,children:t})]})},Au=m.forwardRef(({bsPrefix:e,className:t,as:n="div",...r},a)=>{const i=H(e,"row"),l=D1(),o=B1(),s=`${i}-cols`,u=[];return l.forEach(c=>{const d=r[c];delete r[c];let f;d!=null&&typeof d=="object"?{cols:f}=d:f=d;const h=c!==o?`-${c}`:"";f!=null&&u.push(`${s}${h}-${f}`)}),g.jsx(n,{ref:a,...r,className:z(t,i,...u)})});Au.displayName="Row";const IS=m.forwardRef(({bsPrefix:e,variant:t,animation:n="border",size:r,as:a="div",className:i,...l},o)=>{e=H(e,"spinner");const s=`${e}-${n}`;return g.jsx(a,{ref:o,...l,className:z(i,s,r&&`${s}-${r}`,t&&`text-${t}`)})});IS.displayName="Spinner";const JR={[Vt]:"showing",[ji]:"showing show"},eE=m.forwardRef((e,t)=>g.jsx(uo,{...e,ref:t,transitionClasses:JR}));eE.displayName="ToastFade";const tE=m.createContext({onClose(){}}),nE=m.forwardRef(({bsPrefix:e,closeLabel:t="Close",closeVariant:n,closeButton:r=!0,className:a,children:i,...l},o)=>{e=H(e,"toast-header");const s=m.useContext(tE),u=Ii(c=>{s==null||s.onClose==null||s.onClose(c)});return g.jsxs("div",{ref:o,...l,className:z(e,a),children:[i,r&&g.jsx(wc,{"aria-label":t,variant:n,onClick:u,"data-dismiss":"toast"})]})});nE.displayName="ToastHeader";const rE=m.forwardRef(({className:e,bsPrefix:t,as:n="div",...r},a)=>(t=H(t,"toast-body"),g.jsx(n,{ref:a,className:z(e,t),...r})));rE.displayName="ToastBody";const aE=m.forwardRef(({bsPrefix:e,className:t,transition:n=eE,show:r=!0,animation:a=!0,delay:i=5e3,autohide:l=!1,onClose:o,onEntered:s,onExit:u,onExiting:c,onEnter:d,onEntering:f,onExited:h,bg:b,...S},x)=>{e=H(e,"toast");const y=m.useRef(i),p=m.useRef(o);m.useEffect(()=>{y.current=i,p.current=o},[i,o]);const v=oS(),E=!!(l&&r),_=m.useCallback(()=>{E&&(p.current==null||p.current())},[E]);m.useEffect(()=>{v.set(_,y.current)},[v,_]);const O=m.useMemo(()=>({onClose:o}),[o]),T=!!(n&&a),w=g.jsx("div",{...S,ref:x,className:z(e,t,b&&`bg-${b}`,!T&&(r?"show":"hide")),role:"alert","aria-live":"assertive","aria-atomic":"true"});return g.jsx(tE.Provider,{value:O,children:T&&n?g.jsx(n,{in:r,onEnter:d,onEntering:f,onEntered:s,onExit:u,onExiting:c,onExited:h,unmountOnExit:!0,children:w}):w})});aE.displayName="Toast";const Gl=Object.assign(aE,{Body:rE,Header:nE}),IR={"top-start":"top-0 start-0","top-center":"top-0 start-50 translate-middle-x","top-end":"top-0 end-0","middle-start":"top-50 start-0 translate-middle-y","middle-center":"top-50 start-50 translate-middle","middle-end":"top-50 end-0 translate-middle-y","bottom-start":"bottom-0 start-0","bottom-center":"bottom-0 start-50 translate-middle-x","bottom-end":"bottom-0 end-0"},Sy=m.forwardRef(({bsPrefix:e,position:t,containerPosition:n,className:r,as:a="div",...i},l)=>(e=H(e,"toast-container"),g.jsx(a,{ref:l,...i,className:z(e,t&&IR[t],n&&`position-${n}`,r)})));Sy.displayName="ToastContainer";const eC=()=>{},Ey=m.forwardRef(({bsPrefix:e,name:t,className:n,checked:r,type:a,onChange:i,value:l,disabled:o,id:s,inputRef:u,...c},d)=>(e=H(e,"btn-check"),g.jsxs(g.Fragment,{children:[g.jsx("input",{className:e,name:t,type:a,value:l,ref:u,autoComplete:"off",checked:!!r,disabled:!!o,onChange:i||eC,id:s}),g.jsx(Lo,{...c,ref:d,className:z(n,o&&"disabled"),type:void 0,role:void 0,as:"label",htmlFor:s})]})));Ey.displayName="ToggleButton";const Ln=Object.create(null);Ln.open="0";Ln.close="1";Ln.ping="2";Ln.pong="3";Ln.message="4";Ln.upgrade="5";Ln.noop="6";const Ys=Object.create(null);Object.keys(Ln).forEach(e=>{Ys[Ln[e]]=e});const Sp={type:"error",data:"parser error"},iE=typeof Blob=="function"||typeof Blob<"u"&&Object.prototype.toString.call(Blob)==="[object BlobConstructor]",lE=typeof ArrayBuffer=="function",oE=e=>typeof ArrayBuffer.isView=="function"?ArrayBuffer.isView(e):e&&e.buffer instanceof ArrayBuffer,xy=({type:e,data:t},n,r)=>iE&&t instanceof Blob?n?r(t):Xv(t,r):lE&&(t instanceof ArrayBuffer||oE(t))?n?r(t):Xv(new Blob([t]),r):r(Ln[e]+(t||"")),Xv=(e,t)=>{const n=new FileReader;return n.onload=function(){const r=n.result.split(",")[1];t("b"+(r||""))},n.readAsDataURL(e)};function Pv(e){return e instanceof Uint8Array?e:e instanceof ArrayBuffer?new Uint8Array(e):new Uint8Array(e.buffer,e.byteOffset,e.byteLength)}let Wf;function tC(e,t){if(iE&&e.data instanceof Blob)return e.data.arrayBuffer().then(Pv).then(t);if(lE&&(e.data instanceof ArrayBuffer||oE(e.data)))return t(Pv(e.data));xy(e,!1,n=>{Wf||(Wf=new TextEncoder),t(Wf.encode(n))})}const Fv="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",Nl=typeof Uint8Array>"u"?[]:new Uint8Array(256);for(let e=0;e<Fv.length;e++)Nl[Fv.charCodeAt(e)]=e;const nC=e=>{let t=e.length*.75,n=e.length,r,a=0,i,l,o,s;e[e.length-1]==="="&&(t--,e[e.length-2]==="="&&t--);const u=new ArrayBuffer(t),c=new Uint8Array(u);for(r=0;r<n;r+=4)i=Nl[e.charCodeAt(r)],l=Nl[e.charCodeAt(r+1)],o=Nl[e.charCodeAt(r+2)],s=Nl[e.charCodeAt(r+3)],c[a++]=i<<2|l>>4,c[a++]=(l&15)<<4|o>>2,c[a++]=(o&3)<<6|s&63;return u},rC=typeof ArrayBuffer=="function",Oy=(e,t)=>{if(typeof e!="string")return{type:"message",data:sE(e,t)};const n=e.charAt(0);return n==="b"?{type:"message",data:aC(e.substring(1),t)}:Ys[n]?e.length>1?{type:Ys[n],data:e.substring(1)}:{type:Ys[n]}:Sp},aC=(e,t)=>{if(rC){const n=nC(e);return sE(n,t)}else return{base64:!0,data:e}},sE=(e,t)=>{switch(t){case"blob":return e instanceof Blob?e:new Blob([e]);case"arraybuffer":default:return e instanceof ArrayBuffer?e:e.buffer}},uE="",iC=(e,t)=>{const n=e.length,r=new Array(n);let a=0;e.forEach((i,l)=>{xy(i,!1,o=>{r[l]=o,++a===n&&t(r.join(uE))})})},lC=(e,t)=>{const n=e.split(uE),r=[];for(let a=0;a<n.length;a++){const i=Oy(n[a],t);if(r.push(i),i.type==="error")break}return r};function oC(){return new TransformStream({transform(e,t){tC(e,n=>{const r=n.length;let a;if(r<126)a=new Uint8Array(1),new DataView(a.buffer).setUint8(0,r);else if(r<65536){a=new Uint8Array(3);const i=new DataView(a.buffer);i.setUint8(0,126),i.setUint16(1,r)}else{a=new Uint8Array(9);const i=new DataView(a.buffer);i.setUint8(0,127),i.setBigUint64(1,BigInt(r))}e.data&&typeof e.data!="string"&&(a[0]|=128),t.enqueue(a),t.enqueue(n)})}})}let Jf;function hs(e){return e.reduce((t,n)=>t+n.length,0)}function ys(e,t){if(e[0].length===t)return e.shift();const n=new Uint8Array(t);let r=0;for(let a=0;a<t;a++)n[a]=e[0][r++],r===e[0].length&&(e.shift(),r=0);return e.length&&r<e[0].length&&(e[0]=e[0].slice(r)),n}function sC(e,t){Jf||(Jf=new TextDecoder);const n=[];let r=0,a=-1,i=!1;return new TransformStream({transform(l,o){for(n.push(l);;){if(r===0){if(hs(n)<1)break;const s=ys(n,1);i=(s[0]&128)===128,a=s[0]&127,a<126?r=3:a===126?r=1:r=2}else if(r===1){if(hs(n)<2)break;const s=ys(n,2);a=new DataView(s.buffer,s.byteOffset,s.length).getUint16(0),r=3}else if(r===2){if(hs(n)<8)break;const s=ys(n,8),u=new DataView(s.buffer,s.byteOffset,s.length),c=u.getUint32(0);if(c>Math.pow(2,21)-1){o.enqueue(Sp);break}a=c*Math.pow(2,32)+u.getUint32(4),r=3}else{if(hs(n)<a)break;const s=ys(n,a);o.enqueue(Oy(i?s:Jf.decode(s),t)),r=0}if(a===0||a>e){o.enqueue(Sp);break}}}})}const cE=4;function qe(e){if(e)return uC(e)}function uC(e){for(var t in qe.prototype)e[t]=qe.prototype[t];return e}qe.prototype.on=qe.prototype.addEventListener=function(e,t){return this._callbacks=this._callbacks||{},(this._callbacks["$"+e]=this._callbacks["$"+e]||[]).push(t),this};qe.prototype.once=function(e,t){function n(){this.off(e,n),t.apply(this,arguments)}return n.fn=t,this.on(e,n),this};qe.prototype.off=qe.prototype.removeListener=qe.prototype.removeAllListeners=qe.prototype.removeEventListener=function(e,t){if(this._callbacks=this._callbacks||{},arguments.length==0)return this._callbacks={},this;var n=this._callbacks["$"+e];if(!n)return this;if(arguments.length==1)return delete this._callbacks["$"+e],this;for(var r,a=0;a<n.length;a++)if(r=n[a],r===t||r.fn===t){n.splice(a,1);break}return n.length===0&&delete this._callbacks["$"+e],this};qe.prototype.emit=function(e){this._callbacks=this._callbacks||{};for(var t=new Array(arguments.length-1),n=this._callbacks["$"+e],r=1;r<arguments.length;r++)t[r-1]=arguments[r];if(n){n=n.slice(0);for(var r=0,a=n.length;r<a;++r)n[r].apply(this,t)}return this};qe.prototype.emitReserved=qe.prototype.emit;qe.prototype.listeners=function(e){return this._callbacks=this._callbacks||{},this._callbacks["$"+e]||[]};qe.prototype.hasListeners=function(e){return!!this.listeners(e).length};const Nc=typeof Promise=="function"&&typeof Promise.resolve=="function"?t=>Promise.resolve().then(t):(t,n)=>n(t,0),Xt=typeof self<"u"?self:typeof window<"u"?window:Function("return this")(),cC="arraybuffer";function fE(e,...t){return t.reduce((n,r)=>(e.hasOwnProperty(r)&&(n[r]=e[r]),n),{})}const fC=Xt.setTimeout,dC=Xt.clearTimeout;function Mc(e,t){t.useNativeTimers?(e.setTimeoutFn=fC.bind(Xt),e.clearTimeoutFn=dC.bind(Xt)):(e.setTimeoutFn=Xt.setTimeout.bind(Xt),e.clearTimeoutFn=Xt.clearTimeout.bind(Xt))}const pC=1.33;function hC(e){return typeof e=="string"?yC(e):Math.ceil((e.byteLength||e.size)*pC)}function yC(e){let t=0,n=0;for(let r=0,a=e.length;r<a;r++)t=e.charCodeAt(r),t<128?n+=1:t<2048?n+=2:t<55296||t>=57344?n+=3:(r++,n+=4);return n}function dE(){return Date.now().toString(36).substring(3)+Math.random().toString(36).substring(2,5)}function mC(e){let t="";for(let n in e)e.hasOwnProperty(n)&&(t.length&&(t+="&"),t+=encodeURIComponent(n)+"="+encodeURIComponent(e[n]));return t}function vC(e){let t={},n=e.split("&");for(let r=0,a=n.length;r<a;r++){let i=n[r].split("=");t[decodeURIComponent(i[0])]=decodeURIComponent(i[1])}return t}class gC extends Error{constructor(t,n,r){super(t),this.description=n,this.context=r,this.type="TransportError"}}class _y extends qe{constructor(t){super(),this.writable=!1,Mc(this,t),this.opts=t,this.query=t.query,this.socket=t.socket,this.supportsBinary=!t.forceBase64}onError(t,n,r){return super.emitReserved("error",new gC(t,n,r)),this}open(){return this.readyState="opening",this.doOpen(),this}close(){return(this.readyState==="opening"||this.readyState==="open")&&(this.doClose(),this.onClose()),this}send(t){this.readyState==="open"&&this.write(t)}onOpen(){this.readyState="open",this.writable=!0,super.emitReserved("open")}onData(t){const n=Oy(t,this.socket.binaryType);this.onPacket(n)}onPacket(t){super.emitReserved("packet",t)}onClose(t){this.readyState="closed",super.emitReserved("close",t)}pause(t){}createUri(t,n={}){return t+"://"+this._hostname()+this._port()+this.opts.path+this._query(n)}_hostname(){const t=this.opts.hostname;return t.indexOf(":")===-1?t:"["+t+"]"}_port(){return this.opts.port&&(this.opts.secure&&+(this.opts.port!==443)||!this.opts.secure&&Number(this.opts.port)!==80)?":"+this.opts.port:""}_query(t){const n=mC(t);return n.length?"?"+n:""}}class bC extends _y{constructor(){super(...arguments),this._polling=!1}get name(){return"polling"}doOpen(){this._poll()}pause(t){this.readyState="pausing";const n=()=>{this.readyState="paused",t()};if(this._polling||!this.writable){let r=0;this._polling&&(r++,this.once("pollComplete",function(){--r||n()})),this.writable||(r++,this.once("drain",function(){--r||n()}))}else n()}_poll(){this._polling=!0,this.doPoll(),this.emitReserved("poll")}onData(t){const n=r=>{if(this.readyState==="opening"&&r.type==="open"&&this.onOpen(),r.type==="close")return this.onClose({description:"transport closed by the server"}),!1;this.onPacket(r)};lC(t,this.socket.binaryType).forEach(n),this.readyState!=="closed"&&(this._polling=!1,this.emitReserved("pollComplete"),this.readyState==="open"&&this._poll())}doClose(){const t=()=>{this.write([{type:"close"}])};this.readyState==="open"?t():this.once("open",t)}write(t){this.writable=!1,iC(t,n=>{this.doWrite(n,()=>{this.writable=!0,this.emitReserved("drain")})})}uri(){const t=this.opts.secure?"https":"http",n=this.query||{};return this.opts.timestampRequests!==!1&&(n[this.opts.timestampParam]=dE()),!this.supportsBinary&&!n.sid&&(n.b64=1),this.createUri(t,n)}}let pE=!1;try{pE=typeof XMLHttpRequest<"u"&&"withCredentials"in new XMLHttpRequest}catch{}const SC=pE;function EC(){}class xC extends bC{constructor(t){if(super(t),typeof location<"u"){const n=location.protocol==="https:";let r=location.port;r||(r=n?"443":"80"),this.xd=typeof location<"u"&&t.hostname!==location.hostname||r!==t.port}}doWrite(t,n){const r=this.request({method:"POST",data:t});r.on("success",n),r.on("error",(a,i)=>{this.onError("xhr post error",a,i)})}doPoll(){const t=this.request();t.on("data",this.onData.bind(this)),t.on("error",(n,r)=>{this.onError("xhr poll error",n,r)}),this.pollXhr=t}}class Mn extends qe{constructor(t,n,r){super(),this.createRequest=t,Mc(this,r),this._opts=r,this._method=r.method||"GET",this._uri=n,this._data=r.data!==void 0?r.data:null,this._create()}_create(){var t;const n=fE(this._opts,"agent","pfx","key","passphrase","cert","ca","ciphers","rejectUnauthorized","autoUnref");n.xdomain=!!this._opts.xd;const r=this._xhr=this.createRequest(n);try{r.open(this._method,this._uri,!0);try{if(this._opts.extraHeaders){r.setDisableHeaderCheck&&r.setDisableHeaderCheck(!0);for(let a in this._opts.extraHeaders)this._opts.extraHeaders.hasOwnProperty(a)&&r.setRequestHeader(a,this._opts.extraHeaders[a])}}catch{}if(this._method==="POST")try{r.setRequestHeader("Content-type","text/plain;charset=UTF-8")}catch{}try{r.setRequestHeader("Accept","*/*")}catch{}(t=this._opts.cookieJar)===null||t===void 0||t.addCookies(r),"withCredentials"in r&&(r.withCredentials=this._opts.withCredentials),this._opts.requestTimeout&&(r.timeout=this._opts.requestTimeout),r.onreadystatechange=()=>{var a;r.readyState===3&&((a=this._opts.cookieJar)===null||a===void 0||a.parseCookies(r.getResponseHeader("set-cookie"))),r.readyState===4&&(r.status===200||r.status===1223?this._onLoad():this.setTimeoutFn(()=>{this._onError(typeof r.status=="number"?r.status:0)},0))},r.send(this._data)}catch(a){this.setTimeoutFn(()=>{this._onError(a)},0);return}typeof document<"u"&&(this._index=Mn.requestsCount++,Mn.requests[this._index]=this)}_onError(t){this.emitReserved("error",t,this._xhr),this._cleanup(!0)}_cleanup(t){if(!(typeof this._xhr>"u"||this._xhr===null)){if(this._xhr.onreadystatechange=EC,t)try{this._xhr.abort()}catch{}typeof document<"u"&&delete Mn.requests[this._index],this._xhr=null}}_onLoad(){const t=this._xhr.responseText;t!==null&&(this.emitReserved("data",t),this.emitReserved("success"),this._cleanup())}abort(){this._cleanup()}}Mn.requestsCount=0;Mn.requests={};if(typeof document<"u"){if(typeof attachEvent=="function")attachEvent("onunload",Qv);else if(typeof addEventListener=="function"){const e="onpagehide"in Xt?"pagehide":"unload";addEventListener(e,Qv,!1)}}function Qv(){for(let e in Mn.requests)Mn.requests.hasOwnProperty(e)&&Mn.requests[e].abort()}const OC=function(){const e=hE({xdomain:!1});return e&&e.responseType!==null}();class _C extends xC{constructor(t){super(t);const n=t&&t.forceBase64;this.supportsBinary=OC&&!n}request(t={}){return Object.assign(t,{xd:this.xd},this.opts),new Mn(hE,this.uri(),t)}}function hE(e){const t=e.xdomain;try{if(typeof XMLHttpRequest<"u"&&(!t||SC))return new XMLHttpRequest}catch{}if(!t)try{return new Xt[["Active"].concat("Object").join("X")]("Microsoft.XMLHTTP")}catch{}}const yE=typeof navigator<"u"&&typeof navigator.product=="string"&&navigator.product.toLowerCase()==="reactnative";class TC extends _y{get name(){return"websocket"}doOpen(){const t=this.uri(),n=this.opts.protocols,r=yE?{}:fE(this.opts,"agent","perMessageDeflate","pfx","key","passphrase","cert","ca","ciphers","rejectUnauthorized","localAddress","protocolVersion","origin","maxPayload","family","checkServerIdentity");this.opts.extraHeaders&&(r.headers=this.opts.extraHeaders);try{this.ws=this.createSocket(t,n,r)}catch(a){return this.emitReserved("error",a)}this.ws.binaryType=this.socket.binaryType,this.addEventListeners()}addEventListeners(){this.ws.onopen=()=>{this.opts.autoUnref&&this.ws._socket.unref(),this.onOpen()},this.ws.onclose=t=>this.onClose({description:"websocket connection closed",context:t}),this.ws.onmessage=t=>this.onData(t.data),this.ws.onerror=t=>this.onError("websocket error",t)}write(t){this.writable=!1;for(let n=0;n<t.length;n++){const r=t[n],a=n===t.length-1;xy(r,this.supportsBinary,i=>{try{this.doWrite(r,i)}catch{}a&&Nc(()=>{this.writable=!0,this.emitReserved("drain")},this.setTimeoutFn)})}}doClose(){typeof this.ws<"u"&&(this.ws.onerror=()=>{},this.ws.close(),this.ws=null)}uri(){const t=this.opts.secure?"wss":"ws",n=this.query||{};return this.opts.timestampRequests&&(n[this.opts.timestampParam]=dE()),this.supportsBinary||(n.b64=1),this.createUri(t,n)}}const If=Xt.WebSocket||Xt.MozWebSocket;class wC extends TC{createSocket(t,n,r){return yE?new If(t,n,r):n?new If(t,n):new If(t)}doWrite(t,n){this.ws.send(n)}}class AC extends _y{get name(){return"webtransport"}doOpen(){try{this._transport=new WebTransport(this.createUri("https"),this.opts.transportOptions[this.name])}catch(t){return this.emitReserved("error",t)}this._transport.closed.then(()=>{this.onClose()}).catch(t=>{this.onError("webtransport error",t)}),this._transport.ready.then(()=>{this._transport.createBidirectionalStream().then(t=>{const n=sC(Number.MAX_SAFE_INTEGER,this.socket.binaryType),r=t.readable.pipeThrough(n).getReader(),a=oC();a.readable.pipeTo(t.writable),this._writer=a.writable.getWriter();const i=()=>{r.read().then(({done:o,value:s})=>{o||(this.onPacket(s),i())}).catch(o=>{})};i();const l={type:"open"};this.query.sid&&(l.data=`{"sid":"${this.query.sid}"}`),this._writer.write(l).then(()=>this.onOpen())})})}write(t){this.writable=!1;for(let n=0;n<t.length;n++){const r=t[n],a=n===t.length-1;this._writer.write(r).then(()=>{a&&Nc(()=>{this.writable=!0,this.emitReserved("drain")},this.setTimeoutFn)})}}doClose(){var t;(t=this._transport)===null||t===void 0||t.close()}}const RC={websocket:wC,webtransport:AC,polling:_C},CC=/^(?:(?![^:@\/?#]+:[^:@\/]*@)(http|https|ws|wss):\/\/)?((?:(([^:@\/?#]*)(?::([^:@\/?#]*))?)?@)?((?:[a-f0-9]{0,4}:){2,7}[a-f0-9]{0,4}|[^:\/?#]*)(?::(\d*))?)(((\/(?:[^?#](?![^?#\/]*\.[^?#\/.]+(?:[?#]|$)))*\/?)?([^?#\/]*))(?:\?([^#]*))?(?:#(.*))?)/,$C=["source","protocol","authority","userInfo","user","password","host","port","relative","path","directory","file","query","anchor"];function Ep(e){if(e.length>8e3)throw"URI too long";const t=e,n=e.indexOf("["),r=e.indexOf("]");n!=-1&&r!=-1&&(e=e.substring(0,n)+e.substring(n,r).replace(/:/g,";")+e.substring(r,e.length));let a=CC.exec(e||""),i={},l=14;for(;l--;)i[$C[l]]=a[l]||"";return n!=-1&&r!=-1&&(i.source=t,i.host=i.host.substring(1,i.host.length-1).replace(/;/g,":"),i.authority=i.authority.replace("[","").replace("]","").replace(/;/g,":"),i.ipv6uri=!0),i.pathNames=NC(i,i.path),i.queryKey=MC(i,i.query),i}function NC(e,t){const n=/\/{2,9}/g,r=t.replace(n,"/").split("/");return(t.slice(0,1)=="/"||t.length===0)&&r.splice(0,1),t.slice(-1)=="/"&&r.splice(r.length-1,1),r}function MC(e,t){const n={};return t.replace(/(?:^|&)([^&=]*)=?([^&]*)/g,function(r,a,i){a&&(n[a]=i)}),n}const xp=typeof addEventListener=="function"&&typeof removeEventListener=="function",Vs=[];xp&&addEventListener("offline",()=>{Vs.forEach(e=>e())},!1);class qr extends qe{constructor(t,n){if(super(),this.binaryType=cC,this.writeBuffer=[],this._prevBufferLen=0,this._pingInterval=-1,this._pingTimeout=-1,this._maxPayload=-1,this._pingTimeoutTime=1/0,t&&typeof t=="object"&&(n=t,t=null),t){const r=Ep(t);n.hostname=r.host,n.secure=r.protocol==="https"||r.protocol==="wss",n.port=r.port,r.query&&(n.query=r.query)}else n.host&&(n.hostname=Ep(n.host).host);Mc(this,n),this.secure=n.secure!=null?n.secure:typeof location<"u"&&location.protocol==="https:",n.hostname&&!n.port&&(n.port=this.secure?"443":"80"),this.hostname=n.hostname||(typeof location<"u"?location.hostname:"localhost"),this.port=n.port||(typeof location<"u"&&location.port?location.port:this.secure?"443":"80"),this.transports=[],this._transportsByName={},n.transports.forEach(r=>{const a=r.prototype.name;this.transports.push(a),this._transportsByName[a]=r}),this.opts=Object.assign({path:"/engine.io",agent:!1,withCredentials:!1,upgrade:!0,timestampParam:"t",rememberUpgrade:!1,addTrailingSlash:!0,rejectUnauthorized:!0,perMessageDeflate:{threshold:1024},transportOptions:{},closeOnBeforeunload:!1},n),this.opts.path=this.opts.path.replace(/\/$/,"")+(this.opts.addTrailingSlash?"/":""),typeof this.opts.query=="string"&&(this.opts.query=vC(this.opts.query)),xp&&(this.opts.closeOnBeforeunload&&(this._beforeunloadEventListener=()=>{this.transport&&(this.transport.removeAllListeners(),this.transport.close())},addEventListener("beforeunload",this._beforeunloadEventListener,!1)),this.hostname!=="localhost"&&(this._offlineEventListener=()=>{this._onClose("transport close",{description:"network connection lost"})},Vs.push(this._offlineEventListener))),this.opts.withCredentials&&(this._cookieJar=void 0),this._open()}createTransport(t){const n=Object.assign({},this.opts.query);n.EIO=cE,n.transport=t,this.id&&(n.sid=this.id);const r=Object.assign({},this.opts,{query:n,socket:this,hostname:this.hostname,secure:this.secure,port:this.port},this.opts.transportOptions[t]);return new this._transportsByName[t](r)}_open(){if(this.transports.length===0){this.setTimeoutFn(()=>{this.emitReserved("error","No transports available")},0);return}const t=this.opts.rememberUpgrade&&qr.priorWebsocketSuccess&&this.transports.indexOf("websocket")!==-1?"websocket":this.transports[0];this.readyState="opening";const n=this.createTransport(t);n.open(),this.setTransport(n)}setTransport(t){this.transport&&this.transport.removeAllListeners(),this.transport=t,t.on("drain",this._onDrain.bind(this)).on("packet",this._onPacket.bind(this)).on("error",this._onError.bind(this)).on("close",n=>this._onClose("transport close",n))}onOpen(){this.readyState="open",qr.priorWebsocketSuccess=this.transport.name==="websocket",this.emitReserved("open"),this.flush()}_onPacket(t){if(this.readyState==="opening"||this.readyState==="open"||this.readyState==="closing")switch(this.emitReserved("packet",t),this.emitReserved("heartbeat"),t.type){case"open":this.onHandshake(JSON.parse(t.data));break;case"ping":this._sendPacket("pong"),this.emitReserved("ping"),this.emitReserved("pong"),this._resetPingTimeout();break;case"error":const n=new Error("server error");n.code=t.data,this._onError(n);break;case"message":this.emitReserved("data",t.data),this.emitReserved("message",t.data);break}}onHandshake(t){this.emitReserved("handshake",t),this.id=t.sid,this.transport.query.sid=t.sid,this._pingInterval=t.pingInterval,this._pingTimeout=t.pingTimeout,this._maxPayload=t.maxPayload,this.onOpen(),this.readyState!=="closed"&&this._resetPingTimeout()}_resetPingTimeout(){this.clearTimeoutFn(this._pingTimeoutTimer);const t=this._pingInterval+this._pingTimeout;this._pingTimeoutTime=Date.now()+t,this._pingTimeoutTimer=this.setTimeoutFn(()=>{this._onClose("ping timeout")},t),this.opts.autoUnref&&this._pingTimeoutTimer.unref()}_onDrain(){this.writeBuffer.splice(0,this._prevBufferLen),this._prevBufferLen=0,this.writeBuffer.length===0?this.emitReserved("drain"):this.flush()}flush(){if(this.readyState!=="closed"&&this.transport.writable&&!this.upgrading&&this.writeBuffer.length){const t=this._getWritablePackets();this.transport.send(t),this._prevBufferLen=t.length,this.emitReserved("flush")}}_getWritablePackets(){if(!(this._maxPayload&&this.transport.name==="polling"&&this.writeBuffer.length>1))return this.writeBuffer;let n=1;for(let r=0;r<this.writeBuffer.length;r++){const a=this.writeBuffer[r].data;if(a&&(n+=hC(a)),r>0&&n>this._maxPayload)return this.writeBuffer.slice(0,r);n+=2}return this.writeBuffer}_hasPingExpired(){if(!this._pingTimeoutTime)return!0;const t=Date.now()>this._pingTimeoutTime;return t&&(this._pingTimeoutTime=0,Nc(()=>{this._onClose("ping timeout")},this.setTimeoutFn)),t}write(t,n,r){return this._sendPacket("message",t,n,r),this}send(t,n,r){return this._sendPacket("message",t,n,r),this}_sendPacket(t,n,r,a){if(typeof n=="function"&&(a=n,n=void 0),typeof r=="function"&&(a=r,r=null),this.readyState==="closing"||this.readyState==="closed")return;r=r||{},r.compress=r.compress!==!1;const i={type:t,data:n,options:r};this.emitReserved("packetCreate",i),this.writeBuffer.push(i),a&&this.once("flush",a),this.flush()}close(){const t=()=>{this._onClose("forced close"),this.transport.close()},n=()=>{this.off("upgrade",n),this.off("upgradeError",n),t()},r=()=>{this.once("upgrade",n),this.once("upgradeError",n)};return(this.readyState==="opening"||this.readyState==="open")&&(this.readyState="closing",this.writeBuffer.length?this.once("drain",()=>{this.upgrading?r():t()}):this.upgrading?r():t()),this}_onError(t){if(qr.priorWebsocketSuccess=!1,this.opts.tryAllTransports&&this.transports.length>1&&this.readyState==="opening")return this.transports.shift(),this._open();this.emitReserved("error",t),this._onClose("transport error",t)}_onClose(t,n){if(this.readyState==="opening"||this.readyState==="open"||this.readyState==="closing"){if(this.clearTimeoutFn(this._pingTimeoutTimer),this.transport.removeAllListeners("close"),this.transport.close(),this.transport.removeAllListeners(),xp&&(this._beforeunloadEventListener&&removeEventListener("beforeunload",this._beforeunloadEventListener,!1),this._offlineEventListener)){const r=Vs.indexOf(this._offlineEventListener);r!==-1&&Vs.splice(r,1)}this.readyState="closed",this.id=null,this.emitReserved("close",t,n),this.writeBuffer=[],this._prevBufferLen=0}}}qr.protocol=cE;class jC extends qr{constructor(){super(...arguments),this._upgrades=[]}onOpen(){if(super.onOpen(),this.readyState==="open"&&this.opts.upgrade)for(let t=0;t<this._upgrades.length;t++)this._probe(this._upgrades[t])}_probe(t){let n=this.createTransport(t),r=!1;qr.priorWebsocketSuccess=!1;const a=()=>{r||(n.send([{type:"ping",data:"probe"}]),n.once("packet",d=>{if(!r)if(d.type==="pong"&&d.data==="probe"){if(this.upgrading=!0,this.emitReserved("upgrading",n),!n)return;qr.priorWebsocketSuccess=n.name==="websocket",this.transport.pause(()=>{r||this.readyState!=="closed"&&(c(),this.setTransport(n),n.send([{type:"upgrade"}]),this.emitReserved("upgrade",n),n=null,this.upgrading=!1,this.flush())})}else{const f=new Error("probe error");f.transport=n.name,this.emitReserved("upgradeError",f)}}))};function i(){r||(r=!0,c(),n.close(),n=null)}const l=d=>{const f=new Error("probe error: "+d);f.transport=n.name,i(),this.emitReserved("upgradeError",f)};function o(){l("transport closed")}function s(){l("socket closed")}function u(d){n&&d.name!==n.name&&i()}const c=()=>{n.removeListener("open",a),n.removeListener("error",l),n.removeListener("close",o),this.off("close",s),this.off("upgrading",u)};n.once("open",a),n.once("error",l),n.once("close",o),this.once("close",s),this.once("upgrading",u),this._upgrades.indexOf("webtransport")!==-1&&t!=="webtransport"?this.setTimeoutFn(()=>{r||n.open()},200):n.open()}onHandshake(t){this._upgrades=this._filterUpgrades(t.upgrades),super.onHandshake(t)}_filterUpgrades(t){const n=[];for(let r=0;r<t.length;r++)~this.transports.indexOf(t[r])&&n.push(t[r]);return n}}let DC=class extends jC{constructor(t,n={}){const r=typeof t=="object"?t:n;(!r.transports||r.transports&&typeof r.transports[0]=="string")&&(r.transports=(r.transports||["polling","websocket","webtransport"]).map(a=>RC[a]).filter(a=>!!a)),super(t,r)}};function BC(e,t="",n){let r=e;n=n||typeof location<"u"&&location,e==null&&(e=n.protocol+"//"+n.host),typeof e=="string"&&(e.charAt(0)==="/"&&(e.charAt(1)==="/"?e=n.protocol+e:e=n.host+e),/^(https?|wss?):\/\//.test(e)||(typeof n<"u"?e=n.protocol+"//"+e:e="https://"+e),r=Ep(e)),r.port||(/^(http|ws)$/.test(r.protocol)?r.port="80":/^(http|ws)s$/.test(r.protocol)&&(r.port="443")),r.path=r.path||"/";const i=r.host.indexOf(":")!==-1?"["+r.host+"]":r.host;return r.id=r.protocol+"://"+i+":"+r.port+t,r.href=r.protocol+"://"+i+(n&&n.port===r.port?"":":"+r.port),r}const kC=typeof ArrayBuffer=="function",zC=e=>typeof ArrayBuffer.isView=="function"?ArrayBuffer.isView(e):e.buffer instanceof ArrayBuffer,mE=Object.prototype.toString,LC=typeof Blob=="function"||typeof Blob<"u"&&mE.call(Blob)==="[object BlobConstructor]",UC=typeof File=="function"||typeof File<"u"&&mE.call(File)==="[object FileConstructor]";function Ty(e){return kC&&(e instanceof ArrayBuffer||zC(e))||LC&&e instanceof Blob||UC&&e instanceof File}function Xs(e,t){if(!e||typeof e!="object")return!1;if(Array.isArray(e)){for(let n=0,r=e.length;n<r;n++)if(Xs(e[n]))return!0;return!1}if(Ty(e))return!0;if(e.toJSON&&typeof e.toJSON=="function"&&arguments.length===1)return Xs(e.toJSON(),!0);for(const n in e)if(Object.prototype.hasOwnProperty.call(e,n)&&Xs(e[n]))return!0;return!1}function HC(e){const t=[],n=e.data,r=e;return r.data=Op(n,t),r.attachments=t.length,{packet:r,buffers:t}}function Op(e,t){if(!e)return e;if(Ty(e)){const n={_placeholder:!0,num:t.length};return t.push(e),n}else if(Array.isArray(e)){const n=new Array(e.length);for(let r=0;r<e.length;r++)n[r]=Op(e[r],t);return n}else if(typeof e=="object"&&!(e instanceof Date)){const n={};for(const r in e)Object.prototype.hasOwnProperty.call(e,r)&&(n[r]=Op(e[r],t));return n}return e}function qC(e,t){return e.data=_p(e.data,t),delete e.attachments,e}function _p(e,t){if(!e)return e;if(e&&e._placeholder===!0){if(typeof e.num=="number"&&e.num>=0&&e.num<t.length)return t[e.num];throw new Error("illegal attachments")}else if(Array.isArray(e))for(let n=0;n<e.length;n++)e[n]=_p(e[n],t);else if(typeof e=="object")for(const n in e)Object.prototype.hasOwnProperty.call(e,n)&&(e[n]=_p(e[n],t));return e}const GC=["connect","connect_error","disconnect","disconnecting","newListener","removeListener"],YC=5;var I;(function(e){e[e.CONNECT=0]="CONNECT",e[e.DISCONNECT=1]="DISCONNECT",e[e.EVENT=2]="EVENT",e[e.ACK=3]="ACK",e[e.CONNECT_ERROR=4]="CONNECT_ERROR",e[e.BINARY_EVENT=5]="BINARY_EVENT",e[e.BINARY_ACK=6]="BINARY_ACK"})(I||(I={}));class VC{constructor(t){this.replacer=t}encode(t){return(t.type===I.EVENT||t.type===I.ACK)&&Xs(t)?this.encodeAsBinary({type:t.type===I.EVENT?I.BINARY_EVENT:I.BINARY_ACK,nsp:t.nsp,data:t.data,id:t.id}):[this.encodeAsString(t)]}encodeAsString(t){let n=""+t.type;return(t.type===I.BINARY_EVENT||t.type===I.BINARY_ACK)&&(n+=t.attachments+"-"),t.nsp&&t.nsp!=="/"&&(n+=t.nsp+","),t.id!=null&&(n+=t.id),t.data!=null&&(n+=JSON.stringify(t.data,this.replacer)),n}encodeAsBinary(t){const n=HC(t),r=this.encodeAsString(n.packet),a=n.buffers;return a.unshift(r),a}}function Kv(e){return Object.prototype.toString.call(e)==="[object Object]"}class wy extends qe{constructor(t){super(),this.reviver=t}add(t){let n;if(typeof t=="string"){if(this.reconstructor)throw new Error("got plaintext data when reconstructing a packet");n=this.decodeString(t);const r=n.type===I.BINARY_EVENT;r||n.type===I.BINARY_ACK?(n.type=r?I.EVENT:I.ACK,this.reconstructor=new XC(n),n.attachments===0&&super.emitReserved("decoded",n)):super.emitReserved("decoded",n)}else if(Ty(t)||t.base64)if(this.reconstructor)n=this.reconstructor.takeBinaryData(t),n&&(this.reconstructor=null,super.emitReserved("decoded",n));else throw new Error("got binary data when not reconstructing a packet");else throw new Error("Unknown type: "+t)}decodeString(t){let n=0;const r={type:Number(t.charAt(0))};if(I[r.type]===void 0)throw new Error("unknown packet type "+r.type);if(r.type===I.BINARY_EVENT||r.type===I.BINARY_ACK){const i=n+1;for(;t.charAt(++n)!=="-"&&n!=t.length;);const l=t.substring(i,n);if(l!=Number(l)||t.charAt(n)!=="-")throw new Error("Illegal attachments");r.attachments=Number(l)}if(t.charAt(n+1)==="/"){const i=n+1;for(;++n&&!(t.charAt(n)===","||n===t.length););r.nsp=t.substring(i,n)}else r.nsp="/";const a=t.charAt(n+1);if(a!==""&&Number(a)==a){const i=n+1;for(;++n;){const l=t.charAt(n);if(l==null||Number(l)!=l){--n;break}if(n===t.length)break}r.id=Number(t.substring(i,n+1))}if(t.charAt(++n)){const i=this.tryParse(t.substr(n));if(wy.isPayloadValid(r.type,i))r.data=i;else throw new Error("invalid payload")}return r}tryParse(t){try{return JSON.parse(t,this.reviver)}catch{return!1}}static isPayloadValid(t,n){switch(t){case I.CONNECT:return Kv(n);case I.DISCONNECT:return n===void 0;case I.CONNECT_ERROR:return typeof n=="string"||Kv(n);case I.EVENT:case I.BINARY_EVENT:return Array.isArray(n)&&(typeof n[0]=="number"||typeof n[0]=="string"&&GC.indexOf(n[0])===-1);case I.ACK:case I.BINARY_ACK:return Array.isArray(n)}}destroy(){this.reconstructor&&(this.reconstructor.finishedReconstruction(),this.reconstructor=null)}}class XC{constructor(t){this.packet=t,this.buffers=[],this.reconPack=t}takeBinaryData(t){if(this.buffers.push(t),this.buffers.length===this.reconPack.attachments){const n=qC(this.reconPack,this.buffers);return this.finishedReconstruction(),n}return null}finishedReconstruction(){this.reconPack=null,this.buffers=[]}}const PC=Object.freeze(Object.defineProperty({__proto__:null,Decoder:wy,Encoder:VC,get PacketType(){return I},protocol:YC},Symbol.toStringTag,{value:"Module"}));function pn(e,t,n){return e.on(t,n),function(){e.off(t,n)}}const FC=Object.freeze({connect:1,connect_error:1,disconnect:1,disconnecting:1,newListener:1,removeListener:1});class vE extends qe{constructor(t,n,r){super(),this.connected=!1,this.recovered=!1,this.receiveBuffer=[],this.sendBuffer=[],this._queue=[],this._queueSeq=0,this.ids=0,this.acks={},this.flags={},this.io=t,this.nsp=n,r&&r.auth&&(this.auth=r.auth),this._opts=Object.assign({},r),this.io._autoConnect&&this.open()}get disconnected(){return!this.connected}subEvents(){if(this.subs)return;const t=this.io;this.subs=[pn(t,"open",this.onopen.bind(this)),pn(t,"packet",this.onpacket.bind(this)),pn(t,"error",this.onerror.bind(this)),pn(t,"close",this.onclose.bind(this))]}get active(){return!!this.subs}connect(){return this.connected?this:(this.subEvents(),this.io._reconnecting||this.io.open(),this.io._readyState==="open"&&this.onopen(),this)}open(){return this.connect()}send(...t){return t.unshift("message"),this.emit.apply(this,t),this}emit(t,...n){var r,a,i;if(FC.hasOwnProperty(t))throw new Error('"'+t.toString()+'" is a reserved event name');if(n.unshift(t),this._opts.retries&&!this.flags.fromQueue&&!this.flags.volatile)return this._addToQueue(n),this;const l={type:I.EVENT,data:n};if(l.options={},l.options.compress=this.flags.compress!==!1,typeof n[n.length-1]=="function"){const c=this.ids++,d=n.pop();this._registerAckCallback(c,d),l.id=c}const o=(a=(r=this.io.engine)===null||r===void 0?void 0:r.transport)===null||a===void 0?void 0:a.writable,s=this.connected&&!(!((i=this.io.engine)===null||i===void 0)&&i._hasPingExpired());return this.flags.volatile&&!o||(s?(this.notifyOutgoingListeners(l),this.packet(l)):this.sendBuffer.push(l)),this.flags={},this}_registerAckCallback(t,n){var r;const a=(r=this.flags.timeout)!==null&&r!==void 0?r:this._opts.ackTimeout;if(a===void 0){this.acks[t]=n;return}const i=this.io.setTimeoutFn(()=>{delete this.acks[t];for(let o=0;o<this.sendBuffer.length;o++)this.sendBuffer[o].id===t&&this.sendBuffer.splice(o,1);n.call(this,new Error("operation has timed out"))},a),l=(...o)=>{this.io.clearTimeoutFn(i),n.apply(this,o)};l.withError=!0,this.acks[t]=l}emitWithAck(t,...n){return new Promise((r,a)=>{const i=(l,o)=>l?a(l):r(o);i.withError=!0,n.push(i),this.emit(t,...n)})}_addToQueue(t){let n;typeof t[t.length-1]=="function"&&(n=t.pop());const r={id:this._queueSeq++,tryCount:0,pending:!1,args:t,flags:Object.assign({fromQueue:!0},this.flags)};t.push((a,...i)=>r!==this._queue[0]?void 0:(a!==null?r.tryCount>this._opts.retries&&(this._queue.shift(),n&&n(a)):(this._queue.shift(),n&&n(null,...i)),r.pending=!1,this._drainQueue())),this._queue.push(r),this._drainQueue()}_drainQueue(t=!1){if(!this.connected||this._queue.length===0)return;const n=this._queue[0];n.pending&&!t||(n.pending=!0,n.tryCount++,this.flags=n.flags,this.emit.apply(this,n.args))}packet(t){t.nsp=this.nsp,this.io._packet(t)}onopen(){typeof this.auth=="function"?this.auth(t=>{this._sendConnectPacket(t)}):this._sendConnectPacket(this.auth)}_sendConnectPacket(t){this.packet({type:I.CONNECT,data:this._pid?Object.assign({pid:this._pid,offset:this._lastOffset},t):t})}onerror(t){this.connected||this.emitReserved("connect_error",t)}onclose(t,n){this.connected=!1,delete this.id,this.emitReserved("disconnect",t,n),this._clearAcks()}_clearAcks(){Object.keys(this.acks).forEach(t=>{if(!this.sendBuffer.some(r=>String(r.id)===t)){const r=this.acks[t];delete this.acks[t],r.withError&&r.call(this,new Error("socket has been disconnected"))}})}onpacket(t){if(t.nsp===this.nsp)switch(t.type){case I.CONNECT:t.data&&t.data.sid?this.onconnect(t.data.sid,t.data.pid):this.emitReserved("connect_error",new Error("It seems you are trying to reach a Socket.IO server in v2.x with a v3.x client, but they are not compatible (more information here: https://socket.io/docs/v3/migrating-from-2-x-to-3-0/)"));break;case I.EVENT:case I.BINARY_EVENT:this.onevent(t);break;case I.ACK:case I.BINARY_ACK:this.onack(t);break;case I.DISCONNECT:this.ondisconnect();break;case I.CONNECT_ERROR:this.destroy();const r=new Error(t.data.message);r.data=t.data.data,this.emitReserved("connect_error",r);break}}onevent(t){const n=t.data||[];t.id!=null&&n.push(this.ack(t.id)),this.connected?this.emitEvent(n):this.receiveBuffer.push(Object.freeze(n))}emitEvent(t){if(this._anyListeners&&this._anyListeners.length){const n=this._anyListeners.slice();for(const r of n)r.apply(this,t)}super.emit.apply(this,t),this._pid&&t.length&&typeof t[t.length-1]=="string"&&(this._lastOffset=t[t.length-1])}ack(t){const n=this;let r=!1;return function(...a){r||(r=!0,n.packet({type:I.ACK,id:t,data:a}))}}onack(t){const n=this.acks[t.id];typeof n=="function"&&(delete this.acks[t.id],n.withError&&t.data.unshift(null),n.apply(this,t.data))}onconnect(t,n){this.id=t,this.recovered=n&&this._pid===n,this._pid=n,this.connected=!0,this.emitBuffered(),this.emitReserved("connect"),this._drainQueue(!0)}emitBuffered(){this.receiveBuffer.forEach(t=>this.emitEvent(t)),this.receiveBuffer=[],this.sendBuffer.forEach(t=>{this.notifyOutgoingListeners(t),this.packet(t)}),this.sendBuffer=[]}ondisconnect(){this.destroy(),this.onclose("io server disconnect")}destroy(){this.subs&&(this.subs.forEach(t=>t()),this.subs=void 0),this.io._destroy(this)}disconnect(){return this.connected&&this.packet({type:I.DISCONNECT}),this.destroy(),this.connected&&this.onclose("io client disconnect"),this}close(){return this.disconnect()}compress(t){return this.flags.compress=t,this}get volatile(){return this.flags.volatile=!0,this}timeout(t){return this.flags.timeout=t,this}onAny(t){return this._anyListeners=this._anyListeners||[],this._anyListeners.push(t),this}prependAny(t){return this._anyListeners=this._anyListeners||[],this._anyListeners.unshift(t),this}offAny(t){if(!this._anyListeners)return this;if(t){const n=this._anyListeners;for(let r=0;r<n.length;r++)if(t===n[r])return n.splice(r,1),this}else this._anyListeners=[];return this}listenersAny(){return this._anyListeners||[]}onAnyOutgoing(t){return this._anyOutgoingListeners=this._anyOutgoingListeners||[],this._anyOutgoingListeners.push(t),this}prependAnyOutgoing(t){return this._anyOutgoingListeners=this._anyOutgoingListeners||[],this._anyOutgoingListeners.unshift(t),this}offAnyOutgoing(t){if(!this._anyOutgoingListeners)return this;if(t){const n=this._anyOutgoingListeners;for(let r=0;r<n.length;r++)if(t===n[r])return n.splice(r,1),this}else this._anyOutgoingListeners=[];return this}listenersAnyOutgoing(){return this._anyOutgoingListeners||[]}notifyOutgoingListeners(t){if(this._anyOutgoingListeners&&this._anyOutgoingListeners.length){const n=this._anyOutgoingListeners.slice();for(const r of n)r.apply(this,t.data)}}}function el(e){e=e||{},this.ms=e.min||100,this.max=e.max||1e4,this.factor=e.factor||2,this.jitter=e.jitter>0&&e.jitter<=1?e.jitter:0,this.attempts=0}el.prototype.duration=function(){var e=this.ms*Math.pow(this.factor,this.attempts++);if(this.jitter){var t=Math.random(),n=Math.floor(t*this.jitter*e);e=Math.floor(t*10)&1?e+n:e-n}return Math.min(e,this.max)|0};el.prototype.reset=function(){this.attempts=0};el.prototype.setMin=function(e){this.ms=e};el.prototype.setMax=function(e){this.max=e};el.prototype.setJitter=function(e){this.jitter=e};class Tp extends qe{constructor(t,n){var r;super(),this.nsps={},this.subs=[],t&&typeof t=="object"&&(n=t,t=void 0),n=n||{},n.path=n.path||"/socket.io",this.opts=n,Mc(this,n),this.reconnection(n.reconnection!==!1),this.reconnectionAttempts(n.reconnectionAttempts||1/0),this.reconnectionDelay(n.reconnectionDelay||1e3),this.reconnectionDelayMax(n.reconnectionDelayMax||5e3),this.randomizationFactor((r=n.randomizationFactor)!==null&&r!==void 0?r:.5),this.backoff=new el({min:this.reconnectionDelay(),max:this.reconnectionDelayMax(),jitter:this.randomizationFactor()}),this.timeout(n.timeout==null?2e4:n.timeout),this._readyState="closed",this.uri=t;const a=n.parser||PC;this.encoder=new a.Encoder,this.decoder=new a.Decoder,this._autoConnect=n.autoConnect!==!1,this._autoConnect&&this.open()}reconnection(t){return arguments.length?(this._reconnection=!!t,t||(this.skipReconnect=!0),this):this._reconnection}reconnectionAttempts(t){return t===void 0?this._reconnectionAttempts:(this._reconnectionAttempts=t,this)}reconnectionDelay(t){var n;return t===void 0?this._reconnectionDelay:(this._reconnectionDelay=t,(n=this.backoff)===null||n===void 0||n.setMin(t),this)}randomizationFactor(t){var n;return t===void 0?this._randomizationFactor:(this._randomizationFactor=t,(n=this.backoff)===null||n===void 0||n.setJitter(t),this)}reconnectionDelayMax(t){var n;return t===void 0?this._reconnectionDelayMax:(this._reconnectionDelayMax=t,(n=this.backoff)===null||n===void 0||n.setMax(t),this)}timeout(t){return arguments.length?(this._timeout=t,this):this._timeout}maybeReconnectOnOpen(){!this._reconnecting&&this._reconnection&&this.backoff.attempts===0&&this.reconnect()}open(t){if(~this._readyState.indexOf("open"))return this;this.engine=new DC(this.uri,this.opts);const n=this.engine,r=this;this._readyState="opening",this.skipReconnect=!1;const a=pn(n,"open",function(){r.onopen(),t&&t()}),i=o=>{this.cleanup(),this._readyState="closed",this.emitReserved("error",o),t?t(o):this.maybeReconnectOnOpen()},l=pn(n,"error",i);if(this._timeout!==!1){const o=this._timeout,s=this.setTimeoutFn(()=>{a(),i(new Error("timeout")),n.close()},o);this.opts.autoUnref&&s.unref(),this.subs.push(()=>{this.clearTimeoutFn(s)})}return this.subs.push(a),this.subs.push(l),this}connect(t){return this.open(t)}onopen(){this.cleanup(),this._readyState="open",this.emitReserved("open");const t=this.engine;this.subs.push(pn(t,"ping",this.onping.bind(this)),pn(t,"data",this.ondata.bind(this)),pn(t,"error",this.onerror.bind(this)),pn(t,"close",this.onclose.bind(this)),pn(this.decoder,"decoded",this.ondecoded.bind(this)))}onping(){this.emitReserved("ping")}ondata(t){try{this.decoder.add(t)}catch(n){this.onclose("parse error",n)}}ondecoded(t){Nc(()=>{this.emitReserved("packet",t)},this.setTimeoutFn)}onerror(t){this.emitReserved("error",t)}socket(t,n){let r=this.nsps[t];return r?this._autoConnect&&!r.active&&r.connect():(r=new vE(this,t,n),this.nsps[t]=r),r}_destroy(t){const n=Object.keys(this.nsps);for(const r of n)if(this.nsps[r].active)return;this._close()}_packet(t){const n=this.encoder.encode(t);for(let r=0;r<n.length;r++)this.engine.write(n[r],t.options)}cleanup(){this.subs.forEach(t=>t()),this.subs.length=0,this.decoder.destroy()}_close(){this.skipReconnect=!0,this._reconnecting=!1,this.onclose("forced close")}disconnect(){return this._close()}onclose(t,n){var r;this.cleanup(),(r=this.engine)===null||r===void 0||r.close(),this.backoff.reset(),this._readyState="closed",this.emitReserved("close",t,n),this._reconnection&&!this.skipReconnect&&this.reconnect()}reconnect(){if(this._reconnecting||this.skipReconnect)return this;const t=this;if(this.backoff.attempts>=this._reconnectionAttempts)this.backoff.reset(),this.emitReserved("reconnect_failed"),this._reconnecting=!1;else{const n=this.backoff.duration();this._reconnecting=!0;const r=this.setTimeoutFn(()=>{t.skipReconnect||(this.emitReserved("reconnect_attempt",t.backoff.attempts),!t.skipReconnect&&t.open(a=>{a?(t._reconnecting=!1,t.reconnect(),this.emitReserved("reconnect_error",a)):t.onreconnect()}))},n);this.opts.autoUnref&&r.unref(),this.subs.push(()=>{this.clearTimeoutFn(r)})}}onreconnect(){const t=this.backoff.attempts;this._reconnecting=!1,this.backoff.reset(),this.emitReserved("reconnect",t)}}const yl={};function Ps(e,t){typeof e=="object"&&(t=e,e=void 0),t=t||{};const n=BC(e,t.path||"/socket.io"),r=n.source,a=n.id,i=n.path,l=yl[a]&&i in yl[a].nsps,o=t.forceNew||t["force new connection"]||t.multiplex===!1||l;let s;return o?s=new Tp(r,t):(yl[a]||(yl[a]=new Tp(r,t)),s=yl[a]),n.query&&!t.query&&(t.query=n.queryKey),s.socket(n.path,t)}Object.assign(Ps,{Manager:Tp,Socket:vE,io:Ps,connect:Ps});const gE=(e,t)=>{if(typeof e=="number")return{full_access_path:t,doc:null,readonly:!1,type:Number.isInteger(e)?"int":"float",value:e};if(typeof e=="boolean")return{full_access_path:t,doc:null,readonly:!1,type:"bool",value:e};if(typeof e=="string")return{full_access_path:t,doc:null,readonly:!1,type:"str",value:e};if(e===null)return{full_access_path:t,doc:null,readonly:!1,type:"None",value:null};throw new Error("Unsupported type for serialization")},QC=(e,t="")=>{const r=e.map((a,i)=>{(typeof a=="number"||typeof a=="boolean"||typeof a=="string"||a===null)&&gE(a,`${t}[${i}]`)});return{full_access_path:t,type:"list",value:r,readonly:!1,doc:null}},KC=(e,t="")=>{const r=Object.entries(e).reduce((a,[i,l])=>{const o=`${t}["${i}"]`;return(typeof l=="number"||typeof l=="boolean"||typeof l=="string"||l===null)&&(a[i]=gE(l,o)),a},{});return{full_access_path:t,type:"dict",value:r,readonly:!1,doc:null}},bE=window.location.hostname,SE=window.location.port,EE=window.__FORWARDED_PREFIX__||"",Ml=window.__FORWARDED_PROTO__||"http",ms=`${bE}:${SE}${EE}`,ZC=Ml==="http"?"ws":"wss",WC=`${ZC}://${bE}:${SE}/`,er=Ps(WC,{path:`${EE}/ws/socket.io`,transports:["websocket"]}),JC=(e,t)=>{t?er.emit("update_value",{access_path:e.full_access_path,value:e},t):er.emit("update_value",{access_path:e.full_access_path,value:e})},xE=(e,t=[],n={},r)=>{const a=QC(t),i=KC(n);er.emit("trigger_method",{access_path:e,args:a,kwargs:i})},OE=ne.memo(e=>{const{showNotification:t,notifications:n,removeNotificationById:r}=e;return g.jsx(Sy,{className:"navbarOffset toastContainer",position:"top-end",children:n.map(a=>a.levelname==="ERROR"||a.levelname==="CRITICAL"||t&&["WARNING","INFO","DEBUG"].includes(a.levelname)?g.jsxs(Gl,{className:a.levelname.toLowerCase()+"Toast",onClose:()=>r(a.id),onClick:()=>r(a.id),onMouseLeave:()=>{a.levelname!=="ERROR"&&r(a.id)},show:!0,autohide:a.levelname==="WARNING"||a.levelname==="INFO"||a.levelname==="DEBUG",delay:a.levelname==="WARNING"||a.levelname==="INFO"||a.levelname==="DEBUG"?2e3:void 0,children:[g.jsxs(Gl.Header,{closeButton:!1,className:a.levelname.toLowerCase()+"Toast text-right",children:[g.jsx("strong",{className:"me-auto",children:a.levelname}),g.jsx("small",{children:a.timeStamp})]}),g.jsx(Gl.Body,{children:a.message})]},a.id):null)})});OE.displayName="Notifications";const wp=ne.memo(({connectionStatus:e})=>{const[t,n]=m.useState(!0);m.useEffect(()=>{n(!0)},[e]);const r=()=>n(!1),a=()=>{switch(e){case"connecting":return{message:"Connecting...",bg:"info",delay:void 0};case"connected":return{message:"Connected",bg:"success",delay:1e3};case"disconnected":return{message:"Disconnected",bg:"danger",delay:void 0};case"reconnecting":return{message:"Reconnecting...",bg:"info",delay:void 0};default:return{message:"",bg:"info",delay:void 0}}},{message:i,bg:l,delay:o}=a();return g.jsx(Sy,{position:"bottom-center",className:"toastContainer",children:g.jsx(Gl,{show:t,onClose:r,delay:o,autohide:o!==void 0,bg:l,children:g.jsxs(Gl.Body,{className:"d-flex justify-content-between",children:[i,g.jsx(Lo,{variant:"close",size:"sm",onClick:r})]})})})});wp.displayName="ConnectionToast";function _E(e){const t=/\w+|\[\d+\.\d+\]|\[\d+\]|\["[^"]*"\]|\['[^']*'\]/g;return e.match(t)??[]}function IC(e){if(e.startsWith("[")&&e.endsWith("]")&&(e=e.slice(1,-1)),e.startsWith("'")&&e.endsWith("'")||e.startsWith('"')&&e.endsWith('"'))return e.slice(1,-1);const t=parseFloat(e);return isNaN(t)?e:t}function e5(e,t,n){if(t in e)return e[t];if(Array.isArray(e)){if(n&&t===e.length)return e.push(Wv()),e[t];throw new Error(`Index out of bounds: ${t}`)}else{if(n)return e[t]=Wv(),e[t];throw new Error(`Key not found: ${t}`)}}function Zv(e,t,n=!1){const r=IC(t);try{return e5(e,r,n)}catch(a){throw a instanceof RangeError?new Error(`Index '${r}': ${a.message}`):a instanceof Error?new Error(`Key '${r}': ${a.message}`):a}}function t5(e,t,n){const r=_E(t),a=JSON.parse(JSON.stringify(e));let i=a;try{for(let s=0;s<r.length-1;s++){const u=r[s];i=Zv(i,u,!1).value}const l=r[r.length-1],o=Zv(i,l,!0);return Object.assign(o,n),a}catch{}return{}}function Wv(){return{full_access_path:"",value:null,type:"None",doc:null,readonly:!1}}const Ay=m.createContext({}),Sn=ne.memo(e=>{const{docString:t}=e;if(!t)return null;const n=g.jsx(WS,{id:"tooltip",children:t});return g.jsx(WR,{placement:"bottom",overlay:n,children:g.jsx(K1,{pill:!0,className:"tooltip-trigger",bg:"light",text:"dark",children:"?"})})});Sn.displayName="DocStringComponent";function Hn(){const e=m.useRef(0);return m.useEffect(()=>{e.current+=1}),e.current}const TE=ne.memo(e=>{const{value:t,fullAccessPath:n,readOnly:r,docString:a,addNotification:i,changeCallback:l=()=>{},displayName:o,id:s}=e;Hn(),m.useEffect(()=>{i(`${n} changed to ${t}.`)},[e.value]);const u=c=>{l({type:"bool",value:c,full_access_path:n,readonly:r,doc:a})};return g.jsxs("div",{className:"component buttonComponent",id:s,children:[!1,g.jsxs(Ey,{id:`toggle-check-${s}`,type:"checkbox",variant:t?"success":"secondary",checked:t,value:o,disabled:r,onChange:c=>u(c.currentTarget.checked),children:[o,g.jsx(Sn,{docString:a})]})]})});TE.displayName="ButtonComponent";const n5=(e,t,n)=>{const r=t.split("."),a=r[0].length,i=r[1]?r[1].length:0,l=n>a;let o=0;l?o=Math.pow(10,a+1-n):o=Math.pow(10,a-n);const u=(parseFloat(t)+(e==="ArrowUp"?o:-o)).toFixed(i),c=u.split(".")[0].length;return c>a?n+=1:c<a&&(n-=1),{value:u,selectionStart:n}},r5=(e,t,n)=>n>t?{value:e.slice(0,t)+e.slice(n),selectionStart:t}:t>0?{value:e.slice(0,t-1)+e.slice(t),selectionStart:t-1}:{value:e,selectionStart:t},a5=(e,t,n)=>n>t?{value:e.slice(0,t)+e.slice(n),selectionStart:t}:t<e.length?{value:e.slice(0,t)+e.slice(t+1),selectionStart:t}:{value:e,selectionStart:t},Jv=(e,t,n,r)=>{let a=t;return e==="."&&t.includes(".")?{value:t,selectionStart:n}:e==="-"?(n===0&&r>n?(a="-"+t.slice(r),n=1):n===0&&!t.startsWith("-")?(a="-"+t,n=1):(n===0||n===1)&&t.startsWith("-")&&(a=t.slice(1),n=0),{value:a,selectionStart:n}):(r>n?a=t.slice(0,n)+e+t.slice(r):a=t.slice(0,n)+e+t.slice(n),{value:a,selectionStart:n+1})},i5=(e,t)=>Math.max(0,e-t),l5=(e,t,n)=>Math.min(n,e+t),Ru=ne.memo(e=>{const{fullAccessPath:t,value:n,readOnly:r,type:a,docString:i,isInstantUpdate:l,unit:o,addNotification:s,changeCallback:u=()=>{},displayName:c,id:d}=e,f=m.useRef(null),[h,b]=m.useState(n.toString());Hn();const S=y=>{const{key:p,target:v}=y,E=v,{value:_}=E,O=_.length,T=E.selectionEnd??0;let w=E.selectionStart??0;if(p==="F1"||p==="F5"||p==="F12"||p==="Tab")return;if(p==="ArrowLeft"||p==="ArrowRight"){if(T>w&&!y.shiftKey){const N=p==="ArrowLeft"?w:T;f.current=N}else{const N=p==="ArrowLeft"?i5(w,1):l5(T,1,O);f.current=N}return}y.preventDefault();let $=_;if(y.ctrlKey&&p==="a"){E.setSelectionRange(0,_.length);return}else if(p>="0"&&p<="9"||p==="-")({value:$,selectionStart:w}=Jv(p,_,w,T));else if(p==="."&&(a==="float"||a==="Quantity"))({value:$,selectionStart:w}=Jv(p,_,w,T));else if(p==="ArrowUp"||p==="ArrowDown")({value:$,selectionStart:w}=n5(p,_,w));else if(p==="Backspace")({value:$,selectionStart:w}=r5(_,w,T));else if(p==="Delete")({value:$,selectionStart:w}=a5(_,w,T));else if(p==="Enter"&&!l){let R;a==="Quantity"?R={type:"Quantity",value:{magnitude:Number($),unit:o},full_access_path:t,readonly:r,doc:i}:R={type:a,value:Number($),full_access_path:t,readonly:r,doc:i},u(R);return}else return;if(l){let R;a==="Quantity"?R={type:"Quantity",value:{magnitude:Number($),unit:o},full_access_path:t,readonly:r,doc:i}:R={type:a,value:Number($),full_access_path:t,readonly:r,doc:i},u(R)}b($),f.current=w},x=()=>{if(!l){let y;a==="Quantity"?y={type:"Quantity",value:{magnitude:Number(h),unit:o},full_access_path:t,readonly:r,doc:i}:y={type:a,value:Number(h),full_access_path:t,readonly:r,doc:i},u(y)}};return m.useEffect(()=>{const y=a==="int"?parseInt(h):parseFloat(h);n!==y&&b(n.toString());let p=`${t} changed to ${e.value}`;o===void 0?p+=".":p+=` ${o}.`,s(p)},[n]),m.useEffect(()=>{const y=document.getElementsByName(d)[0];y&&f.current!==null&&y.setSelectionRange(f.current,f.current)}),g.jsxs("div",{className:"component numberComponent",id:d,children:[!1,g.jsxs(dr,{children:[c&&g.jsxs(dr.Text,{children:[c,g.jsx(Sn,{docString:i})]}),g.jsx(pt.Control,{type:"text",value:h,disabled:r,onChange:()=>{},name:d,onKeyDown:S,onBlur:x,className:l&&!r?"instantUpdate":""}),o&&g.jsx(dr.Text,{children:o})]})]})});Ru.displayName="NumberComponent";const ho={black:"#000",white:"#fff"},Pa={50:"#ffebee",100:"#ffcdd2",200:"#ef9a9a",300:"#e57373",400:"#ef5350",500:"#f44336",600:"#e53935",700:"#d32f2f",800:"#c62828",900:"#b71c1c",A100:"#ff8a80",A200:"#ff5252",A400:"#ff1744",A700:"#d50000"},Fa={50:"#f3e5f5",100:"#e1bee7",200:"#ce93d8",300:"#ba68c8",400:"#ab47bc",500:"#9c27b0",600:"#8e24aa",700:"#7b1fa2",800:"#6a1b9a",900:"#4a148c",A100:"#ea80fc",A200:"#e040fb",A400:"#d500f9",A700:"#aa00ff"},Qa={50:"#e3f2fd",100:"#bbdefb",200:"#90caf9",300:"#64b5f6",400:"#42a5f5",500:"#2196f3",600:"#1e88e5",700:"#1976d2",800:"#1565c0",900:"#0d47a1",A100:"#82b1ff",A200:"#448aff",A400:"#2979ff",A700:"#2962ff"},Ka={50:"#e1f5fe",100:"#b3e5fc",200:"#81d4fa",300:"#4fc3f7",400:"#29b6f6",500:"#03a9f4",600:"#039be5",700:"#0288d1",800:"#0277bd",900:"#01579b",A100:"#80d8ff",A200:"#40c4ff",A400:"#00b0ff",A700:"#0091ea"},Za={50:"#e8f5e9",100:"#c8e6c9",200:"#a5d6a7",300:"#81c784",400:"#66bb6a",500:"#4caf50",600:"#43a047",700:"#388e3c",800:"#2e7d32",900:"#1b5e20",A100:"#b9f6ca",A200:"#69f0ae",A400:"#00e676",A700:"#00c853"},ml={50:"#fff3e0",100:"#ffe0b2",200:"#ffcc80",300:"#ffb74d",400:"#ffa726",500:"#ff9800",600:"#fb8c00",700:"#f57c00",800:"#ef6c00",900:"#e65100",A100:"#ffd180",A200:"#ffab40",A400:"#ff9100",A700:"#ff6d00"},o5={50:"#fafafa",100:"#f5f5f5",200:"#eeeeee",300:"#e0e0e0",400:"#bdbdbd",500:"#9e9e9e",600:"#757575",700:"#616161",800:"#424242",900:"#212121",A100:"#f5f5f5",A200:"#eeeeee",A400:"#bdbdbd",A700:"#616161"};function yo(e){let t="https://mui.com/production-error/?code="+e;for(let n=1;n<arguments.length;n+=1)t+="&args[]="+encodeURIComponent(arguments[n]);return"Minified MUI error #"+e+"; visit "+t+" for the full message."}const s5=Object.freeze(Object.defineProperty({__proto__:null,default:yo},Symbol.toStringTag,{value:"Module"})),u5="$$material";function c5(e){if(e.sheet)return e.sheet;for(var t=0;t<document.styleSheets.length;t++)if(document.styleSheets[t].ownerNode===e)return document.styleSheets[t]}function f5(e){var t=document.createElement("style");return t.setAttribute("data-emotion",e.key),e.nonce!==void 0&&t.setAttribute("nonce",e.nonce),t.appendChild(document.createTextNode("")),t.setAttribute("data-s",""),t}var d5=function(){function e(n){var r=this;this._insertTag=function(a){var i;r.tags.length===0?r.insertionPoint?i=r.insertionPoint.nextSibling:r.prepend?i=r.container.firstChild:i=r.before:i=r.tags[r.tags.length-1].nextSibling,r.container.insertBefore(a,i),r.tags.push(a)},this.isSpeedy=n.speedy===void 0?!0:n.speedy,this.tags=[],this.ctr=0,this.nonce=n.nonce,this.key=n.key,this.container=n.container,this.prepend=n.prepend,this.insertionPoint=n.insertionPoint,this.before=null}var t=e.prototype;return t.hydrate=function(r){r.forEach(this._insertTag)},t.insert=function(r){this.ctr%(this.isSpeedy?65e3:1)===0&&this._insertTag(f5(this));var a=this.tags[this.tags.length-1];if(this.isSpeedy){var i=c5(a);try{i.insertRule(r,i.cssRules.length)}catch{}}else a.appendChild(document.createTextNode(r));this.ctr++},t.flush=function(){this.tags.forEach(function(r){var a;return(a=r.parentNode)==null?void 0:a.removeChild(r)}),this.tags=[],this.ctr=0},e}(),ot="-ms-",Cu="-moz-",oe="-webkit-",wE="comm",Ry="rule",Cy="decl",p5="@import",AE="@keyframes",h5="@layer",y5=Math.abs,jc=String.fromCharCode,m5=Object.assign;function v5(e,t){return nt(e,0)^45?(((t<<2^nt(e,0))<<2^nt(e,1))<<2^nt(e,2))<<2^nt(e,3):0}function RE(e){return e.trim()}function g5(e,t){return(e=t.exec(e))?e[0]:e}function se(e,t,n){return e.replace(t,n)}function Ap(e,t){return e.indexOf(t)}function nt(e,t){return e.charCodeAt(t)|0}function mo(e,t,n){return e.slice(t,n)}function _n(e){return e.length}function $y(e){return e.length}function vs(e,t){return t.push(e),e}function b5(e,t){return e.map(t).join("")}var Dc=1,Li=1,CE=0,Et=0,He=0,tl="";function Bc(e,t,n,r,a,i,l){return{value:e,root:t,parent:n,type:r,props:a,children:i,line:Dc,column:Li,length:l,return:""}}function vl(e,t){return m5(Bc("",null,null,"",null,null,0),e,{length:-e.length},t)}function S5(){return He}function E5(){return He=Et>0?nt(tl,--Et):0,Li--,He===10&&(Li=1,Dc--),He}function $t(){return He=Et<CE?nt(tl,Et++):0,Li++,He===10&&(Li=1,Dc++),He}function jn(){return nt(tl,Et)}function Fs(){return Et}function Yo(e,t){return mo(tl,e,t)}function vo(e){switch(e){case 0:case 9:case 10:case 13:case 32:return 5;case 33:case 43:case 44:case 47:case 62:case 64:case 126:case 59:case 123:case 125:return 4;case 58:return 3;case 34:case 39:case 40:case 91:return 2;case 41:case 93:return 1}return 0}function $E(e){return Dc=Li=1,CE=_n(tl=e),Et=0,[]}function NE(e){return tl="",e}function Qs(e){return RE(Yo(Et-1,Rp(e===91?e+2:e===40?e+1:e)))}function x5(e){for(;(He=jn())&&He<33;)$t();return vo(e)>2||vo(He)>3?"":" "}function O5(e,t){for(;--t&&$t()&&!(He<48||He>102||He>57&&He<65||He>70&&He<97););return Yo(e,Fs()+(t<6&&jn()==32&&$t()==32))}function Rp(e){for(;$t();)switch(He){case e:return Et;case 34:case 39:e!==34&&e!==39&&Rp(He);break;case 40:e===41&&Rp(e);break;case 92:$t();break}return Et}function _5(e,t){for(;$t()&&e+He!==57;)if(e+He===84&&jn()===47)break;return"/*"+Yo(t,Et-1)+"*"+jc(e===47?e:$t())}function T5(e){for(;!vo(jn());)$t();return Yo(e,Et)}function w5(e){return NE(Ks("",null,null,null,[""],e=$E(e),0,[0],e))}function Ks(e,t,n,r,a,i,l,o,s){for(var u=0,c=0,d=l,f=0,h=0,b=0,S=1,x=1,y=1,p=0,v="",E=a,_=i,O=r,T=v;x;)switch(b=p,p=$t()){case 40:if(b!=108&&nt(T,d-1)==58){Ap(T+=se(Qs(p),"&","&\f"),"&\f")!=-1&&(y=-1);break}case 34:case 39:case 91:T+=Qs(p);break;case 9:case 10:case 13:case 32:T+=x5(b);break;case 92:T+=O5(Fs()-1,7);continue;case 47:switch(jn()){case 42:case 47:vs(A5(_5($t(),Fs()),t,n),s);break;default:T+="/"}break;case 123*S:o[u++]=_n(T)*y;case 125*S:case 59:case 0:switch(p){case 0:case 125:x=0;case 59+c:y==-1&&(T=se(T,/\f/g,"")),h>0&&_n(T)-d&&vs(h>32?eg(T+";",r,n,d-1):eg(se(T," ","")+";",r,n,d-2),s);break;case 59:T+=";";default:if(vs(O=Iv(T,t,n,u,c,a,o,v,E=[],_=[],d),i),p===123)if(c===0)Ks(T,t,O,O,E,i,d,o,_);else switch(f===99&&nt(T,3)===110?100:f){case 100:case 108:case 109:case 115:Ks(e,O,O,r&&vs(Iv(e,O,O,0,0,a,o,v,a,E=[],d),_),a,_,d,o,r?E:_);break;default:Ks(T,O,O,O,[""],_,0,o,_)}}u=c=h=0,S=y=1,v=T="",d=l;break;case 58:d=1+_n(T),h=b;default:if(S<1){if(p==123)--S;else if(p==125&&S++==0&&E5()==125)continue}switch(T+=jc(p),p*S){case 38:y=c>0?1:(T+="\f",-1);break;case 44:o[u++]=(_n(T)-1)*y,y=1;break;case 64:jn()===45&&(T+=Qs($t())),f=jn(),c=d=_n(v=T+=T5(Fs())),p++;break;case 45:b===45&&_n(T)==2&&(S=0)}}return i}function Iv(e,t,n,r,a,i,l,o,s,u,c){for(var d=a-1,f=a===0?i:[""],h=$y(f),b=0,S=0,x=0;b<r;++b)for(var y=0,p=mo(e,d+1,d=y5(S=l[b])),v=e;y<h;++y)(v=RE(S>0?f[y]+" "+p:se(p,/&\f/g,f[y])))&&(s[x++]=v);return Bc(e,t,n,a===0?Ry:o,s,u,c)}function A5(e,t,n){return Bc(e,t,n,wE,jc(S5()),mo(e,2,-2),0)}function eg(e,t,n,r){return Bc(e,t,n,Cy,mo(e,0,r),mo(e,r+1,-1),r)}function xi(e,t){for(var n="",r=$y(e),a=0;a<r;a++)n+=t(e[a],a,e,t)||"";return n}function R5(e,t,n,r){switch(e.type){case h5:if(e.children.length)break;case p5:case Cy:return e.return=e.return||e.value;case wE:return"";case AE:return e.return=e.value+"{"+xi(e.children,r)+"}";case Ry:e.value=e.props.join(",")}return _n(n=xi(e.children,r))?e.return=e.value+"{"+n+"}":""}function C5(e){var t=$y(e);return function(n,r,a,i){for(var l="",o=0;o<t;o++)l+=e[o](n,r,a,i)||"";return l}}function $5(e){return function(t){t.root||(t=t.return)&&e(t)}}function ME(e){var t=Object.create(null);return function(n){return t[n]===void 0&&(t[n]=e(n)),t[n]}}var N5=function(t,n,r){for(var a=0,i=0;a=i,i=jn(),a===38&&i===12&&(n[r]=1),!vo(i);)$t();return Yo(t,Et)},M5=function(t,n){var r=-1,a=44;do switch(vo(a)){case 0:a===38&&jn()===12&&(n[r]=1),t[r]+=N5(Et-1,n,r);break;case 2:t[r]+=Qs(a);break;case 4:if(a===44){t[++r]=jn()===58?"&\f":"",n[r]=t[r].length;break}default:t[r]+=jc(a)}while(a=$t());return t},j5=function(t,n){return NE(M5($E(t),n))},tg=new WeakMap,D5=function(t){if(!(t.type!=="rule"||!t.parent||t.length<1)){for(var n=t.value,r=t.parent,a=t.column===r.column&&t.line===r.line;r.type!=="rule";)if(r=r.parent,!r)return;if(!(t.props.length===1&&n.charCodeAt(0)!==58&&!tg.get(r))&&!a){tg.set(t,!0);for(var i=[],l=j5(n,i),o=r.props,s=0,u=0;s<l.length;s++)for(var c=0;c<o.length;c++,u++)t.props[u]=i[s]?l[s].replace(/&\f/g,o[c]):o[c]+" "+l[s]}}},B5=function(t){if(t.type==="decl"){var n=t.value;n.charCodeAt(0)===108&&n.charCodeAt(2)===98&&(t.return="",t.value="")}};function jE(e,t){switch(v5(e,t)){case 5103:return oe+"print-"+e+e;case 5737:case 4201:case 3177:case 3433:case 1641:case 4457:case 2921:case 5572:case 6356:case 5844:case 3191:case 6645:case 3005:case 6391:case 5879:case 5623:case 6135:case 4599:case 4855:case 4215:case 6389:case 5109:case 5365:case 5621:case 3829:return oe+e+e;case 5349:case 4246:case 4810:case 6968:case 2756:return oe+e+Cu+e+ot+e+e;case 6828:case 4268:return oe+e+ot+e+e;case 6165:return oe+e+ot+"flex-"+e+e;case 5187:return oe+e+se(e,/(\w+).+(:[^]+)/,oe+"box-$1$2"+ot+"flex-$1$2")+e;case 5443:return oe+e+ot+"flex-item-"+se(e,/flex-|-self/,"")+e;case 4675:return oe+e+ot+"flex-line-pack"+se(e,/align-content|flex-|-self/,"")+e;case 5548:return oe+e+ot+se(e,"shrink","negative")+e;case 5292:return oe+e+ot+se(e,"basis","preferred-size")+e;case 6060:return oe+"box-"+se(e,"-grow","")+oe+e+ot+se(e,"grow","positive")+e;case 4554:return oe+se(e,/([^-])(transform)/g,"$1"+oe+"$2")+e;case 6187:return se(se(se(e,/(zoom-|grab)/,oe+"$1"),/(image-set)/,oe+"$1"),e,"")+e;case 5495:case 3959:return se(e,/(image-set\([^]*)/,oe+"$1$`$1");case 4968:return se(se(e,/(.+:)(flex-)?(.*)/,oe+"box-pack:$3"+ot+"flex-pack:$3"),/s.+-b[^;]+/,"justify")+oe+e+e;case 4095:case 3583:case 4068:case 2532:return se(e,/(.+)-inline(.+)/,oe+"$1$2")+e;case 8116:case 7059:case 5753:case 5535:case 5445:case 5701:case 4933:case 4677:case 5533:case 5789:case 5021:case 4765:if(_n(e)-1-t>6)switch(nt(e,t+1)){case 109:if(nt(e,t+4)!==45)break;case 102:return se(e,/(.+:)(.+)-([^]+)/,"$1"+oe+"$2-$3$1"+Cu+(nt(e,t+3)==108?"$3":"$2-$3"))+e;case 115:return~Ap(e,"stretch")?jE(se(e,"stretch","fill-available"),t)+e:e}break;case 4949:if(nt(e,t+1)!==115)break;case 6444:switch(nt(e,_n(e)-3-(~Ap(e,"!important")&&10))){case 107:return se(e,":",":"+oe)+e;case 101:return se(e,/(.+:)([^;!]+)(;|!.+)?/,"$1"+oe+(nt(e,14)===45?"inline-":"")+"box$3$1"+oe+"$2$3$1"+ot+"$2box$3")+e}break;case 5936:switch(nt(e,t+11)){case 114:return oe+e+ot+se(e,/[svh]\w+-[tblr]{2}/,"tb")+e;case 108:return oe+e+ot+se(e,/[svh]\w+-[tblr]{2}/,"tb-rl")+e;case 45:return oe+e+ot+se(e,/[svh]\w+-[tblr]{2}/,"lr")+e}return oe+e+ot+e+e}return e}var k5=function(t,n,r,a){if(t.length>-1&&!t.return)switch(t.type){case Cy:t.return=jE(t.value,t.length);break;case AE:return xi([vl(t,{value:se(t.value,"@","@"+oe)})],a);case Ry:if(t.length)return b5(t.props,function(i){switch(g5(i,/(::plac\w+|:read-\w+)/)){case":read-only":case":read-write":return xi([vl(t,{props:[se(i,/:(read-\w+)/,":"+Cu+"$1")]})],a);case"::placeholder":return xi([vl(t,{props:[se(i,/:(plac\w+)/,":"+oe+"input-$1")]}),vl(t,{props:[se(i,/:(plac\w+)/,":"+Cu+"$1")]}),vl(t,{props:[se(i,/:(plac\w+)/,ot+"input-$1")]})],a)}return""})}},z5=[k5],DE=function(t){var n=t.key;if(n==="css"){var r=document.querySelectorAll("style[data-emotion]:not([data-s])");Array.prototype.forEach.call(r,function(S){var x=S.getAttribute("data-emotion");x.indexOf(" ")!==-1&&(document.head.appendChild(S),S.setAttribute("data-s",""))})}var a=t.stylisPlugins||z5,i={},l,o=[];l=t.container||document.head,Array.prototype.forEach.call(document.querySelectorAll('style[data-emotion^="'+n+' "]'),function(S){for(var x=S.getAttribute("data-emotion").split(" "),y=1;y<x.length;y++)i[x[y]]=!0;o.push(S)});var s,u=[D5,B5];{var c,d=[R5,$5(function(S){c.insert(S)})],f=C5(u.concat(a,d)),h=function(x){return xi(w5(x),f)};s=function(x,y,p,v){c=p,h(x?x+"{"+y.styles+"}":y.styles),v&&(b.inserted[y.name]=!0)}}var b={key:n,sheet:new d5({key:n,container:l,nonce:t.nonce,speedy:t.speedy,prepend:t.prepend,insertionPoint:t.insertionPoint}),nonce:t.nonce,inserted:i,registered:{},insert:s};return b.sheet.hydrate(o),b},BE={exports:{}},ce={};/** @license React v16.13.1
 * react-is.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Je=typeof Symbol=="function"&&Symbol.for,Ny=Je?Symbol.for("react.element"):60103,My=Je?Symbol.for("react.portal"):60106,kc=Je?Symbol.for("react.fragment"):60107,zc=Je?Symbol.for("react.strict_mode"):60108,Lc=Je?Symbol.for("react.profiler"):60114,Uc=Je?Symbol.for("react.provider"):60109,Hc=Je?Symbol.for("react.context"):60110,jy=Je?Symbol.for("react.async_mode"):60111,qc=Je?Symbol.for("react.concurrent_mode"):60111,Gc=Je?Symbol.for("react.forward_ref"):60112,Yc=Je?Symbol.for("react.suspense"):60113,L5=Je?Symbol.for("react.suspense_list"):60120,Vc=Je?Symbol.for("react.memo"):60115,Xc=Je?Symbol.for("react.lazy"):60116,U5=Je?Symbol.for("react.block"):60121,H5=Je?Symbol.for("react.fundamental"):60117,q5=Je?Symbol.for("react.responder"):60118,G5=Je?Symbol.for("react.scope"):60119;function kt(e){if(typeof e=="object"&&e!==null){var t=e.$$typeof;switch(t){case Ny:switch(e=e.type,e){case jy:case qc:case kc:case Lc:case zc:case Yc:return e;default:switch(e=e&&e.$$typeof,e){case Hc:case Gc:case Xc:case Vc:case Uc:return e;default:return t}}case My:return t}}}function kE(e){return kt(e)===qc}ce.AsyncMode=jy;ce.ConcurrentMode=qc;ce.ContextConsumer=Hc;ce.ContextProvider=Uc;ce.Element=Ny;ce.ForwardRef=Gc;ce.Fragment=kc;ce.Lazy=Xc;ce.Memo=Vc;ce.Portal=My;ce.Profiler=Lc;ce.StrictMode=zc;ce.Suspense=Yc;ce.isAsyncMode=function(e){return kE(e)||kt(e)===jy};ce.isConcurrentMode=kE;ce.isContextConsumer=function(e){return kt(e)===Hc};ce.isContextProvider=function(e){return kt(e)===Uc};ce.isElement=function(e){return typeof e=="object"&&e!==null&&e.$$typeof===Ny};ce.isForwardRef=function(e){return kt(e)===Gc};ce.isFragment=function(e){return kt(e)===kc};ce.isLazy=function(e){return kt(e)===Xc};ce.isMemo=function(e){return kt(e)===Vc};ce.isPortal=function(e){return kt(e)===My};ce.isProfiler=function(e){return kt(e)===Lc};ce.isStrictMode=function(e){return kt(e)===zc};ce.isSuspense=function(e){return kt(e)===Yc};ce.isValidElementType=function(e){return typeof e=="string"||typeof e=="function"||e===kc||e===qc||e===Lc||e===zc||e===Yc||e===L5||typeof e=="object"&&e!==null&&(e.$$typeof===Xc||e.$$typeof===Vc||e.$$typeof===Uc||e.$$typeof===Hc||e.$$typeof===Gc||e.$$typeof===H5||e.$$typeof===q5||e.$$typeof===G5||e.$$typeof===U5)};ce.typeOf=kt;BE.exports=ce;var Y5=BE.exports,zE=Y5,V5={$$typeof:!0,render:!0,defaultProps:!0,displayName:!0,propTypes:!0},X5={$$typeof:!0,compare:!0,defaultProps:!0,displayName:!0,propTypes:!0,type:!0},LE={};LE[zE.ForwardRef]=V5;LE[zE.Memo]=X5;var P5=!0;function UE(e,t,n){var r="";return n.split(" ").forEach(function(a){e[a]!==void 0?t.push(e[a]+";"):a&&(r+=a+" ")}),r}var Dy=function(t,n,r){var a=t.key+"-"+n.name;(r===!1||P5===!1)&&t.registered[a]===void 0&&(t.registered[a]=n.styles)},By=function(t,n,r){Dy(t,n,r);var a=t.key+"-"+n.name;if(t.inserted[n.name]===void 0){var i=n;do t.insert(n===i?"."+a:"",i,t.sheet,!0),i=i.next;while(i!==void 0)}};function F5(e){for(var t=0,n,r=0,a=e.length;a>=4;++r,a-=4)n=e.charCodeAt(r)&255|(e.charCodeAt(++r)&255)<<8|(e.charCodeAt(++r)&255)<<16|(e.charCodeAt(++r)&255)<<24,n=(n&65535)*1540483477+((n>>>16)*59797<<16),n^=n>>>24,t=(n&65535)*1540483477+((n>>>16)*59797<<16)^(t&65535)*1540483477+((t>>>16)*59797<<16);switch(a){case 3:t^=(e.charCodeAt(r+2)&255)<<16;case 2:t^=(e.charCodeAt(r+1)&255)<<8;case 1:t^=e.charCodeAt(r)&255,t=(t&65535)*1540483477+((t>>>16)*59797<<16)}return t^=t>>>13,t=(t&65535)*1540483477+((t>>>16)*59797<<16),((t^t>>>15)>>>0).toString(36)}var Q5={animationIterationCount:1,aspectRatio:1,borderImageOutset:1,borderImageSlice:1,borderImageWidth:1,boxFlex:1,boxFlexGroup:1,boxOrdinalGroup:1,columnCount:1,columns:1,flex:1,flexGrow:1,flexPositive:1,flexShrink:1,flexNegative:1,flexOrder:1,gridRow:1,gridRowEnd:1,gridRowSpan:1,gridRowStart:1,gridColumn:1,gridColumnEnd:1,gridColumnSpan:1,gridColumnStart:1,msGridRow:1,msGridRowSpan:1,msGridColumn:1,msGridColumnSpan:1,fontWeight:1,lineHeight:1,opacity:1,order:1,orphans:1,scale:1,tabSize:1,widows:1,zIndex:1,zoom:1,WebkitLineClamp:1,fillOpacity:1,floodOpacity:1,stopOpacity:1,strokeDasharray:1,strokeDashoffset:1,strokeMiterlimit:1,strokeOpacity:1,strokeWidth:1},K5=/[A-Z]|^ms/g,Z5=/_EMO_([^_]+?)_([^]*?)_EMO_/g,HE=function(t){return t.charCodeAt(1)===45},ng=function(t){return t!=null&&typeof t!="boolean"},ed=ME(function(e){return HE(e)?e:e.replace(K5,"-$&").toLowerCase()}),rg=function(t,n){switch(t){case"animation":case"animationName":if(typeof n=="string")return n.replace(Z5,function(r,a,i){return Tn={name:a,styles:i,next:Tn},a})}return Q5[t]!==1&&!HE(t)&&typeof n=="number"&&n!==0?n+"px":n};function go(e,t,n){if(n==null)return"";var r=n;if(r.__emotion_styles!==void 0)return r;switch(typeof n){case"boolean":return"";case"object":{var a=n;if(a.anim===1)return Tn={name:a.name,styles:a.styles,next:Tn},a.name;var i=n;if(i.styles!==void 0){var l=i.next;if(l!==void 0)for(;l!==void 0;)Tn={name:l.name,styles:l.styles,next:Tn},l=l.next;var o=i.styles+";";return o}return W5(e,t,n)}case"function":{if(e!==void 0){var s=Tn,u=n(e);return Tn=s,go(e,t,u)}break}}var c=n;if(t==null)return c;var d=t[c];return d!==void 0?d:c}function W5(e,t,n){var r="";if(Array.isArray(n))for(var a=0;a<n.length;a++)r+=go(e,t,n[a])+";";else for(var i in n){var l=n[i];if(typeof l!="object"){var o=l;t!=null&&t[o]!==void 0?r+=i+"{"+t[o]+"}":ng(o)&&(r+=ed(i)+":"+rg(i,o)+";")}else if(Array.isArray(l)&&typeof l[0]=="string"&&(t==null||t[l[0]]===void 0))for(var s=0;s<l.length;s++)ng(l[s])&&(r+=ed(i)+":"+rg(i,l[s])+";");else{var u=go(e,t,l);switch(i){case"animation":case"animationName":{r+=ed(i)+":"+u+";";break}default:r+=i+"{"+u+"}"}}}return r}var ag=/label:\s*([^\s;{]+)\s*(;|$)/g,Tn;function Pc(e,t,n){if(e.length===1&&typeof e[0]=="object"&&e[0]!==null&&e[0].styles!==void 0)return e[0];var r=!0,a="";Tn=void 0;var i=e[0];if(i==null||i.raw===void 0)r=!1,a+=go(n,t,i);else{var l=i;a+=l[0]}for(var o=1;o<e.length;o++)if(a+=go(n,t,e[o]),r){var s=i;a+=s[o]}ag.lastIndex=0;for(var u="",c;(c=ag.exec(a))!==null;)u+="-"+c[1];var d=F5(a)+u;return{name:d,styles:a,next:Tn}}var J5=function(t){return t()},qE=Sv.useInsertionEffect?Sv.useInsertionEffect:!1,GE=qE||J5,ig=qE||m.useLayoutEffect,YE=m.createContext(typeof HTMLElement<"u"?DE({key:"css"}):null),I5=YE.Provider,ky=function(t){return m.forwardRef(function(n,r){var a=m.useContext(YE);return t(n,a,r)})},Fc=m.createContext({}),zy={}.hasOwnProperty,Cp="__EMOTION_TYPE_PLEASE_DO_NOT_USE__",e$=function(t,n){var r={};for(var a in n)zy.call(n,a)&&(r[a]=n[a]);return r[Cp]=t,r},t$=function(t){var n=t.cache,r=t.serialized,a=t.isStringTag;return Dy(n,r,a),GE(function(){return By(n,r,a)}),null},n$=ky(function(e,t,n){var r=e.css;typeof r=="string"&&t.registered[r]!==void 0&&(r=t.registered[r]);var a=e[Cp],i=[r],l="";typeof e.className=="string"?l=UE(t.registered,i,e.className):e.className!=null&&(l=e.className+" ");var o=Pc(i,void 0,m.useContext(Fc));l+=t.key+"-"+o.name;var s={};for(var u in e)zy.call(e,u)&&u!=="css"&&u!==Cp&&(s[u]=e[u]);return s.className=l,n&&(s.ref=n),m.createElement(m.Fragment,null,m.createElement(t$,{cache:t,serialized:o,isStringTag:typeof a=="string"}),m.createElement(a,s))}),r$=n$,td={exports:{}},lg;function VE(){return lg||(lg=1,function(e){function t(){return e.exports=t=Object.assign?Object.assign.bind():function(n){for(var r=1;r<arguments.length;r++){var a=arguments[r];for(var i in a)({}).hasOwnProperty.call(a,i)&&(n[i]=a[i])}return n},e.exports.__esModule=!0,e.exports.default=e.exports,t.apply(null,arguments)}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports}(td)),td.exports}VE();var a$=function(t,n){var r=arguments;if(n==null||!zy.call(n,"css"))return m.createElement.apply(void 0,r);var a=r.length,i=new Array(a);i[0]=r$,i[1]=e$(t,n);for(var l=2;l<a;l++)i[l]=r[l];return m.createElement.apply(null,i)};(function(e){var t;t||(t=e.JSX||(e.JSX={}))})(a$);var i$=ky(function(e,t){var n=e.styles,r=Pc([n],void 0,m.useContext(Fc)),a=m.useRef();return ig(function(){var i=t.key+"-global",l=new t.sheet.constructor({key:i,nonce:t.sheet.nonce,container:t.sheet.container,speedy:t.sheet.isSpeedy}),o=!1,s=document.querySelector('style[data-emotion="'+i+" "+r.name+'"]');return t.sheet.tags.length&&(l.before=t.sheet.tags[0]),s!==null&&(o=!0,s.setAttribute("data-emotion",i),l.hydrate([s])),a.current=[l,o],function(){l.flush()}},[t]),ig(function(){var i=a.current,l=i[0],o=i[1];if(o){i[1]=!1;return}if(r.next!==void 0&&By(t,r.next,!0),l.tags.length){var s=l.tags[l.tags.length-1].nextElementSibling;l.before=s,l.flush()}t.insert("",r,l,!1)},[t,r.name]),null});function XE(){for(var e=arguments.length,t=new Array(e),n=0;n<e;n++)t[n]=arguments[n];return Pc(t)}function l$(){var e=XE.apply(void 0,arguments),t="animation-"+e.name;return{name:t,styles:"@keyframes "+t+"{"+e.styles+"}",anim:1,toString:function(){return"_EMO_"+this.name+"_"+this.styles+"_EMO_"}}}var o$=/^((children|dangerouslySetInnerHTML|key|ref|autoFocus|defaultValue|defaultChecked|innerHTML|suppressContentEditableWarning|suppressHydrationWarning|valueLink|abbr|accept|acceptCharset|accessKey|action|allow|allowUserMedia|allowPaymentRequest|allowFullScreen|allowTransparency|alt|async|autoComplete|autoPlay|capture|cellPadding|cellSpacing|challenge|charSet|checked|cite|classID|className|cols|colSpan|content|contentEditable|contextMenu|controls|controlsList|coords|crossOrigin|data|dateTime|decoding|default|defer|dir|disabled|disablePictureInPicture|disableRemotePlayback|download|draggable|encType|enterKeyHint|fetchpriority|fetchPriority|form|formAction|formEncType|formMethod|formNoValidate|formTarget|frameBorder|headers|height|hidden|high|href|hrefLang|htmlFor|httpEquiv|id|inputMode|integrity|is|keyParams|keyType|kind|label|lang|list|loading|loop|low|marginHeight|marginWidth|max|maxLength|media|mediaGroup|method|min|minLength|multiple|muted|name|nonce|noValidate|open|optimum|pattern|placeholder|playsInline|poster|preload|profile|radioGroup|readOnly|referrerPolicy|rel|required|reversed|role|rows|rowSpan|sandbox|scope|scoped|scrolling|seamless|selected|shape|size|sizes|slot|span|spellCheck|src|srcDoc|srcLang|srcSet|start|step|style|summary|tabIndex|target|title|translate|type|useMap|value|width|wmode|wrap|about|datatype|inlist|prefix|property|resource|typeof|vocab|autoCapitalize|autoCorrect|autoSave|color|incremental|fallback|inert|itemProp|itemScope|itemType|itemID|itemRef|on|option|results|security|unselectable|accentHeight|accumulate|additive|alignmentBaseline|allowReorder|alphabetic|amplitude|arabicForm|ascent|attributeName|attributeType|autoReverse|azimuth|baseFrequency|baselineShift|baseProfile|bbox|begin|bias|by|calcMode|capHeight|clip|clipPathUnits|clipPath|clipRule|colorInterpolation|colorInterpolationFilters|colorProfile|colorRendering|contentScriptType|contentStyleType|cursor|cx|cy|d|decelerate|descent|diffuseConstant|direction|display|divisor|dominantBaseline|dur|dx|dy|edgeMode|elevation|enableBackground|end|exponent|externalResourcesRequired|fill|fillOpacity|fillRule|filter|filterRes|filterUnits|floodColor|floodOpacity|focusable|fontFamily|fontSize|fontSizeAdjust|fontStretch|fontStyle|fontVariant|fontWeight|format|from|fr|fx|fy|g1|g2|glyphName|glyphOrientationHorizontal|glyphOrientationVertical|glyphRef|gradientTransform|gradientUnits|hanging|horizAdvX|horizOriginX|ideographic|imageRendering|in|in2|intercept|k|k1|k2|k3|k4|kernelMatrix|kernelUnitLength|kerning|keyPoints|keySplines|keyTimes|lengthAdjust|letterSpacing|lightingColor|limitingConeAngle|local|markerEnd|markerMid|markerStart|markerHeight|markerUnits|markerWidth|mask|maskContentUnits|maskUnits|mathematical|mode|numOctaves|offset|opacity|operator|order|orient|orientation|origin|overflow|overlinePosition|overlineThickness|panose1|paintOrder|pathLength|patternContentUnits|patternTransform|patternUnits|pointerEvents|points|pointsAtX|pointsAtY|pointsAtZ|preserveAlpha|preserveAspectRatio|primitiveUnits|r|radius|refX|refY|renderingIntent|repeatCount|repeatDur|requiredExtensions|requiredFeatures|restart|result|rotate|rx|ry|scale|seed|shapeRendering|slope|spacing|specularConstant|specularExponent|speed|spreadMethod|startOffset|stdDeviation|stemh|stemv|stitchTiles|stopColor|stopOpacity|strikethroughPosition|strikethroughThickness|string|stroke|strokeDasharray|strokeDashoffset|strokeLinecap|strokeLinejoin|strokeMiterlimit|strokeOpacity|strokeWidth|surfaceScale|systemLanguage|tableValues|targetX|targetY|textAnchor|textDecoration|textRendering|textLength|to|transform|u1|u2|underlinePosition|underlineThickness|unicode|unicodeBidi|unicodeRange|unitsPerEm|vAlphabetic|vHanging|vIdeographic|vMathematical|values|vectorEffect|version|vertAdvY|vertOriginX|vertOriginY|viewBox|viewTarget|visibility|widths|wordSpacing|writingMode|x|xHeight|x1|x2|xChannelSelector|xlinkActuate|xlinkArcrole|xlinkHref|xlinkRole|xlinkShow|xlinkTitle|xlinkType|xmlBase|xmlns|xmlnsXlink|xmlLang|xmlSpace|y|y1|y2|yChannelSelector|z|zoomAndPan|for|class|autofocus)|(([Dd][Aa][Tt][Aa]|[Aa][Rr][Ii][Aa]|x)-.*))$/,s$=ME(function(e){return o$.test(e)||e.charCodeAt(0)===111&&e.charCodeAt(1)===110&&e.charCodeAt(2)<91}),u$=s$,c$=function(t){return t!=="theme"},og=function(t){return typeof t=="string"&&t.charCodeAt(0)>96?u$:c$},sg=function(t,n,r){var a;if(n){var i=n.shouldForwardProp;a=t.__emotion_forwardProp&&i?function(l){return t.__emotion_forwardProp(l)&&i(l)}:i}return typeof a!="function"&&r&&(a=t.__emotion_forwardProp),a},f$=function(t){var n=t.cache,r=t.serialized,a=t.isStringTag;return Dy(n,r,a),GE(function(){return By(n,r,a)}),null},d$=function e(t,n){var r=t.__emotion_real===t,a=r&&t.__emotion_base||t,i,l;n!==void 0&&(i=n.label,l=n.target);var o=sg(t,n,r),s=o||og(a),u=!s("as");return function(){var c=arguments,d=r&&t.__emotion_styles!==void 0?t.__emotion_styles.slice(0):[];if(i!==void 0&&d.push("label:"+i+";"),c[0]==null||c[0].raw===void 0)d.push.apply(d,c);else{var f=c[0];d.push(f[0]);for(var h=c.length,b=1;b<h;b++)d.push(c[b],f[b])}var S=ky(function(x,y,p){var v=u&&x.as||a,E="",_=[],O=x;if(x.theme==null){O={};for(var T in x)O[T]=x[T];O.theme=m.useContext(Fc)}typeof x.className=="string"?E=UE(y.registered,_,x.className):x.className!=null&&(E=x.className+" ");var w=Pc(d.concat(_),y.registered,O);E+=y.key+"-"+w.name,l!==void 0&&(E+=" "+l);var $=u&&o===void 0?og(v):s,R={};for(var N in x)u&&N==="as"||$(N)&&(R[N]=x[N]);return R.className=E,p&&(R.ref=p),m.createElement(m.Fragment,null,m.createElement(f$,{cache:y,serialized:w,isStringTag:typeof v=="string"}),m.createElement(v,R))});return S.displayName=i!==void 0?i:"Styled("+(typeof a=="string"?a:a.displayName||a.name||"Component")+")",S.defaultProps=t.defaultProps,S.__emotion_real=S,S.__emotion_base=a,S.__emotion_styles=d,S.__emotion_forwardProp=o,Object.defineProperty(S,"toString",{value:function(){return"."+l}}),S.withComponent=function(x,y){var p=e(x,k({},n,y,{shouldForwardProp:sg(S,y,!0)}));return p.apply(void 0,d)},S}},p$=["a","abbr","address","area","article","aside","audio","b","base","bdi","bdo","big","blockquote","body","br","button","canvas","caption","cite","code","col","colgroup","data","datalist","dd","del","details","dfn","dialog","div","dl","dt","em","embed","fieldset","figcaption","figure","footer","form","h1","h2","h3","h4","h5","h6","head","header","hgroup","hr","html","i","iframe","img","input","ins","kbd","keygen","label","legend","li","link","main","map","mark","marquee","menu","menuitem","meta","meter","nav","noscript","object","ol","optgroup","option","output","p","param","picture","pre","progress","q","rp","rt","ruby","s","samp","script","section","select","small","source","span","strong","style","sub","summary","sup","table","tbody","td","textarea","tfoot","th","thead","time","title","tr","track","u","ul","var","video","wbr","circle","clipPath","defs","ellipse","foreignObject","g","image","line","linearGradient","mask","path","pattern","polygon","polyline","radialGradient","rect","stop","svg","text","tspan"],$p=d$.bind(null);p$.forEach(function(e){$p[e]=$p(e)});let Np;typeof document=="object"&&(Np=DE({key:"css",prepend:!0}));function h$(e){const{injectFirst:t,children:n}=e;return t&&Np?g.jsx(I5,{value:Np,children:n}):n}function y$(e){return e==null||Object.keys(e).length===0}function m$(e){const{styles:t,defaultTheme:n={}}=e,r=typeof t=="function"?a=>t(y$(a)?n:a):t;return g.jsx(i$,{styles:r})}function v$(e,t){return $p(e,t)}const g$=(e,t)=>{Array.isArray(e.__emotion_styles)&&(e.__emotion_styles=t(e.__emotion_styles))},b$=Object.freeze(Object.defineProperty({__proto__:null,GlobalStyles:m$,StyledEngineProvider:h$,ThemeContext:Fc,css:XE,default:v$,internal_processStyles:g$,keyframes:l$},Symbol.toStringTag,{value:"Module"}));function jr(e){if(typeof e!="object"||e===null)return!1;const t=Object.getPrototypeOf(e);return(t===null||t===Object.prototype||Object.getPrototypeOf(t)===null)&&!(Symbol.toStringTag in e)&&!(Symbol.iterator in e)}function PE(e){if(m.isValidElement(e)||!jr(e))return e;const t={};return Object.keys(e).forEach(n=>{t[n]=PE(e[n])}),t}function Dn(e,t,n={clone:!0}){const r=n.clone?k({},e):e;return jr(e)&&jr(t)&&Object.keys(t).forEach(a=>{m.isValidElement(t[a])?r[a]=t[a]:jr(t[a])&&Object.prototype.hasOwnProperty.call(e,a)&&jr(e[a])?r[a]=Dn(e[a],t[a],n):n.clone?r[a]=jr(t[a])?PE(t[a]):t[a]:r[a]=t[a]}),r}const S$=Object.freeze(Object.defineProperty({__proto__:null,default:Dn,isPlainObject:jr},Symbol.toStringTag,{value:"Module"})),E$=["values","unit","step"],x$=e=>{const t=Object.keys(e).map(n=>({key:n,val:e[n]}))||[];return t.sort((n,r)=>n.val-r.val),t.reduce((n,r)=>k({},n,{[r.key]:r.val}),{})};function FE(e){const{values:t={xs:0,sm:600,md:900,lg:1200,xl:1536},unit:n="px",step:r=5}=e,a=bn(e,E$),i=x$(t),l=Object.keys(i);function o(f){return`@media (min-width:${typeof t[f]=="number"?t[f]:f}${n})`}function s(f){return`@media (max-width:${(typeof t[f]=="number"?t[f]:f)-r/100}${n})`}function u(f,h){const b=l.indexOf(h);return`@media (min-width:${typeof t[f]=="number"?t[f]:f}${n}) and (max-width:${(b!==-1&&typeof t[l[b]]=="number"?t[l[b]]:h)-r/100}${n})`}function c(f){return l.indexOf(f)+1<l.length?u(f,l[l.indexOf(f)+1]):o(f)}function d(f){const h=l.indexOf(f);return h===0?o(l[1]):h===l.length-1?s(l[h]):u(f,l[l.indexOf(f)+1]).replace("@media","@media not all and")}return k({keys:l,values:i,up:o,down:s,between:u,only:c,not:d,unit:n},a)}const O$={borderRadius:4};function Yl(e,t){return t?Dn(e,t,{clone:!1}):e}const Ly={xs:0,sm:600,md:900,lg:1200,xl:1536},ug={keys:["xs","sm","md","lg","xl"],up:e=>`@media (min-width:${Ly[e]}px)`};function mr(e,t,n){const r=e.theme||{};if(Array.isArray(t)){const i=r.breakpoints||ug;return t.reduce((l,o,s)=>(l[i.up(i.keys[s])]=n(t[s]),l),{})}if(typeof t=="object"){const i=r.breakpoints||ug;return Object.keys(t).reduce((l,o)=>{if(Object.keys(i.values||Ly).indexOf(o)!==-1){const s=i.up(o);l[s]=n(t[o],o)}else{const s=o;l[s]=t[s]}return l},{})}return n(t)}function _$(e={}){var t;return((t=e.keys)==null?void 0:t.reduce((r,a)=>{const i=e.up(a);return r[i]={},r},{}))||{}}function T$(e,t){return e.reduce((n,r)=>{const a=n[r];return(!a||Object.keys(a).length===0)&&delete n[r],n},t)}function mn(e){if(typeof e!="string")throw new Error(yo(7));return e.charAt(0).toUpperCase()+e.slice(1)}const w$=Object.freeze(Object.defineProperty({__proto__:null,default:mn},Symbol.toStringTag,{value:"Module"}));function Qc(e,t,n=!0){if(!t||typeof t!="string")return null;if(e&&e.vars&&n){const r=`vars.${t}`.split(".").reduce((a,i)=>a&&a[i]?a[i]:null,e);if(r!=null)return r}return t.split(".").reduce((r,a)=>r&&r[a]!=null?r[a]:null,e)}function $u(e,t,n,r=n){let a;return typeof e=="function"?a=e(n):Array.isArray(e)?a=e[n]||r:a=Qc(e,n)||r,t&&(a=t(a,r,e)),a}function De(e){const{prop:t,cssProperty:n=e.prop,themeKey:r,transform:a}=e,i=l=>{if(l[t]==null)return null;const o=l[t],s=l.theme,u=Qc(s,r)||{};return mr(l,o,d=>{let f=$u(u,a,d);return d===f&&typeof d=="string"&&(f=$u(u,a,`${t}${d==="default"?"":mn(d)}`,d)),n===!1?f:{[n]:f}})};return i.propTypes={},i.filterProps=[t],i}function A$(e){const t={};return n=>(t[n]===void 0&&(t[n]=e(n)),t[n])}const R$={m:"margin",p:"padding"},C$={t:"Top",r:"Right",b:"Bottom",l:"Left",x:["Left","Right"],y:["Top","Bottom"]},cg={marginX:"mx",marginY:"my",paddingX:"px",paddingY:"py"},$$=A$(e=>{if(e.length>2)if(cg[e])e=cg[e];else return[e];const[t,n]=e.split(""),r=R$[t],a=C$[n]||"";return Array.isArray(a)?a.map(i=>r+i):[r+a]}),Uy=["m","mt","mr","mb","ml","mx","my","margin","marginTop","marginRight","marginBottom","marginLeft","marginX","marginY","marginInline","marginInlineStart","marginInlineEnd","marginBlock","marginBlockStart","marginBlockEnd"],Hy=["p","pt","pr","pb","pl","px","py","padding","paddingTop","paddingRight","paddingBottom","paddingLeft","paddingX","paddingY","paddingInline","paddingInlineStart","paddingInlineEnd","paddingBlock","paddingBlockStart","paddingBlockEnd"];[...Uy,...Hy];function Vo(e,t,n,r){var a;const i=(a=Qc(e,t,!1))!=null?a:n;return typeof i=="number"?l=>typeof l=="string"?l:i*l:Array.isArray(i)?l=>typeof l=="string"?l:i[l]:typeof i=="function"?i:()=>{}}function QE(e){return Vo(e,"spacing",8)}function Xo(e,t){if(typeof t=="string"||t==null)return t;const n=Math.abs(t),r=e(n);return t>=0?r:typeof r=="number"?-r:`-${r}`}function N$(e,t){return n=>e.reduce((r,a)=>(r[a]=Xo(t,n),r),{})}function M$(e,t,n,r){if(t.indexOf(n)===-1)return null;const a=$$(n),i=N$(a,r),l=e[n];return mr(e,l,i)}function KE(e,t){const n=QE(e.theme);return Object.keys(e).map(r=>M$(e,t,r,n)).reduce(Yl,{})}function Ce(e){return KE(e,Uy)}Ce.propTypes={};Ce.filterProps=Uy;function $e(e){return KE(e,Hy)}$e.propTypes={};$e.filterProps=Hy;function j$(e=8){if(e.mui)return e;const t=QE({spacing:e}),n=(...r)=>(r.length===0?[1]:r).map(i=>{const l=t(i);return typeof l=="number"?`${l}px`:l}).join(" ");return n.mui=!0,n}function Kc(...e){const t=e.reduce((r,a)=>(a.filterProps.forEach(i=>{r[i]=a}),r),{}),n=r=>Object.keys(r).reduce((a,i)=>t[i]?Yl(a,t[i](r)):a,{});return n.propTypes={},n.filterProps=e.reduce((r,a)=>r.concat(a.filterProps),[]),n}function Pt(e){return typeof e!="number"?e:`${e}px solid`}function on(e,t){return De({prop:e,themeKey:"borders",transform:t})}const D$=on("border",Pt),B$=on("borderTop",Pt),k$=on("borderRight",Pt),z$=on("borderBottom",Pt),L$=on("borderLeft",Pt),U$=on("borderColor"),H$=on("borderTopColor"),q$=on("borderRightColor"),G$=on("borderBottomColor"),Y$=on("borderLeftColor"),V$=on("outline",Pt),X$=on("outlineColor"),Zc=e=>{if(e.borderRadius!==void 0&&e.borderRadius!==null){const t=Vo(e.theme,"shape.borderRadius",4),n=r=>({borderRadius:Xo(t,r)});return mr(e,e.borderRadius,n)}return null};Zc.propTypes={};Zc.filterProps=["borderRadius"];Kc(D$,B$,k$,z$,L$,U$,H$,q$,G$,Y$,Zc,V$,X$);const Wc=e=>{if(e.gap!==void 0&&e.gap!==null){const t=Vo(e.theme,"spacing",8),n=r=>({gap:Xo(t,r)});return mr(e,e.gap,n)}return null};Wc.propTypes={};Wc.filterProps=["gap"];const Jc=e=>{if(e.columnGap!==void 0&&e.columnGap!==null){const t=Vo(e.theme,"spacing",8),n=r=>({columnGap:Xo(t,r)});return mr(e,e.columnGap,n)}return null};Jc.propTypes={};Jc.filterProps=["columnGap"];const Ic=e=>{if(e.rowGap!==void 0&&e.rowGap!==null){const t=Vo(e.theme,"spacing",8),n=r=>({rowGap:Xo(t,r)});return mr(e,e.rowGap,n)}return null};Ic.propTypes={};Ic.filterProps=["rowGap"];const P$=De({prop:"gridColumn"}),F$=De({prop:"gridRow"}),Q$=De({prop:"gridAutoFlow"}),K$=De({prop:"gridAutoColumns"}),Z$=De({prop:"gridAutoRows"}),W$=De({prop:"gridTemplateColumns"}),J$=De({prop:"gridTemplateRows"}),I$=De({prop:"gridTemplateAreas"}),e3=De({prop:"gridArea"});Kc(Wc,Jc,Ic,P$,F$,Q$,K$,Z$,W$,J$,I$,e3);function Oi(e,t){return t==="grey"?t:e}const t3=De({prop:"color",themeKey:"palette",transform:Oi}),n3=De({prop:"bgcolor",cssProperty:"backgroundColor",themeKey:"palette",transform:Oi}),r3=De({prop:"backgroundColor",themeKey:"palette",transform:Oi});Kc(t3,n3,r3);function Rt(e){return e<=1&&e!==0?`${e*100}%`:e}const a3=De({prop:"width",transform:Rt}),qy=e=>{if(e.maxWidth!==void 0&&e.maxWidth!==null){const t=n=>{var r,a;const i=((r=e.theme)==null||(r=r.breakpoints)==null||(r=r.values)==null?void 0:r[n])||Ly[n];return i?((a=e.theme)==null||(a=a.breakpoints)==null?void 0:a.unit)!=="px"?{maxWidth:`${i}${e.theme.breakpoints.unit}`}:{maxWidth:i}:{maxWidth:Rt(n)}};return mr(e,e.maxWidth,t)}return null};qy.filterProps=["maxWidth"];const i3=De({prop:"minWidth",transform:Rt}),l3=De({prop:"height",transform:Rt}),o3=De({prop:"maxHeight",transform:Rt}),s3=De({prop:"minHeight",transform:Rt});De({prop:"size",cssProperty:"width",transform:Rt});De({prop:"size",cssProperty:"height",transform:Rt});const u3=De({prop:"boxSizing"});Kc(a3,qy,i3,l3,o3,s3,u3);const Po={border:{themeKey:"borders",transform:Pt},borderTop:{themeKey:"borders",transform:Pt},borderRight:{themeKey:"borders",transform:Pt},borderBottom:{themeKey:"borders",transform:Pt},borderLeft:{themeKey:"borders",transform:Pt},borderColor:{themeKey:"palette"},borderTopColor:{themeKey:"palette"},borderRightColor:{themeKey:"palette"},borderBottomColor:{themeKey:"palette"},borderLeftColor:{themeKey:"palette"},outline:{themeKey:"borders",transform:Pt},outlineColor:{themeKey:"palette"},borderRadius:{themeKey:"shape.borderRadius",style:Zc},color:{themeKey:"palette",transform:Oi},bgcolor:{themeKey:"palette",cssProperty:"backgroundColor",transform:Oi},backgroundColor:{themeKey:"palette",transform:Oi},p:{style:$e},pt:{style:$e},pr:{style:$e},pb:{style:$e},pl:{style:$e},px:{style:$e},py:{style:$e},padding:{style:$e},paddingTop:{style:$e},paddingRight:{style:$e},paddingBottom:{style:$e},paddingLeft:{style:$e},paddingX:{style:$e},paddingY:{style:$e},paddingInline:{style:$e},paddingInlineStart:{style:$e},paddingInlineEnd:{style:$e},paddingBlock:{style:$e},paddingBlockStart:{style:$e},paddingBlockEnd:{style:$e},m:{style:Ce},mt:{style:Ce},mr:{style:Ce},mb:{style:Ce},ml:{style:Ce},mx:{style:Ce},my:{style:Ce},margin:{style:Ce},marginTop:{style:Ce},marginRight:{style:Ce},marginBottom:{style:Ce},marginLeft:{style:Ce},marginX:{style:Ce},marginY:{style:Ce},marginInline:{style:Ce},marginInlineStart:{style:Ce},marginInlineEnd:{style:Ce},marginBlock:{style:Ce},marginBlockStart:{style:Ce},marginBlockEnd:{style:Ce},displayPrint:{cssProperty:!1,transform:e=>({"@media print":{display:e}})},display:{},overflow:{},textOverflow:{},visibility:{},whiteSpace:{},flexBasis:{},flexDirection:{},flexWrap:{},justifyContent:{},alignItems:{},alignContent:{},order:{},flex:{},flexGrow:{},flexShrink:{},alignSelf:{},justifyItems:{},justifySelf:{},gap:{style:Wc},rowGap:{style:Ic},columnGap:{style:Jc},gridColumn:{},gridRow:{},gridAutoFlow:{},gridAutoColumns:{},gridAutoRows:{},gridTemplateColumns:{},gridTemplateRows:{},gridTemplateAreas:{},gridArea:{},position:{},zIndex:{themeKey:"zIndex"},top:{},right:{},bottom:{},left:{},boxShadow:{themeKey:"shadows"},width:{transform:Rt},maxWidth:{style:qy},minWidth:{transform:Rt},height:{transform:Rt},maxHeight:{transform:Rt},minHeight:{transform:Rt},boxSizing:{},fontFamily:{themeKey:"typography"},fontSize:{themeKey:"typography"},fontStyle:{themeKey:"typography"},fontWeight:{themeKey:"typography"},letterSpacing:{},textTransform:{},lineHeight:{},textAlign:{},typography:{cssProperty:!1,themeKey:"typography"}};function c3(...e){const t=e.reduce((r,a)=>r.concat(Object.keys(a)),[]),n=new Set(t);return e.every(r=>n.size===Object.keys(r).length)}function f3(e,t){return typeof e=="function"?e(t):e}function ZE(){function e(n,r,a,i){const l={[n]:r,theme:a},o=i[n];if(!o)return{[n]:r};const{cssProperty:s=n,themeKey:u,transform:c,style:d}=o;if(r==null)return null;if(u==="typography"&&r==="inherit")return{[n]:r};const f=Qc(a,u)||{};return d?d(l):mr(l,r,b=>{let S=$u(f,c,b);return b===S&&typeof b=="string"&&(S=$u(f,c,`${n}${b==="default"?"":mn(b)}`,b)),s===!1?S:{[s]:S}})}function t(n){var r;const{sx:a,theme:i={}}=n||{};if(!a)return null;const l=(r=i.unstable_sxConfig)!=null?r:Po;function o(s){let u=s;if(typeof s=="function")u=s(i);else if(typeof s!="object")return s;if(!u)return null;const c=_$(i.breakpoints),d=Object.keys(c);let f=c;return Object.keys(u).forEach(h=>{const b=f3(u[h],i);if(b!=null)if(typeof b=="object")if(l[h])f=Yl(f,e(h,b,i,l));else{const S=mr({theme:i},b,x=>({[h]:x}));c3(S,b)?f[h]=t({sx:b,theme:i}):f=Yl(f,S)}else f=Yl(f,e(h,b,i,l))}),T$(d,f)}return Array.isArray(a)?a.map(o):o(a)}return t}const ef=ZE();ef.filterProps=["sx"];function WE(e,t){const n=this;return n.vars&&typeof n.getColorSchemeSelector=="function"?{[n.getColorSchemeSelector(e).replace(/(\[[^\]]+\])/,"*:where($1)")]:t}:n.palette.mode===e?t:{}}const d3=["breakpoints","palette","spacing","shape"];function JE(e={},...t){const{breakpoints:n={},palette:r={},spacing:a,shape:i={}}=e,l=bn(e,d3),o=FE(n),s=j$(a);let u=Dn({breakpoints:o,direction:"ltr",components:{},palette:k({mode:"light"},r),spacing:s,shape:k({},O$,i)},l);return u.applyStyles=WE,u=t.reduce((c,d)=>Dn(c,d),u),u.unstable_sxConfig=k({},Po,l==null?void 0:l.unstable_sxConfig),u.unstable_sx=function(d){return ef({sx:d,theme:this})},u}const p3=Object.freeze(Object.defineProperty({__proto__:null,default:JE,private_createBreakpoints:FE,unstable_applyStyles:WE},Symbol.toStringTag,{value:"Module"})),h3=["sx"],y3=e=>{var t,n;const r={systemProps:{},otherProps:{}},a=(t=e==null||(n=e.theme)==null?void 0:n.unstable_sxConfig)!=null?t:Po;return Object.keys(e).forEach(i=>{a[i]?r.systemProps[i]=e[i]:r.otherProps[i]=e[i]}),r};function m3(e){const{sx:t}=e,n=bn(e,h3),{systemProps:r,otherProps:a}=y3(n);let i;return Array.isArray(t)?i=[r,...t]:typeof t=="function"?i=(...l)=>{const o=t(...l);return jr(o)?k({},r,o):r}:i=k({},r,t),k({},a,{sx:i})}const v3=Object.freeze(Object.defineProperty({__proto__:null,default:ef,extendSxProp:m3,unstable_createStyleFunctionSx:ZE,unstable_defaultSxConfig:Po},Symbol.toStringTag,{value:"Module"})),fg=e=>e,g3=()=>{let e=fg;return{configure(t){e=t},generate(t){return e(t)},reset(){e=fg}}},b3=g3();function IE(e){var t,n,r="";if(typeof e=="string"||typeof e=="number")r+=e;else if(typeof e=="object")if(Array.isArray(e)){var a=e.length;for(t=0;t<a;t++)e[t]&&(n=IE(e[t]))&&(r&&(r+=" "),r+=n)}else for(n in e)e[n]&&(r&&(r+=" "),r+=n);return r}function Gr(){for(var e,t,n=0,r="",a=arguments.length;n<a;n++)(e=arguments[n])&&(t=IE(e))&&(r&&(r+=" "),r+=t);return r}const S3={active:"active",checked:"checked",completed:"completed",disabled:"disabled",error:"error",expanded:"expanded",focused:"focused",focusVisible:"focusVisible",open:"open",readOnly:"readOnly",required:"required",selected:"selected"};function ex(e,t,n="Mui"){const r=S3[t];return r?`${n}-${r}`:`${b3.generate(e)}-${t}`}function E3(e,t,n="Mui"){const r={};return t.forEach(a=>{r[a]=ex(e,a,n)}),r}var tx={exports:{}},me={};/**
 * @license React
 * react-is.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Gy=Symbol.for("react.transitional.element"),Yy=Symbol.for("react.portal"),tf=Symbol.for("react.fragment"),nf=Symbol.for("react.strict_mode"),rf=Symbol.for("react.profiler"),af=Symbol.for("react.consumer"),lf=Symbol.for("react.context"),of=Symbol.for("react.forward_ref"),sf=Symbol.for("react.suspense"),uf=Symbol.for("react.suspense_list"),cf=Symbol.for("react.memo"),ff=Symbol.for("react.lazy"),x3=Symbol.for("react.offscreen"),O3=Symbol.for("react.client.reference");function sn(e){if(typeof e=="object"&&e!==null){var t=e.$$typeof;switch(t){case Gy:switch(e=e.type,e){case tf:case rf:case nf:case sf:case uf:return e;default:switch(e=e&&e.$$typeof,e){case lf:case of:case ff:case cf:return e;case af:return e;default:return t}}case Yy:return t}}}me.ContextConsumer=af;me.ContextProvider=lf;me.Element=Gy;me.ForwardRef=of;me.Fragment=tf;me.Lazy=ff;me.Memo=cf;me.Portal=Yy;me.Profiler=rf;me.StrictMode=nf;me.Suspense=sf;me.SuspenseList=uf;me.isContextConsumer=function(e){return sn(e)===af};me.isContextProvider=function(e){return sn(e)===lf};me.isElement=function(e){return typeof e=="object"&&e!==null&&e.$$typeof===Gy};me.isForwardRef=function(e){return sn(e)===of};me.isFragment=function(e){return sn(e)===tf};me.isLazy=function(e){return sn(e)===ff};me.isMemo=function(e){return sn(e)===cf};me.isPortal=function(e){return sn(e)===Yy};me.isProfiler=function(e){return sn(e)===rf};me.isStrictMode=function(e){return sn(e)===nf};me.isSuspense=function(e){return sn(e)===sf};me.isSuspenseList=function(e){return sn(e)===uf};me.isValidElementType=function(e){return typeof e=="string"||typeof e=="function"||e===tf||e===rf||e===nf||e===sf||e===uf||e===x3||typeof e=="object"&&e!==null&&(e.$$typeof===ff||e.$$typeof===cf||e.$$typeof===lf||e.$$typeof===af||e.$$typeof===of||e.$$typeof===O3||e.getModuleId!==void 0)};me.typeOf=sn;tx.exports=me;var dg=tx.exports;const _3=/^\s*function(?:\s|\s*\/\*.*\*\/\s*)+([^(\s/]*)\s*/;function nx(e){const t=`${e}`.match(_3);return t&&t[1]||""}function rx(e,t=""){return e.displayName||e.name||nx(e)||t}function pg(e,t,n){const r=rx(t);return e.displayName||(r!==""?`${n}(${r})`:n)}function T3(e){if(e!=null){if(typeof e=="string")return e;if(typeof e=="function")return rx(e,"Component");if(typeof e=="object")switch(e.$$typeof){case dg.ForwardRef:return pg(e,e.render,"ForwardRef");case dg.Memo:return pg(e,e.type,"memo");default:return}}}const w3=Object.freeze(Object.defineProperty({__proto__:null,default:T3,getFunctionName:nx},Symbol.toStringTag,{value:"Module"}));function Mp(e,t){const n=k({},t);return Object.keys(e).forEach(r=>{if(r.toString().match(/^(components|slots)$/))n[r]=k({},e[r],n[r]);else if(r.toString().match(/^(componentsProps|slotProps)$/)){const a=e[r]||{},i=t[r];n[r]={},!i||!Object.keys(i)?n[r]=a:!a||!Object.keys(a)?n[r]=i:(n[r]=k({},i),Object.keys(a).forEach(l=>{n[r][l]=Mp(a[l],i[l])}))}else n[r]===void 0&&(n[r]=e[r])}),n}const ax=typeof window<"u"?m.useLayoutEffect:m.useEffect;function ai(e,t=Number.MIN_SAFE_INTEGER,n=Number.MAX_SAFE_INTEGER){return Math.max(t,Math.min(e,n))}const A3=Object.freeze(Object.defineProperty({__proto__:null,default:ai},Symbol.toStringTag,{value:"Module"}));function Zs(e){return e&&e.ownerDocument||document}function R3(e,t){typeof e=="function"?e(t):e&&(e.current=t)}function C3({controlled:e,default:t,name:n,state:r="value"}){const{current:a}=m.useRef(e!==void 0),[i,l]=m.useState(t),o=a?e:i,s=m.useCallback(u=>{a||l(u)},[]);return[o,s]}function nd(e){const t=m.useRef(e);return ax(()=>{t.current=e}),m.useRef((...n)=>(0,t.current)(...n)).current}function jp(...e){return m.useMemo(()=>e.every(t=>t==null)?null:t=>{e.forEach(n=>{R3(n,t)})},e)}class Vy{constructor(){this.currentId=null,this.clear=()=>{this.currentId!==null&&(clearTimeout(this.currentId),this.currentId=null)},this.disposeEffect=()=>this.clear}static create(){return new Vy}start(t,n){this.clear(),this.currentId=setTimeout(()=>{this.currentId=null,n()},t)}}let df=!0,Dp=!1;const $3=new Vy,N3={text:!0,search:!0,url:!0,tel:!0,email:!0,password:!0,number:!0,date:!0,month:!0,week:!0,time:!0,datetime:!0,"datetime-local":!0};function M3(e){const{type:t,tagName:n}=e;return!!(n==="INPUT"&&N3[t]&&!e.readOnly||n==="TEXTAREA"&&!e.readOnly||e.isContentEditable)}function j3(e){e.metaKey||e.altKey||e.ctrlKey||(df=!0)}function rd(){df=!1}function D3(){this.visibilityState==="hidden"&&Dp&&(df=!0)}function B3(e){e.addEventListener("keydown",j3,!0),e.addEventListener("mousedown",rd,!0),e.addEventListener("pointerdown",rd,!0),e.addEventListener("touchstart",rd,!0),e.addEventListener("visibilitychange",D3,!0)}function k3(e){const{target:t}=e;try{return t.matches(":focus-visible")}catch{}return df||M3(t)}function z3(){const e=m.useCallback(a=>{a!=null&&B3(a.ownerDocument)},[]),t=m.useRef(!1);function n(){return t.current?(Dp=!0,$3.start(100,()=>{Dp=!1}),t.current=!1,!0):!1}function r(a){return k3(a)?(t.current=!0,!0):!1}return{isFocusVisibleRef:t,onFocus:r,onBlur:n,ref:e}}const L3={border:0,clip:"rect(0 0 0 0)",height:"1px",margin:"-1px",overflow:"hidden",padding:0,position:"absolute",whiteSpace:"nowrap",width:"1px"};function U3(e,t,n=void 0){const r={};return Object.keys(e).forEach(a=>{r[a]=e[a].reduce((i,l)=>{if(l){const o=t(l);o!==""&&i.push(o),n&&n[l]&&i.push(n[l])}return i},[]).join(" ")}),r}function Vl(e){return typeof e=="string"}function H3(e,t,n){return e===void 0||Vl(e)?t:k({},t,{ownerState:k({},t.ownerState,n)})}function Ws(e,t=[]){if(e===void 0)return{};const n={};return Object.keys(e).filter(r=>r.match(/^on[A-Z]/)&&typeof e[r]=="function"&&!t.includes(r)).forEach(r=>{n[r]=e[r]}),n}function hg(e){if(e===void 0)return{};const t={};return Object.keys(e).filter(n=>!(n.match(/^on[A-Z]/)&&typeof e[n]=="function")).forEach(n=>{t[n]=e[n]}),t}function q3(e){const{getSlotProps:t,additionalProps:n,externalSlotProps:r,externalForwardedProps:a,className:i}=e;if(!t){const h=Gr(n==null?void 0:n.className,i,a==null?void 0:a.className,r==null?void 0:r.className),b=k({},n==null?void 0:n.style,a==null?void 0:a.style,r==null?void 0:r.style),S=k({},n,a,r);return h.length>0&&(S.className=h),Object.keys(b).length>0&&(S.style=b),{props:S,internalRef:void 0}}const l=Ws(k({},a,r)),o=hg(r),s=hg(a),u=t(l),c=Gr(u==null?void 0:u.className,n==null?void 0:n.className,i,a==null?void 0:a.className,r==null?void 0:r.className),d=k({},u==null?void 0:u.style,n==null?void 0:n.style,a==null?void 0:a.style,r==null?void 0:r.style),f=k({},u,n,s,o);return c.length>0&&(f.className=c),Object.keys(d).length>0&&(f.style=d),{props:f,internalRef:u.ref}}function G3(e,t,n){return typeof e=="function"?e(t,n):e}const Y3=["elementType","externalSlotProps","ownerState","skipResolvingSlotProps"];function Ar(e){var t;const{elementType:n,externalSlotProps:r,ownerState:a,skipResolvingSlotProps:i=!1}=e,l=bn(e,Y3),o=i?{}:G3(r,a),{props:s,internalRef:u}=q3(k({},l,{externalSlotProps:o})),c=jp(u,o==null?void 0:o.ref,(t=e.additionalProps)==null?void 0:t.ref);return H3(n,k({},s,{ref:c}),a)}const V3=m.createContext(),X3=()=>{const e=m.useContext(V3);return e??!1},P3=m.createContext(void 0);function F3(e){const{theme:t,name:n,props:r}=e;if(!t||!t.components||!t.components[n])return r;const a=t.components[n];return a.defaultProps?Mp(a.defaultProps,r):!a.styleOverrides&&!a.variants?Mp(a,r):r}function Q3({props:e,name:t}){const n=m.useContext(P3);return F3({props:e,name:t,theme:{components:n}})}function K3(e,t){return k({toolbar:{minHeight:56,[e.up("xs")]:{"@media (orientation: landscape)":{minHeight:48}},[e.up("sm")]:{minHeight:64}}},t)}var Be={},ix={exports:{}};(function(e){function t(n){return n&&n.__esModule?n:{default:n}}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports})(ix);var lx=ix.exports;const Z3=Er(s5),W3=Er(A3);var ox=lx;Object.defineProperty(Be,"__esModule",{value:!0});var yg=Be.alpha=fx;Be.blend=sN;Be.colorChannel=void 0;var Bp=Be.darken=Py;Be.decomposeColor=rn;Be.emphasize=dx;var J3=Be.getContrastRatio=rN;Be.getLuminance=Nu;Be.hexToRgb=sx;Be.hslToRgb=cx;var kp=Be.lighten=Fy;Be.private_safeAlpha=aN;Be.private_safeColorChannel=void 0;Be.private_safeDarken=iN;Be.private_safeEmphasize=oN;Be.private_safeLighten=lN;Be.recomposeColor=nl;Be.rgbToHex=nN;var mg=ox(Z3),I3=ox(W3);function Xy(e,t=0,n=1){return(0,I3.default)(e,t,n)}function sx(e){e=e.slice(1);const t=new RegExp(`.{1,${e.length>=6?2:1}}`,"g");let n=e.match(t);return n&&n[0].length===1&&(n=n.map(r=>r+r)),n?`rgb${n.length===4?"a":""}(${n.map((r,a)=>a<3?parseInt(r,16):Math.round(parseInt(r,16)/255*1e3)/1e3).join(", ")})`:""}function eN(e){const t=e.toString(16);return t.length===1?`0${t}`:t}function rn(e){if(e.type)return e;if(e.charAt(0)==="#")return rn(sx(e));const t=e.indexOf("("),n=e.substring(0,t);if(["rgb","rgba","hsl","hsla","color"].indexOf(n)===-1)throw new Error((0,mg.default)(9,e));let r=e.substring(t+1,e.length-1),a;if(n==="color"){if(r=r.split(" "),a=r.shift(),r.length===4&&r[3].charAt(0)==="/"&&(r[3]=r[3].slice(1)),["srgb","display-p3","a98-rgb","prophoto-rgb","rec-2020"].indexOf(a)===-1)throw new Error((0,mg.default)(10,a))}else r=r.split(",");return r=r.map(i=>parseFloat(i)),{type:n,values:r,colorSpace:a}}const ux=e=>{const t=rn(e);return t.values.slice(0,3).map((n,r)=>t.type.indexOf("hsl")!==-1&&r!==0?`${n}%`:n).join(" ")};Be.colorChannel=ux;const tN=(e,t)=>{try{return ux(e)}catch{return e}};Be.private_safeColorChannel=tN;function nl(e){const{type:t,colorSpace:n}=e;let{values:r}=e;return t.indexOf("rgb")!==-1?r=r.map((a,i)=>i<3?parseInt(a,10):a):t.indexOf("hsl")!==-1&&(r[1]=`${r[1]}%`,r[2]=`${r[2]}%`),t.indexOf("color")!==-1?r=`${n} ${r.join(" ")}`:r=`${r.join(", ")}`,`${t}(${r})`}function nN(e){if(e.indexOf("#")===0)return e;const{values:t}=rn(e);return`#${t.map((n,r)=>eN(r===3?Math.round(255*n):n)).join("")}`}function cx(e){e=rn(e);const{values:t}=e,n=t[0],r=t[1]/100,a=t[2]/100,i=r*Math.min(a,1-a),l=(u,c=(u+n/30)%12)=>a-i*Math.max(Math.min(c-3,9-c,1),-1);let o="rgb";const s=[Math.round(l(0)*255),Math.round(l(8)*255),Math.round(l(4)*255)];return e.type==="hsla"&&(o+="a",s.push(t[3])),nl({type:o,values:s})}function Nu(e){e=rn(e);let t=e.type==="hsl"||e.type==="hsla"?rn(cx(e)).values:e.values;return t=t.map(n=>(e.type!=="color"&&(n/=255),n<=.03928?n/12.92:((n+.055)/1.055)**2.4)),Number((.2126*t[0]+.7152*t[1]+.0722*t[2]).toFixed(3))}function rN(e,t){const n=Nu(e),r=Nu(t);return(Math.max(n,r)+.05)/(Math.min(n,r)+.05)}function fx(e,t){return e=rn(e),t=Xy(t),(e.type==="rgb"||e.type==="hsl")&&(e.type+="a"),e.type==="color"?e.values[3]=`/${t}`:e.values[3]=t,nl(e)}function aN(e,t,n){try{return fx(e,t)}catch{return e}}function Py(e,t){if(e=rn(e),t=Xy(t),e.type.indexOf("hsl")!==-1)e.values[2]*=1-t;else if(e.type.indexOf("rgb")!==-1||e.type.indexOf("color")!==-1)for(let n=0;n<3;n+=1)e.values[n]*=1-t;return nl(e)}function iN(e,t,n){try{return Py(e,t)}catch{return e}}function Fy(e,t){if(e=rn(e),t=Xy(t),e.type.indexOf("hsl")!==-1)e.values[2]+=(100-e.values[2])*t;else if(e.type.indexOf("rgb")!==-1)for(let n=0;n<3;n+=1)e.values[n]+=(255-e.values[n])*t;else if(e.type.indexOf("color")!==-1)for(let n=0;n<3;n+=1)e.values[n]+=(1-e.values[n])*t;return nl(e)}function lN(e,t,n){try{return Fy(e,t)}catch{return e}}function dx(e,t=.15){return Nu(e)>.5?Py(e,t):Fy(e,t)}function oN(e,t,n){try{return dx(e,t)}catch{return e}}function sN(e,t,n,r=1){const a=(s,u)=>Math.round((s**(1/r)*(1-n)+u**(1/r)*n)**r),i=rn(e),l=rn(t),o=[a(i.values[0],l.values[0]),a(i.values[1],l.values[1]),a(i.values[2],l.values[2])];return nl({type:"rgb",values:o})}const uN=["mode","contrastThreshold","tonalOffset"],vg={text:{primary:"rgba(0, 0, 0, 0.87)",secondary:"rgba(0, 0, 0, 0.6)",disabled:"rgba(0, 0, 0, 0.38)"},divider:"rgba(0, 0, 0, 0.12)",background:{paper:ho.white,default:ho.white},action:{active:"rgba(0, 0, 0, 0.54)",hover:"rgba(0, 0, 0, 0.04)",hoverOpacity:.04,selected:"rgba(0, 0, 0, 0.08)",selectedOpacity:.08,disabled:"rgba(0, 0, 0, 0.26)",disabledBackground:"rgba(0, 0, 0, 0.12)",disabledOpacity:.38,focus:"rgba(0, 0, 0, 0.12)",focusOpacity:.12,activatedOpacity:.12}},ad={text:{primary:ho.white,secondary:"rgba(255, 255, 255, 0.7)",disabled:"rgba(255, 255, 255, 0.5)",icon:"rgba(255, 255, 255, 0.5)"},divider:"rgba(255, 255, 255, 0.12)",background:{paper:"#121212",default:"#121212"},action:{active:ho.white,hover:"rgba(255, 255, 255, 0.08)",hoverOpacity:.08,selected:"rgba(255, 255, 255, 0.16)",selectedOpacity:.16,disabled:"rgba(255, 255, 255, 0.3)",disabledBackground:"rgba(255, 255, 255, 0.12)",disabledOpacity:.38,focus:"rgba(255, 255, 255, 0.12)",focusOpacity:.12,activatedOpacity:.24}};function gg(e,t,n,r){const a=r.light||r,i=r.dark||r*1.5;e[t]||(e.hasOwnProperty(n)?e[t]=e[n]:t==="light"?e.light=kp(e.main,a):t==="dark"&&(e.dark=Bp(e.main,i)))}function cN(e="light"){return e==="dark"?{main:Qa[200],light:Qa[50],dark:Qa[400]}:{main:Qa[700],light:Qa[400],dark:Qa[800]}}function fN(e="light"){return e==="dark"?{main:Fa[200],light:Fa[50],dark:Fa[400]}:{main:Fa[500],light:Fa[300],dark:Fa[700]}}function dN(e="light"){return e==="dark"?{main:Pa[500],light:Pa[300],dark:Pa[700]}:{main:Pa[700],light:Pa[400],dark:Pa[800]}}function pN(e="light"){return e==="dark"?{main:Ka[400],light:Ka[300],dark:Ka[700]}:{main:Ka[700],light:Ka[500],dark:Ka[900]}}function hN(e="light"){return e==="dark"?{main:Za[400],light:Za[300],dark:Za[700]}:{main:Za[800],light:Za[500],dark:Za[900]}}function yN(e="light"){return e==="dark"?{main:ml[400],light:ml[300],dark:ml[700]}:{main:"#ed6c02",light:ml[500],dark:ml[900]}}function mN(e){const{mode:t="light",contrastThreshold:n=3,tonalOffset:r=.2}=e,a=bn(e,uN),i=e.primary||cN(t),l=e.secondary||fN(t),o=e.error||dN(t),s=e.info||pN(t),u=e.success||hN(t),c=e.warning||yN(t);function d(S){return J3(S,ad.text.primary)>=n?ad.text.primary:vg.text.primary}const f=({color:S,name:x,mainShade:y=500,lightShade:p=300,darkShade:v=700})=>{if(S=k({},S),!S.main&&S[y]&&(S.main=S[y]),!S.hasOwnProperty("main"))throw new Error(yo(11,x?` (${x})`:"",y));if(typeof S.main!="string")throw new Error(yo(12,x?` (${x})`:"",JSON.stringify(S.main)));return gg(S,"light",p,r),gg(S,"dark",v,r),S.contrastText||(S.contrastText=d(S.main)),S},h={dark:ad,light:vg};return Dn(k({common:k({},ho),mode:t,primary:f({color:i,name:"primary"}),secondary:f({color:l,name:"secondary",mainShade:"A400",lightShade:"A200",darkShade:"A700"}),error:f({color:o,name:"error"}),warning:f({color:c,name:"warning"}),info:f({color:s,name:"info"}),success:f({color:u,name:"success"}),grey:o5,contrastThreshold:n,getContrastText:d,augmentColor:f,tonalOffset:r},h[t]),a)}const vN=["fontFamily","fontSize","fontWeightLight","fontWeightRegular","fontWeightMedium","fontWeightBold","htmlFontSize","allVariants","pxToRem"];function gN(e){return Math.round(e*1e5)/1e5}const bg={textTransform:"uppercase"},Sg='"Roboto", "Helvetica", "Arial", sans-serif';function bN(e,t){const n=typeof t=="function"?t(e):t,{fontFamily:r=Sg,fontSize:a=14,fontWeightLight:i=300,fontWeightRegular:l=400,fontWeightMedium:o=500,fontWeightBold:s=700,htmlFontSize:u=16,allVariants:c,pxToRem:d}=n,f=bn(n,vN),h=a/14,b=d||(y=>`${y/u*h}rem`),S=(y,p,v,E,_)=>k({fontFamily:r,fontWeight:y,fontSize:b(p),lineHeight:v},r===Sg?{letterSpacing:`${gN(E/p)}em`}:{},_,c),x={h1:S(i,96,1.167,-1.5),h2:S(i,60,1.2,-.5),h3:S(l,48,1.167,0),h4:S(l,34,1.235,.25),h5:S(l,24,1.334,0),h6:S(o,20,1.6,.15),subtitle1:S(l,16,1.75,.15),subtitle2:S(o,14,1.57,.1),body1:S(l,16,1.5,.15),body2:S(l,14,1.43,.15),button:S(o,14,1.75,.4,bg),caption:S(l,12,1.66,.4),overline:S(l,12,2.66,1,bg),inherit:{fontFamily:"inherit",fontWeight:"inherit",fontSize:"inherit",lineHeight:"inherit",letterSpacing:"inherit"}};return Dn(k({htmlFontSize:u,pxToRem:b,fontFamily:r,fontSize:a,fontWeightLight:i,fontWeightRegular:l,fontWeightMedium:o,fontWeightBold:s},x),f,{clone:!1})}const SN=.2,EN=.14,xN=.12;function _e(...e){return[`${e[0]}px ${e[1]}px ${e[2]}px ${e[3]}px rgba(0,0,0,${SN})`,`${e[4]}px ${e[5]}px ${e[6]}px ${e[7]}px rgba(0,0,0,${EN})`,`${e[8]}px ${e[9]}px ${e[10]}px ${e[11]}px rgba(0,0,0,${xN})`].join(",")}const ON=["none",_e(0,2,1,-1,0,1,1,0,0,1,3,0),_e(0,3,1,-2,0,2,2,0,0,1,5,0),_e(0,3,3,-2,0,3,4,0,0,1,8,0),_e(0,2,4,-1,0,4,5,0,0,1,10,0),_e(0,3,5,-1,0,5,8,0,0,1,14,0),_e(0,3,5,-1,0,6,10,0,0,1,18,0),_e(0,4,5,-2,0,7,10,1,0,2,16,1),_e(0,5,5,-3,0,8,10,1,0,3,14,2),_e(0,5,6,-3,0,9,12,1,0,3,16,2),_e(0,6,6,-3,0,10,14,1,0,4,18,3),_e(0,6,7,-4,0,11,15,1,0,4,20,3),_e(0,7,8,-4,0,12,17,2,0,5,22,4),_e(0,7,8,-4,0,13,19,2,0,5,24,4),_e(0,7,9,-4,0,14,21,2,0,5,26,4),_e(0,8,9,-5,0,15,22,2,0,6,28,5),_e(0,8,10,-5,0,16,24,2,0,6,30,5),_e(0,8,11,-5,0,17,26,2,0,6,32,5),_e(0,9,11,-5,0,18,28,2,0,7,34,6),_e(0,9,12,-6,0,19,29,2,0,7,36,6),_e(0,10,13,-6,0,20,31,3,0,8,38,7),_e(0,10,13,-6,0,21,33,3,0,8,40,7),_e(0,10,14,-6,0,22,35,3,0,8,42,7),_e(0,11,14,-7,0,23,36,3,0,9,44,8),_e(0,11,15,-7,0,24,38,3,0,9,46,8)],_N=["duration","easing","delay"],TN={easeInOut:"cubic-bezier(0.4, 0, 0.2, 1)",easeOut:"cubic-bezier(0.0, 0, 0.2, 1)",easeIn:"cubic-bezier(0.4, 0, 1, 1)",sharp:"cubic-bezier(0.4, 0, 0.6, 1)"},wN={shortest:150,shorter:200,short:250,standard:300,complex:375,enteringScreen:225,leavingScreen:195};function Eg(e){return`${Math.round(e)}ms`}function AN(e){if(!e)return 0;const t=e/36;return Math.round((4+15*t**.25+t/5)*10)}function RN(e){const t=k({},TN,e.easing),n=k({},wN,e.duration);return k({getAutoHeightDuration:AN,create:(a=["all"],i={})=>{const{duration:l=n.standard,easing:o=t.easeInOut,delay:s=0}=i;return bn(i,_N),(Array.isArray(a)?a:[a]).map(u=>`${u} ${typeof l=="string"?l:Eg(l)} ${o} ${typeof s=="string"?s:Eg(s)}`).join(",")}},e,{easing:t,duration:n})}const CN={mobileStepper:1e3,fab:1050,speedDial:1050,appBar:1100,drawer:1200,modal:1300,snackbar:1400,tooltip:1500},$N=["breakpoints","mixins","spacing","palette","transitions","typography","shape"];function NN(e={},...t){const{mixins:n={},palette:r={},transitions:a={},typography:i={}}=e,l=bn(e,$N);if(e.vars)throw new Error(yo(18));const o=mN(r),s=JE(e);let u=Dn(s,{mixins:K3(s.breakpoints,n),palette:o,shadows:ON.slice(),typography:bN(o,i),transitions:RN(a),zIndex:k({},CN)});return u=Dn(u,l),u=t.reduce((c,d)=>Dn(c,d),u),u.unstable_sxConfig=k({},Po,l==null?void 0:l.unstable_sxConfig),u.unstable_sx=function(d){return ef({sx:d,theme:this})},u}const MN=NN();var Fo={},id={exports:{}},xg;function jN(){return xg||(xg=1,function(e){function t(n,r){if(n==null)return{};var a={};for(var i in n)if({}.hasOwnProperty.call(n,i)){if(r.includes(i))continue;a[i]=n[i]}return a}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports}(id)),id.exports}const DN=Er(b$),BN=Er(S$),kN=Er(w$),zN=Er(w3),LN=Er(p3),UN=Er(v3);var rl=lx;Object.defineProperty(Fo,"__esModule",{value:!0});var HN=Fo.default=IN;Fo.shouldForwardProp=Js;Fo.systemDefaultTheme=void 0;var Gt=rl(VE()),zp=rl(jN()),Og=FN(DN),qN=BN;rl(kN);rl(zN);var GN=rl(LN),YN=rl(UN);const VN=["ownerState"],XN=["variants"],PN=["name","slot","skipVariantsResolver","skipSx","overridesResolver"];function px(e){if(typeof WeakMap!="function")return null;var t=new WeakMap,n=new WeakMap;return(px=function(r){return r?n:t})(e)}function FN(e,t){if(e&&e.__esModule)return e;if(e===null||typeof e!="object"&&typeof e!="function")return{default:e};var n=px(t);if(n&&n.has(e))return n.get(e);var r={__proto__:null},a=Object.defineProperty&&Object.getOwnPropertyDescriptor;for(var i in e)if(i!=="default"&&Object.prototype.hasOwnProperty.call(e,i)){var l=a?Object.getOwnPropertyDescriptor(e,i):null;l&&(l.get||l.set)?Object.defineProperty(r,i,l):r[i]=e[i]}return r.default=e,n&&n.set(e,r),r}function QN(e){return Object.keys(e).length===0}function KN(e){return typeof e=="string"&&e.charCodeAt(0)>96}function Js(e){return e!=="ownerState"&&e!=="theme"&&e!=="sx"&&e!=="as"}const ZN=Fo.systemDefaultTheme=(0,GN.default)(),WN=e=>e&&e.charAt(0).toLowerCase()+e.slice(1);function gs({defaultTheme:e,theme:t,themeId:n}){return QN(t)?e:t[n]||t}function JN(e){return e?(t,n)=>n[e]:null}function Is(e,t){let{ownerState:n}=t,r=(0,zp.default)(t,VN);const a=typeof e=="function"?e((0,Gt.default)({ownerState:n},r)):e;if(Array.isArray(a))return a.flatMap(i=>Is(i,(0,Gt.default)({ownerState:n},r)));if(a&&typeof a=="object"&&Array.isArray(a.variants)){const{variants:i=[]}=a;let o=(0,zp.default)(a,XN);return i.forEach(s=>{let u=!0;typeof s.props=="function"?u=s.props((0,Gt.default)({ownerState:n},r,n)):Object.keys(s.props).forEach(c=>{(n==null?void 0:n[c])!==s.props[c]&&r[c]!==s.props[c]&&(u=!1)}),u&&(Array.isArray(o)||(o=[o]),o.push(typeof s.style=="function"?s.style((0,Gt.default)({ownerState:n},r,n)):s.style))}),o}return a}function IN(e={}){const{themeId:t,defaultTheme:n=ZN,rootShouldForwardProp:r=Js,slotShouldForwardProp:a=Js}=e,i=l=>(0,YN.default)((0,Gt.default)({},l,{theme:gs((0,Gt.default)({},l,{defaultTheme:n,themeId:t}))}));return i.__mui_systemSx=!0,(l,o={})=>{(0,Og.internal_processStyles)(l,_=>_.filter(O=>!(O!=null&&O.__mui_systemSx)));const{name:s,slot:u,skipVariantsResolver:c,skipSx:d,overridesResolver:f=JN(WN(u))}=o,h=(0,zp.default)(o,PN),b=c!==void 0?c:u&&u!=="Root"&&u!=="root"||!1,S=d||!1;let x,y=Js;u==="Root"||u==="root"?y=r:u?y=a:KN(l)&&(y=void 0);const p=(0,Og.default)(l,(0,Gt.default)({shouldForwardProp:y,label:x},h)),v=_=>typeof _=="function"&&_.__emotion_real!==_||(0,qN.isPlainObject)(_)?O=>Is(_,(0,Gt.default)({},O,{theme:gs({theme:O.theme,defaultTheme:n,themeId:t})})):_,E=(_,...O)=>{let T=v(_);const w=O?O.map(v):[];s&&f&&w.push(N=>{const G=gs((0,Gt.default)({},N,{defaultTheme:n,themeId:t}));if(!G.components||!G.components[s]||!G.components[s].styleOverrides)return null;const U=G.components[s].styleOverrides,X={};return Object.entries(U).forEach(([q,C])=>{X[q]=Is(C,(0,Gt.default)({},N,{theme:G}))}),f(N,X)}),s&&!b&&w.push(N=>{var G;const U=gs((0,Gt.default)({},N,{defaultTheme:n,themeId:t})),X=U==null||(G=U.components)==null||(G=G[s])==null?void 0:G.variants;return Is({variants:X},(0,Gt.default)({},N,{theme:U}))}),S||w.push(i);const $=w.length-O.length;if(Array.isArray(_)&&$>0){const N=new Array($).fill("");T=[..._,...N],T.raw=[..._.raw,...N]}const R=p(T,...w);return l.muiName&&(R.muiName=l.muiName),R};return p.withConfig&&(E.withConfig=p.withConfig),E}}function Qy(e){return e!=="ownerState"&&e!=="theme"&&e!=="sx"&&e!=="as"}const e4=e=>Qy(e)&&e!=="classes",Ca=HN({themeId:u5,defaultTheme:MN,rootShouldForwardProp:e4});function t4(e){return Q3(e)}function n4(e,t,n=(r,a)=>r===a){return e.length===t.length&&e.every((r,a)=>n(r,t[a]))}const r4=2;function hx(e,t){return e-t}function _g(e,t){var n;const{index:r}=(n=e.reduce((a,i,l)=>{const o=Math.abs(t-i);return a===null||o<a.distance||o===a.distance?{distance:o,index:l}:a},null))!=null?n:{};return r}function bs(e,t){if(t.current!==void 0&&e.changedTouches){const n=e;for(let r=0;r<n.changedTouches.length;r+=1){const a=n.changedTouches[r];if(a.identifier===t.current)return{x:a.clientX,y:a.clientY}}return!1}return{x:e.clientX,y:e.clientY}}function Mu(e,t,n){return(e-t)*100/(n-t)}function a4(e,t,n){return(n-t)*e+t}function i4(e){if(Math.abs(e)<1){const n=e.toExponential().split("e-"),r=n[0].split(".")[1];return(r?r.length:0)+parseInt(n[1],10)}const t=e.toString().split(".")[1];return t?t.length:0}function l4(e,t,n){const r=Math.round((e-n)/t)*t+n;return Number(r.toFixed(i4(t)))}function Tg({values:e,newValue:t,index:n}){const r=e.slice();return r[n]=t,r.sort(hx)}function Ss({sliderRef:e,activeIndex:t,setActive:n}){var r,a;const i=Zs(e.current);if(!((r=e.current)!=null&&r.contains(i.activeElement))||Number(i==null||(a=i.activeElement)==null?void 0:a.getAttribute("data-index"))!==t){var l;(l=e.current)==null||l.querySelector(`[type="range"][data-index="${t}"]`).focus()}n&&n(t)}function Es(e,t){return typeof e=="number"&&typeof t=="number"?e===t:typeof e=="object"&&typeof t=="object"?n4(e,t):!1}const o4={horizontal:{offset:e=>({left:`${e}%`}),leap:e=>({width:`${e}%`})},"horizontal-reverse":{offset:e=>({right:`${e}%`}),leap:e=>({width:`${e}%`})},vertical:{offset:e=>({bottom:`${e}%`}),leap:e=>({height:`${e}%`})}},s4=e=>e;let xs;function wg(){return xs===void 0&&(typeof CSS<"u"&&typeof CSS.supports=="function"?xs=CSS.supports("touch-action","none"):xs=!0),xs}function u4(e){const{"aria-labelledby":t,defaultValue:n,disabled:r=!1,disableSwap:a=!1,isRtl:i=!1,marks:l=!1,max:o=100,min:s=0,name:u,onChange:c,onChangeCommitted:d,orientation:f="horizontal",rootRef:h,scale:b=s4,step:S=1,shiftStep:x=10,tabIndex:y,value:p}=e,v=m.useRef(void 0),[E,_]=m.useState(-1),[O,T]=m.useState(-1),[w,$]=m.useState(!1),R=m.useRef(0),[N,G]=C3({controlled:p,default:n??s,name:"Slider"}),U=c&&((M,B,L)=>{const re=M.nativeEvent||M,le=new re.constructor(re.type,re);Object.defineProperty(le,"target",{writable:!0,value:{value:B,name:u}}),c(le,B,L)}),X=Array.isArray(N);let q=X?N.slice().sort(hx):[N];q=q.map(M=>M==null?s:ai(M,s,o));const C=l===!0&&S!==null?[...Array(Math.floor((o-s)/S)+1)].map((M,B)=>({value:s+S*B})):l||[],j=C.map(M=>M.value),{isFocusVisibleRef:D,onBlur:K,onFocus:ie,ref:Pe}=z3(),[ke,xe]=m.useState(-1),V=m.useRef(null),ve=jp(Pe,V),zt=jp(h,ve),_t=M=>B=>{var L;const re=Number(B.currentTarget.getAttribute("data-index"));ie(B),D.current===!0&&xe(re),T(re),M==null||(L=M.onFocus)==null||L.call(M,B)},it=M=>B=>{var L;K(B),D.current===!1&&xe(-1),T(-1),M==null||(L=M.onBlur)==null||L.call(M,B)},Lt=(M,B)=>{const L=Number(M.currentTarget.getAttribute("data-index")),re=q[L],le=j.indexOf(re);let ee=B;if(C&&S==null){const fn=j[j.length-1];ee>fn?ee=fn:ee<j[0]?ee=j[0]:ee=ee<re?j[le-1]:j[le+1]}if(ee=ai(ee,s,o),X){a&&(ee=ai(ee,q[L-1]||-1/0,q[L+1]||1/0));const fn=ee;ee=Tg({values:q,newValue:ee,index:L});let xn=L;a||(xn=ee.indexOf(fn)),Ss({sliderRef:V,activeIndex:xn})}G(ee),xe(L),U&&!Es(ee,N)&&U(M,ee,L),d&&d(M,ee)},Vn=M=>B=>{var L;if(S!==null){const re=Number(B.currentTarget.getAttribute("data-index")),le=q[re];let ee=null;(B.key==="ArrowLeft"||B.key==="ArrowDown")&&B.shiftKey||B.key==="PageDown"?ee=Math.max(le-x,s):((B.key==="ArrowRight"||B.key==="ArrowUp")&&B.shiftKey||B.key==="PageUp")&&(ee=Math.min(le+x,o)),ee!==null&&(Lt(B,ee),B.preventDefault())}M==null||(L=M.onKeyDown)==null||L.call(M,B)};ax(()=>{if(r&&V.current.contains(document.activeElement)){var M;(M=document.activeElement)==null||M.blur()}},[r]),r&&E!==-1&&_(-1),r&&ke!==-1&&xe(-1);const Xn=M=>B=>{var L;(L=M.onChange)==null||L.call(M,B),Lt(B,B.target.valueAsNumber)},Pn=m.useRef(void 0);let lt=f;i&&f==="horizontal"&&(lt+="-reverse");const fe=({finger:M,move:B=!1})=>{const{current:L}=V,{width:re,height:le,bottom:ee,left:fn}=L.getBoundingClientRect();let xn;lt.indexOf("vertical")===0?xn=(ee-M.y)/le:xn=(M.x-fn)/re,lt.indexOf("-reverse")!==-1&&(xn=1-xn);let ue;if(ue=a4(xn,s,o),S)ue=l4(ue,S,s);else{const Ga=_g(j,ue);ue=j[Ga]}ue=ai(ue,s,o);let Ut=0;if(X){B?Ut=Pn.current:Ut=_g(q,ue),a&&(ue=ai(ue,q[Ut-1]||-1/0,q[Ut+1]||1/0));const Ga=ue;ue=Tg({values:q,newValue:ue,index:Ut}),a&&B||(Ut=ue.indexOf(Ga),Pn.current=Ut)}return{newValue:ue,activeIndex:Ut}},Y=nd(M=>{const B=bs(M,v);if(!B)return;if(R.current+=1,M.type==="mousemove"&&M.buttons===0){Ie(M);return}const{newValue:L,activeIndex:re}=fe({finger:B,move:!0});Ss({sliderRef:V,activeIndex:re,setActive:_}),G(L),!w&&R.current>r4&&$(!0),U&&!Es(L,N)&&U(M,L,re)}),Ie=nd(M=>{const B=bs(M,v);if($(!1),!B)return;const{newValue:L}=fe({finger:B,move:!0});_(-1),M.type==="touchend"&&T(-1),d&&d(M,L),v.current=void 0,Tt()}),dt=nd(M=>{if(r)return;wg()||M.preventDefault();const B=M.changedTouches[0];B!=null&&(v.current=B.identifier);const L=bs(M,v);if(L!==!1){const{newValue:le,activeIndex:ee}=fe({finger:L});Ss({sliderRef:V,activeIndex:ee,setActive:_}),G(le),U&&!Es(le,N)&&U(M,le,ee)}R.current=0;const re=Zs(V.current);re.addEventListener("touchmove",Y,{passive:!0}),re.addEventListener("touchend",Ie,{passive:!0})}),Tt=m.useCallback(()=>{const M=Zs(V.current);M.removeEventListener("mousemove",Y),M.removeEventListener("mouseup",Ie),M.removeEventListener("touchmove",Y),M.removeEventListener("touchend",Ie)},[Ie,Y]);m.useEffect(()=>{const{current:M}=V;return M.addEventListener("touchstart",dt,{passive:wg()}),()=>{M.removeEventListener("touchstart",dt),Tt()}},[Tt,dt]),m.useEffect(()=>{r&&Tt()},[r,Tt]);const fl=M=>B=>{var L;if((L=M.onMouseDown)==null||L.call(M,B),r||B.defaultPrevented||B.button!==0)return;B.preventDefault();const re=bs(B,v);if(re!==!1){const{newValue:ee,activeIndex:fn}=fe({finger:re});Ss({sliderRef:V,activeIndex:fn,setActive:_}),G(ee),U&&!Es(ee,N)&&U(B,ee,fn)}R.current=0;const le=Zs(V.current);le.addEventListener("mousemove",Y,{passive:!0}),le.addEventListener("mouseup",Ie)},Oe=Mu(X?q[0]:s,s,o),cn=Mu(q[q.length-1],s,o)-Oe,Ua=(M={})=>{const B=Ws(M),L={onMouseDown:fl(B||{})},re=k({},B,L);return k({},M,{ref:zt},re)},Ha=M=>B=>{var L;(L=M.onMouseOver)==null||L.call(M,B);const re=Number(B.currentTarget.getAttribute("data-index"));T(re)},la=M=>B=>{var L;(L=M.onMouseLeave)==null||L.call(M,B),T(-1)};return{active:E,axis:lt,axisProps:o4,dragging:w,focusedThumbIndex:ke,getHiddenInputProps:(M={})=>{var B;const L=Ws(M),re={onChange:Xn(L||{}),onFocus:_t(L||{}),onBlur:it(L||{}),onKeyDown:Vn(L||{})},le=k({},L,re);return k({tabIndex:y,"aria-labelledby":t,"aria-orientation":f,"aria-valuemax":b(o),"aria-valuemin":b(s),name:u,type:"range",min:e.min,max:e.max,step:e.step===null&&e.marks?"any":(B=e.step)!=null?B:void 0,disabled:r},M,le,{style:k({},L3,{direction:i?"rtl":"ltr",width:"100%",height:"100%"})})},getRootProps:Ua,getThumbProps:(M={})=>{const B=Ws(M),L={onMouseOver:Ha(B||{}),onMouseLeave:la(B||{})};return k({},M,B,L)},marks:C,open:O,range:X,rootRef:zt,trackLeap:cn,trackOffset:Oe,values:q,getThumbStyle:M=>({pointerEvents:E!==-1&&E!==M?"none":void 0})}}const c4=e=>!e||!Vl(e);function f4(e){return ex("MuiSlider",e)}const Jt=E3("MuiSlider",["root","active","colorPrimary","colorSecondary","colorError","colorInfo","colorSuccess","colorWarning","disabled","dragging","focusVisible","mark","markActive","marked","markLabel","markLabelActive","rail","sizeSmall","thumb","thumbColorPrimary","thumbColorSecondary","thumbColorError","thumbColorSuccess","thumbColorInfo","thumbColorWarning","track","trackInverted","trackFalse","thumbSizeSmall","valueLabel","valueLabelOpen","valueLabelCircle","valueLabelLabel","vertical"]),d4=e=>{const{open:t}=e;return{offset:Gr(t&&Jt.valueLabelOpen),circle:Jt.valueLabelCircle,label:Jt.valueLabelLabel}};function p4(e){const{children:t,className:n,value:r}=e,a=d4(e);return t?m.cloneElement(t,{className:Gr(t.props.className)},g.jsxs(m.Fragment,{children:[t.props.children,g.jsx("span",{className:Gr(a.offset,n),"aria-hidden":!0,children:g.jsx("span",{className:a.circle,children:g.jsx("span",{className:a.label,children:r})})})]})):null}const h4=["aria-label","aria-valuetext","aria-labelledby","component","components","componentsProps","color","classes","className","disableSwap","disabled","getAriaLabel","getAriaValueText","marks","max","min","name","onChange","onChangeCommitted","orientation","shiftStep","size","step","scale","slotProps","slots","tabIndex","track","value","valueLabelDisplay","valueLabelFormat"];function Ag(e){return e}const y4=Ca("span",{name:"MuiSlider",slot:"Root",overridesResolver:(e,t)=>{const{ownerState:n}=e;return[t.root,t[`color${mn(n.color)}`],n.size!=="medium"&&t[`size${mn(n.size)}`],n.marked&&t.marked,n.orientation==="vertical"&&t.vertical,n.track==="inverted"&&t.trackInverted,n.track===!1&&t.trackFalse]}})(({theme:e})=>{var t;return{borderRadius:12,boxSizing:"content-box",display:"inline-block",position:"relative",cursor:"pointer",touchAction:"none",WebkitTapHighlightColor:"transparent","@media print":{colorAdjust:"exact"},[`&.${Jt.disabled}`]:{pointerEvents:"none",cursor:"default",color:(e.vars||e).palette.grey[400]},[`&.${Jt.dragging}`]:{[`& .${Jt.thumb}, & .${Jt.track}`]:{transition:"none"}},variants:[...Object.keys(((t=e.vars)!=null?t:e).palette).filter(n=>{var r;return((r=e.vars)!=null?r:e).palette[n].main}).map(n=>({props:{color:n},style:{color:(e.vars||e).palette[n].main}})),{props:{orientation:"horizontal"},style:{height:4,width:"100%",padding:"13px 0","@media (pointer: coarse)":{padding:"20px 0"}}},{props:{orientation:"horizontal",size:"small"},style:{height:2}},{props:{orientation:"horizontal",marked:!0},style:{marginBottom:20}},{props:{orientation:"vertical"},style:{height:"100%",width:4,padding:"0 13px","@media (pointer: coarse)":{padding:"0 20px"}}},{props:{orientation:"vertical",size:"small"},style:{width:2}},{props:{orientation:"vertical",marked:!0},style:{marginRight:44}}]}}),m4=Ca("span",{name:"MuiSlider",slot:"Rail",overridesResolver:(e,t)=>t.rail})({display:"block",position:"absolute",borderRadius:"inherit",backgroundColor:"currentColor",opacity:.38,variants:[{props:{orientation:"horizontal"},style:{width:"100%",height:"inherit",top:"50%",transform:"translateY(-50%)"}},{props:{orientation:"vertical"},style:{height:"100%",width:"inherit",left:"50%",transform:"translateX(-50%)"}},{props:{track:"inverted"},style:{opacity:1}}]}),v4=Ca("span",{name:"MuiSlider",slot:"Track",overridesResolver:(e,t)=>t.track})(({theme:e})=>{var t;return{display:"block",position:"absolute",borderRadius:"inherit",border:"1px solid currentColor",backgroundColor:"currentColor",transition:e.transitions.create(["left","width","bottom","height"],{duration:e.transitions.duration.shortest}),variants:[{props:{size:"small"},style:{border:"none"}},{props:{orientation:"horizontal"},style:{height:"inherit",top:"50%",transform:"translateY(-50%)"}},{props:{orientation:"vertical"},style:{width:"inherit",left:"50%",transform:"translateX(-50%)"}},{props:{track:!1},style:{display:"none"}},...Object.keys(((t=e.vars)!=null?t:e).palette).filter(n=>{var r;return((r=e.vars)!=null?r:e).palette[n].main}).map(n=>({props:{color:n,track:"inverted"},style:k({},e.vars?{backgroundColor:e.vars.palette.Slider[`${n}Track`],borderColor:e.vars.palette.Slider[`${n}Track`]}:k({backgroundColor:kp(e.palette[n].main,.62),borderColor:kp(e.palette[n].main,.62)},e.applyStyles("dark",{backgroundColor:Bp(e.palette[n].main,.5)}),e.applyStyles("dark",{borderColor:Bp(e.palette[n].main,.5)})))}))]}}),g4=Ca("span",{name:"MuiSlider",slot:"Thumb",overridesResolver:(e,t)=>{const{ownerState:n}=e;return[t.thumb,t[`thumbColor${mn(n.color)}`],n.size!=="medium"&&t[`thumbSize${mn(n.size)}`]]}})(({theme:e})=>{var t;return{position:"absolute",width:20,height:20,boxSizing:"border-box",borderRadius:"50%",outline:0,backgroundColor:"currentColor",display:"flex",alignItems:"center",justifyContent:"center",transition:e.transitions.create(["box-shadow","left","bottom"],{duration:e.transitions.duration.shortest}),"&::before":{position:"absolute",content:'""',borderRadius:"inherit",width:"100%",height:"100%",boxShadow:(e.vars||e).shadows[2]},"&::after":{position:"absolute",content:'""',borderRadius:"50%",width:42,height:42,top:"50%",left:"50%",transform:"translate(-50%, -50%)"},[`&.${Jt.disabled}`]:{"&:hover":{boxShadow:"none"}},variants:[{props:{size:"small"},style:{width:12,height:12,"&::before":{boxShadow:"none"}}},{props:{orientation:"horizontal"},style:{top:"50%",transform:"translate(-50%, -50%)"}},{props:{orientation:"vertical"},style:{left:"50%",transform:"translate(-50%, 50%)"}},...Object.keys(((t=e.vars)!=null?t:e).palette).filter(n=>{var r;return((r=e.vars)!=null?r:e).palette[n].main}).map(n=>({props:{color:n},style:{[`&:hover, &.${Jt.focusVisible}`]:k({},e.vars?{boxShadow:`0px 0px 0px 8px rgba(${e.vars.palette[n].mainChannel} / 0.16)`}:{boxShadow:`0px 0px 0px 8px ${yg(e.palette[n].main,.16)}`},{"@media (hover: none)":{boxShadow:"none"}}),[`&.${Jt.active}`]:k({},e.vars?{boxShadow:`0px 0px 0px 14px rgba(${e.vars.palette[n].mainChannel} / 0.16)`}:{boxShadow:`0px 0px 0px 14px ${yg(e.palette[n].main,.16)}`})}}))]}}),b4=Ca(p4,{name:"MuiSlider",slot:"ValueLabel",overridesResolver:(e,t)=>t.valueLabel})(({theme:e})=>k({zIndex:1,whiteSpace:"nowrap"},e.typography.body2,{fontWeight:500,transition:e.transitions.create(["transform"],{duration:e.transitions.duration.shortest}),position:"absolute",backgroundColor:(e.vars||e).palette.grey[600],borderRadius:2,color:(e.vars||e).palette.common.white,display:"flex",alignItems:"center",justifyContent:"center",padding:"0.25rem 0.75rem",variants:[{props:{orientation:"horizontal"},style:{transform:"translateY(-100%) scale(0)",top:"-10px",transformOrigin:"bottom center","&::before":{position:"absolute",content:'""',width:8,height:8,transform:"translate(-50%, 50%) rotate(45deg)",backgroundColor:"inherit",bottom:0,left:"50%"},[`&.${Jt.valueLabelOpen}`]:{transform:"translateY(-100%) scale(1)"}}},{props:{orientation:"vertical"},style:{transform:"translateY(-50%) scale(0)",right:"30px",top:"50%",transformOrigin:"right center","&::before":{position:"absolute",content:'""',width:8,height:8,transform:"translate(-50%, -50%) rotate(45deg)",backgroundColor:"inherit",right:-8,top:"50%"},[`&.${Jt.valueLabelOpen}`]:{transform:"translateY(-50%) scale(1)"}}},{props:{size:"small"},style:{fontSize:e.typography.pxToRem(12),padding:"0.25rem 0.5rem"}},{props:{orientation:"vertical",size:"small"},style:{right:"20px"}}]})),S4=Ca("span",{name:"MuiSlider",slot:"Mark",shouldForwardProp:e=>Qy(e)&&e!=="markActive",overridesResolver:(e,t)=>{const{markActive:n}=e;return[t.mark,n&&t.markActive]}})(({theme:e})=>({position:"absolute",width:2,height:2,borderRadius:1,backgroundColor:"currentColor",variants:[{props:{orientation:"horizontal"},style:{top:"50%",transform:"translate(-1px, -50%)"}},{props:{orientation:"vertical"},style:{left:"50%",transform:"translate(-50%, 1px)"}},{props:{markActive:!0},style:{backgroundColor:(e.vars||e).palette.background.paper,opacity:.8}}]})),E4=Ca("span",{name:"MuiSlider",slot:"MarkLabel",shouldForwardProp:e=>Qy(e)&&e!=="markLabelActive",overridesResolver:(e,t)=>t.markLabel})(({theme:e})=>k({},e.typography.body2,{color:(e.vars||e).palette.text.secondary,position:"absolute",whiteSpace:"nowrap",variants:[{props:{orientation:"horizontal"},style:{top:30,transform:"translateX(-50%)","@media (pointer: coarse)":{top:40}}},{props:{orientation:"vertical"},style:{left:36,transform:"translateY(50%)","@media (pointer: coarse)":{left:44}}},{props:{markLabelActive:!0},style:{color:(e.vars||e).palette.text.primary}}]})),x4=e=>{const{disabled:t,dragging:n,marked:r,orientation:a,track:i,classes:l,color:o,size:s}=e,u={root:["root",t&&"disabled",n&&"dragging",r&&"marked",a==="vertical"&&"vertical",i==="inverted"&&"trackInverted",i===!1&&"trackFalse",o&&`color${mn(o)}`,s&&`size${mn(s)}`],rail:["rail"],track:["track"],mark:["mark"],markActive:["markActive"],markLabel:["markLabel"],markLabelActive:["markLabelActive"],valueLabel:["valueLabel"],thumb:["thumb",t&&"disabled",s&&`thumbSize${mn(s)}`,o&&`thumbColor${mn(o)}`],active:["active"],disabled:["disabled"],focusVisible:["focusVisible"]};return U3(u,f4,l)},O4=({children:e})=>e,_4=m.forwardRef(function(t,n){var r,a,i,l,o,s,u,c,d,f,h,b,S,x,y,p,v,E,_,O,T,w,$,R;const N=t4({props:t,name:"MuiSlider"}),G=X3(),{"aria-label":U,"aria-valuetext":X,"aria-labelledby":q,component:C="span",components:j={},componentsProps:D={},color:K="primary",classes:ie,className:Pe,disableSwap:ke=!1,disabled:xe=!1,getAriaLabel:V,getAriaValueText:ve,marks:zt=!1,max:_t=100,min:it=0,orientation:Lt="horizontal",shiftStep:Vn=10,size:Xn="medium",step:Pn=1,scale:lt=Ag,slotProps:fe,slots:Y,track:Ie="normal",valueLabelDisplay:dt="off",valueLabelFormat:Tt=Ag}=N,fl=bn(N,h4),Oe=k({},N,{isRtl:G,max:_t,min:it,classes:ie,disabled:xe,disableSwap:ke,orientation:Lt,marks:zt,color:K,size:Xn,step:Pn,shiftStep:Vn,scale:lt,track:Ie,valueLabelDisplay:dt,valueLabelFormat:Tt}),{axisProps:cn,getRootProps:Ua,getHiddenInputProps:Ha,getThumbProps:la,open:dl,active:qa,axis:Tr,focusedThumbIndex:M,range:B,dragging:L,marks:re,values:le,trackOffset:ee,trackLeap:fn,getThumbStyle:xn}=u4(k({},Oe,{rootRef:n}));Oe.marked=re.length>0&&re.some(ze=>ze.label),Oe.dragging=L,Oe.focusedThumbIndex=M;const ue=x4(Oe),Ut=(r=(a=Y==null?void 0:Y.root)!=null?a:j.Root)!=null?r:y4,Ga=(i=(l=Y==null?void 0:Y.rail)!=null?l:j.Rail)!=null?i:m4,fv=(o=(s=Y==null?void 0:Y.track)!=null?s:j.Track)!=null?o:v4,dv=(u=(c=Y==null?void 0:Y.thumb)!=null?c:j.Thumb)!=null?u:g4,pv=(d=(f=Y==null?void 0:Y.valueLabel)!=null?f:j.ValueLabel)!=null?d:b4,Df=(h=(b=Y==null?void 0:Y.mark)!=null?b:j.Mark)!=null?h:S4,Bf=(S=(x=Y==null?void 0:Y.markLabel)!=null?x:j.MarkLabel)!=null?S:E4,hv=(y=(p=Y==null?void 0:Y.input)!=null?p:j.Input)!=null?y:"input",kf=(v=fe==null?void 0:fe.root)!=null?v:D.root,hT=(E=fe==null?void 0:fe.rail)!=null?E:D.rail,zf=(_=fe==null?void 0:fe.track)!=null?_:D.track,Lf=(O=fe==null?void 0:fe.thumb)!=null?O:D.thumb,Uf=(T=fe==null?void 0:fe.valueLabel)!=null?T:D.valueLabel,yT=(w=fe==null?void 0:fe.mark)!=null?w:D.mark,mT=($=fe==null?void 0:fe.markLabel)!=null?$:D.markLabel,vT=(R=fe==null?void 0:fe.input)!=null?R:D.input,gT=Ar({elementType:Ut,getSlotProps:Ua,externalSlotProps:kf,externalForwardedProps:fl,additionalProps:k({},c4(Ut)&&{as:C}),ownerState:k({},Oe,kf==null?void 0:kf.ownerState),className:[ue.root,Pe]}),bT=Ar({elementType:Ga,externalSlotProps:hT,ownerState:Oe,className:ue.rail}),ST=Ar({elementType:fv,externalSlotProps:zf,additionalProps:{style:k({},cn[Tr].offset(ee),cn[Tr].leap(fn))},ownerState:k({},Oe,zf==null?void 0:zf.ownerState),className:ue.track}),Hf=Ar({elementType:dv,getSlotProps:la,externalSlotProps:Lf,ownerState:k({},Oe,Lf==null?void 0:Lf.ownerState),className:ue.thumb}),ET=Ar({elementType:pv,externalSlotProps:Uf,ownerState:k({},Oe,Uf==null?void 0:Uf.ownerState),className:ue.valueLabel}),qf=Ar({elementType:Df,externalSlotProps:yT,ownerState:Oe,className:ue.mark}),Gf=Ar({elementType:Bf,externalSlotProps:mT,ownerState:Oe,className:ue.markLabel}),xT=Ar({elementType:hv,getSlotProps:Ha,externalSlotProps:vT,ownerState:Oe});return g.jsxs(Ut,k({},gT,{children:[g.jsx(Ga,k({},bT)),g.jsx(fv,k({},ST)),re.filter(ze=>ze.value>=it&&ze.value<=_t).map((ze,et)=>{const Yf=Mu(ze.value,it,_t),cs=cn[Tr].offset(Yf);let Fn;return Ie===!1?Fn=le.indexOf(ze.value)!==-1:Fn=Ie==="normal"&&(B?ze.value>=le[0]&&ze.value<=le[le.length-1]:ze.value<=le[0])||Ie==="inverted"&&(B?ze.value<=le[0]||ze.value>=le[le.length-1]:ze.value>=le[0]),g.jsxs(m.Fragment,{children:[g.jsx(Df,k({"data-index":et},qf,!Vl(Df)&&{markActive:Fn},{style:k({},cs,qf.style),className:Gr(qf.className,Fn&&ue.markActive)})),ze.label!=null?g.jsx(Bf,k({"aria-hidden":!0,"data-index":et},Gf,!Vl(Bf)&&{markLabelActive:Fn},{style:k({},cs,Gf.style),className:Gr(ue.markLabel,Gf.className,Fn&&ue.markLabelActive),children:ze.label})):null]},et)}),le.map((ze,et)=>{const Yf=Mu(ze,it,_t),cs=cn[Tr].offset(Yf),Fn=dt==="off"?O4:pv;return g.jsx(Fn,k({},!Vl(Fn)&&{valueLabelFormat:Tt,valueLabelDisplay:dt,value:typeof Tt=="function"?Tt(lt(ze),et):Tt,index:et,open:dl===et||qa===et||dt==="on",disabled:xe},ET,{children:g.jsx(dv,k({"data-index":et},Hf,{className:Gr(ue.thumb,Hf.className,qa===et&&ue.active,M===et&&ue.focusVisible),style:k({},cs,xn(et),Hf.style),children:g.jsx(hv,k({"data-index":et,"aria-label":V?V(et):U,"aria-valuenow":lt(ze),"aria-labelledby":q,"aria-valuetext":ve?ve(lt(ze),et):X,value:le[et]},xT))}))}),et)})]}))});var Rg=Object.prototype.toString,yx=function(t){var n=Rg.call(t),r=n==="[object Arguments]";return r||(r=n!=="[object Array]"&&t!==null&&typeof t=="object"&&typeof t.length=="number"&&t.length>=0&&Rg.call(t.callee)==="[object Function]"),r},ld,Cg;function T4(){if(Cg)return ld;Cg=1;var e;if(!Object.keys){var t=Object.prototype.hasOwnProperty,n=Object.prototype.toString,r=yx,a=Object.prototype.propertyIsEnumerable,i=!a.call({toString:null},"toString"),l=a.call(function(){},"prototype"),o=["toString","toLocaleString","valueOf","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","constructor"],s=function(f){var h=f.constructor;return h&&h.prototype===f},u={$applicationCache:!0,$console:!0,$external:!0,$frame:!0,$frameElement:!0,$frames:!0,$innerHeight:!0,$innerWidth:!0,$onmozfullscreenchange:!0,$onmozfullscreenerror:!0,$outerHeight:!0,$outerWidth:!0,$pageXOffset:!0,$pageYOffset:!0,$parent:!0,$scrollLeft:!0,$scrollTop:!0,$scrollX:!0,$scrollY:!0,$self:!0,$webkitIndexedDB:!0,$webkitStorageInfo:!0,$window:!0},c=function(){if(typeof window>"u")return!1;for(var f in window)try{if(!u["$"+f]&&t.call(window,f)&&window[f]!==null&&typeof window[f]=="object")try{s(window[f])}catch{return!0}}catch{return!0}return!1}(),d=function(f){if(typeof window>"u"||!c)return s(f);try{return s(f)}catch{return!1}};e=function(h){var b=h!==null&&typeof h=="object",S=n.call(h)==="[object Function]",x=r(h),y=b&&n.call(h)==="[object String]",p=[];if(!b&&!S&&!x)throw new TypeError("Object.keys called on a non-object");var v=l&&S;if(y&&h.length>0&&!t.call(h,0))for(var E=0;E<h.length;++E)p.push(String(E));if(x&&h.length>0)for(var _=0;_<h.length;++_)p.push(String(_));else for(var O in h)!(v&&O==="prototype")&&t.call(h,O)&&p.push(String(O));if(i)for(var T=d(h),w=0;w<o.length;++w)!(T&&o[w]==="constructor")&&t.call(h,o[w])&&p.push(o[w]);return p}}return ld=e,ld}var w4=Array.prototype.slice,A4=yx,$g=Object.keys,eu=$g?function(t){return $g(t)}:T4(),Ng=Object.keys;eu.shim=function(){if(Object.keys){var t=function(){var n=Object.keys(arguments);return n&&n.length===arguments.length}(1,2);t||(Object.keys=function(r){return A4(r)?Ng(w4.call(r)):Ng(r)})}else Object.keys=eu;return Object.keys||eu};var Ky=eu,tu=Object.defineProperty||!1;if(tu)try{tu({},"a",{value:1})}catch{tu=!1}var pf=tu,Zy=SyntaxError,un=TypeError,R4=Object.getOwnPropertyDescriptor,nu=R4;if(nu)try{nu([],"length")}catch{nu=null}var $a=nu,Mg=pf,C4=Zy,Wa=un,jg=$a,Wy=function(t,n,r){if(!t||typeof t!="object"&&typeof t!="function")throw new Wa("`obj` must be an object or a function`");if(typeof n!="string"&&typeof n!="symbol")throw new Wa("`property` must be a string or a symbol`");if(arguments.length>3&&typeof arguments[3]!="boolean"&&arguments[3]!==null)throw new Wa("`nonEnumerable`, if provided, must be a boolean or null");if(arguments.length>4&&typeof arguments[4]!="boolean"&&arguments[4]!==null)throw new Wa("`nonWritable`, if provided, must be a boolean or null");if(arguments.length>5&&typeof arguments[5]!="boolean"&&arguments[5]!==null)throw new Wa("`nonConfigurable`, if provided, must be a boolean or null");if(arguments.length>6&&typeof arguments[6]!="boolean")throw new Wa("`loose`, if provided, must be a boolean");var a=arguments.length>3?arguments[3]:null,i=arguments.length>4?arguments[4]:null,l=arguments.length>5?arguments[5]:null,o=arguments.length>6?arguments[6]:!1,s=!!jg&&jg(t,n);if(Mg)Mg(t,n,{configurable:l===null&&s?s.configurable:!l,enumerable:a===null&&s?s.enumerable:!a,value:r,writable:i===null&&s?s.writable:!i});else if(o||!a&&!i&&!l)t[n]=r;else throw new C4("This environment does not support defining a property as non-configurable, non-writable, or non-enumerable.")},Lp=pf,mx=function(){return!!Lp};mx.hasArrayLengthDefineBug=function(){if(!Lp)return null;try{return Lp([],"length",{value:1}).length!==1}catch{return!0}};var Jy=mx,$4=Ky,N4=typeof Symbol=="function"&&typeof Symbol("foo")=="symbol",M4=Object.prototype.toString,j4=Array.prototype.concat,Dg=Wy,D4=function(e){return typeof e=="function"&&M4.call(e)==="[object Function]"},vx=Jy(),B4=function(e,t,n,r){if(t in e){if(r===!0){if(e[t]===n)return}else if(!D4(r)||!r())return}vx?Dg(e,t,n,!0):Dg(e,t,n)},gx=function(e,t){var n=arguments.length>2?arguments[2]:{},r=$4(t);N4&&(r=j4.call(r,Object.getOwnPropertySymbols(t)));for(var a=0;a<r.length;a+=1)B4(e,r[a],t[r[a]],n[r[a]])};gx.supportsDescriptors=!!vx;var Na=gx,bx={exports:{}},Iy=Object,Sx=Error,k4=EvalError,z4=RangeError,L4=ReferenceError,U4=URIError,H4=Math.abs,q4=Math.floor,G4=Math.max,Y4=Math.min,V4=Math.pow,X4=Math.round,P4=Number.isNaN||function(t){return t!==t},F4=P4,Q4=function(t){return F4(t)||t===0?t:t<0?-1:1},hf=function(){if(typeof Symbol!="function"||typeof Object.getOwnPropertySymbols!="function")return!1;if(typeof Symbol.iterator=="symbol")return!0;var t={},n=Symbol("test"),r=Object(n);if(typeof n=="string"||Object.prototype.toString.call(n)!=="[object Symbol]"||Object.prototype.toString.call(r)!=="[object Symbol]")return!1;var a=42;t[n]=a;for(var i in t)return!1;if(typeof Object.keys=="function"&&Object.keys(t).length!==0||typeof Object.getOwnPropertyNames=="function"&&Object.getOwnPropertyNames(t).length!==0)return!1;var l=Object.getOwnPropertySymbols(t);if(l.length!==1||l[0]!==n||!Object.prototype.propertyIsEnumerable.call(t,n))return!1;if(typeof Object.getOwnPropertyDescriptor=="function"){var o=Object.getOwnPropertyDescriptor(t,n);if(o.value!==a||o.enumerable!==!0)return!1}return!0},Bg=typeof Symbol<"u"&&Symbol,K4=hf,em=function(){return typeof Bg!="function"||typeof Symbol!="function"||typeof Bg("foo")!="symbol"||typeof Symbol("bar")!="symbol"?!1:K4()},Ex=typeof Reflect<"u"&&Reflect.getPrototypeOf||null,Z4=Iy,xx=Z4.getPrototypeOf||null,od,kg;function W4(){if(kg)return od;kg=1;var e="Function.prototype.bind called on incompatible ",t=Object.prototype.toString,n=Math.max,r="[object Function]",a=function(s,u){for(var c=[],d=0;d<s.length;d+=1)c[d]=s[d];for(var f=0;f<u.length;f+=1)c[f+s.length]=u[f];return c},i=function(s,u){for(var c=[],d=u,f=0;d<s.length;d+=1,f+=1)c[f]=s[d];return c},l=function(o,s){for(var u="",c=0;c<o.length;c+=1)u+=o[c],c+1<o.length&&(u+=s);return u};return od=function(s){var u=this;if(typeof u!="function"||t.apply(u)!==r)throw new TypeError(e+u);for(var c=i(arguments,1),d,f=function(){if(this instanceof d){var y=u.apply(this,a(c,arguments));return Object(y)===y?y:this}return u.apply(s,a(c,arguments))},h=n(0,u.length-c.length),b=[],S=0;S<h;S++)b[S]="$"+S;if(d=Function("binder","return function ("+l(b,",")+"){ return binder.apply(this,arguments); }")(f),u.prototype){var x=function(){};x.prototype=u.prototype,d.prototype=new x,x.prototype=null}return d},od}var sd,zg;function Qo(){if(zg)return sd;zg=1;var e=W4();return sd=Function.prototype.bind||e,sd}var ud,Lg;function tm(){return Lg||(Lg=1,ud=Function.prototype.call),ud}var cd,Ug;function nm(){return Ug||(Ug=1,cd=Function.prototype.apply),cd}var J4=typeof Reflect<"u"&&Reflect&&Reflect.apply,I4=Qo(),eM=nm(),tM=tm(),nM=J4,Ox=nM||I4.call(tM,eM),rM=Qo(),aM=un,iM=tm(),lM=Ox,rm=function(t){if(t.length<1||typeof t[0]!="function")throw new aM("a function is required");return lM(rM,iM,t)},oM=rm,Hg=$a,_x;try{_x=[].__proto__===Array.prototype}catch(e){if(!e||typeof e!="object"||!("code"in e)||e.code!=="ERR_PROTO_ACCESS")throw e}var fd=!!_x&&Hg&&Hg(Object.prototype,"__proto__"),Tx=Object,qg=Tx.getPrototypeOf,sM=fd&&typeof fd.get=="function"?oM([fd.get]):typeof qg=="function"?function(t){return qg(t==null?t:Tx(t))}:!1,Gg=Ex,Yg=xx,Vg=sM,wx=Gg?function(t){return Gg(t)}:Yg?function(t){if(!t||typeof t!="object"&&typeof t!="function")throw new TypeError("getProto: not an object");return Yg(t)}:Vg?function(t){return Vg(t)}:null,uM=Function.prototype.call,cM=Object.prototype.hasOwnProperty,fM=Qo(),am=fM.call(uM,cM),W,dM=Iy,pM=Sx,hM=k4,yM=z4,mM=L4,Ui=Zy,_i=un,vM=U4,gM=H4,bM=q4,SM=G4,EM=Y4,xM=V4,OM=X4,_M=Q4,Ax=Function,dd=function(e){try{return Ax('"use strict"; return ('+e+").constructor;")()}catch{}},bo=$a,TM=pf,pd=function(){throw new _i},wM=bo?function(){try{return arguments.callee,pd}catch{try{return bo(arguments,"callee").get}catch{return pd}}}():pd,Ja=em(),Fe=wx,AM=xx,RM=Ex,Rx=nm(),Ko=tm(),ii={},CM=typeof Uint8Array>"u"||!Fe?W:Fe(Uint8Array),ya={__proto__:null,"%AggregateError%":typeof AggregateError>"u"?W:AggregateError,"%Array%":Array,"%ArrayBuffer%":typeof ArrayBuffer>"u"?W:ArrayBuffer,"%ArrayIteratorPrototype%":Ja&&Fe?Fe([][Symbol.iterator]()):W,"%AsyncFromSyncIteratorPrototype%":W,"%AsyncFunction%":ii,"%AsyncGenerator%":ii,"%AsyncGeneratorFunction%":ii,"%AsyncIteratorPrototype%":ii,"%Atomics%":typeof Atomics>"u"?W:Atomics,"%BigInt%":typeof BigInt>"u"?W:BigInt,"%BigInt64Array%":typeof BigInt64Array>"u"?W:BigInt64Array,"%BigUint64Array%":typeof BigUint64Array>"u"?W:BigUint64Array,"%Boolean%":Boolean,"%DataView%":typeof DataView>"u"?W:DataView,"%Date%":Date,"%decodeURI%":decodeURI,"%decodeURIComponent%":decodeURIComponent,"%encodeURI%":encodeURI,"%encodeURIComponent%":encodeURIComponent,"%Error%":pM,"%eval%":eval,"%EvalError%":hM,"%Float32Array%":typeof Float32Array>"u"?W:Float32Array,"%Float64Array%":typeof Float64Array>"u"?W:Float64Array,"%FinalizationRegistry%":typeof FinalizationRegistry>"u"?W:FinalizationRegistry,"%Function%":Ax,"%GeneratorFunction%":ii,"%Int8Array%":typeof Int8Array>"u"?W:Int8Array,"%Int16Array%":typeof Int16Array>"u"?W:Int16Array,"%Int32Array%":typeof Int32Array>"u"?W:Int32Array,"%isFinite%":isFinite,"%isNaN%":isNaN,"%IteratorPrototype%":Ja&&Fe?Fe(Fe([][Symbol.iterator]())):W,"%JSON%":typeof JSON=="object"?JSON:W,"%Map%":typeof Map>"u"?W:Map,"%MapIteratorPrototype%":typeof Map>"u"||!Ja||!Fe?W:Fe(new Map()[Symbol.iterator]()),"%Math%":Math,"%Number%":Number,"%Object%":dM,"%Object.getOwnPropertyDescriptor%":bo,"%parseFloat%":parseFloat,"%parseInt%":parseInt,"%Promise%":typeof Promise>"u"?W:Promise,"%Proxy%":typeof Proxy>"u"?W:Proxy,"%RangeError%":yM,"%ReferenceError%":mM,"%Reflect%":typeof Reflect>"u"?W:Reflect,"%RegExp%":RegExp,"%Set%":typeof Set>"u"?W:Set,"%SetIteratorPrototype%":typeof Set>"u"||!Ja||!Fe?W:Fe(new Set()[Symbol.iterator]()),"%SharedArrayBuffer%":typeof SharedArrayBuffer>"u"?W:SharedArrayBuffer,"%String%":String,"%StringIteratorPrototype%":Ja&&Fe?Fe(""[Symbol.iterator]()):W,"%Symbol%":Ja?Symbol:W,"%SyntaxError%":Ui,"%ThrowTypeError%":wM,"%TypedArray%":CM,"%TypeError%":_i,"%Uint8Array%":typeof Uint8Array>"u"?W:Uint8Array,"%Uint8ClampedArray%":typeof Uint8ClampedArray>"u"?W:Uint8ClampedArray,"%Uint16Array%":typeof Uint16Array>"u"?W:Uint16Array,"%Uint32Array%":typeof Uint32Array>"u"?W:Uint32Array,"%URIError%":vM,"%WeakMap%":typeof WeakMap>"u"?W:WeakMap,"%WeakRef%":typeof WeakRef>"u"?W:WeakRef,"%WeakSet%":typeof WeakSet>"u"?W:WeakSet,"%Function.prototype.call%":Ko,"%Function.prototype.apply%":Rx,"%Object.defineProperty%":TM,"%Object.getPrototypeOf%":AM,"%Math.abs%":gM,"%Math.floor%":bM,"%Math.max%":SM,"%Math.min%":EM,"%Math.pow%":xM,"%Math.round%":OM,"%Math.sign%":_M,"%Reflect.getPrototypeOf%":RM};if(Fe)try{null.error}catch(e){var $M=Fe(Fe(e));ya["%Error.prototype%"]=$M}var NM=function e(t){var n;if(t==="%AsyncFunction%")n=dd("async function () {}");else if(t==="%GeneratorFunction%")n=dd("function* () {}");else if(t==="%AsyncGeneratorFunction%")n=dd("async function* () {}");else if(t==="%AsyncGenerator%"){var r=e("%AsyncGeneratorFunction%");r&&(n=r.prototype)}else if(t==="%AsyncIteratorPrototype%"){var a=e("%AsyncGenerator%");a&&Fe&&(n=Fe(a.prototype))}return ya[t]=n,n},Xg={__proto__:null,"%ArrayBufferPrototype%":["ArrayBuffer","prototype"],"%ArrayPrototype%":["Array","prototype"],"%ArrayProto_entries%":["Array","prototype","entries"],"%ArrayProto_forEach%":["Array","prototype","forEach"],"%ArrayProto_keys%":["Array","prototype","keys"],"%ArrayProto_values%":["Array","prototype","values"],"%AsyncFunctionPrototype%":["AsyncFunction","prototype"],"%AsyncGenerator%":["AsyncGeneratorFunction","prototype"],"%AsyncGeneratorPrototype%":["AsyncGeneratorFunction","prototype","prototype"],"%BooleanPrototype%":["Boolean","prototype"],"%DataViewPrototype%":["DataView","prototype"],"%DatePrototype%":["Date","prototype"],"%ErrorPrototype%":["Error","prototype"],"%EvalErrorPrototype%":["EvalError","prototype"],"%Float32ArrayPrototype%":["Float32Array","prototype"],"%Float64ArrayPrototype%":["Float64Array","prototype"],"%FunctionPrototype%":["Function","prototype"],"%Generator%":["GeneratorFunction","prototype"],"%GeneratorPrototype%":["GeneratorFunction","prototype","prototype"],"%Int8ArrayPrototype%":["Int8Array","prototype"],"%Int16ArrayPrototype%":["Int16Array","prototype"],"%Int32ArrayPrototype%":["Int32Array","prototype"],"%JSONParse%":["JSON","parse"],"%JSONStringify%":["JSON","stringify"],"%MapPrototype%":["Map","prototype"],"%NumberPrototype%":["Number","prototype"],"%ObjectPrototype%":["Object","prototype"],"%ObjProto_toString%":["Object","prototype","toString"],"%ObjProto_valueOf%":["Object","prototype","valueOf"],"%PromisePrototype%":["Promise","prototype"],"%PromiseProto_then%":["Promise","prototype","then"],"%Promise_all%":["Promise","all"],"%Promise_reject%":["Promise","reject"],"%Promise_resolve%":["Promise","resolve"],"%RangeErrorPrototype%":["RangeError","prototype"],"%ReferenceErrorPrototype%":["ReferenceError","prototype"],"%RegExpPrototype%":["RegExp","prototype"],"%SetPrototype%":["Set","prototype"],"%SharedArrayBufferPrototype%":["SharedArrayBuffer","prototype"],"%StringPrototype%":["String","prototype"],"%SymbolPrototype%":["Symbol","prototype"],"%SyntaxErrorPrototype%":["SyntaxError","prototype"],"%TypedArrayPrototype%":["TypedArray","prototype"],"%TypeErrorPrototype%":["TypeError","prototype"],"%Uint8ArrayPrototype%":["Uint8Array","prototype"],"%Uint8ClampedArrayPrototype%":["Uint8ClampedArray","prototype"],"%Uint16ArrayPrototype%":["Uint16Array","prototype"],"%Uint32ArrayPrototype%":["Uint32Array","prototype"],"%URIErrorPrototype%":["URIError","prototype"],"%WeakMapPrototype%":["WeakMap","prototype"],"%WeakSetPrototype%":["WeakSet","prototype"]},Zo=Qo(),ju=am,MM=Zo.call(Ko,Array.prototype.concat),jM=Zo.call(Rx,Array.prototype.splice),Pg=Zo.call(Ko,String.prototype.replace),Du=Zo.call(Ko,String.prototype.slice),DM=Zo.call(Ko,RegExp.prototype.exec),BM=/[^%.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|%$))/g,kM=/\\(\\)?/g,zM=function(t){var n=Du(t,0,1),r=Du(t,-1);if(n==="%"&&r!=="%")throw new Ui("invalid intrinsic syntax, expected closing `%`");if(r==="%"&&n!=="%")throw new Ui("invalid intrinsic syntax, expected opening `%`");var a=[];return Pg(t,BM,function(i,l,o,s){a[a.length]=o?Pg(s,kM,"$1"):l||i}),a},LM=function(t,n){var r=t,a;if(ju(Xg,r)&&(a=Xg[r],r="%"+a[0]+"%"),ju(ya,r)){var i=ya[r];if(i===ii&&(i=NM(r)),typeof i>"u"&&!n)throw new _i("intrinsic "+t+" exists, but is not available. Please file an issue!");return{alias:a,name:r,value:i}}throw new Ui("intrinsic "+t+" does not exist!")},Or=function(t,n){if(typeof t!="string"||t.length===0)throw new _i("intrinsic name must be a non-empty string");if(arguments.length>1&&typeof n!="boolean")throw new _i('"allowMissing" argument must be a boolean');if(DM(/^%?[^%]*%?$/,t)===null)throw new Ui("`%` may not be present anywhere but at the beginning and end of the intrinsic name");var r=zM(t),a=r.length>0?r[0]:"",i=LM("%"+a+"%",n),l=i.name,o=i.value,s=!1,u=i.alias;u&&(a=u[0],jM(r,MM([0,1],u)));for(var c=1,d=!0;c<r.length;c+=1){var f=r[c],h=Du(f,0,1),b=Du(f,-1);if((h==='"'||h==="'"||h==="`"||b==='"'||b==="'"||b==="`")&&h!==b)throw new Ui("property names with quotes must have matching quotes");if((f==="constructor"||!d)&&(s=!0),a+="."+f,l="%"+a+"%",ju(ya,l))o=ya[l];else if(o!=null){if(!(f in o)){if(!n)throw new _i("base intrinsic for "+t+" exists, but the property is not available.");return}if(bo&&c+1>=r.length){var S=bo(o,f);d=!!S,d&&"get"in S&&!("originalValue"in S.get)?o=S.get:o=o[f]}else d=ju(o,f),o=o[f];d&&!s&&(ya[l]=o)}}return o},UM=Or,Fg=Wy,HM=Jy(),Qg=$a,Kg=un,qM=UM("%Math.floor%"),GM=function(t,n){if(typeof t!="function")throw new Kg("`fn` is not a function");if(typeof n!="number"||n<0||n>4294967295||qM(n)!==n)throw new Kg("`length` must be a positive 32-bit integer");var r=arguments.length>2&&!!arguments[2],a=!0,i=!0;if("length"in t&&Qg){var l=Qg(t,"length");l&&!l.configurable&&(a=!1),l&&!l.writable&&(i=!1)}return(a||i||!r)&&(HM?Fg(t,"length",n,!0,!0):Fg(t,"length",n)),t},YM=Qo(),VM=nm(),XM=Ox,PM=function(){return XM(YM,VM,arguments)};(function(e){var t=GM,n=pf,r=rm,a=PM;e.exports=function(l){var o=r(arguments),s=l.length-(arguments.length-1);return t(o,1+(s>0?s:0),!0)},n?n(e.exports,"apply",{value:a}):e.exports.apply=a})(bx);var al=bx.exports,Cx=Or,$x=rm,FM=$x([Cx("%String.prototype.indexOf%")]),ft=function(t,n){var r=Cx(t,!!n);return typeof r=="function"&&FM(t,".prototype.")>-1?$x([r]):r},QM=Ky,Nx=hf(),Mx=ft,ru=Iy,KM=Mx("Array.prototype.push"),Zg=Mx("Object.prototype.propertyIsEnumerable"),ZM=Nx?ru.getOwnPropertySymbols:null,jx=function(t,n){if(t==null)throw new TypeError("target must be an object");var r=ru(t);if(arguments.length===1)return r;for(var a=1;a<arguments.length;++a){var i=ru(arguments[a]),l=QM(i),o=Nx&&(ru.getOwnPropertySymbols||ZM);if(o)for(var s=o(i),u=0;u<s.length;++u){var c=s[u];Zg(i,c)&&KM(l,c)}for(var d=0;d<l.length;++d){var f=l[d];if(Zg(i,f)){var h=i[f];r[f]=h}}}return r},hd=jx,WM=function(){if(!Object.assign)return!1;for(var e="abcdefghijklmnopqrst",t=e.split(""),n={},r=0;r<t.length;++r)n[t[r]]=t[r];var a=Object.assign({},n),i="";for(var l in a)i+=l;return e!==i},JM=function(){if(!Object.assign||!Object.preventExtensions)return!1;var e=Object.preventExtensions({1:2});try{Object.assign(e,"xy")}catch{return e[1]==="y"}return!1},Dx=function(){return!Object.assign||WM()||JM()?hd:Object.assign},IM=Na,e6=Dx,t6=function(){var t=e6();return IM(Object,{assign:t},{assign:function(){return Object.assign!==t}}),t},n6=Na,r6=al,a6=jx,Bx=Dx,i6=t6,l6=r6.apply(Bx()),kx=function(t,n){return l6(Object,arguments)};n6(kx,{getPolyfill:Bx,implementation:a6,shim:i6});var o6=kx,zx=Or,Lx=al,s6=Lx(zx("String.prototype.indexOf")),Ux=function(t,n){var r=zx(t,!!n);return typeof r=="function"&&s6(t,".prototype.")>-1?Lx(r):r},So=function(){return typeof(function(){}).name=="string"},Xl=Object.getOwnPropertyDescriptor;if(Xl)try{Xl([],"length")}catch{Xl=null}So.functionsHaveConfigurableNames=function(){if(!So()||!Xl)return!1;var t=Xl(function(){},"name");return!!t&&!!t.configurable};var u6=Function.prototype.bind;So.boundFunctionsHaveNames=function(){return So()&&typeof u6=="function"&&(function(){}).bind().name!==""};var c6=So,Wg=Wy,f6=Jy(),d6=c6.functionsHaveConfigurableNames(),p6=un,h6=function(t,n){if(typeof t!="function")throw new p6("`fn` is not a function");var r=arguments.length>2&&!!arguments[2];return(!r||d6)&&(f6?Wg(t,"name",n,!0,!0):Wg(t,"name",n)),t},y6=h6,m6=un,v6=Object,Hx=y6(function(){if(this==null||this!==v6(this))throw new m6("RegExp.prototype.flags getter called on non-object");var t="";return this.hasIndices&&(t+="d"),this.global&&(t+="g"),this.ignoreCase&&(t+="i"),this.multiline&&(t+="m"),this.dotAll&&(t+="s"),this.unicode&&(t+="u"),this.unicodeSets&&(t+="v"),this.sticky&&(t+="y"),t},"get flags",!0),g6=Hx,b6=Na.supportsDescriptors,S6=Object.getOwnPropertyDescriptor,qx=function(){if(b6&&/a/mig.flags==="gim"){var t=S6(RegExp.prototype,"flags");if(t&&typeof t.get=="function"&&"dotAll"in RegExp.prototype&&"hasIndices"in RegExp.prototype){var n="",r={};if(Object.defineProperty(r,"hasIndices",{get:function(){n+="d"}}),Object.defineProperty(r,"sticky",{get:function(){n+="y"}}),t.get.call(r),n==="dy")return t.get}}return g6},E6=Na.supportsDescriptors,x6=qx,O6=$a,_6=Object.defineProperty,T6=Sx,Jg=wx,w6=/a/,A6=function(){if(!E6||!Jg)throw new T6("RegExp.prototype.flags requires a true ES5 environment that supports property descriptors");var t=x6(),n=Jg(w6),r=O6(n,"flags");return(!r||r.get!==t)&&_6(n,"flags",{configurable:!0,enumerable:!1,get:t}),t},R6=Na,C6=al,$6=Hx,Gx=qx,N6=A6,Yx=C6(Gx());R6(Yx,{getPolyfill:Gx,implementation:$6,shim:N6});var M6=Yx,au={exports:{}},j6=hf,Ma=function(){return j6()&&!!Symbol.toStringTag},D6=Ma(),B6=ft,Up=B6("Object.prototype.toString"),yf=function(t){return D6&&t&&typeof t=="object"&&Symbol.toStringTag in t?!1:Up(t)==="[object Arguments]"},Vx=function(t){return yf(t)?!0:t!==null&&typeof t=="object"&&"length"in t&&typeof t.length=="number"&&t.length>=0&&Up(t)!=="[object Array]"&&"callee"in t&&Up(t.callee)==="[object Function]"},k6=function(){return yf(arguments)}();yf.isLegacyArguments=Vx;var Xx=k6?yf:Vx;const z6={},L6=Object.freeze(Object.defineProperty({__proto__:null,default:z6},Symbol.toStringTag,{value:"Module"})),U6=Er(L6);var im=typeof Map=="function"&&Map.prototype,yd=Object.getOwnPropertyDescriptor&&im?Object.getOwnPropertyDescriptor(Map.prototype,"size"):null,Bu=im&&yd&&typeof yd.get=="function"?yd.get:null,Ig=im&&Map.prototype.forEach,lm=typeof Set=="function"&&Set.prototype,md=Object.getOwnPropertyDescriptor&&lm?Object.getOwnPropertyDescriptor(Set.prototype,"size"):null,ku=lm&&md&&typeof md.get=="function"?md.get:null,e0=lm&&Set.prototype.forEach,H6=typeof WeakMap=="function"&&WeakMap.prototype,Pl=H6?WeakMap.prototype.has:null,q6=typeof WeakSet=="function"&&WeakSet.prototype,Fl=q6?WeakSet.prototype.has:null,G6=typeof WeakRef=="function"&&WeakRef.prototype,t0=G6?WeakRef.prototype.deref:null,Y6=Boolean.prototype.valueOf,V6=Object.prototype.toString,X6=Function.prototype.toString,P6=String.prototype.match,om=String.prototype.slice,Lr=String.prototype.replace,F6=String.prototype.toUpperCase,n0=String.prototype.toLowerCase,Px=RegExp.prototype.test,r0=Array.prototype.concat,wn=Array.prototype.join,Q6=Array.prototype.slice,a0=Math.floor,Hp=typeof BigInt=="function"?BigInt.prototype.valueOf:null,vd=Object.getOwnPropertySymbols,qp=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?Symbol.prototype.toString:null,Hi=typeof Symbol=="function"&&typeof Symbol.iterator=="object",ct=typeof Symbol=="function"&&Symbol.toStringTag&&(typeof Symbol.toStringTag===Hi||!0)?Symbol.toStringTag:null,Fx=Object.prototype.propertyIsEnumerable,i0=(typeof Reflect=="function"?Reflect.getPrototypeOf:Object.getPrototypeOf)||([].__proto__===Array.prototype?function(e){return e.__proto__}:null);function l0(e,t){if(e===1/0||e===-1/0||e!==e||e&&e>-1e3&&e<1e3||Px.call(/e/,t))return t;var n=/[0-9](?=(?:[0-9]{3})+(?![0-9]))/g;if(typeof e=="number"){var r=e<0?-a0(-e):a0(e);if(r!==e){var a=String(r),i=om.call(t,a.length+1);return Lr.call(a,n,"$&_")+"."+Lr.call(Lr.call(i,/([0-9]{3})/g,"$&_"),/_$/,"")}}return Lr.call(t,n,"$&_")}var Gp=U6,o0=Gp.custom,s0=Zx(o0)?o0:null,Qx={__proto__:null,double:'"',single:"'"},K6={__proto__:null,double:/(["\\])/g,single:/(['\\])/g},mf=function e(t,n,r,a){var i=n||{};if(In(i,"quoteStyle")&&!In(Qx,i.quoteStyle))throw new TypeError('option "quoteStyle" must be "single" or "double"');if(In(i,"maxStringLength")&&(typeof i.maxStringLength=="number"?i.maxStringLength<0&&i.maxStringLength!==1/0:i.maxStringLength!==null))throw new TypeError('option "maxStringLength", if provided, must be a positive integer, Infinity, or `null`');var l=In(i,"customInspect")?i.customInspect:!0;if(typeof l!="boolean"&&l!=="symbol")throw new TypeError("option \"customInspect\", if provided, must be `true`, `false`, or `'symbol'`");if(In(i,"indent")&&i.indent!==null&&i.indent!=="	"&&!(parseInt(i.indent,10)===i.indent&&i.indent>0))throw new TypeError('option "indent" must be "\\t", an integer > 0, or `null`');if(In(i,"numericSeparator")&&typeof i.numericSeparator!="boolean")throw new TypeError('option "numericSeparator", if provided, must be `true` or `false`');var o=i.numericSeparator;if(typeof t>"u")return"undefined";if(t===null)return"null";if(typeof t=="boolean")return t?"true":"false";if(typeof t=="string")return Jx(t,i);if(typeof t=="number"){if(t===0)return 1/0/t>0?"0":"-0";var s=String(t);return o?l0(t,s):s}if(typeof t=="bigint"){var u=String(t)+"n";return o?l0(t,u):u}var c=typeof i.depth>"u"?5:i.depth;if(typeof r>"u"&&(r=0),r>=c&&c>0&&typeof t=="object")return Yp(t)?"[Array]":"[Object]";var d=pj(i,r);if(typeof a>"u")a=[];else if(Wx(a,t)>=0)return"[Circular]";function f(U,X,q){if(X&&(a=Q6.call(a),a.push(X)),q){var C={depth:i.depth};return In(i,"quoteStyle")&&(C.quoteStyle=i.quoteStyle),e(U,C,r+1,a)}return e(U,i,r+1,a)}if(typeof t=="function"&&!u0(t)){var h=aj(t),b=Os(t,f);return"[Function"+(h?": "+h:" (anonymous)")+"]"+(b.length>0?" { "+wn.call(b,", ")+" }":"")}if(Zx(t)){var S=Hi?Lr.call(String(t),/^(Symbol\(.*\))_[^)]*$/,"$1"):qp.call(t);return typeof t=="object"&&!Hi?gl(S):S}if(cj(t)){for(var x="<"+n0.call(String(t.nodeName)),y=t.attributes||[],p=0;p<y.length;p++)x+=" "+y[p].name+"="+Kx(Z6(y[p].value),"double",i);return x+=">",t.childNodes&&t.childNodes.length&&(x+="..."),x+="</"+n0.call(String(t.nodeName))+">",x}if(Yp(t)){if(t.length===0)return"[]";var v=Os(t,f);return d&&!dj(v)?"["+Vp(v,d)+"]":"[ "+wn.call(v,", ")+" ]"}if(J6(t)){var E=Os(t,f);return!("cause"in Error.prototype)&&"cause"in t&&!Fx.call(t,"cause")?"{ ["+String(t)+"] "+wn.call(r0.call("[cause]: "+f(t.cause),E),", ")+" }":E.length===0?"["+String(t)+"]":"{ ["+String(t)+"] "+wn.call(E,", ")+" }"}if(typeof t=="object"&&l){if(s0&&typeof t[s0]=="function"&&Gp)return Gp(t,{depth:c-r});if(l!=="symbol"&&typeof t.inspect=="function")return t.inspect()}if(ij(t)){var _=[];return Ig&&Ig.call(t,function(U,X){_.push(f(X,t,!0)+" => "+f(U,t))}),c0("Map",Bu.call(t),_,d)}if(sj(t)){var O=[];return e0&&e0.call(t,function(U){O.push(f(U,t))}),c0("Set",ku.call(t),O,d)}if(lj(t))return gd("WeakMap");if(uj(t))return gd("WeakSet");if(oj(t))return gd("WeakRef");if(ej(t))return gl(f(Number(t)));if(nj(t))return gl(f(Hp.call(t)));if(tj(t))return gl(Y6.call(t));if(I6(t))return gl(f(String(t)));if(typeof window<"u"&&t===window)return"{ [object Window] }";if(typeof globalThis<"u"&&t===globalThis||typeof xu<"u"&&t===xu)return"{ [object globalThis] }";if(!W6(t)&&!u0(t)){var T=Os(t,f),w=i0?i0(t)===Object.prototype:t instanceof Object||t.constructor===Object,$=t instanceof Object?"":"null prototype",R=!w&&ct&&Object(t)===t&&ct in t?om.call(ra(t),8,-1):$?"Object":"",N=w||typeof t.constructor!="function"?"":t.constructor.name?t.constructor.name+" ":"",G=N+(R||$?"["+wn.call(r0.call([],R||[],$||[]),": ")+"] ":"");return T.length===0?G+"{}":d?G+"{"+Vp(T,d)+"}":G+"{ "+wn.call(T,", ")+" }"}return String(t)};function Kx(e,t,n){var r=n.quoteStyle||t,a=Qx[r];return a+e+a}function Z6(e){return Lr.call(String(e),/"/g,"&quot;")}function Yp(e){return ra(e)==="[object Array]"&&(!ct||!(typeof e=="object"&&ct in e))}function W6(e){return ra(e)==="[object Date]"&&(!ct||!(typeof e=="object"&&ct in e))}function u0(e){return ra(e)==="[object RegExp]"&&(!ct||!(typeof e=="object"&&ct in e))}function J6(e){return ra(e)==="[object Error]"&&(!ct||!(typeof e=="object"&&ct in e))}function I6(e){return ra(e)==="[object String]"&&(!ct||!(typeof e=="object"&&ct in e))}function ej(e){return ra(e)==="[object Number]"&&(!ct||!(typeof e=="object"&&ct in e))}function tj(e){return ra(e)==="[object Boolean]"&&(!ct||!(typeof e=="object"&&ct in e))}function Zx(e){if(Hi)return e&&typeof e=="object"&&e instanceof Symbol;if(typeof e=="symbol")return!0;if(!e||typeof e!="object"||!qp)return!1;try{return qp.call(e),!0}catch{}return!1}function nj(e){if(!e||typeof e!="object"||!Hp)return!1;try{return Hp.call(e),!0}catch{}return!1}var rj=Object.prototype.hasOwnProperty||function(e){return e in this};function In(e,t){return rj.call(e,t)}function ra(e){return V6.call(e)}function aj(e){if(e.name)return e.name;var t=P6.call(X6.call(e),/^function\s*([\w$]+)/);return t?t[1]:null}function Wx(e,t){if(e.indexOf)return e.indexOf(t);for(var n=0,r=e.length;n<r;n++)if(e[n]===t)return n;return-1}function ij(e){if(!Bu||!e||typeof e!="object")return!1;try{Bu.call(e);try{ku.call(e)}catch{return!0}return e instanceof Map}catch{}return!1}function lj(e){if(!Pl||!e||typeof e!="object")return!1;try{Pl.call(e,Pl);try{Fl.call(e,Fl)}catch{return!0}return e instanceof WeakMap}catch{}return!1}function oj(e){if(!t0||!e||typeof e!="object")return!1;try{return t0.call(e),!0}catch{}return!1}function sj(e){if(!ku||!e||typeof e!="object")return!1;try{ku.call(e);try{Bu.call(e)}catch{return!0}return e instanceof Set}catch{}return!1}function uj(e){if(!Fl||!e||typeof e!="object")return!1;try{Fl.call(e,Fl);try{Pl.call(e,Pl)}catch{return!0}return e instanceof WeakSet}catch{}return!1}function cj(e){return!e||typeof e!="object"?!1:typeof HTMLElement<"u"&&e instanceof HTMLElement?!0:typeof e.nodeName=="string"&&typeof e.getAttribute=="function"}function Jx(e,t){if(e.length>t.maxStringLength){var n=e.length-t.maxStringLength,r="... "+n+" more character"+(n>1?"s":"");return Jx(om.call(e,0,t.maxStringLength),t)+r}var a=K6[t.quoteStyle||"single"];a.lastIndex=0;var i=Lr.call(Lr.call(e,a,"\\$1"),/[\x00-\x1f]/g,fj);return Kx(i,"single",t)}function fj(e){var t=e.charCodeAt(0),n={8:"b",9:"t",10:"n",12:"f",13:"r"}[t];return n?"\\"+n:"\\x"+(t<16?"0":"")+F6.call(t.toString(16))}function gl(e){return"Object("+e+")"}function gd(e){return e+" { ? }"}function c0(e,t,n,r){var a=r?Vp(n,r):wn.call(n,", ");return e+" ("+t+") {"+a+"}"}function dj(e){for(var t=0;t<e.length;t++)if(Wx(e[t],`
`)>=0)return!1;return!0}function pj(e,t){var n;if(e.indent==="	")n="	";else if(typeof e.indent=="number"&&e.indent>0)n=wn.call(Array(e.indent+1)," ");else return null;return{base:n,prev:wn.call(Array(t+1),n)}}function Vp(e,t){if(e.length===0)return"";var n=`
`+t.prev+t.base;return n+wn.call(e,","+n)+`
`+t.prev}function Os(e,t){var n=Yp(e),r=[];if(n){r.length=e.length;for(var a=0;a<e.length;a++)r[a]=In(e,a)?t(e[a],e):""}var i=typeof vd=="function"?vd(e):[],l;if(Hi){l={};for(var o=0;o<i.length;o++)l["$"+i[o]]=i[o]}for(var s in e)In(e,s)&&(n&&String(Number(s))===s&&s<e.length||Hi&&l["$"+s]instanceof Symbol||(Px.call(/[^\w$]/,s)?r.push(t(s,e)+": "+t(e[s],e)):r.push(s+": "+t(e[s],e))));if(typeof vd=="function")for(var u=0;u<i.length;u++)Fx.call(e,i[u])&&r.push("["+t(i[u])+"]: "+t(e[i[u]],e));return r}var hj=mf,yj=un,vf=function(e,t,n){for(var r=e,a;(a=r.next)!=null;r=a)if(a.key===t)return r.next=a.next,n||(a.next=e.next,e.next=a),a},mj=function(e,t){if(e){var n=vf(e,t);return n&&n.value}},vj=function(e,t,n){var r=vf(e,t);r?r.value=n:e.next={key:t,next:e.next,value:n}},gj=function(e,t){return e?!!vf(e,t):!1},bj=function(e,t){if(e)return vf(e,t,!0)},Sj=function(){var t,n={assert:function(r){if(!n.has(r))throw new yj("Side channel does not contain "+hj(r))},delete:function(r){var a=t&&t.next,i=bj(t,r);return i&&a&&a===i&&(t=void 0),!!i},get:function(r){return mj(t,r)},has:function(r){return gj(t,r)},set:function(r,a){t||(t={next:void 0}),vj(t,r,a)}};return n},Ej=Or,Wo=ft,xj=mf,Oj=un,f0=Ej("%Map%",!0),_j=Wo("Map.prototype.get",!0),Tj=Wo("Map.prototype.set",!0),wj=Wo("Map.prototype.has",!0),Aj=Wo("Map.prototype.delete",!0),Rj=Wo("Map.prototype.size",!0),Ix=!!f0&&function(){var t,n={assert:function(r){if(!n.has(r))throw new Oj("Side channel does not contain "+xj(r))},delete:function(r){if(t){var a=Aj(t,r);return Rj(t)===0&&(t=void 0),a}return!1},get:function(r){if(t)return _j(t,r)},has:function(r){return t?wj(t,r):!1},set:function(r,a){t||(t=new f0),Tj(t,r,a)}};return n},Cj=Or,gf=ft,$j=mf,_s=Ix,Nj=un,Ia=Cj("%WeakMap%",!0),Mj=gf("WeakMap.prototype.get",!0),jj=gf("WeakMap.prototype.set",!0),Dj=gf("WeakMap.prototype.has",!0),Bj=gf("WeakMap.prototype.delete",!0),kj=Ia?function(){var t,n,r={assert:function(a){if(!r.has(a))throw new Nj("Side channel does not contain "+$j(a))},delete:function(a){if(Ia&&a&&(typeof a=="object"||typeof a=="function")){if(t)return Bj(t,a)}else if(_s&&n)return n.delete(a);return!1},get:function(a){return Ia&&a&&(typeof a=="object"||typeof a=="function")&&t?Mj(t,a):n&&n.get(a)},has:function(a){return Ia&&a&&(typeof a=="object"||typeof a=="function")&&t?Dj(t,a):!!n&&n.has(a)},set:function(a,i){Ia&&a&&(typeof a=="object"||typeof a=="function")?(t||(t=new Ia),jj(t,a,i)):_s&&(n||(n=_s()),n.set(a,i))}};return r}:_s,zj=un,Lj=mf,Uj=Sj,Hj=Ix,qj=kj,Gj=qj||Hj||Uj,eO=function(){var t,n={assert:function(r){if(!n.has(r))throw new zj("Side channel does not contain "+Lj(r))},delete:function(r){return!!t&&t.delete(r)},get:function(r){return t&&t.get(r)},has:function(r){return!!t&&t.has(r)},set:function(r,a){t||(t=Gj()),t.set(r,a)}};return n},Yj=am,bl=eO(),Qn=un,sm={assert:function(e,t){if(!e||typeof e!="object"&&typeof e!="function")throw new Qn("`O` is not an object");if(typeof t!="string")throw new Qn("`slot` must be a string");if(bl.assert(e),!sm.has(e,t))throw new Qn("`"+t+"` is not present on `O`")},get:function(e,t){if(!e||typeof e!="object"&&typeof e!="function")throw new Qn("`O` is not an object");if(typeof t!="string")throw new Qn("`slot` must be a string");var n=bl.get(e);return n&&n["$"+t]},has:function(e,t){if(!e||typeof e!="object"&&typeof e!="function")throw new Qn("`O` is not an object");if(typeof t!="string")throw new Qn("`slot` must be a string");var n=bl.get(e);return!!n&&Yj(n,"$"+t)},set:function(e,t,n){if(!e||typeof e!="object"&&typeof e!="function")throw new Qn("`O` is not an object");if(typeof t!="string")throw new Qn("`slot` must be a string");var r=bl.get(e);r||(r={},bl.set(e,r)),r["$"+t]=n}};Object.freeze&&Object.freeze(sm);var Vj=sm,Sl=Vj,Xj=Zy,d0=typeof StopIteration=="object"?StopIteration:null,Pj=function(t){if(!d0)throw new Xj("this environment lacks StopIteration");Sl.set(t,"[[Done]]",!1);var n={next:function(){var a=Sl.get(this,"[[Iterator]]"),i=!!Sl.get(a,"[[Done]]");try{return{done:i,value:i?void 0:a.next()}}catch(l){if(Sl.set(a,"[[Done]]",!0),l!==d0)throw l;return{done:!0,value:void 0}}}};return Sl.set(n,"[[Iterator]]",t),n},Fj={}.toString,tO=Array.isArray||function(e){return Fj.call(e)=="[object Array]"},nO=ft,Qj=nO("String.prototype.valueOf"),Kj=function(t){try{return Qj(t),!0}catch{return!1}},Zj=nO("Object.prototype.toString"),Wj="[object String]",Jj=Ma(),rO=function(t){return typeof t=="string"?!0:!t||typeof t!="object"?!1:Jj?Kj(t):Zj(t)===Wj},um=typeof Map=="function"&&Map.prototype?Map:null,Ij=typeof Set=="function"&&Set.prototype?Set:null,zu;um||(zu=function(t){return!1});var aO=um?Map.prototype.has:null,p0=Ij?Set.prototype.has:null;!zu&&!aO&&(zu=function(t){return!1});var iO=zu||function(t){if(!t||typeof t!="object")return!1;try{if(aO.call(t),p0)try{p0.call(t)}catch{return!0}return t instanceof um}catch{}return!1},eD=typeof Map=="function"&&Map.prototype?Map:null,cm=typeof Set=="function"&&Set.prototype?Set:null,Lu;cm||(Lu=function(t){return!1});var h0=eD?Map.prototype.has:null,lO=cm?Set.prototype.has:null;!Lu&&!lO&&(Lu=function(t){return!1});var oO=Lu||function(t){if(!t||typeof t!="object")return!1;try{if(lO.call(t),h0)try{h0.call(t)}catch{return!0}return t instanceof cm}catch{}return!1},y0=Xx,m0=Pj;if(em()||hf()){var bd=Symbol.iterator;au.exports=function(t){if(t!=null&&typeof t[bd]<"u")return t[bd]();if(y0(t))return Array.prototype[bd].call(t)}}else{var tD=tO,nD=rO,v0=Or,rD=v0("%Map%",!0),aD=v0("%Set%",!0),dn=Ux,g0=dn("Array.prototype.push"),b0=dn("String.prototype.charCodeAt"),iD=dn("String.prototype.slice"),lD=function(t,n){var r=t.length;if(n+1>=r)return n+1;var a=b0(t,n);if(a<55296||a>56319)return n+1;var i=b0(t,n+1);return i<56320||i>57343?n+1:n+2},Sd=function(t){var n=0;return{next:function(){var a=n>=t.length,i;return a||(i=t[n],n+=1),{done:a,value:i}}}},S0=function(t,n){if(tD(t)||y0(t))return Sd(t);if(nD(t)){var r=0;return{next:function(){var i=lD(t,r),l=iD(t,r,i);return r=i,{done:i>t.length,value:l}}}}if(n&&typeof t["_es6-shim iterator_"]<"u")return t["_es6-shim iterator_"]()};if(!rD&&!aD)au.exports=function(t){if(t!=null)return S0(t,!0)};else{var oD=iO,sD=oO,E0=dn("Map.prototype.forEach",!0),x0=dn("Set.prototype.forEach",!0);if(typeof process>"u"||!process.versions||!process.versions.node)var O0=dn("Map.prototype.iterator",!0),_0=dn("Set.prototype.iterator",!0);var T0=dn("Map.prototype.@@iterator",!0)||dn("Map.prototype._es6-shim iterator_",!0),w0=dn("Set.prototype.@@iterator",!0)||dn("Set.prototype._es6-shim iterator_",!0),uD=function(t){if(oD(t)){if(O0)return m0(O0(t));if(T0)return T0(t);if(E0){var n=[];return E0(t,function(a,i){g0(n,[i,a])}),Sd(n)}}if(sD(t)){if(_0)return m0(_0(t));if(w0)return w0(t);if(x0){var r=[];return x0(t,function(a){g0(r,a)}),Sd(r)}}};au.exports=function(t){return uD(t)||S0(t)}}}var cD=au.exports,A0=function(e){return e!==e},sO=function(t,n){return t===0&&n===0?1/t===1/n:!!(t===n||A0(t)&&A0(n))},fD=sO,uO=function(){return typeof Object.is=="function"?Object.is:fD},dD=uO,pD=Na,hD=function(){var t=dD();return pD(Object,{is:t},{is:function(){return Object.is!==t}}),t},yD=Na,mD=al,vD=sO,cO=uO,gD=hD,fO=mD(cO(),Object);yD(fO,{getPolyfill:cO,implementation:vD,shim:gD});var bD=fO,SD=al,dO=ft,ED=Or,Xp=ED("%ArrayBuffer%",!0),iu=dO("ArrayBuffer.prototype.byteLength",!0),xD=dO("Object.prototype.toString"),R0=!!Xp&&!iu&&new Xp(0).slice,C0=!!R0&&SD(R0),pO=iu||C0?function(t){if(!t||typeof t!="object")return!1;try{return iu?iu(t):C0(t,0),!0}catch{return!1}}:Xp?function(t){return xD(t)==="[object ArrayBuffer]"}:function(t){return!1},hO=ft,OD=hO("Date.prototype.getDay"),_D=function(t){try{return OD(t),!0}catch{return!1}},TD=hO("Object.prototype.toString"),wD="[object Date]",AD=Ma(),RD=function(t){return typeof t!="object"||t===null?!1:AD?_D(t):TD(t)===wD},$0=ft,CD=Ma(),$D=am,ND=$a,Pp;if(CD){var MD=$0("RegExp.prototype.exec"),N0={},Ed=function(){throw N0},M0={toString:Ed,valueOf:Ed};typeof Symbol.toPrimitive=="symbol"&&(M0[Symbol.toPrimitive]=Ed),Pp=function(t){if(!t||typeof t!="object")return!1;var n=ND(t,"lastIndex"),r=n&&$D(n,"value");if(!r)return!1;try{MD(t,M0)}catch(a){return a===N0}}}else{var jD=$0("Object.prototype.toString"),DD="[object RegExp]";Pp=function(t){return!t||typeof t!="object"&&typeof t!="function"?!1:jD(t)===DD}}var yO=Pp,BD=ft,j0=BD("SharedArrayBuffer.prototype.byteLength",!0),kD=j0?function(t){if(!t||typeof t!="object")return!1;try{return j0(t),!0}catch{return!1}}:function(t){return!1},mO=ft,zD=mO("Number.prototype.toString"),LD=function(t){try{return zD(t),!0}catch{return!1}},UD=mO("Object.prototype.toString"),HD="[object Number]",qD=Ma(),GD=function(t){return typeof t=="number"?!0:!t||typeof t!="object"?!1:qD?LD(t):UD(t)===HD},vO=ft,YD=vO("Boolean.prototype.toString"),VD=vO("Object.prototype.toString"),XD=function(t){try{return YD(t),!0}catch{return!1}},PD="[object Boolean]",FD=Ma(),QD=function(t){return typeof t=="boolean"?!0:t===null||typeof t!="object"?!1:FD&&Symbol.toStringTag in t?XD(t):VD(t)===PD},Fp={exports:{}},xd,D0;function KD(){if(D0)return xd;D0=1;var e=ft,t=yO,n=e("RegExp.prototype.exec"),r=un;return xd=function(i){if(!t(i))throw new r("`regex` must be a RegExp");return function(o){return n(i,o)!==null}},xd}var gO=ft,ZD=gO("Object.prototype.toString"),WD=em(),JD=KD();if(WD){var ID=gO("Symbol.prototype.toString"),eB=JD(/^Symbol\(.*\)$/),tB=function(t){return typeof t.valueOf()!="symbol"?!1:eB(ID(t))};Fp.exports=function(t){if(typeof t=="symbol")return!0;if(!t||typeof t!="object"||ZD(t)!=="[object Symbol]")return!1;try{return tB(t)}catch{return!1}}}else Fp.exports=function(t){return!1};var nB=Fp.exports,Qp={exports:{}},B0=typeof BigInt<"u"&&BigInt,rB=function(){return typeof B0=="function"&&typeof BigInt=="function"&&typeof B0(42)=="bigint"&&typeof BigInt(42)=="bigint"},aB=rB();if(aB){var iB=BigInt.prototype.valueOf,lB=function(t){try{return iB.call(t),!0}catch{}return!1};Qp.exports=function(t){return t===null||typeof t>"u"||typeof t=="boolean"||typeof t=="string"||typeof t=="number"||typeof t=="symbol"||typeof t=="function"?!1:typeof t=="bigint"?!0:lB(t)}}else Qp.exports=function(t){return!1};var oB=Qp.exports,sB=rO,uB=GD,cB=QD,fB=nB,dB=oB,pB=function(t){if(t==null||typeof t!="object"&&typeof t!="function")return null;if(sB(t))return"String";if(uB(t))return"Number";if(cB(t))return"Boolean";if(fB(t))return"Symbol";if(dB(t))return"BigInt"},Uu=typeof WeakMap=="function"&&WeakMap.prototype?WeakMap:null,k0=typeof WeakSet=="function"&&WeakSet.prototype?WeakSet:null,Hu;Uu||(Hu=function(t){return!1});var Kp=Uu?Uu.prototype.has:null,Od=k0?k0.prototype.has:null;!Hu&&!Kp&&(Hu=function(t){return!1});var hB=Hu||function(t){if(!t||typeof t!="object")return!1;try{if(Kp.call(t,Kp),Od)try{Od.call(t,Od)}catch{return!0}return t instanceof Uu}catch{}return!1},Zp={exports:{}},yB=Or,bO=ft,mB=yB("%WeakSet%",!0),_d=bO("WeakSet.prototype.has",!0);if(_d){var Td=bO("WeakMap.prototype.has",!0);Zp.exports=function(t){if(!t||typeof t!="object")return!1;try{if(_d(t,_d),Td)try{Td(t,Td)}catch{return!0}return t instanceof mB}catch{}return!1}}else Zp.exports=function(t){return!1};var vB=Zp.exports,gB=iO,bB=oO,SB=hB,EB=vB,xB=function(t){if(t&&typeof t=="object"){if(gB(t))return"Map";if(bB(t))return"Set";if(SB(t))return"WeakMap";if(EB(t))return"WeakSet"}return!1},SO=Function.prototype.toString,ui=typeof Reflect=="object"&&Reflect!==null&&Reflect.apply,Wp,lu;if(typeof ui=="function"&&typeof Object.defineProperty=="function")try{Wp=Object.defineProperty({},"length",{get:function(){throw lu}}),lu={},ui(function(){throw 42},null,Wp)}catch(e){e!==lu&&(ui=null)}else ui=null;var OB=/^\s*class\b/,Jp=function(t){try{var n=SO.call(t);return OB.test(n)}catch{return!1}},wd=function(t){try{return Jp(t)?!1:(SO.call(t),!0)}catch{return!1}},ou=Object.prototype.toString,_B="[object Object]",TB="[object Function]",wB="[object GeneratorFunction]",AB="[object HTMLAllCollection]",RB="[object HTML document.all class]",CB="[object HTMLCollection]",$B=typeof Symbol=="function"&&!!Symbol.toStringTag,NB=!(0 in[,]),Ip=function(){return!1};if(typeof document=="object"){var MB=document.all;ou.call(MB)===ou.call(document.all)&&(Ip=function(t){if((NB||!t)&&(typeof t>"u"||typeof t=="object"))try{var n=ou.call(t);return(n===AB||n===RB||n===CB||n===_B)&&t("")==null}catch{}return!1})}var jB=ui?function(t){if(Ip(t))return!0;if(!t||typeof t!="function"&&typeof t!="object")return!1;try{ui(t,null,Wp)}catch(n){if(n!==lu)return!1}return!Jp(t)&&wd(t)}:function(t){if(Ip(t))return!0;if(!t||typeof t!="function"&&typeof t!="object")return!1;if($B)return wd(t);if(Jp(t))return!1;var n=ou.call(t);return n!==TB&&n!==wB&&!/^\[object HTML/.test(n)?!1:wd(t)},DB=jB,BB=Object.prototype.toString,EO=Object.prototype.hasOwnProperty,kB=function(t,n,r){for(var a=0,i=t.length;a<i;a++)EO.call(t,a)&&(r==null?n(t[a],a,t):n.call(r,t[a],a,t))},zB=function(t,n,r){for(var a=0,i=t.length;a<i;a++)r==null?n(t.charAt(a),a,t):n.call(r,t.charAt(a),a,t)},LB=function(t,n,r){for(var a in t)EO.call(t,a)&&(r==null?n(t[a],a,t):n.call(r,t[a],a,t))},UB=function(t,n,r){if(!DB(n))throw new TypeError("iterator must be a function");var a;arguments.length>=3&&(a=r),BB.call(t)==="[object Array]"?kB(t,n,a):typeof t=="string"?zB(t,n,a):LB(t,n,a)},HB=UB,qB=["Float32Array","Float64Array","Int8Array","Int16Array","Int32Array","Uint8Array","Uint8ClampedArray","Uint16Array","Uint32Array","BigInt64Array","BigUint64Array"],Ad=qB,GB=typeof globalThis>"u"?xu:globalThis,YB=function(){for(var t=[],n=0;n<Ad.length;n++)typeof GB[Ad[n]]=="function"&&(t[t.length]=Ad[n]);return t},qu=HB,VB=YB,z0=al,fm=ft,su=$a,XB=fm("Object.prototype.toString"),xO=Ma(),L0=typeof globalThis>"u"?xu:globalThis,eh=VB(),dm=fm("String.prototype.slice"),Rd=Object.getPrototypeOf,PB=fm("Array.prototype.indexOf",!0)||function(t,n){for(var r=0;r<t.length;r+=1)if(t[r]===n)return r;return-1},Gu={__proto__:null};xO&&su&&Rd?qu(eh,function(e){var t=new L0[e];if(Symbol.toStringTag in t){var n=Rd(t),r=su(n,Symbol.toStringTag);if(!r){var a=Rd(n);r=su(a,Symbol.toStringTag)}Gu["$"+e]=z0(r.get)}}):qu(eh,function(e){var t=new L0[e],n=t.slice||t.set;n&&(Gu["$"+e]=z0(n))});var FB=function(t){var n=!1;return qu(Gu,function(r,a){if(!n)try{"$"+r(t)===a&&(n=dm(a,1))}catch{}}),n},QB=function(t){var n=!1;return qu(Gu,function(r,a){if(!n)try{r(t),n=dm(a,1)}catch{}}),n},KB=function(t){if(!t||typeof t!="object")return!1;if(!xO){var n=dm(XB(t),8,-1);return PB(eh,n)>-1?n:n!=="Object"?!1:QB(t)}return su?FB(t):null},ZB=ft,U0=ZB("ArrayBuffer.prototype.byteLength",!0),WB=pO,JB=function(t){return WB(t)?U0?U0(t):t.byteLength:NaN},OO=o6,qn=Ux,H0=M6,IB=Or,qi=cD,e8=eO,q0=bD,G0=Xx,Y0=tO,V0=pO,X0=RD,P0=yO,F0=kD,Q0=Ky,K0=pB,Z0=xB,W0=KB,J0=JB,I0=qn("SharedArrayBuffer.prototype.byteLength",!0),eb=qn("Date.prototype.getTime"),Cd=Object.getPrototypeOf,tb=qn("Object.prototype.toString"),Yu=IB("%Set%",!0),th=qn("Map.prototype.has",!0),Vu=qn("Map.prototype.get",!0),nb=qn("Map.prototype.size",!0),Xu=qn("Set.prototype.add",!0),_O=qn("Set.prototype.delete",!0),Pu=qn("Set.prototype.has",!0),uu=qn("Set.prototype.size",!0);function rb(e,t,n,r){for(var a=qi(e),i;(i=a.next())&&!i.done;)if(vn(t,i.value,n,r))return _O(e,i.value),!0;return!1}function TO(e){if(typeof e>"u")return null;if(typeof e!="object")return typeof e=="symbol"?!1:typeof e=="string"||typeof e=="number"?+e==+e:!0}function t8(e,t,n,r,a,i){var l=TO(n);if(l!=null)return l;var o=Vu(t,l),s=OO({},a,{strict:!1});return typeof o>"u"&&!th(t,l)||!vn(r,o,s,i)?!1:!th(e,l)&&vn(r,o,s,i)}function n8(e,t,n){var r=TO(n);return r??(Pu(t,r)&&!Pu(e,r))}function ab(e,t,n,r,a,i){for(var l=qi(e),o,s;(o=l.next())&&!o.done;)if(s=o.value,vn(n,s,a,i)&&vn(r,Vu(t,s),a,i))return _O(e,s),!0;return!1}function vn(e,t,n,r){var a=n||{};if(a.strict?q0(e,t):e===t)return!0;var i=K0(e),l=K0(t);if(i!==l)return!1;if(!e||!t||typeof e!="object"&&typeof t!="object")return a.strict?q0(e,t):e==t;var o=r.has(e),s=r.has(t),u;if(o&&s){if(r.get(e)===r.get(t))return!0}else u={};return o||r.set(e,u),s||r.set(t,u),i8(e,t,a,r)}function ib(e){return!e||typeof e!="object"||typeof e.length!="number"||typeof e.copy!="function"||typeof e.slice!="function"||e.length>0&&typeof e[0]!="number"?!1:!!(e.constructor&&e.constructor.isBuffer&&e.constructor.isBuffer(e))}function r8(e,t,n,r){if(uu(e)!==uu(t))return!1;for(var a=qi(e),i=qi(t),l,o,s;(l=a.next())&&!l.done;)if(l.value&&typeof l.value=="object")s||(s=new Yu),Xu(s,l.value);else if(!Pu(t,l.value)){if(n.strict||!n8(e,t,l.value))return!1;s||(s=new Yu),Xu(s,l.value)}if(s){for(;(o=i.next())&&!o.done;)if(o.value&&typeof o.value=="object"){if(!rb(s,o.value,n.strict,r))return!1}else if(!n.strict&&!Pu(e,o.value)&&!rb(s,o.value,n.strict,r))return!1;return uu(s)===0}return!0}function a8(e,t,n,r){if(nb(e)!==nb(t))return!1;for(var a=qi(e),i=qi(t),l,o,s,u,c,d;(l=a.next())&&!l.done;)if(u=l.value[0],c=l.value[1],u&&typeof u=="object")s||(s=new Yu),Xu(s,u);else if(d=Vu(t,u),typeof d>"u"&&!th(t,u)||!vn(c,d,n,r)){if(n.strict||!t8(e,t,u,c,n,r))return!1;s||(s=new Yu),Xu(s,u)}if(s){for(;(o=i.next())&&!o.done;)if(u=o.value[0],d=o.value[1],u&&typeof u=="object"){if(!ab(s,e,u,d,n,r))return!1}else if(!n.strict&&(!e.has(u)||!vn(Vu(e,u),d,n,r))&&!ab(s,e,u,d,OO({},n,{strict:!1}),r))return!1;return uu(s)===0}return!0}function i8(e,t,n,r){var a,i;if(typeof e!=typeof t||e==null||t==null||tb(e)!==tb(t)||G0(e)!==G0(t))return!1;var l=Y0(e),o=Y0(t);if(l!==o)return!1;var s=e instanceof Error,u=t instanceof Error;if(s!==u||(s||u)&&(e.name!==t.name||e.message!==t.message))return!1;var c=P0(e),d=P0(t);if(c!==d||(c||d)&&(e.source!==t.source||H0(e)!==H0(t)))return!1;var f=X0(e),h=X0(t);if(f!==h||(f||h)&&eb(e)!==eb(t)||n.strict&&Cd&&Cd(e)!==Cd(t))return!1;var b=W0(e),S=W0(t);if(b!==S)return!1;if(b||S){if(e.length!==t.length)return!1;for(a=0;a<e.length;a++)if(e[a]!==t[a])return!1;return!0}var x=ib(e),y=ib(t);if(x!==y)return!1;if(x||y){if(e.length!==t.length)return!1;for(a=0;a<e.length;a++)if(e[a]!==t[a])return!1;return!0}var p=V0(e),v=V0(t);if(p!==v)return!1;if(p||v)return J0(e)!==J0(t)?!1:typeof Uint8Array=="function"&&vn(new Uint8Array(e),new Uint8Array(t),n,r);var E=F0(e),_=F0(t);if(E!==_)return!1;if(E||_)return I0(e)!==I0(t)?!1:typeof Uint8Array=="function"&&vn(new Uint8Array(e),new Uint8Array(t),n,r);if(typeof e!=typeof t)return!1;var O=Q0(e),T=Q0(t);if(O.length!==T.length)return!1;for(O.sort(),T.sort(),a=O.length-1;a>=0;a--)if(O[a]!=T[a])return!1;for(a=O.length-1;a>=0;a--)if(i=O[a],!vn(e[i],t[i],n,r))return!1;var w=Z0(e),$=Z0(t);return w!==$?!1:w==="Set"||$==="Set"?r8(e,t,n,r):w==="Map"?a8(e,t,n,r):!0}var l8=function(t,n,r){return vn(t,n,r,e8())};const o8=Ra(l8),pm=(e,t)=>{for(const n in t)if(typeof t[n]=="object"){if(!o8(e[n],t[n]))return!1}else if(!Object.is(e[n],t[n]))return!1;return!0},Ts=e=>{let t=0,n;const r=e.readonly;return e.type==="int"||e.type==="float"?t=e.value:e.type==="Quantity"&&(t=e.value.magnitude,n=e.value.unit),[t,r,n]},wO=ne.memo(e=>{Hn();const[t,n]=m.useState(!1),{fullAccessPath:r,value:a,min:i,max:l,stepSize:o,docString:s,isInstantUpdate:u,addNotification:c,changeCallback:d=()=>{},displayName:f,id:h}=e;m.useEffect(()=>{c(`${r} changed to ${a.value}.`)},[e.value.value]),m.useEffect(()=>{c(`${r}.min changed to ${i.value}.`)},[e.min.value,e.min.type]),m.useEffect(()=>{c(`${r}.max changed to ${l.value}.`)},[e.max.value,e.max.type]),m.useEffect(()=>{c(`${r}.stepSize changed to ${o.value}.`)},[e.stepSize.value,e.stepSize.type]);const b=($,R)=>{Array.isArray(R)&&(R=R[0]);let N;a.type==="Quantity"?N={type:"Quantity",value:{magnitude:R,unit:a.value.unit},full_access_path:`${r}.value`,readonly:a.readonly,doc:s}:N={type:a.type,value:R,full_access_path:`${r}.value`,readonly:a.readonly,doc:s},d(N)},S=($,R,N)=>{let G;N.type==="Quantity"?G={type:N.type,value:{magnitude:$,unit:N.value.unit},full_access_path:`${r}.${R}`,readonly:N.readonly,doc:null}:G={type:N.type,value:$,full_access_path:`${r}.${R}`,readonly:N.readonly,doc:null},d(G)},[x,y,p]=Ts(a),[v,E]=Ts(i),[_,O]=Ts(l),[T,w]=Ts(o);return g.jsxs("div",{className:"component sliderComponent",id:h,children:[!1,g.jsxs(Au,{children:[g.jsx(On,{xs:"auto",xl:"auto",children:g.jsxs(dr.Text,{children:[f,g.jsx(Sn,{docString:s})]})}),g.jsx(On,{xs:"5",xl:!0,children:g.jsx(_4,{style:{margin:"0px 0px 10px 0px"},"aria-label":"Always visible",disabled:y,value:x,onChange:($,R)=>b($,R),min:v,max:_,step:T,marks:[{value:v,label:`${v}`},{value:_,label:`${_}`}]})}),g.jsx(On,{xs:"3",xl:!0,children:g.jsx(Ru,{isInstantUpdate:u,fullAccessPath:`${r}.value`,docString:s,readOnly:y,type:a.type,value:x,unit:p,addNotification:()=>{},changeCallback:d,id:h+"-value"})}),g.jsx(On,{xs:"auto",children:g.jsx(Ey,{id:`button-${h}`,onClick:()=>n(!t),type:"checkbox",checked:t,value:"",className:"btn",variant:"light","aria-controls":"slider-settings","aria-expanded":t,children:g.jsxs("svg",{xmlns:"http://www.w3.org/2000/svg",width:"16",height:"16",className:"bi bi-gear",viewBox:"0 0 16 16",children:[g.jsx("path",{d:"M8 4.754a3.246 3.246 0 1 0 0 6.492 3.246 3.246 0 0 0 0-6.492zM5.754 8a2.246 2.246 0 1 1 4.492 0 2.246 2.246 0 0 1-4.492 0z"}),g.jsx("path",{d:"M9.796 1.343c-.527-1.79-3.065-1.79-3.592 0l-.094.319a.873.873 0 0 1-1.255.52l-.292-.16c-1.64-.892-3.433.902-2.54 2.541l.159.292a.873.873 0 0 1-.52 1.255l-.319.094c-1.79.527-1.79 3.065 0 3.592l.319.094a.873.873 0 0 1 .52 1.255l-.16.292c-.892 1.64.901 3.434 2.541 2.54l.292-.159a.873.873 0 0 1 1.255.52l.094.319c.527 1.79 3.065 1.79 3.592 0l.094-.319a.873.873 0 0 1 1.255-.52l.292.16c1.64.893 3.434-.902 2.54-2.541l-.159-.292a.873.873 0 0 1 .52-1.255l.319-.094c1.79-.527 1.79-3.065 0-3.592l-.319-.094a.873.873 0 0 1-.52-1.255l.16-.292c.893-1.64-.902-3.433-2.541-2.54l-.292.159a.873.873 0 0 1-1.255-.52l-.094-.319zm-2.633.283c.246-.835 1.428-.835 1.674 0l.094.319a1.873 1.873 0 0 0 2.693 1.115l.291-.16c.764-.415 1.6.42 1.184 1.185l-.159.292a1.873 1.873 0 0 0 1.116 2.692l.318.094c.835.246.835 1.428 0 1.674l-.319.094a1.873 1.873 0 0 0-1.115 2.693l.16.291c.415.764-.42 1.6-1.185 1.184l-.291-.159a1.873 1.873 0 0 0-2.693 1.116l-.094.318c-.246.835-1.428.835-1.674 0l-.094-.319a1.873 1.873 0 0 0-2.692-1.115l-.292.16c-.764.415-1.6-.42-1.184-1.185l.159-.291A1.873 1.873 0 0 0 1.945 8.93l-.319-.094c-.835-.246-.835-1.428 0-1.674l.319-.094A1.873 1.873 0 0 0 3.06 4.377l-.16-.292c-.415-.764.42-1.6 1.185-1.184l.292.159a1.873 1.873 0 0 0 2.692-1.115l.094-.319z"})]})})})]}),g.jsx(Tc,{in:t,children:g.jsx(pt.Group,{children:g.jsxs(Au,{className:"justify-content-center",style:{paddingTop:"20px",margin:"10px"},children:[g.jsxs(On,{xs:"auto",children:[g.jsx(pt.Label,{children:"Min Value"}),g.jsx(pt.Control,{type:"number",value:v,disabled:E,onChange:$=>S(Number($.target.value),"min",i)})]}),g.jsxs(On,{xs:"auto",children:[g.jsx(pt.Label,{children:"Max Value"}),g.jsx(pt.Control,{type:"number",value:_,disabled:O,onChange:$=>S(Number($.target.value),"max",l)})]}),g.jsxs(On,{xs:"auto",children:[g.jsx(pt.Label,{children:"Step Size"}),g.jsx(pt.Control,{type:"number",value:T,disabled:w,onChange:$=>S(Number($.target.value),"step_size",o)})]})]})})})]})},pm);wO.displayName="SliderComponent";const AO=ne.memo(e=>{const{addNotification:t,displayName:n,id:r,value:a,full_access_path:i,enum:l,doc:o,readonly:s,changeCallback:u}=e;return Hn(),m.useEffect(()=>{t(`${i} changed to ${a}.`)},[a]),g.jsxs("div",{className:"component enumComponent",id:r,children:[!1,g.jsx(Au,{children:g.jsxs(On,{className:"d-flex align-items-center",children:[g.jsxs(dr.Text,{children:[n,g.jsx(Sn,{docString:o})]}),s?g.jsx(pt.Control,{style:e.type=="ColouredEnum"?{backgroundColor:l[a]}:{},value:e.type=="ColouredEnum"?a:l[a],name:i,disabled:!0}):g.jsx(pt.Select,{"aria-label":"example-select",value:a,name:i,style:e.type=="ColouredEnum"?{backgroundColor:l[a]}:{},onChange:c=>u({type:e.type,name:e.name,enum:l,value:c.target.value,full_access_path:i,readonly:e.readonly,doc:e.doc}),children:Object.entries(l).map(([c,d])=>g.jsx("option",{value:c,children:e.type=="ColouredEnum"?c:d},c))})]})})]})},pm);AO.displayName="EnumComponent";const hm=ne.memo(e=>{const{fullAccessPath:t,docString:n,addNotification:r,displayName:a,id:i}=e;if(!e.render)return null;Hn();const l=m.useRef(null),o=()=>{const u=`Method ${t} was triggered.`;r(u)},s=async u=>{u.preventDefault(),xE(t),o()};return g.jsxs("div",{className:"component methodComponent",id:i,children:[!1,g.jsx(pt,{onSubmit:s,ref:l,children:g.jsxs(Lo,{className:"component",variant:"primary",type:"submit",children:[`${a} `,g.jsx(Sn,{docString:n})]})})]})},pm);hm.displayName="MethodComponent";const RO=ne.memo(e=>{const{fullAccessPath:t,readOnly:n,docString:r,isInstantUpdate:a,addNotification:i,changeCallback:l=()=>{},displayName:o,id:s}=e;Hn();const[u,c]=m.useState(e.value);m.useEffect(()=>{e.value!==u&&c(e.value),i(`${t} changed to ${e.value}.`)},[e.value]);const d=b=>{c(b.target.value),a&&l({type:"str",value:b.target.value,full_access_path:t,readonly:n,doc:r})},f=b=>{b.key==="Enter"&&!a&&(l({type:"str",value:u,full_access_path:t,readonly:n,doc:r}),b.preventDefault())},h=()=>{a||l({type:"str",value:u,full_access_path:t,readonly:n,doc:r})};return g.jsxs("div",{className:"component stringComponent",id:s,children:[!1,g.jsxs(dr,{children:[g.jsxs(dr.Text,{children:[o,g.jsx(Sn,{docString:r})]}),g.jsx(pt.Control,{type:"text",name:s,value:u,disabled:n,onChange:d,onKeyDown:f,onBlur:h,className:a&&!n?"instantUpdate":""})]})]})});RO.displayName="StringComponent";function ym(e){const t=m.useContext(Ay),n=a=>{var i;return((i=t[a])==null?void 0:i.displayOrder)??Number.MAX_SAFE_INTEGER};let r;return Array.isArray(e)?r=[...e].sort((a,i)=>n(a.full_access_path)-n(i.full_access_path)):r=Object.values(e).sort((a,i)=>n(a.full_access_path)-n(i.full_access_path)),r}const CO=ne.memo(e=>{const{docString:t,isInstantUpdate:n,addNotification:r,id:a}=e,i=ym(e.value);return Hn(),g.jsxs("div",{className:"listComponent",id:a,children:[!1,g.jsx(Sn,{docString:t}),i.map(l=>g.jsx(Gi,{attribute:l,isInstantUpdate:n,addNotification:r},l.full_access_path))]})});CO.displayName="ListComponent";var s8=["color","size","title","className"];function nh(){return nh=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},nh.apply(null,arguments)}function u8(e,t){if(e==null)return{};var n,r,a=c8(e,t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(e);for(r=0;r<i.length;r++)n=i[r],t.includes(n)||{}.propertyIsEnumerable.call(e,n)&&(a[n]=e[n])}return a}function c8(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.includes(r))continue;n[r]=e[r]}return n}var mm=m.forwardRef(function(e,t){var n=e.color,r=n===void 0?"currentColor":n,a=e.size,i=a===void 0?"1em":a,l=e.title,o=l===void 0?null:l,s=e.className,u=s===void 0?"":s,c=u8(e,s8);return ne.createElement("svg",nh({ref:t,xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 16 16",width:i,height:i,fill:r,className:["bi","bi-chevron-down",u].filter(Boolean).join(" ")},c),o?ne.createElement("title",null,o):null,ne.createElement("path",{fillRule:"evenodd",d:"M1.646 4.646a.5.5 0 0 1 .708 0L8 10.293l5.646-5.647a.5.5 0 0 1 .708.708l-6 6a.5.5 0 0 1-.708 0l-6-6a.5.5 0 0 1 0-.708"}))});mm.propTypes={color:he.string,size:he.oneOfType([he.string,he.number]),title:he.string,className:he.string};var f8=["color","size","title","className"];function rh(){return rh=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},rh.apply(null,arguments)}function d8(e,t){if(e==null)return{};var n,r,a=p8(e,t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(e);for(r=0;r<i.length;r++)n=i[r],t.includes(n)||{}.propertyIsEnumerable.call(e,n)&&(a[n]=e[n])}return a}function p8(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.includes(r))continue;n[r]=e[r]}return n}var vm=m.forwardRef(function(e,t){var n=e.color,r=n===void 0?"currentColor":n,a=e.size,i=a===void 0?"1em":a,l=e.title,o=l===void 0?null:l,s=e.className,u=s===void 0?"":s,c=d8(e,f8);return ne.createElement("svg",rh({ref:t,xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 16 16",width:i,height:i,fill:r,className:["bi","bi-chevron-right",u].filter(Boolean).join(" ")},c),o?ne.createElement("title",null,o):null,ne.createElement("path",{fillRule:"evenodd",d:"M4.646 1.646a.5.5 0 0 1 .708 0l6 6a.5.5 0 0 1 0 .708l-6 6a.5.5 0 0 1-.708-.708L10.293 8 4.646 2.354a.5.5 0 0 1 0-.708"}))});vm.propTypes={color:he.string,size:he.oneOfType([he.string,he.number]),title:he.string,className:he.string};function ah(e,t){const[n,r]=m.useState(()=>{const a=localStorage.getItem(e);return a?JSON.parse(a):t});return m.useEffect(()=>{n!==void 0&&localStorage.setItem(e,JSON.stringify(n))},[n,e]),[n,r]}const gm=ne.memo(({props:e,isInstantUpdate:t,addNotification:n,displayName:r,id:a})=>{const[i,l]=ah(`dataServiceComponent-${a}-open`,!0),o=ym(e);return r!==""?g.jsx("div",{className:"component dataServiceComponent",id:a,children:g.jsxs(Ei,{children:[g.jsxs(Ei.Header,{onClick:()=>l(!i),style:{cursor:"pointer"},children:[r," ",i?g.jsx(mm,{}):g.jsx(vm,{})]}),g.jsx(Tc,{in:i,children:g.jsx(Ei.Body,{children:o.map(s=>g.jsx(Gi,{attribute:s,isInstantUpdate:t,addNotification:n},s.full_access_path))})})]})}):g.jsx("div",{className:"component dataServiceComponent",id:a,children:o.map(s=>g.jsx(Gi,{attribute:s,isInstantUpdate:t,addNotification:n},s.full_access_path))})});gm.displayName="DataServiceComponent";const $O=ne.memo(({fullAccessPath:e,props:t,isInstantUpdate:n,addNotification:r,displayName:a,id:i})=>{const{connected:l,connect:o,...s}=t,u=l.value;return g.jsxs("div",{className:"deviceConnectionComponent",id:i,children:[!u&&g.jsxs("div",{className:"overlayContent",children:[g.jsxs("div",{children:[a!=""?a:"Device"," is currently not available!"]}),g.jsx(hm,{fullAccessPath:`${e}.connect`,docString:o.doc,addNotification:r,displayName:"reconnect",id:i+"-connect",render:!0})]}),g.jsx(gm,{props:s,isInstantUpdate:n,addNotification:r,displayName:a,id:i})]})});$O.displayName="DeviceConnectionComponent";const NO=ne.memo(e=>{const{fullAccessPath:t,value:n,docString:r,format:a,addNotification:i,displayName:l,id:o}=e;Hn();const[s,u]=m.useState(!0);return m.useEffect(()=>{i(`${t} changed.`)},[e.value]),g.jsx("div",{className:"component imageComponent",id:o,children:g.jsxs(Ei,{children:[g.jsxs(Ei.Header,{onClick:()=>u(!s),style:{cursor:"pointer"},children:[l,g.jsx(Sn,{docString:r}),s?g.jsx(mm,{}):g.jsx(vm,{})]}),g.jsx(Tc,{in:s,children:g.jsxs(Ei.Body,{children:[!1,a===""&&n===""?g.jsx("p",{children:"No image set in the backend."}):g.jsx(ES,{src:`data:image/${a.toLowerCase()};base64,${n}`})]})})]})})});NO.displayName="ImageComponent";function h8(e){if(e){let t=e.replace(/\]\./g,"-");return t=t.replace(/[^\w_]+/g,"-"),t=t.replace(/-+$/,""),t}else return"main"}const MO=ne.memo(e=>{const{docString:t,isInstantUpdate:n,addNotification:r,id:a}=e,i=ym(e.value);return Hn(),g.jsxs("div",{className:"listComponent",id:a,children:[!1,g.jsx(Sn,{docString:t}),i.map(l=>g.jsx(Gi,{attribute:l,isInstantUpdate:n,addNotification:r},l.full_access_path))]})});MO.displayName="DictComponent";const jO=ne.memo(e=>{const{fullAccessPath:t,docString:n,status:r,addNotification:a,displayName:i,id:l}=e;Hn();const o=m.useRef(null),[s,u]=m.useState(!1);m.useEffect(()=>{let d;r==="RUNNING"?d=`${t} was started.`:d=`${t} was stopped.`,a(d),u(!1)},[r]);const c=async d=>{d.preventDefault();const f=r=="RUNNING"?"stop":"start",h=[t,f].filter(b=>b).join(".");u(!0),xE(h)};return g.jsxs("div",{className:"component taskComponent",id:l,children:[!1,g.jsx(pt,{onSubmit:c,ref:o,children:g.jsxs(dr,{children:[g.jsxs(dr.Text,{children:[i,g.jsx(Sn,{docString:n})]}),g.jsx(Lo,{id:`button-${l}`,type:"submit",children:s?g.jsx(IS,{size:"sm",role:"status","aria-hidden":"true"}):r==="RUNNING"?"Stop ":"Start "})]})})]})});jO.displayName="TaskComponent";const y8=e=>{let t="";for(const n of e)!n.startsWith("[")&&t!==""&&(t+="."),t+=n;return t},m8=e=>{const t=[],n=_E(e);for(let r=n.length-1;r>=0;r--){const a=n[r];if(t.unshift(a),!a.startsWith("["))break}return y8(t)};function ei(e,t=()=>{}){JC(e,t)}const Gi=ne.memo(({attribute:e,isInstantUpdate:t,addNotification:n})=>{const{full_access_path:r}=e,a=h8(r),i=m.useContext(Ay);let l=m8(r);if(i[r]){if(i[r].display===!1)return null;i[r].displayName&&(l=i[r].displayName)}return e.type==="bool"?g.jsx(TE,{fullAccessPath:r,docString:e.doc,readOnly:e.readonly,value:!!e.value,addNotification:n,changeCallback:ei,displayName:l,id:a}):e.type==="float"||e.type==="int"?g.jsx(Ru,{type:e.type,fullAccessPath:r,docString:e.doc,readOnly:e.readonly,value:Number(e.value),isInstantUpdate:t,addNotification:n,changeCallback:ei,displayName:l,id:a}):e.type==="Quantity"?g.jsx(Ru,{type:"Quantity",fullAccessPath:r,docString:e.doc,readOnly:e.readonly,value:Number(e.value.magnitude),unit:e.value.unit,isInstantUpdate:t,addNotification:n,changeCallback:ei,displayName:l,id:a}):e.type==="NumberSlider"?g.jsx(wO,{fullAccessPath:r,docString:e.value.value.doc,readOnly:e.readonly,value:e.value.value,min:e.value.min,max:e.value.max,stepSize:e.value.step_size,isInstantUpdate:t,addNotification:n,changeCallback:ei,displayName:l,id:a}):e.type==="Enum"||e.type==="ColouredEnum"?g.jsx(AO,{...e,addNotification:n,changeCallback:ei,displayName:l,id:a}):e.type==="method"?g.jsx(hm,{fullAccessPath:r,docString:e.doc,addNotification:n,displayName:l,id:a,render:e.frontend_render}):e.type==="str"?g.jsx(RO,{fullAccessPath:r,value:e.value,readOnly:e.readonly,docString:e.doc,isInstantUpdate:t,addNotification:n,changeCallback:ei,displayName:l,id:a}):e.type=="Task"?g.jsx(jO,{fullAccessPath:r,docString:e.doc,status:e.value.status.value,addNotification:n,displayName:l,id:a}):e.type==="DataService"?g.jsx(gm,{props:e.value,isInstantUpdate:t,addNotification:n,displayName:l,id:a}):e.type==="DeviceConnection"?g.jsx($O,{fullAccessPath:r,props:e.value,isInstantUpdate:t,addNotification:n,displayName:l,id:a}):e.type==="list"?g.jsx(CO,{value:e.value,docString:e.doc,isInstantUpdate:t,addNotification:n,id:a}):e.type==="dict"?g.jsx(MO,{value:e.value,docString:e.doc,isInstantUpdate:t,addNotification:n,id:a}):e.type==="Image"?g.jsx(NO,{fullAccessPath:r,docString:e.value.value.doc,displayName:l,id:a,addNotification:n,value:e.value.value.value,format:e.value.format.value}):g.jsx("div",{children:r},r)});Gi.displayName="GenericComponent";const v8=(e,t)=>{switch(t.type){case"SET_DATA":return t.data;case"UPDATE_ATTRIBUTE":return e===null?null:{...e,value:t5(e.value,t.fullAccessPath,t.newValue)};default:throw new Error}},g8=()=>{const[e,t]=m.useReducer(v8,null),[n,r]=m.useState(null),[a,i]=m.useState({}),[l,o]=ah("isInstantUpdate",!1),[s,u]=m.useState(!1),[c,d]=ah("showNotification",!1),[f,h]=m.useState([]),[b,S]=m.useState("connecting");m.useEffect(()=>(fetch(`${Ml}://${ms}/custom.css`,{credentials:"include"}).then(O=>{if(O.ok){const T=document.createElement("link");T.href=`${Ml}://${ms}/custom.css`,T.type="text/css",T.rel="stylesheet",document.head.appendChild(T)}}).catch(console.error),er.on("connect",()=>{fetch(`${Ml}://${ms}/service-properties`,{credentials:"include"}).then(O=>O.json()).then(O=>{t({type:"SET_DATA",data:O}),r(O.name),document.title=O.name}),fetch(`${Ml}://${ms}/web-settings`,{credentials:"include"}).then(O=>O.json()).then(O=>i(O)),S("connected")}),er.on("disconnect",()=>{S("disconnected"),setTimeout(()=>{S(O=>O==="disconnected"?"reconnecting":O)},2e3)}),er.on("notify",E),er.on("log",_),()=>{er.off("notify",E),er.off("log",_)}),[]);const x=m.useCallback((O,T="DEBUG")=>{const w=new Date().toISOString().substring(11,19),$=Math.random();h(R=>[{levelname:T,id:$,message:O,timeStamp:w},...R])},[]),y=O=>{h(T=>T.filter(w=>w.id!==O))},p=()=>u(!1),v=()=>u(!0);function E(O){const{full_access_path:T,value:w}=O.data;t({type:"UPDATE_ATTRIBUTE",fullAccessPath:T,newValue:w})}function _(O){x(O.message,O.levelname)}return e?g.jsxs(g.Fragment,{children:[g.jsx(Zf,{expand:!1,bg:"primary",variant:"dark",fixed:"top",children:g.jsxs(sS,{fluid:!0,children:[g.jsx(Zf.Brand,{children:n}),g.jsx(Zf.Toggle,{"aria-controls":"offcanvasNavbar",onClick:v})]})}),g.jsx(OE,{showNotification:c,notifications:f,removeNotificationById:y}),g.jsxs($l,{show:s,onHide:p,placement:"end",style:{zIndex:9999},children:[g.jsx($l.Header,{closeButton:!0,children:g.jsx($l.Title,{children:"Settings"})}),g.jsxs($l.Body,{children:[g.jsx(pt.Check,{checked:l,onChange:O=>o(O.target.checked),type:"switch",label:"Enable Instant Update"}),g.jsx(pt.Check,{checked:c,onChange:O=>d(O.target.checked),type:"switch",label:"Show Notifications"})]})]}),g.jsx("div",{className:"App navbarOffset",children:g.jsx(Ay.Provider,{value:a,children:g.jsx(Gi,{attribute:e,isInstantUpdate:l,addNotification:x})})}),g.jsx(wp,{connectionStatus:b})]}):g.jsx(wp,{connectionStatus:b})};var DO={exports:{}},bf={},BO={exports:{}},kO={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */(function(e){function t(C,j){var D=C.length;C.push(j);e:for(;0<D;){var K=D-1>>>1,ie=C[K];if(0<a(ie,j))C[K]=j,C[D]=ie,D=K;else break e}}function n(C){return C.length===0?null:C[0]}function r(C){if(C.length===0)return null;var j=C[0],D=C.pop();if(D!==j){C[0]=D;e:for(var K=0,ie=C.length,Pe=ie>>>1;K<Pe;){var ke=2*(K+1)-1,xe=C[ke],V=ke+1,ve=C[V];if(0>a(xe,D))V<ie&&0>a(ve,xe)?(C[K]=ve,C[V]=D,K=V):(C[K]=xe,C[ke]=D,K=ke);else if(V<ie&&0>a(ve,D))C[K]=ve,C[V]=D,K=V;else break e}}return j}function a(C,j){var D=C.sortIndex-j.sortIndex;return D!==0?D:C.id-j.id}if(e.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var i=performance;e.unstable_now=function(){return i.now()}}else{var l=Date,o=l.now();e.unstable_now=function(){return l.now()-o}}var s=[],u=[],c=1,d=null,f=3,h=!1,b=!1,S=!1,x=typeof setTimeout=="function"?setTimeout:null,y=typeof clearTimeout=="function"?clearTimeout:null,p=typeof setImmediate<"u"?setImmediate:null;function v(C){for(var j=n(u);j!==null;){if(j.callback===null)r(u);else if(j.startTime<=C)r(u),j.sortIndex=j.expirationTime,t(s,j);else break;j=n(u)}}function E(C){if(S=!1,v(C),!b)if(n(s)!==null)b=!0,X();else{var j=n(u);j!==null&&q(E,j.startTime-C)}}var _=!1,O=-1,T=5,w=-1;function $(){return!(e.unstable_now()-w<T)}function R(){if(_){var C=e.unstable_now();w=C;var j=!0;try{e:{b=!1,S&&(S=!1,y(O),O=-1),h=!0;var D=f;try{t:{for(v(C),d=n(s);d!==null&&!(d.expirationTime>C&&$());){var K=d.callback;if(typeof K=="function"){d.callback=null,f=d.priorityLevel;var ie=K(d.expirationTime<=C);if(C=e.unstable_now(),typeof ie=="function"){d.callback=ie,v(C),j=!0;break t}d===n(s)&&r(s),v(C)}else r(s);d=n(s)}if(d!==null)j=!0;else{var Pe=n(u);Pe!==null&&q(E,Pe.startTime-C),j=!1}}break e}finally{d=null,f=D,h=!1}j=void 0}}finally{j?N():_=!1}}}var N;if(typeof p=="function")N=function(){p(R)};else if(typeof MessageChannel<"u"){var G=new MessageChannel,U=G.port2;G.port1.onmessage=R,N=function(){U.postMessage(null)}}else N=function(){x(R,0)};function X(){_||(_=!0,N())}function q(C,j){O=x(function(){C(e.unstable_now())},j)}e.unstable_IdlePriority=5,e.unstable_ImmediatePriority=1,e.unstable_LowPriority=4,e.unstable_NormalPriority=3,e.unstable_Profiling=null,e.unstable_UserBlockingPriority=2,e.unstable_cancelCallback=function(C){C.callback=null},e.unstable_continueExecution=function(){b||h||(b=!0,X())},e.unstable_forceFrameRate=function(C){0>C||125<C||(T=0<C?Math.floor(1e3/C):5)},e.unstable_getCurrentPriorityLevel=function(){return f},e.unstable_getFirstCallbackNode=function(){return n(s)},e.unstable_next=function(C){switch(f){case 1:case 2:case 3:var j=3;break;default:j=f}var D=f;f=j;try{return C()}finally{f=D}},e.unstable_pauseExecution=function(){},e.unstable_requestPaint=function(){},e.unstable_runWithPriority=function(C,j){switch(C){case 1:case 2:case 3:case 4:case 5:break;default:C=3}var D=f;f=C;try{return j()}finally{f=D}},e.unstable_scheduleCallback=function(C,j,D){var K=e.unstable_now();switch(typeof D=="object"&&D!==null?(D=D.delay,D=typeof D=="number"&&0<D?K+D:K):D=K,C){case 1:var ie=-1;break;case 2:ie=250;break;case 5:ie=1073741823;break;case 4:ie=1e4;break;default:ie=5e3}return ie=D+ie,C={id:c++,callback:j,priorityLevel:C,startTime:D,expirationTime:ie,sortIndex:-1},D>K?(C.sortIndex=D,t(u,C),n(s)===null&&C===n(u)&&(S?(y(O),O=-1):S=!0,q(E,D-K))):(C.sortIndex=ie,t(s,C),b||h||(b=!0,X())),C},e.unstable_shouldYield=$,e.unstable_wrapCallback=function(C){var j=f;return function(){var D=f;f=j;try{return C.apply(this,arguments)}finally{f=D}}}})(kO);BO.exports=kO;var b8=BO.exports;/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Xe=b8,zO=m,S8=Y1;function A(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function LO(e){return!(!e||e.nodeType!==1&&e.nodeType!==9&&e.nodeType!==11)}var E8=Symbol.for("react.element"),ws=Symbol.for("react.transitional.element"),jl=Symbol.for("react.portal"),ci=Symbol.for("react.fragment"),UO=Symbol.for("react.strict_mode"),ih=Symbol.for("react.profiler"),x8=Symbol.for("react.provider"),HO=Symbol.for("react.consumer"),lr=Symbol.for("react.context"),bm=Symbol.for("react.forward_ref"),lh=Symbol.for("react.suspense"),oh=Symbol.for("react.suspense_list"),Sm=Symbol.for("react.memo"),Nr=Symbol.for("react.lazy"),qO=Symbol.for("react.offscreen"),O8=Symbol.for("react.memo_cache_sentinel"),lb=Symbol.iterator;function El(e){return e===null||typeof e!="object"?null:(e=lb&&e[lb]||e["@@iterator"],typeof e=="function"?e:null)}var _8=Symbol.for("react.client.reference");function sh(e){if(e==null)return null;if(typeof e=="function")return e.$$typeof===_8?null:e.displayName||e.name||null;if(typeof e=="string")return e;switch(e){case ci:return"Fragment";case jl:return"Portal";case ih:return"Profiler";case UO:return"StrictMode";case lh:return"Suspense";case oh:return"SuspenseList"}if(typeof e=="object")switch(e.$$typeof){case lr:return(e.displayName||"Context")+".Provider";case HO:return(e._context.displayName||"Context")+".Consumer";case bm:var t=e.render;return e=e.displayName,e||(e=t.displayName||t.name||"",e=e!==""?"ForwardRef("+e+")":"ForwardRef"),e;case Sm:return t=e.displayName||null,t!==null?t:sh(e.type)||"Memo";case Nr:t=e._payload,e=e._init;try{return sh(e(t))}catch{}}return null}var P=zO.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,Ee=Object.assign,$d,ob;function Dl(e){if($d===void 0)try{throw Error()}catch(n){var t=n.stack.trim().match(/\n( *(at )?)/);$d=t&&t[1]||"",ob=-1<n.stack.indexOf(`
    at`)?" (<anonymous>)":-1<n.stack.indexOf("@")?"@unknown:0:0":""}return`
`+$d+e+ob}var Nd=!1;function Md(e,t){if(!e||Nd)return"";Nd=!0;var n=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var r={DetermineComponentFrameRoot:function(){try{if(t){var d=function(){throw Error()};if(Object.defineProperty(d.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(d,[])}catch(h){var f=h}Reflect.construct(e,[],d)}else{try{d.call()}catch(h){f=h}e.call(d.prototype)}}else{try{throw Error()}catch(h){f=h}(d=e())&&typeof d.catch=="function"&&d.catch(function(){})}}catch(h){if(h&&f&&typeof h.stack=="string")return[h.stack,f.stack]}return[null,null]}};r.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var a=Object.getOwnPropertyDescriptor(r.DetermineComponentFrameRoot,"name");a&&a.configurable&&Object.defineProperty(r.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var i=r.DetermineComponentFrameRoot(),l=i[0],o=i[1];if(l&&o){var s=l.split(`
`),u=o.split(`
`);for(a=r=0;r<s.length&&!s[r].includes("DetermineComponentFrameRoot");)r++;for(;a<u.length&&!u[a].includes("DetermineComponentFrameRoot");)a++;if(r===s.length||a===u.length)for(r=s.length-1,a=u.length-1;1<=r&&0<=a&&s[r]!==u[a];)a--;for(;1<=r&&0<=a;r--,a--)if(s[r]!==u[a]){if(r!==1||a!==1)do if(r--,a--,0>a||s[r]!==u[a]){var c=`
`+s[r].replace(" at new "," at ");return e.displayName&&c.includes("<anonymous>")&&(c=c.replace("<anonymous>",e.displayName)),c}while(1<=r&&0<=a);break}}}finally{Nd=!1,Error.prepareStackTrace=n}return(n=e?e.displayName||e.name:"")?Dl(n):""}function T8(e){switch(e.tag){case 26:case 27:case 5:return Dl(e.type);case 16:return Dl("Lazy");case 13:return Dl("Suspense");case 19:return Dl("SuspenseList");case 0:case 15:return e=Md(e.type,!1),e;case 11:return e=Md(e.type.render,!1),e;case 1:return e=Md(e.type,!0),e;default:return""}}function sb(e){try{var t="";do t+=T8(e),e=e.return;while(e);return t}catch(n){return`
Error generating stack: `+n.message+`
`+n.stack}}function il(e){var t=e,n=e;if(e.alternate)for(;t.return;)t=t.return;else{e=t;do t=e,t.flags&4098&&(n=t.return),e=t.return;while(e)}return t.tag===3?n:null}function GO(e){if(e.tag===13){var t=e.memoizedState;if(t===null&&(e=e.alternate,e!==null&&(t=e.memoizedState)),t!==null)return t.dehydrated}return null}function ub(e){if(il(e)!==e)throw Error(A(188))}function w8(e){var t=e.alternate;if(!t){if(t=il(e),t===null)throw Error(A(188));return t!==e?null:e}for(var n=e,r=t;;){var a=n.return;if(a===null)break;var i=a.alternate;if(i===null){if(r=a.return,r!==null){n=r;continue}break}if(a.child===i.child){for(i=a.child;i;){if(i===n)return ub(a),e;if(i===r)return ub(a),t;i=i.sibling}throw Error(A(188))}if(n.return!==r.return)n=a,r=i;else{for(var l=!1,o=a.child;o;){if(o===n){l=!0,n=a,r=i;break}if(o===r){l=!0,r=a,n=i;break}o=o.sibling}if(!l){for(o=i.child;o;){if(o===n){l=!0,n=i,r=a;break}if(o===r){l=!0,r=i,n=a;break}o=o.sibling}if(!l)throw Error(A(189))}}if(n.alternate!==r)throw Error(A(190))}if(n.tag!==3)throw Error(A(188));return n.stateNode.current===n?e:t}function YO(e){var t=e.tag;if(t===5||t===26||t===27||t===6)return e;for(e=e.child;e!==null;){if(t=YO(e),t!==null)return t;e=e.sibling}return null}var Bl=Array.isArray,Se=S8.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,ma={pending:!1,data:null,method:null,action:null},uh=[],fi=-1;function Gn(e){return{current:e}}function We(e){0>fi||(e.current=uh[fi],uh[fi]=null,fi--)}function Te(e,t){fi++,uh[fi]=e.current,e.current=t}var Bn=Gn(null),Eo=Gn(null),Yr=Gn(null),Fu=Gn(null);function Qu(e,t){switch(Te(Yr,t),Te(Eo,e),Te(Bn,null),e=t.nodeType,e){case 9:case 11:t=(t=t.documentElement)&&(t=t.namespaceURI)?c1(t):0;break;default:if(e=e===8?t.parentNode:t,t=e.tagName,e=e.namespaceURI)e=c1(e),t=tT(e,t);else switch(t){case"svg":t=1;break;case"math":t=2;break;default:t=0}}We(Bn),Te(Bn,t)}function Yi(){We(Bn),We(Eo),We(Yr)}function ch(e){e.memoizedState!==null&&Te(Fu,e);var t=Bn.current,n=tT(t,e.type);t!==n&&(Te(Eo,e),Te(Bn,n))}function Ku(e){Eo.current===e&&(We(Bn),We(Eo)),Fu.current===e&&(We(Fu),Mo._currentValue=ma)}var fh=Object.prototype.hasOwnProperty,Em=Xe.unstable_scheduleCallback,jd=Xe.unstable_cancelCallback,A8=Xe.unstable_shouldYield,R8=Xe.unstable_requestPaint,kn=Xe.unstable_now,C8=Xe.unstable_getCurrentPriorityLevel,VO=Xe.unstable_ImmediatePriority,XO=Xe.unstable_UserBlockingPriority,Zu=Xe.unstable_NormalPriority,$8=Xe.unstable_LowPriority,PO=Xe.unstable_IdlePriority,N8=Xe.log,M8=Xe.unstable_setDisableYieldValue,Jo=null,Nt=null;function j8(e){if(Nt&&typeof Nt.onCommitFiberRoot=="function")try{Nt.onCommitFiberRoot(Jo,e,void 0,(e.current.flags&128)===128)}catch{}}function Ur(e){if(typeof N8=="function"&&M8(e),Nt&&typeof Nt.setStrictMode=="function")try{Nt.setStrictMode(Jo,e)}catch{}}var Mt=Math.clz32?Math.clz32:k8,D8=Math.log,B8=Math.LN2;function k8(e){return e>>>=0,e===0?32:31-(D8(e)/B8|0)|0}var As=128,Rs=4194304;function sa(e){var t=e&42;if(t!==0)return t;switch(e&-e){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return e&4194176;case 4194304:case 8388608:case 16777216:case 33554432:return e&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return e}}function Sf(e,t){var n=e.pendingLanes;if(n===0)return 0;var r=0,a=e.suspendedLanes,i=e.pingedLanes,l=e.warmLanes;e=e.finishedLanes!==0;var o=n&134217727;return o!==0?(n=o&~a,n!==0?r=sa(n):(i&=o,i!==0?r=sa(i):e||(l=o&~l,l!==0&&(r=sa(l))))):(o=n&~a,o!==0?r=sa(o):i!==0?r=sa(i):e||(l=n&~l,l!==0&&(r=sa(l)))),r===0?0:t!==0&&t!==r&&!(t&a)&&(a=r&-r,l=t&-t,a>=l||a===32&&(l&4194176)!==0)?t:r}function Io(e,t){return(e.pendingLanes&~(e.suspendedLanes&~e.pingedLanes)&t)===0}function z8(e,t){switch(e){case 1:case 2:case 4:case 8:return t+250;case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return t+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function FO(){var e=As;return As<<=1,!(As&4194176)&&(As=128),e}function QO(){var e=Rs;return Rs<<=1,!(Rs&62914560)&&(Rs=4194304),e}function Dd(e){for(var t=[],n=0;31>n;n++)t.push(e);return t}function es(e,t){e.pendingLanes|=t,t!==268435456&&(e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0)}function L8(e,t,n,r,a,i){var l=e.pendingLanes;e.pendingLanes=n,e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0,e.expiredLanes&=n,e.entangledLanes&=n,e.errorRecoveryDisabledLanes&=n,e.shellSuspendCounter=0;var o=e.entanglements,s=e.expirationTimes,u=e.hiddenUpdates;for(n=l&~n;0<n;){var c=31-Mt(n),d=1<<c;o[c]=0,s[c]=-1;var f=u[c];if(f!==null)for(u[c]=null,c=0;c<f.length;c++){var h=f[c];h!==null&&(h.lane&=-536870913)}n&=~d}r!==0&&KO(e,r,0),i!==0&&a===0&&e.tag!==0&&(e.suspendedLanes|=i&~(l&~t))}function KO(e,t,n){e.pendingLanes|=t,e.suspendedLanes&=~t;var r=31-Mt(t);e.entangledLanes|=t,e.entanglements[r]=e.entanglements[r]|1073741824|n&4194218}function ZO(e,t){var n=e.entangledLanes|=t;for(e=e.entanglements;n;){var r=31-Mt(n),a=1<<r;a&t|e[r]&t&&(e[r]|=t),n&=~a}}function WO(e){return e&=-e,2<e?8<e?e&134217727?32:268435456:8:2}function JO(){var e=Se.p;return e!==0?e:(e=window.event,e===void 0?32:fT(e.type))}function U8(e,t){var n=Se.p;try{return Se.p=e,t()}finally{Se.p=n}}var aa=Math.random().toString(36).slice(2),st="__reactFiber$"+aa,xt="__reactProps$"+aa,ll="__reactContainer$"+aa,dh="__reactEvents$"+aa,H8="__reactListeners$"+aa,q8="__reactHandles$"+aa,cb="__reactResources$"+aa,xo="__reactMarker$"+aa;function xm(e){delete e[st],delete e[xt],delete e[dh],delete e[H8],delete e[q8]}function da(e){var t=e[st];if(t)return t;for(var n=e.parentNode;n;){if(t=n[ll]||n[st]){if(n=t.alternate,t.child!==null||n!==null&&n.child!==null)for(e=d1(e);e!==null;){if(n=e[st])return n;e=d1(e)}return t}e=n,n=e.parentNode}return null}function ol(e){if(e=e[st]||e[ll]){var t=e.tag;if(t===5||t===6||t===13||t===26||t===27||t===3)return e}return null}function kl(e){var t=e.tag;if(t===5||t===26||t===27||t===6)return e.stateNode;throw Error(A(33))}function Ti(e){var t=e[cb];return t||(t=e[cb]={hoistableStyles:new Map,hoistableScripts:new Map}),t}function Ke(e){e[xo]=!0}var IO=new Set,e_={};function ja(e,t){Vi(e,t),Vi(e+"Capture",t)}function Vi(e,t){for(e_[e]=t,e=0;e<t.length;e++)IO.add(t[e])}var vr=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),G8=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),fb={},db={};function Y8(e){return fh.call(db,e)?!0:fh.call(fb,e)?!1:G8.test(e)?db[e]=!0:(fb[e]=!0,!1)}function cu(e,t,n){if(Y8(t))if(n===null)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":e.removeAttribute(t);return;case"boolean":var r=t.toLowerCase().slice(0,5);if(r!=="data-"&&r!=="aria-"){e.removeAttribute(t);return}}e.setAttribute(t,""+n)}}function Cs(e,t,n){if(n===null)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":case"boolean":e.removeAttribute(t);return}e.setAttribute(t,""+n)}}function Kn(e,t,n,r){if(r===null)e.removeAttribute(n);else{switch(typeof r){case"undefined":case"function":case"symbol":case"boolean":e.removeAttribute(n);return}e.setAttributeNS(t,n,""+r)}}function Ft(e){switch(typeof e){case"bigint":case"boolean":case"number":case"string":case"undefined":return e;case"object":return e;default:return""}}function t_(e){var t=e.type;return(e=e.nodeName)&&e.toLowerCase()==="input"&&(t==="checkbox"||t==="radio")}function V8(e){var t=t_(e)?"checked":"value",n=Object.getOwnPropertyDescriptor(e.constructor.prototype,t),r=""+e[t];if(!e.hasOwnProperty(t)&&typeof n<"u"&&typeof n.get=="function"&&typeof n.set=="function"){var a=n.get,i=n.set;return Object.defineProperty(e,t,{configurable:!0,get:function(){return a.call(this)},set:function(l){r=""+l,i.call(this,l)}}),Object.defineProperty(e,t,{enumerable:n.enumerable}),{getValue:function(){return r},setValue:function(l){r=""+l},stopTracking:function(){e._valueTracker=null,delete e[t]}}}}function Wu(e){e._valueTracker||(e._valueTracker=V8(e))}function n_(e){if(!e)return!1;var t=e._valueTracker;if(!t)return!0;var n=t.getValue(),r="";return e&&(r=t_(e)?e.checked?"true":"false":e.value),e=r,e!==n?(t.setValue(e),!0):!1}function Ju(e){if(e=e||(typeof document<"u"?document:void 0),typeof e>"u")return null;try{return e.activeElement||e.body}catch{return e.body}}var X8=/[\n"\\]/g;function Zt(e){return e.replace(X8,function(t){return"\\"+t.charCodeAt(0).toString(16)+" "})}function ph(e,t,n,r,a,i,l,o){e.name="",l!=null&&typeof l!="function"&&typeof l!="symbol"&&typeof l!="boolean"?e.type=l:e.removeAttribute("type"),t!=null?l==="number"?(t===0&&e.value===""||e.value!=t)&&(e.value=""+Ft(t)):e.value!==""+Ft(t)&&(e.value=""+Ft(t)):l!=="submit"&&l!=="reset"||e.removeAttribute("value"),t!=null?hh(e,l,Ft(t)):n!=null?hh(e,l,Ft(n)):r!=null&&e.removeAttribute("value"),a==null&&i!=null&&(e.defaultChecked=!!i),a!=null&&(e.checked=a&&typeof a!="function"&&typeof a!="symbol"),o!=null&&typeof o!="function"&&typeof o!="symbol"&&typeof o!="boolean"?e.name=""+Ft(o):e.removeAttribute("name")}function r_(e,t,n,r,a,i,l,o){if(i!=null&&typeof i!="function"&&typeof i!="symbol"&&typeof i!="boolean"&&(e.type=i),t!=null||n!=null){if(!(i!=="submit"&&i!=="reset"||t!=null))return;n=n!=null?""+Ft(n):"",t=t!=null?""+Ft(t):n,o||t===e.value||(e.value=t),e.defaultValue=t}r=r??a,r=typeof r!="function"&&typeof r!="symbol"&&!!r,e.checked=o?e.checked:!!r,e.defaultChecked=!!r,l!=null&&typeof l!="function"&&typeof l!="symbol"&&typeof l!="boolean"&&(e.name=l)}function hh(e,t,n){t==="number"&&Ju(e.ownerDocument)===e||e.defaultValue===""+n||(e.defaultValue=""+n)}function wi(e,t,n,r){if(e=e.options,t){t={};for(var a=0;a<n.length;a++)t["$"+n[a]]=!0;for(n=0;n<e.length;n++)a=t.hasOwnProperty("$"+e[n].value),e[n].selected!==a&&(e[n].selected=a),a&&r&&(e[n].defaultSelected=!0)}else{for(n=""+Ft(n),t=null,a=0;a<e.length;a++){if(e[a].value===n){e[a].selected=!0,r&&(e[a].defaultSelected=!0);return}t!==null||e[a].disabled||(t=e[a])}t!==null&&(t.selected=!0)}}function a_(e,t,n){if(t!=null&&(t=""+Ft(t),t!==e.value&&(e.value=t),n==null)){e.defaultValue!==t&&(e.defaultValue=t);return}e.defaultValue=n!=null?""+Ft(n):""}function i_(e,t,n,r){if(t==null){if(r!=null){if(n!=null)throw Error(A(92));if(Bl(r)){if(1<r.length)throw Error(A(93));r=r[0]}n=r}n==null&&(n=""),t=n}n=Ft(t),e.defaultValue=n,r=e.textContent,r===n&&r!==""&&r!==null&&(e.value=r)}function Xi(e,t){if(t){var n=e.firstChild;if(n&&n===e.lastChild&&n.nodeType===3){n.nodeValue=t;return}}e.textContent=t}var P8=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function pb(e,t,n){var r=t.indexOf("--")===0;n==null||typeof n=="boolean"||n===""?r?e.setProperty(t,""):t==="float"?e.cssFloat="":e[t]="":r?e.setProperty(t,n):typeof n!="number"||n===0||P8.has(t)?t==="float"?e.cssFloat=n:e[t]=(""+n).trim():e[t]=n+"px"}function l_(e,t,n){if(t!=null&&typeof t!="object")throw Error(A(62));if(e=e.style,n!=null){for(var r in n)!n.hasOwnProperty(r)||t!=null&&t.hasOwnProperty(r)||(r.indexOf("--")===0?e.setProperty(r,""):r==="float"?e.cssFloat="":e[r]="");for(var a in t)r=t[a],t.hasOwnProperty(a)&&n[a]!==r&&pb(e,a,r)}else for(var i in t)t.hasOwnProperty(i)&&pb(e,i,t[i])}function Om(e){if(e.indexOf("-")===-1)return!1;switch(e){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var F8=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),Q8=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function fu(e){return Q8.test(""+e)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":e}var yh=null;function _m(e){return e=e.target||e.srcElement||window,e.correspondingUseElement&&(e=e.correspondingUseElement),e.nodeType===3?e.parentNode:e}var di=null,Ai=null;function hb(e){var t=ol(e);if(t&&(e=t.stateNode)){var n=e[xt]||null;e:switch(e=t.stateNode,t.type){case"input":if(ph(e,n.value,n.defaultValue,n.defaultValue,n.checked,n.defaultChecked,n.type,n.name),t=n.name,n.type==="radio"&&t!=null){for(n=e;n.parentNode;)n=n.parentNode;for(n=n.querySelectorAll('input[name="'+Zt(""+t)+'"][type="radio"]'),t=0;t<n.length;t++){var r=n[t];if(r!==e&&r.form===e.form){var a=r[xt]||null;if(!a)throw Error(A(90));ph(r,a.value,a.defaultValue,a.defaultValue,a.checked,a.defaultChecked,a.type,a.name)}}for(t=0;t<n.length;t++)r=n[t],r.form===e.form&&n_(r)}break e;case"textarea":a_(e,n.value,n.defaultValue);break e;case"select":t=n.value,t!=null&&wi(e,!!n.multiple,t,!1)}}}var Bd=!1;function o_(e,t,n){if(Bd)return e(t,n);Bd=!0;try{var r=e(t);return r}finally{if(Bd=!1,(di!==null||Ai!==null)&&($f(),di&&(t=di,e=Ai,Ai=di=null,hb(t),e)))for(t=0;t<e.length;t++)hb(e[t])}}function Oo(e,t){var n=e.stateNode;if(n===null)return null;var r=n[xt]||null;if(r===null)return null;n=r[t];e:switch(t){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(r=!r.disabled)||(e=e.type,r=!(e==="button"||e==="input"||e==="select"||e==="textarea")),e=!r;break e;default:e=!1}if(e)return null;if(n&&typeof n!="function")throw Error(A(231,t,typeof n));return n}var mh=!1;if(vr)try{var xl={};Object.defineProperty(xl,"passive",{get:function(){mh=!0}}),window.addEventListener("test",xl,xl),window.removeEventListener("test",xl,xl)}catch{mh=!1}var Hr=null,Tm=null,du=null;function s_(){if(du)return du;var e,t=Tm,n=t.length,r,a="value"in Hr?Hr.value:Hr.textContent,i=a.length;for(e=0;e<n&&t[e]===a[e];e++);var l=n-e;for(r=1;r<=l&&t[n-r]===a[i-r];r++);return du=a.slice(e,1<r?1-r:void 0)}function pu(e){var t=e.keyCode;return"charCode"in e?(e=e.charCode,e===0&&t===13&&(e=13)):e=t,e===10&&(e=13),32<=e||e===13?e:0}function $s(){return!0}function yb(){return!1}function Ot(e){function t(n,r,a,i,l){this._reactName=n,this._targetInst=a,this.type=r,this.nativeEvent=i,this.target=l,this.currentTarget=null;for(var o in e)e.hasOwnProperty(o)&&(n=e[o],this[o]=n?n(i):i[o]);return this.isDefaultPrevented=(i.defaultPrevented!=null?i.defaultPrevented:i.returnValue===!1)?$s:yb,this.isPropagationStopped=yb,this}return Ee(t.prototype,{preventDefault:function(){this.defaultPrevented=!0;var n=this.nativeEvent;n&&(n.preventDefault?n.preventDefault():typeof n.returnValue!="unknown"&&(n.returnValue=!1),this.isDefaultPrevented=$s)},stopPropagation:function(){var n=this.nativeEvent;n&&(n.stopPropagation?n.stopPropagation():typeof n.cancelBubble!="unknown"&&(n.cancelBubble=!0),this.isPropagationStopped=$s)},persist:function(){},isPersistent:$s}),t}var Da={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(e){return e.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},Ef=Ot(Da),ts=Ee({},Da,{view:0,detail:0}),K8=Ot(ts),kd,zd,Ol,xf=Ee({},ts,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:wm,button:0,buttons:0,relatedTarget:function(e){return e.relatedTarget===void 0?e.fromElement===e.srcElement?e.toElement:e.fromElement:e.relatedTarget},movementX:function(e){return"movementX"in e?e.movementX:(e!==Ol&&(Ol&&e.type==="mousemove"?(kd=e.screenX-Ol.screenX,zd=e.screenY-Ol.screenY):zd=kd=0,Ol=e),kd)},movementY:function(e){return"movementY"in e?e.movementY:zd}}),mb=Ot(xf),Z8=Ee({},xf,{dataTransfer:0}),W8=Ot(Z8),J8=Ee({},ts,{relatedTarget:0}),Ld=Ot(J8),I8=Ee({},Da,{animationName:0,elapsedTime:0,pseudoElement:0}),ek=Ot(I8),tk=Ee({},Da,{clipboardData:function(e){return"clipboardData"in e?e.clipboardData:window.clipboardData}}),nk=Ot(tk),rk=Ee({},Da,{data:0}),vb=Ot(rk),ak={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},ik={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},lk={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function ok(e){var t=this.nativeEvent;return t.getModifierState?t.getModifierState(e):(e=lk[e])?!!t[e]:!1}function wm(){return ok}var sk=Ee({},ts,{key:function(e){if(e.key){var t=ak[e.key]||e.key;if(t!=="Unidentified")return t}return e.type==="keypress"?(e=pu(e),e===13?"Enter":String.fromCharCode(e)):e.type==="keydown"||e.type==="keyup"?ik[e.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:wm,charCode:function(e){return e.type==="keypress"?pu(e):0},keyCode:function(e){return e.type==="keydown"||e.type==="keyup"?e.keyCode:0},which:function(e){return e.type==="keypress"?pu(e):e.type==="keydown"||e.type==="keyup"?e.keyCode:0}}),uk=Ot(sk),ck=Ee({},xf,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),gb=Ot(ck),fk=Ee({},ts,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:wm}),dk=Ot(fk),pk=Ee({},Da,{propertyName:0,elapsedTime:0,pseudoElement:0}),hk=Ot(pk),yk=Ee({},xf,{deltaX:function(e){return"deltaX"in e?e.deltaX:"wheelDeltaX"in e?-e.wheelDeltaX:0},deltaY:function(e){return"deltaY"in e?e.deltaY:"wheelDeltaY"in e?-e.wheelDeltaY:"wheelDelta"in e?-e.wheelDelta:0},deltaZ:0,deltaMode:0}),mk=Ot(yk),vk=Ee({},Da,{newState:0,oldState:0}),gk=Ot(vk),bk=[9,13,27,32],Am=vr&&"CompositionEvent"in window,Ql=null;vr&&"documentMode"in document&&(Ql=document.documentMode);var Sk=vr&&"TextEvent"in window&&!Ql,u_=vr&&(!Am||Ql&&8<Ql&&11>=Ql),bb=" ",Sb=!1;function c_(e,t){switch(e){case"keyup":return bk.indexOf(t.keyCode)!==-1;case"keydown":return t.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function f_(e){return e=e.detail,typeof e=="object"&&"data"in e?e.data:null}var pi=!1;function Ek(e,t){switch(e){case"compositionend":return f_(t);case"keypress":return t.which!==32?null:(Sb=!0,bb);case"textInput":return e=t.data,e===bb&&Sb?null:e;default:return null}}function xk(e,t){if(pi)return e==="compositionend"||!Am&&c_(e,t)?(e=s_(),du=Tm=Hr=null,pi=!1,e):null;switch(e){case"paste":return null;case"keypress":if(!(t.ctrlKey||t.altKey||t.metaKey)||t.ctrlKey&&t.altKey){if(t.char&&1<t.char.length)return t.char;if(t.which)return String.fromCharCode(t.which)}return null;case"compositionend":return u_&&t.locale!=="ko"?null:t.data;default:return null}}var Ok={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Eb(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return t==="input"?!!Ok[e.type]:t==="textarea"}function d_(e,t,n,r){di?Ai?Ai.push(r):Ai=[r]:di=r,t=hc(t,"onChange"),0<t.length&&(n=new Ef("onChange","change",null,n,r),e.push({event:n,listeners:t}))}var Kl=null,_o=null;function _k(e){J2(e,0)}function Of(e){var t=kl(e);if(n_(t))return e}function xb(e,t){if(e==="change")return t}var p_=!1;if(vr){var Ud;if(vr){var Hd="oninput"in document;if(!Hd){var Ob=document.createElement("div");Ob.setAttribute("oninput","return;"),Hd=typeof Ob.oninput=="function"}Ud=Hd}else Ud=!1;p_=Ud&&(!document.documentMode||9<document.documentMode)}function _b(){Kl&&(Kl.detachEvent("onpropertychange",h_),_o=Kl=null)}function h_(e){if(e.propertyName==="value"&&Of(_o)){var t=[];d_(t,_o,e,_m(e)),o_(_k,t)}}function Tk(e,t,n){e==="focusin"?(_b(),Kl=t,_o=n,Kl.attachEvent("onpropertychange",h_)):e==="focusout"&&_b()}function wk(e){if(e==="selectionchange"||e==="keyup"||e==="keydown")return Of(_o)}function Ak(e,t){if(e==="click")return Of(t)}function Rk(e,t){if(e==="input"||e==="change")return Of(t)}function Ck(e,t){return e===t&&(e!==0||1/e===1/t)||e!==e&&t!==t}var Bt=typeof Object.is=="function"?Object.is:Ck;function To(e,t){if(Bt(e,t))return!0;if(typeof e!="object"||e===null||typeof t!="object"||t===null)return!1;var n=Object.keys(e),r=Object.keys(t);if(n.length!==r.length)return!1;for(r=0;r<n.length;r++){var a=n[r];if(!fh.call(t,a)||!Bt(e[a],t[a]))return!1}return!0}function Tb(e){for(;e&&e.firstChild;)e=e.firstChild;return e}function wb(e,t){var n=Tb(e);e=0;for(var r;n;){if(n.nodeType===3){if(r=e+n.textContent.length,e<=t&&r>=t)return{node:n,offset:t-e};e=r}e:{for(;n;){if(n.nextSibling){n=n.nextSibling;break e}n=n.parentNode}n=void 0}n=Tb(n)}}function y_(e,t){return e&&t?e===t?!0:e&&e.nodeType===3?!1:t&&t.nodeType===3?y_(e,t.parentNode):"contains"in e?e.contains(t):e.compareDocumentPosition?!!(e.compareDocumentPosition(t)&16):!1:!1}function m_(e){e=e!=null&&e.ownerDocument!=null&&e.ownerDocument.defaultView!=null?e.ownerDocument.defaultView:window;for(var t=Ju(e.document);t instanceof e.HTMLIFrameElement;){try{var n=typeof t.contentWindow.location.href=="string"}catch{n=!1}if(n)e=t.contentWindow;else break;t=Ju(e.document)}return t}function Rm(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return t&&(t==="input"&&(e.type==="text"||e.type==="search"||e.type==="tel"||e.type==="url"||e.type==="password")||t==="textarea"||e.contentEditable==="true")}function $k(e,t){var n=m_(t);t=e.focusedElem;var r=e.selectionRange;if(n!==t&&t&&t.ownerDocument&&y_(t.ownerDocument.documentElement,t)){if(r!==null&&Rm(t)){if(e=r.start,n=r.end,n===void 0&&(n=e),"selectionStart"in t)t.selectionStart=e,t.selectionEnd=Math.min(n,t.value.length);else if(n=(e=t.ownerDocument||document)&&e.defaultView||window,n.getSelection){n=n.getSelection();var a=t.textContent.length,i=Math.min(r.start,a);r=r.end===void 0?i:Math.min(r.end,a),!n.extend&&i>r&&(a=r,r=i,i=a),a=wb(t,i);var l=wb(t,r);a&&l&&(n.rangeCount!==1||n.anchorNode!==a.node||n.anchorOffset!==a.offset||n.focusNode!==l.node||n.focusOffset!==l.offset)&&(e=e.createRange(),e.setStart(a.node,a.offset),n.removeAllRanges(),i>r?(n.addRange(e),n.extend(l.node,l.offset)):(e.setEnd(l.node,l.offset),n.addRange(e)))}}for(e=[],n=t;n=n.parentNode;)n.nodeType===1&&e.push({element:n,left:n.scrollLeft,top:n.scrollTop});for(typeof t.focus=="function"&&t.focus(),t=0;t<e.length;t++)n=e[t],n.element.scrollLeft=n.left,n.element.scrollTop=n.top}}var Nk=vr&&"documentMode"in document&&11>=document.documentMode,hi=null,vh=null,Zl=null,gh=!1;function Ab(e,t,n){var r=n.window===n?n.document:n.nodeType===9?n:n.ownerDocument;gh||hi==null||hi!==Ju(r)||(r=hi,"selectionStart"in r&&Rm(r)?r={start:r.selectionStart,end:r.selectionEnd}:(r=(r.ownerDocument&&r.ownerDocument.defaultView||window).getSelection(),r={anchorNode:r.anchorNode,anchorOffset:r.anchorOffset,focusNode:r.focusNode,focusOffset:r.focusOffset}),Zl&&To(Zl,r)||(Zl=r,r=hc(vh,"onSelect"),0<r.length&&(t=new Ef("onSelect","select",null,t,n),e.push({event:t,listeners:r}),t.target=hi)))}function oa(e,t){var n={};return n[e.toLowerCase()]=t.toLowerCase(),n["Webkit"+e]="webkit"+t,n["Moz"+e]="moz"+t,n}var yi={animationend:oa("Animation","AnimationEnd"),animationiteration:oa("Animation","AnimationIteration"),animationstart:oa("Animation","AnimationStart"),transitionrun:oa("Transition","TransitionRun"),transitionstart:oa("Transition","TransitionStart"),transitioncancel:oa("Transition","TransitionCancel"),transitionend:oa("Transition","TransitionEnd")},qd={},v_={};vr&&(v_=document.createElement("div").style,"AnimationEvent"in window||(delete yi.animationend.animation,delete yi.animationiteration.animation,delete yi.animationstart.animation),"TransitionEvent"in window||delete yi.transitionend.transition);function Ba(e){if(qd[e])return qd[e];if(!yi[e])return e;var t=yi[e],n;for(n in t)if(t.hasOwnProperty(n)&&n in v_)return qd[e]=t[n];return e}var g_=Ba("animationend"),b_=Ba("animationiteration"),S_=Ba("animationstart"),Mk=Ba("transitionrun"),jk=Ba("transitionstart"),Dk=Ba("transitioncancel"),E_=Ba("transitionend"),x_=new Map,Rb="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll scrollEnd toggle touchMove waiting wheel".split(" ");function En(e,t){x_.set(e,t),ja(t,[e])}var Yt=[],mi=0,Cm=0;function _f(){for(var e=mi,t=Cm=mi=0;t<e;){var n=Yt[t];Yt[t++]=null;var r=Yt[t];Yt[t++]=null;var a=Yt[t];Yt[t++]=null;var i=Yt[t];if(Yt[t++]=null,r!==null&&a!==null){var l=r.pending;l===null?a.next=a:(a.next=l.next,l.next=a),r.pending=a}i!==0&&O_(n,a,i)}}function Tf(e,t,n,r){Yt[mi++]=e,Yt[mi++]=t,Yt[mi++]=n,Yt[mi++]=r,Cm|=r,e.lanes|=r,e=e.alternate,e!==null&&(e.lanes|=r)}function $m(e,t,n,r){return Tf(e,t,n,r),Iu(e)}function Jr(e,t){return Tf(e,null,null,t),Iu(e)}function O_(e,t,n){e.lanes|=n;var r=e.alternate;r!==null&&(r.lanes|=n);for(var a=!1,i=e.return;i!==null;)i.childLanes|=n,r=i.alternate,r!==null&&(r.childLanes|=n),i.tag===22&&(e=i.stateNode,e===null||e._visibility&1||(a=!0)),e=i,i=i.return;a&&t!==null&&e.tag===3&&(i=e.stateNode,a=31-Mt(n),i=i.hiddenUpdates,e=i[a],e===null?i[a]=[t]:e.push(t),t.lane=n|536870912)}function Iu(e){if(50<so)throw so=0,Hh=null,Error(A(185));for(var t=e.return;t!==null;)e=t,t=e.return;return e.tag===3?e.stateNode:null}var vi={},Cb=new WeakMap;function Wt(e,t){if(typeof e=="object"&&e!==null){var n=Cb.get(e);return n!==void 0?n:(t={value:e,source:t,stack:sb(t)},Cb.set(e,t),t)}return{value:e,source:t,stack:sb(t)}}var gi=[],bi=0,ec=null,tc=0,Qt=[],Kt=0,va=null,or=1,sr="";function ua(e,t){gi[bi++]=tc,gi[bi++]=ec,ec=e,tc=t}function __(e,t,n){Qt[Kt++]=or,Qt[Kt++]=sr,Qt[Kt++]=va,va=e;var r=or;e=sr;var a=32-Mt(r)-1;r&=~(1<<a),n+=1;var i=32-Mt(t)+a;if(30<i){var l=a-a%5;i=(r&(1<<l)-1).toString(32),r>>=l,a-=l,or=1<<32-Mt(t)+a|n<<a|r,sr=i+e}else or=1<<i|n<<a|r,sr=e}function Nm(e){e.return!==null&&(ua(e,1),__(e,1,0))}function Mm(e){for(;e===ec;)ec=gi[--bi],gi[bi]=null,tc=gi[--bi],gi[bi]=null;for(;e===va;)va=Qt[--Kt],Qt[Kt]=null,sr=Qt[--Kt],Qt[Kt]=null,or=Qt[--Kt],Qt[Kt]=null}var ht=null,rt=null,ae=!1,yn=null,Rn=!1,bh=Error(A(519));function Oa(e){var t=Error(A(418,""));throw wo(Wt(t,e)),bh}function $b(e){var t=e.stateNode,n=e.type,r=e.memoizedProps;switch(t[st]=e,t[xt]=r,n){case"dialog":J("cancel",t),J("close",t);break;case"iframe":case"object":case"embed":J("load",t);break;case"video":case"audio":for(n=0;n<Co.length;n++)J(Co[n],t);break;case"source":J("error",t);break;case"img":case"image":case"link":J("error",t),J("load",t);break;case"details":J("toggle",t);break;case"input":J("invalid",t),r_(t,r.value,r.defaultValue,r.checked,r.defaultChecked,r.type,r.name,!0),Wu(t);break;case"select":J("invalid",t);break;case"textarea":J("invalid",t),i_(t,r.value,r.defaultValue,r.children),Wu(t)}n=r.children,typeof n!="string"&&typeof n!="number"&&typeof n!="bigint"||t.textContent===""+n||r.suppressHydrationWarning===!0||eT(t.textContent,n)?(r.popover!=null&&(J("beforetoggle",t),J("toggle",t)),r.onScroll!=null&&J("scroll",t),r.onScrollEnd!=null&&J("scrollend",t),r.onClick!=null&&(t.onclick=Mf),t=!0):t=!1,t||Oa(e)}function Nb(e){for(ht=e.return;ht;)switch(ht.tag){case 3:case 27:Rn=!0;return;case 5:case 13:Rn=!1;return;default:ht=ht.return}}function _l(e){if(e!==ht)return!1;if(!ae)return Nb(e),ae=!0,!1;var t=!1,n;if((n=e.tag!==3&&e.tag!==27)&&((n=e.tag===5)&&(n=e.type,n=!(n!=="form"&&n!=="button")||Fh(e.type,e.memoizedProps)),n=!n),n&&(t=!0),t&&rt&&Oa(e),Nb(e),e.tag===13){if(e=e.memoizedState,e=e!==null?e.dehydrated:null,!e)throw Error(A(317));e:{for(e=e.nextSibling,t=0;e;){if(e.nodeType===8)if(n=e.data,n==="/$"){if(t===0){rt=gn(e.nextSibling);break e}t--}else n!=="$"&&n!=="$!"&&n!=="$?"||t++;e=e.nextSibling}rt=null}}else rt=ht?gn(e.stateNode.nextSibling):null;return!0}function ns(){rt=ht=null,ae=!1}function wo(e){yn===null?yn=[e]:yn.push(e)}var Wl=Error(A(460)),T_=Error(A(474)),Sh={then:function(){}};function Mb(e){return e=e.status,e==="fulfilled"||e==="rejected"}function Ns(){}function w_(e,t,n){switch(n=e[n],n===void 0?e.push(t):n!==t&&(t.then(Ns,Ns),t=n),t.status){case"fulfilled":return t.value;case"rejected":throw e=t.reason,e===Wl?Error(A(483)):e;default:if(typeof t.status=="string")t.then(Ns,Ns);else{if(e=be,e!==null&&100<e.shellSuspendCounter)throw Error(A(482));e=t,e.status="pending",e.then(function(r){if(t.status==="pending"){var a=t;a.status="fulfilled",a.value=r}},function(r){if(t.status==="pending"){var a=t;a.status="rejected",a.reason=r}})}switch(t.status){case"fulfilled":return t.value;case"rejected":throw e=t.reason,e===Wl?Error(A(483)):e}throw Jl=t,Wl}}var Jl=null;function jb(){if(Jl===null)throw Error(A(459));var e=Jl;return Jl=null,e}var Ri=null,Ao=0;function Ms(e){var t=Ao;return Ao+=1,Ri===null&&(Ri=[]),w_(Ri,e,t)}function Tl(e,t){t=t.props.ref,e.ref=t!==void 0?t:null}function js(e,t){throw t.$$typeof===E8?Error(A(525)):(e=Object.prototype.toString.call(t),Error(A(31,e==="[object Object]"?"object with keys {"+Object.keys(t).join(", ")+"}":e)))}function Db(e){var t=e._init;return t(e._payload)}function A_(e){function t(y,p){if(e){var v=y.deletions;v===null?(y.deletions=[p],y.flags|=16):v.push(p)}}function n(y,p){if(!e)return null;for(;p!==null;)t(y,p),p=p.sibling;return null}function r(y){for(var p=new Map;y!==null;)y.key!==null?p.set(y.key,y):p.set(y.index,y),y=y.sibling;return p}function a(y,p){return y=Pr(y,p),y.index=0,y.sibling=null,y}function i(y,p,v){return y.index=v,e?(v=y.alternate,v!==null?(v=v.index,v<p?(y.flags|=33554434,p):v):(y.flags|=33554434,p)):(y.flags|=1048576,p)}function l(y){return e&&y.alternate===null&&(y.flags|=33554434),y}function o(y,p,v,E){return p===null||p.tag!==6?(p=Id(v,y.mode,E),p.return=y,p):(p=a(p,v),p.return=y,p)}function s(y,p,v,E){var _=v.type;return _===ci?c(y,p,v.props.children,E,v.key):p!==null&&(p.elementType===_||typeof _=="object"&&_!==null&&_.$$typeof===Nr&&Db(_)===p.type)?(p=a(p,v.props),Tl(p,v),p.return=y,p):(p=gu(v.type,v.key,v.props,null,y.mode,E),Tl(p,v),p.return=y,p)}function u(y,p,v,E){return p===null||p.tag!==4||p.stateNode.containerInfo!==v.containerInfo||p.stateNode.implementation!==v.implementation?(p=ep(v,y.mode,E),p.return=y,p):(p=a(p,v.children||[]),p.return=y,p)}function c(y,p,v,E,_){return p===null||p.tag!==7?(p=ba(v,y.mode,E,_),p.return=y,p):(p=a(p,v),p.return=y,p)}function d(y,p,v){if(typeof p=="string"&&p!==""||typeof p=="number"||typeof p=="bigint")return p=Id(""+p,y.mode,v),p.return=y,p;if(typeof p=="object"&&p!==null){switch(p.$$typeof){case ws:return v=gu(p.type,p.key,p.props,null,y.mode,v),Tl(v,p),v.return=y,v;case jl:return p=ep(p,y.mode,v),p.return=y,p;case Nr:var E=p._init;return p=E(p._payload),d(y,p,v)}if(Bl(p)||El(p))return p=ba(p,y.mode,v,null),p.return=y,p;if(typeof p.then=="function")return d(y,Ms(p),v);if(p.$$typeof===lr)return d(y,Ds(y,p),v);js(y,p)}return null}function f(y,p,v,E){var _=p!==null?p.key:null;if(typeof v=="string"&&v!==""||typeof v=="number"||typeof v=="bigint")return _!==null?null:o(y,p,""+v,E);if(typeof v=="object"&&v!==null){switch(v.$$typeof){case ws:return v.key===_?s(y,p,v,E):null;case jl:return v.key===_?u(y,p,v,E):null;case Nr:return _=v._init,v=_(v._payload),f(y,p,v,E)}if(Bl(v)||El(v))return _!==null?null:c(y,p,v,E,null);if(typeof v.then=="function")return f(y,p,Ms(v),E);if(v.$$typeof===lr)return f(y,p,Ds(y,v),E);js(y,v)}return null}function h(y,p,v,E,_){if(typeof E=="string"&&E!==""||typeof E=="number"||typeof E=="bigint")return y=y.get(v)||null,o(p,y,""+E,_);if(typeof E=="object"&&E!==null){switch(E.$$typeof){case ws:return y=y.get(E.key===null?v:E.key)||null,s(p,y,E,_);case jl:return y=y.get(E.key===null?v:E.key)||null,u(p,y,E,_);case Nr:var O=E._init;return E=O(E._payload),h(y,p,v,E,_)}if(Bl(E)||El(E))return y=y.get(v)||null,c(p,y,E,_,null);if(typeof E.then=="function")return h(y,p,v,Ms(E),_);if(E.$$typeof===lr)return h(y,p,v,Ds(p,E),_);js(p,E)}return null}function b(y,p,v,E){for(var _=null,O=null,T=p,w=p=0,$=null;T!==null&&w<v.length;w++){T.index>w?($=T,T=null):$=T.sibling;var R=f(y,T,v[w],E);if(R===null){T===null&&(T=$);break}e&&T&&R.alternate===null&&t(y,T),p=i(R,p,w),O===null?_=R:O.sibling=R,O=R,T=$}if(w===v.length)return n(y,T),ae&&ua(y,w),_;if(T===null){for(;w<v.length;w++)T=d(y,v[w],E),T!==null&&(p=i(T,p,w),O===null?_=T:O.sibling=T,O=T);return ae&&ua(y,w),_}for(T=r(T);w<v.length;w++)$=h(T,y,w,v[w],E),$!==null&&(e&&$.alternate!==null&&T.delete($.key===null?w:$.key),p=i($,p,w),O===null?_=$:O.sibling=$,O=$);return e&&T.forEach(function(N){return t(y,N)}),ae&&ua(y,w),_}function S(y,p,v,E){if(v==null)throw Error(A(151));for(var _=null,O=null,T=p,w=p=0,$=null,R=v.next();T!==null&&!R.done;w++,R=v.next()){T.index>w?($=T,T=null):$=T.sibling;var N=f(y,T,R.value,E);if(N===null){T===null&&(T=$);break}e&&T&&N.alternate===null&&t(y,T),p=i(N,p,w),O===null?_=N:O.sibling=N,O=N,T=$}if(R.done)return n(y,T),ae&&ua(y,w),_;if(T===null){for(;!R.done;w++,R=v.next())R=d(y,R.value,E),R!==null&&(p=i(R,p,w),O===null?_=R:O.sibling=R,O=R);return ae&&ua(y,w),_}for(T=r(T);!R.done;w++,R=v.next())R=h(T,y,w,R.value,E),R!==null&&(e&&R.alternate!==null&&T.delete(R.key===null?w:R.key),p=i(R,p,w),O===null?_=R:O.sibling=R,O=R);return e&&T.forEach(function(G){return t(y,G)}),ae&&ua(y,w),_}function x(y,p,v,E){if(typeof v=="object"&&v!==null&&v.type===ci&&v.key===null&&(v=v.props.children),typeof v=="object"&&v!==null){switch(v.$$typeof){case ws:e:{for(var _=v.key;p!==null;){if(p.key===_){if(_=v.type,_===ci){if(p.tag===7){n(y,p.sibling),E=a(p,v.props.children),E.return=y,y=E;break e}}else if(p.elementType===_||typeof _=="object"&&_!==null&&_.$$typeof===Nr&&Db(_)===p.type){n(y,p.sibling),E=a(p,v.props),Tl(E,v),E.return=y,y=E;break e}n(y,p);break}else t(y,p);p=p.sibling}v.type===ci?(E=ba(v.props.children,y.mode,E,v.key),E.return=y,y=E):(E=gu(v.type,v.key,v.props,null,y.mode,E),Tl(E,v),E.return=y,y=E)}return l(y);case jl:e:{for(_=v.key;p!==null;){if(p.key===_)if(p.tag===4&&p.stateNode.containerInfo===v.containerInfo&&p.stateNode.implementation===v.implementation){n(y,p.sibling),E=a(p,v.children||[]),E.return=y,y=E;break e}else{n(y,p);break}else t(y,p);p=p.sibling}E=ep(v,y.mode,E),E.return=y,y=E}return l(y);case Nr:return _=v._init,v=_(v._payload),x(y,p,v,E)}if(Bl(v))return b(y,p,v,E);if(El(v)){if(_=El(v),typeof _!="function")throw Error(A(150));return v=_.call(v),S(y,p,v,E)}if(typeof v.then=="function")return x(y,p,Ms(v),E);if(v.$$typeof===lr)return x(y,p,Ds(y,v),E);js(y,v)}return typeof v=="string"&&v!==""||typeof v=="number"||typeof v=="bigint"?(v=""+v,p!==null&&p.tag===6?(n(y,p.sibling),E=a(p,v),E.return=y,y=E):(n(y,p),E=Id(v,y.mode,E),E.return=y,y=E),l(y)):n(y,p)}return function(y,p,v,E){try{Ao=0;var _=x(y,p,v,E);return Ri=null,_}catch(T){if(T===Wl)throw T;var O=It(29,T,null,y.mode);return O.lanes=E,O.return=y,O}finally{}}}var _a=A_(!0),R_=A_(!1),Pi=Gn(null),nc=Gn(0);function Bb(e,t){e=Sr,Te(nc,e),Te(Pi,t),Sr=e|t.baseLanes}function Eh(){Te(nc,Sr),Te(Pi,Pi.current)}function jm(){Sr=nc.current,We(Pi),We(nc)}var an=Gn(null),zn=null;function Dr(e){var t=e.alternate;Te(Ve,Ve.current&1),Te(an,e),zn===null&&(t===null||Pi.current!==null||t.memoizedState!==null)&&(zn=e)}function C_(e){if(e.tag===22){if(Te(Ve,Ve.current),Te(an,e),zn===null){var t=e.alternate;t!==null&&t.memoizedState!==null&&(zn=e)}}else Br()}function Br(){Te(Ve,Ve.current),Te(an,an.current)}function ur(e){We(an),zn===e&&(zn=null),We(Ve)}var Ve=Gn(0);function rc(e){for(var t=e;t!==null;){if(t.tag===13){var n=t.memoizedState;if(n!==null&&(n=n.dehydrated,n===null||n.data==="$?"||n.data==="$!"))return t}else if(t.tag===19&&t.memoizedProps.revealOrder!==void 0){if(t.flags&128)return t}else if(t.child!==null){t.child.return=t,t=t.child;continue}if(t===e)break;for(;t.sibling===null;){if(t.return===null||t.return===e)return null;t=t.return}t.sibling.return=t.return,t=t.sibling}return null}var Bk=typeof AbortController<"u"?AbortController:function(){var e=[],t=this.signal={aborted:!1,addEventListener:function(n,r){e.push(r)}};this.abort=function(){t.aborted=!0,e.forEach(function(n){return n()})}},kk=Xe.unstable_scheduleCallback,zk=Xe.unstable_NormalPriority,Ye={$$typeof:lr,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function Dm(){return{controller:new Bk,data:new Map,refCount:0}}function rs(e){e.refCount--,e.refCount===0&&kk(zk,function(){e.controller.abort()})}var Il=null,xh=0,Fi=0,Ci=null;function Lk(e,t){if(Il===null){var n=Il=[];xh=0,Fi=rv(),Ci={status:"pending",value:void 0,then:function(r){n.push(r)}}}return xh++,t.then(kb,kb),t}function kb(){if(--xh===0&&Il!==null){Ci!==null&&(Ci.status="fulfilled");var e=Il;Il=null,Fi=0,Ci=null;for(var t=0;t<e.length;t++)(0,e[t])()}}function Uk(e,t){var n=[],r={status:"pending",value:null,reason:null,then:function(a){n.push(a)}};return e.then(function(){r.status="fulfilled",r.value=t;for(var a=0;a<n.length;a++)(0,n[a])(t)},function(a){for(r.status="rejected",r.reason=a,a=0;a<n.length;a++)(0,n[a])(void 0)}),r}var zb=P.S;P.S=function(e,t){typeof t=="object"&&t!==null&&typeof t.then=="function"&&Lk(e,t),zb!==null&&zb(e,t)};var ga=Gn(null);function Bm(){var e=ga.current;return e!==null?e:be.pooledCache}function hu(e,t){t===null?Te(ga,ga.current):Te(ga,t.pool)}function $_(){var e=Bm();return e===null?null:{parent:Ye._currentValue,pool:e}}var Ir=0,Q=null,pe=null,Ue=null,ac=!1,$i=!1,Ta=!1,ic=0,Ro=0,Ni=null,Hk=0;function je(){throw Error(A(321))}function km(e,t){if(t===null)return!1;for(var n=0;n<t.length&&n<e.length;n++)if(!Bt(e[n],t[n]))return!1;return!0}function zm(e,t,n,r,a,i){return Ir=i,Q=t,t.memoizedState=null,t.updateQueue=null,t.lanes=0,P.H=e===null||e.memoizedState===null?ka:ia,Ta=!1,i=n(r,a),Ta=!1,$i&&(i=M_(t,n,r,a)),N_(e),i}function N_(e){P.H=Un;var t=pe!==null&&pe.next!==null;if(Ir=0,Ue=pe=Q=null,ac=!1,Ro=0,Ni=null,t)throw Error(A(300));e===null||Ze||(e=e.dependencies,e!==null&&sc(e)&&(Ze=!0))}function M_(e,t,n,r){Q=e;var a=0;do{if($i&&(Ni=null),Ro=0,$i=!1,25<=a)throw Error(A(301));if(a+=1,Ue=pe=null,e.updateQueue!=null){var i=e.updateQueue;i.lastEffect=null,i.events=null,i.stores=null,i.memoCache!=null&&(i.memoCache.index=0)}P.H=za,i=t(n,r)}while($i);return i}function qk(){var e=P.H,t=e.useState()[0];return t=typeof t.then=="function"?as(t):t,e=e.useState()[0],(pe!==null?pe.memoizedState:null)!==e&&(Q.flags|=1024),t}function Lm(){var e=ic!==0;return ic=0,e}function Um(e,t,n){t.updateQueue=e.updateQueue,t.flags&=-2053,e.lanes&=~n}function Hm(e){if(ac){for(e=e.memoizedState;e!==null;){var t=e.queue;t!==null&&(t.pending=null),e=e.next}ac=!1}Ir=0,Ue=pe=Q=null,$i=!1,Ro=ic=0,Ni=null}function gt(){var e={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return Ue===null?Q.memoizedState=Ue=e:Ue=Ue.next=e,Ue}function Ge(){if(pe===null){var e=Q.alternate;e=e!==null?e.memoizedState:null}else e=pe.next;var t=Ue===null?Q.memoizedState:Ue.next;if(t!==null)Ue=t,pe=e;else{if(e===null)throw Q.alternate===null?Error(A(467)):Error(A(310));pe=e,e={memoizedState:pe.memoizedState,baseState:pe.baseState,baseQueue:pe.baseQueue,queue:pe.queue,next:null},Ue===null?Q.memoizedState=Ue=e:Ue=Ue.next=e}return Ue}var wf;wf=function(){return{lastEffect:null,events:null,stores:null,memoCache:null}};function as(e){var t=Ro;return Ro+=1,Ni===null&&(Ni=[]),e=w_(Ni,e,t),t=Q,(Ue===null?t.memoizedState:Ue.next)===null&&(t=t.alternate,P.H=t===null||t.memoizedState===null?ka:ia),e}function Af(e){if(e!==null&&typeof e=="object"){if(typeof e.then=="function")return as(e);if(e.$$typeof===lr)return ut(e)}throw Error(A(438,String(e)))}function qm(e){var t=null,n=Q.updateQueue;if(n!==null&&(t=n.memoCache),t==null){var r=Q.alternate;r!==null&&(r=r.updateQueue,r!==null&&(r=r.memoCache,r!=null&&(t={data:r.data.map(function(a){return a.slice()}),index:0})))}if(t==null&&(t={data:[],index:0}),n===null&&(n=wf(),Q.updateQueue=n),n.memoCache=t,n=t.data[t.index],n===void 0)for(n=t.data[t.index]=Array(e),r=0;r<e;r++)n[r]=O8;return t.index++,n}function gr(e,t){return typeof t=="function"?t(e):t}function yu(e){var t=Ge();return Gm(t,pe,e)}function Gm(e,t,n){var r=e.queue;if(r===null)throw Error(A(311));r.lastRenderedReducer=n;var a=e.baseQueue,i=r.pending;if(i!==null){if(a!==null){var l=a.next;a.next=i.next,i.next=l}t.baseQueue=a=i,r.pending=null}if(i=e.baseState,a===null)e.memoizedState=i;else{t=a.next;var o=l=null,s=null,u=t,c=!1;do{var d=u.lane&-536870913;if(d!==u.lane?(te&d)===d:(Ir&d)===d){var f=u.revertLane;if(f===0)s!==null&&(s=s.next={lane:0,revertLane:0,action:u.action,hasEagerState:u.hasEagerState,eagerState:u.eagerState,next:null}),d===Fi&&(c=!0);else if((Ir&f)===f){u=u.next,f===Fi&&(c=!0);continue}else d={lane:0,revertLane:u.revertLane,action:u.action,hasEagerState:u.hasEagerState,eagerState:u.eagerState,next:null},s===null?(o=s=d,l=i):s=s.next=d,Q.lanes|=f,ta|=f;d=u.action,Ta&&n(i,d),i=u.hasEagerState?u.eagerState:n(i,d)}else f={lane:d,revertLane:u.revertLane,action:u.action,hasEagerState:u.hasEagerState,eagerState:u.eagerState,next:null},s===null?(o=s=f,l=i):s=s.next=f,Q.lanes|=d,ta|=d;u=u.next}while(u!==null&&u!==t);if(s===null?l=i:s.next=o,!Bt(i,e.memoizedState)&&(Ze=!0,c&&(n=Ci,n!==null)))throw n;e.memoizedState=i,e.baseState=l,e.baseQueue=s,r.lastRenderedState=i}return a===null&&(r.lanes=0),[e.memoizedState,r.dispatch]}function Gd(e){var t=Ge(),n=t.queue;if(n===null)throw Error(A(311));n.lastRenderedReducer=e;var r=n.dispatch,a=n.pending,i=t.memoizedState;if(a!==null){n.pending=null;var l=a=a.next;do i=e(i,l.action),l=l.next;while(l!==a);Bt(i,t.memoizedState)||(Ze=!0),t.memoizedState=i,t.baseQueue===null&&(t.baseState=i),n.lastRenderedState=i}return[i,r]}function j_(e,t,n){var r=Q,a=Ge(),i=ae;if(i){if(n===void 0)throw Error(A(407));n=n()}else n=t();var l=!Bt((pe||a).memoizedState,n);if(l&&(a.memoizedState=n,Ze=!0),a=a.queue,Ym(k_.bind(null,r,a,e),[e]),a.getSnapshot!==t||l||Ue!==null&&Ue.memoizedState.tag&1){if(r.flags|=2048,Qi(9,B_.bind(null,r,a,n,t),{destroy:void 0},null),be===null)throw Error(A(349));i||Ir&60||D_(r,t,n)}return n}function D_(e,t,n){e.flags|=16384,e={getSnapshot:t,value:n},t=Q.updateQueue,t===null?(t=wf(),Q.updateQueue=t,t.stores=[e]):(n=t.stores,n===null?t.stores=[e]:n.push(e))}function B_(e,t,n,r){t.value=n,t.getSnapshot=r,z_(t)&&L_(e)}function k_(e,t,n){return n(function(){z_(t)&&L_(e)})}function z_(e){var t=e.getSnapshot;e=e.value;try{var n=t();return!Bt(e,n)}catch{return!0}}function L_(e){var t=Jr(e,2);t!==null&&mt(t,e,2)}function Oh(e){var t=gt();if(typeof e=="function"){var n=e;if(e=n(),Ta){Ur(!0);try{n()}finally{Ur(!1)}}}return t.memoizedState=t.baseState=e,t.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:gr,lastRenderedState:e},t}function U_(e,t,n,r){return e.baseState=n,Gm(e,pe,typeof r=="function"?r:gr)}function Gk(e,t,n,r,a){if(Cf(e))throw Error(A(485));if(e=t.action,e!==null){var i={payload:a,action:e,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(l){i.listeners.push(l)}};P.T!==null?n(!0):i.isTransition=!1,r(i),n=t.pending,n===null?(i.next=t.pending=i,H_(t,i)):(i.next=n.next,t.pending=n.next=i)}}function H_(e,t){var n=t.action,r=t.payload,a=e.state;if(t.isTransition){var i=P.T,l={};P.T=l;try{var o=n(a,r),s=P.S;s!==null&&s(l,o),Lb(e,t,o)}catch(u){_h(e,t,u)}finally{P.T=i}}else try{i=n(a,r),Lb(e,t,i)}catch(u){_h(e,t,u)}}function Lb(e,t,n){n!==null&&typeof n=="object"&&typeof n.then=="function"?n.then(function(r){Ub(e,t,r)},function(r){return _h(e,t,r)}):Ub(e,t,n)}function Ub(e,t,n){t.status="fulfilled",t.value=n,q_(t),e.state=n,t=e.pending,t!==null&&(n=t.next,n===t?e.pending=null:(n=n.next,t.next=n,H_(e,n)))}function _h(e,t,n){var r=e.pending;if(e.pending=null,r!==null){r=r.next;do t.status="rejected",t.reason=n,q_(t),t=t.next;while(t!==r)}e.action=null}function q_(e){e=e.listeners;for(var t=0;t<e.length;t++)(0,e[t])()}function G_(e,t){return t}function Y_(e,t){if(ae){var n=be.formState;if(n!==null){e:{var r=Q;if(ae){if(rt){t:{for(var a=rt,i=Rn;a.nodeType!==8;){if(!i){a=null;break t}if(a=gn(a.nextSibling),a===null){a=null;break t}}i=a.data,a=i==="F!"||i==="F"?a:null}if(a){rt=gn(a.nextSibling),r=a.data==="F!";break e}}Oa(r)}r=!1}r&&(t=n[0])}}return n=gt(),n.memoizedState=n.baseState=t,r={pending:null,lanes:0,dispatch:null,lastRenderedReducer:G_,lastRenderedState:t},n.queue=r,n=l2.bind(null,Q,r),r.dispatch=n,r=Oh(!1),i=Fm.bind(null,Q,!1,r.queue),r=gt(),a={state:t,dispatch:null,action:e,pending:null},r.queue=a,n=Gk.bind(null,Q,a,i,n),a.dispatch=n,r.memoizedState=e,[t,n,!1]}function V_(e){var t=Ge();return X_(t,pe,e)}function X_(e,t,n){t=Gm(e,t,G_)[0],e=yu(gr)[0],t=typeof t=="object"&&t!==null&&typeof t.then=="function"?as(t):t;var r=Ge(),a=r.queue,i=a.dispatch;return n!==r.memoizedState&&(Q.flags|=2048,Qi(9,Yk.bind(null,a,n),{destroy:void 0},null)),[t,i,e]}function Yk(e,t){e.action=t}function P_(e){var t=Ge(),n=pe;if(n!==null)return X_(t,n,e);Ge(),t=t.memoizedState,n=Ge();var r=n.queue.dispatch;return n.memoizedState=e,[t,r,!1]}function Qi(e,t,n,r){return e={tag:e,create:t,inst:n,deps:r,next:null},t=Q.updateQueue,t===null&&(t=wf(),Q.updateQueue=t),n=t.lastEffect,n===null?t.lastEffect=e.next=e:(r=n.next,n.next=e,e.next=r,t.lastEffect=e),e}function F_(){return Ge().memoizedState}function mu(e,t,n,r){var a=gt();Q.flags|=e,a.memoizedState=Qi(1|t,n,{destroy:void 0},r===void 0?null:r)}function Rf(e,t,n,r){var a=Ge();r=r===void 0?null:r;var i=a.memoizedState.inst;pe!==null&&r!==null&&km(r,pe.memoizedState.deps)?a.memoizedState=Qi(t,n,i,r):(Q.flags|=e,a.memoizedState=Qi(1|t,n,i,r))}function Hb(e,t){mu(8390656,8,e,t)}function Ym(e,t){Rf(2048,8,e,t)}function Q_(e,t){return Rf(4,2,e,t)}function K_(e,t){return Rf(4,4,e,t)}function Z_(e,t){if(typeof t=="function"){e=e();var n=t(e);return function(){typeof n=="function"?n():t(null)}}if(t!=null)return e=e(),t.current=e,function(){t.current=null}}function W_(e,t,n){n=n!=null?n.concat([e]):null,Rf(4,4,Z_.bind(null,t,e),n)}function Vm(){}function J_(e,t){var n=Ge();t=t===void 0?null:t;var r=n.memoizedState;return t!==null&&km(t,r[1])?r[0]:(n.memoizedState=[e,t],e)}function I_(e,t){var n=Ge();t=t===void 0?null:t;var r=n.memoizedState;if(t!==null&&km(t,r[1]))return r[0];if(r=e(),Ta){Ur(!0);try{e()}finally{Ur(!1)}}return n.memoizedState=[r,t],r}function Xm(e,t,n){return n===void 0||Ir&1073741824?e.memoizedState=t:(e.memoizedState=n,e=q2(),Q.lanes|=e,ta|=e,n)}function e2(e,t,n,r){return Bt(n,t)?n:Pi.current!==null?(e=Xm(e,n,r),Bt(e,t)||(Ze=!0),e):Ir&42?(e=q2(),Q.lanes|=e,ta|=e,t):(Ze=!0,e.memoizedState=n)}function t2(e,t,n,r,a){var i=Se.p;Se.p=i!==0&&8>i?i:8;var l=P.T,o={};P.T=o,Fm(e,!1,t,n);try{var s=a(),u=P.S;if(u!==null&&u(o,s),s!==null&&typeof s=="object"&&typeof s.then=="function"){var c=Uk(s,r);eo(e,t,c,jt(e))}else eo(e,t,r,jt(e))}catch(d){eo(e,t,{then:function(){},status:"rejected",reason:d},jt())}finally{Se.p=i,P.T=l}}function Vk(){}function Th(e,t,n,r){if(e.tag!==5)throw Error(A(476));var a=n2(e).queue;t2(e,a,t,ma,n===null?Vk:function(){return r2(e),n(r)})}function n2(e){var t=e.memoizedState;if(t!==null)return t;t={memoizedState:ma,baseState:ma,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:gr,lastRenderedState:ma},next:null};var n={};return t.next={memoizedState:n,baseState:n,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:gr,lastRenderedState:n},next:null},e.memoizedState=t,e=e.alternate,e!==null&&(e.memoizedState=t),t}function r2(e){var t=n2(e).next.queue;eo(e,t,{},jt())}function Pm(){return ut(Mo)}function a2(){return Ge().memoizedState}function i2(){return Ge().memoizedState}function Xk(e){for(var t=e.return;t!==null;){switch(t.tag){case 24:case 3:var n=jt();e=Vr(n);var r=Xr(t,e,n);r!==null&&(mt(r,t,n),no(r,t,n)),t={cache:Dm()},e.payload=t;return}t=t.return}}function Pk(e,t,n){var r=jt();n={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null},Cf(e)?o2(t,n):(n=$m(e,t,n,r),n!==null&&(mt(n,e,r),s2(n,t,r)))}function l2(e,t,n){var r=jt();eo(e,t,n,r)}function eo(e,t,n,r){var a={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null};if(Cf(e))o2(t,a);else{var i=e.alternate;if(e.lanes===0&&(i===null||i.lanes===0)&&(i=t.lastRenderedReducer,i!==null))try{var l=t.lastRenderedState,o=i(l,n);if(a.hasEagerState=!0,a.eagerState=o,Bt(o,l))return Tf(e,t,a,0),be===null&&_f(),!1}catch{}finally{}if(n=$m(e,t,a,r),n!==null)return mt(n,e,r),s2(n,t,r),!0}return!1}function Fm(e,t,n,r){if(r={lane:2,revertLane:rv(),action:r,hasEagerState:!1,eagerState:null,next:null},Cf(e)){if(t)throw Error(A(479))}else t=$m(e,n,r,2),t!==null&&mt(t,e,2)}function Cf(e){var t=e.alternate;return e===Q||t!==null&&t===Q}function o2(e,t){$i=ac=!0;var n=e.pending;n===null?t.next=t:(t.next=n.next,n.next=t),e.pending=t}function s2(e,t,n){if(n&4194176){var r=t.lanes;r&=e.pendingLanes,n|=r,t.lanes=n,ZO(e,n)}}var Un={readContext:ut,use:Af,useCallback:je,useContext:je,useEffect:je,useImperativeHandle:je,useLayoutEffect:je,useInsertionEffect:je,useMemo:je,useReducer:je,useRef:je,useState:je,useDebugValue:je,useDeferredValue:je,useTransition:je,useSyncExternalStore:je,useId:je};Un.useCacheRefresh=je;Un.useMemoCache=je;Un.useHostTransitionStatus=je;Un.useFormState=je;Un.useActionState=je;Un.useOptimistic=je;var ka={readContext:ut,use:Af,useCallback:function(e,t){return gt().memoizedState=[e,t===void 0?null:t],e},useContext:ut,useEffect:Hb,useImperativeHandle:function(e,t,n){n=n!=null?n.concat([e]):null,mu(4194308,4,Z_.bind(null,t,e),n)},useLayoutEffect:function(e,t){return mu(4194308,4,e,t)},useInsertionEffect:function(e,t){mu(4,2,e,t)},useMemo:function(e,t){var n=gt();t=t===void 0?null:t;var r=e();if(Ta){Ur(!0);try{e()}finally{Ur(!1)}}return n.memoizedState=[r,t],r},useReducer:function(e,t,n){var r=gt();if(n!==void 0){var a=n(t);if(Ta){Ur(!0);try{n(t)}finally{Ur(!1)}}}else a=t;return r.memoizedState=r.baseState=a,e={pending:null,lanes:0,dispatch:null,lastRenderedReducer:e,lastRenderedState:a},r.queue=e,e=e.dispatch=Pk.bind(null,Q,e),[r.memoizedState,e]},useRef:function(e){var t=gt();return e={current:e},t.memoizedState=e},useState:function(e){e=Oh(e);var t=e.queue,n=l2.bind(null,Q,t);return t.dispatch=n,[e.memoizedState,n]},useDebugValue:Vm,useDeferredValue:function(e,t){var n=gt();return Xm(n,e,t)},useTransition:function(){var e=Oh(!1);return e=t2.bind(null,Q,e.queue,!0,!1),gt().memoizedState=e,[!1,e]},useSyncExternalStore:function(e,t,n){var r=Q,a=gt();if(ae){if(n===void 0)throw Error(A(407));n=n()}else{if(n=t(),be===null)throw Error(A(349));te&60||D_(r,t,n)}a.memoizedState=n;var i={value:n,getSnapshot:t};return a.queue=i,Hb(k_.bind(null,r,i,e),[e]),r.flags|=2048,Qi(9,B_.bind(null,r,i,n,t),{destroy:void 0},null),n},useId:function(){var e=gt(),t=be.identifierPrefix;if(ae){var n=sr,r=or;n=(r&~(1<<32-Mt(r)-1)).toString(32)+n,t=":"+t+"R"+n,n=ic++,0<n&&(t+="H"+n.toString(32)),t+=":"}else n=Hk++,t=":"+t+"r"+n.toString(32)+":";return e.memoizedState=t},useCacheRefresh:function(){return gt().memoizedState=Xk.bind(null,Q)}};ka.useMemoCache=qm;ka.useHostTransitionStatus=Pm;ka.useFormState=Y_;ka.useActionState=Y_;ka.useOptimistic=function(e){var t=gt();t.memoizedState=t.baseState=e;var n={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return t.queue=n,t=Fm.bind(null,Q,!0,n),n.dispatch=t,[e,t]};var ia={readContext:ut,use:Af,useCallback:J_,useContext:ut,useEffect:Ym,useImperativeHandle:W_,useInsertionEffect:Q_,useLayoutEffect:K_,useMemo:I_,useReducer:yu,useRef:F_,useState:function(){return yu(gr)},useDebugValue:Vm,useDeferredValue:function(e,t){var n=Ge();return e2(n,pe.memoizedState,e,t)},useTransition:function(){var e=yu(gr)[0],t=Ge().memoizedState;return[typeof e=="boolean"?e:as(e),t]},useSyncExternalStore:j_,useId:a2};ia.useCacheRefresh=i2;ia.useMemoCache=qm;ia.useHostTransitionStatus=Pm;ia.useFormState=V_;ia.useActionState=V_;ia.useOptimistic=function(e,t){var n=Ge();return U_(n,pe,e,t)};var za={readContext:ut,use:Af,useCallback:J_,useContext:ut,useEffect:Ym,useImperativeHandle:W_,useInsertionEffect:Q_,useLayoutEffect:K_,useMemo:I_,useReducer:Gd,useRef:F_,useState:function(){return Gd(gr)},useDebugValue:Vm,useDeferredValue:function(e,t){var n=Ge();return pe===null?Xm(n,e,t):e2(n,pe.memoizedState,e,t)},useTransition:function(){var e=Gd(gr)[0],t=Ge().memoizedState;return[typeof e=="boolean"?e:as(e),t]},useSyncExternalStore:j_,useId:a2};za.useCacheRefresh=i2;za.useMemoCache=qm;za.useHostTransitionStatus=Pm;za.useFormState=P_;za.useActionState=P_;za.useOptimistic=function(e,t){var n=Ge();return pe!==null?U_(n,pe,e,t):(n.baseState=e,[e,n.queue.dispatch])};function Yd(e,t,n,r){t=e.memoizedState,n=n(r,t),n=n==null?t:Ee({},t,n),e.memoizedState=n,e.lanes===0&&(e.updateQueue.baseState=n)}var wh={isMounted:function(e){return(e=e._reactInternals)?il(e)===e:!1},enqueueSetState:function(e,t,n){e=e._reactInternals;var r=jt(),a=Vr(r);a.payload=t,n!=null&&(a.callback=n),t=Xr(e,a,r),t!==null&&(mt(t,e,r),no(t,e,r))},enqueueReplaceState:function(e,t,n){e=e._reactInternals;var r=jt(),a=Vr(r);a.tag=1,a.payload=t,n!=null&&(a.callback=n),t=Xr(e,a,r),t!==null&&(mt(t,e,r),no(t,e,r))},enqueueForceUpdate:function(e,t){e=e._reactInternals;var n=jt(),r=Vr(n);r.tag=2,t!=null&&(r.callback=t),t=Xr(e,r,n),t!==null&&(mt(t,e,n),no(t,e,n))}};function qb(e,t,n,r,a,i,l){return e=e.stateNode,typeof e.shouldComponentUpdate=="function"?e.shouldComponentUpdate(r,i,l):t.prototype&&t.prototype.isPureReactComponent?!To(n,r)||!To(a,i):!0}function Gb(e,t,n,r){e=t.state,typeof t.componentWillReceiveProps=="function"&&t.componentWillReceiveProps(n,r),typeof t.UNSAFE_componentWillReceiveProps=="function"&&t.UNSAFE_componentWillReceiveProps(n,r),t.state!==e&&wh.enqueueReplaceState(t,t.state,null)}function wa(e,t){var n=t;if("ref"in t){n={};for(var r in t)r!=="ref"&&(n[r]=t[r])}if(e=e.defaultProps){n===t&&(n=Ee({},n));for(var a in e)n[a]===void 0&&(n[a]=e[a])}return n}var lc=typeof reportError=="function"?reportError:function(e){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof e=="object"&&e!==null&&typeof e.message=="string"?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",e);return}};function u2(e){lc(e)}function c2(e){}function f2(e){lc(e)}function oc(e,t){try{var n=e.onUncaughtError;n(t.value,{componentStack:t.stack})}catch(r){setTimeout(function(){throw r})}}function Yb(e,t,n){try{var r=e.onCaughtError;r(n.value,{componentStack:n.stack,errorBoundary:t.tag===1?t.stateNode:null})}catch(a){setTimeout(function(){throw a})}}function Ah(e,t,n){return n=Vr(n),n.tag=3,n.payload={element:null},n.callback=function(){oc(e,t)},n}function d2(e){return e=Vr(e),e.tag=3,e}function p2(e,t,n,r){var a=n.type.getDerivedStateFromError;if(typeof a=="function"){var i=r.value;e.payload=function(){return a(i)},e.callback=function(){Yb(t,n,r)}}var l=n.stateNode;l!==null&&typeof l.componentDidCatch=="function"&&(e.callback=function(){Yb(t,n,r),typeof a!="function"&&(Fr===null?Fr=new Set([this]):Fr.add(this));var o=r.stack;this.componentDidCatch(r.value,{componentStack:o!==null?o:""})})}function Fk(e,t,n,r,a){if(n.flags|=32768,r!==null&&typeof r=="object"&&typeof r.then=="function"){if(t=n.alternate,t!==null&&is(t,n,a,!0),n=an.current,n!==null){switch(n.tag){case 13:return zn===null?Gh():n.alternate===null&&Me===0&&(Me=3),n.flags&=-257,n.flags|=65536,n.lanes=a,r===Sh?n.flags|=16384:(t=n.updateQueue,t===null?n.updateQueue=new Set([r]):t.add(r),np(e,r,a)),!1;case 22:return n.flags|=65536,r===Sh?n.flags|=16384:(t=n.updateQueue,t===null?(t={transitions:null,markerInstances:null,retryQueue:new Set([r])},n.updateQueue=t):(n=t.retryQueue,n===null?t.retryQueue=new Set([r]):n.add(r)),np(e,r,a)),!1}throw Error(A(435,n.tag))}return np(e,r,a),Gh(),!1}if(ae)return t=an.current,t!==null?(!(t.flags&65536)&&(t.flags|=256),t.flags|=65536,t.lanes=a,r!==bh&&(e=Error(A(422),{cause:r}),wo(Wt(e,n)))):(r!==bh&&(t=Error(A(423),{cause:r}),wo(Wt(t,n))),e=e.current.alternate,e.flags|=65536,a&=-a,e.lanes|=a,r=Wt(r,n),a=Ah(e.stateNode,r,a),Kd(e,a),Me!==4&&(Me=2)),!1;var i=Error(A(520),{cause:r});if(i=Wt(i,n),lo===null?lo=[i]:lo.push(i),Me!==4&&(Me=2),t===null)return!0;r=Wt(r,n),n=t;do{switch(n.tag){case 3:return n.flags|=65536,e=a&-a,n.lanes|=e,e=Ah(n.stateNode,r,e),Kd(n,e),!1;case 1:if(t=n.type,i=n.stateNode,(n.flags&128)===0&&(typeof t.getDerivedStateFromError=="function"||i!==null&&typeof i.componentDidCatch=="function"&&(Fr===null||!Fr.has(i))))return n.flags|=65536,a&=-a,n.lanes|=a,a=d2(a),p2(a,e,n,r),Kd(n,a),!1}n=n.return}while(n!==null);return!1}var h2=Error(A(461)),Ze=!1;function tt(e,t,n,r){t.child=e===null?R_(t,null,n,r):_a(t,e.child,n,r)}function Vb(e,t,n,r,a){n=n.render;var i=t.ref;if("ref"in r){var l={};for(var o in r)o!=="ref"&&(l[o]=r[o])}else l=r;return Aa(t),r=zm(e,t,n,l,i,a),o=Lm(),e!==null&&!Ze?(Um(e,t,a),br(e,t,a)):(ae&&o&&Nm(t),t.flags|=1,tt(e,t,r,a),t.child)}function Xb(e,t,n,r,a){if(e===null){var i=n.type;return typeof i=="function"&&!Jm(i)&&i.defaultProps===void 0&&n.compare===null?(t.tag=15,t.type=i,y2(e,t,i,r,a)):(e=gu(n.type,null,r,t,t.mode,a),e.ref=t.ref,e.return=t,t.child=e)}if(i=e.child,!Qm(e,a)){var l=i.memoizedProps;if(n=n.compare,n=n!==null?n:To,n(l,r)&&e.ref===t.ref)return br(e,t,a)}return t.flags|=1,e=Pr(i,r),e.ref=t.ref,e.return=t,t.child=e}function y2(e,t,n,r,a){if(e!==null){var i=e.memoizedProps;if(To(i,r)&&e.ref===t.ref)if(Ze=!1,t.pendingProps=r=i,Qm(e,a))e.flags&131072&&(Ze=!0);else return t.lanes=e.lanes,br(e,t,a)}return Rh(e,t,n,r,a)}function m2(e,t,n){var r=t.pendingProps,a=r.children,i=(t.stateNode._pendingVisibility&2)!==0,l=e!==null?e.memoizedState:null;if(to(e,t),r.mode==="hidden"||i){if(t.flags&128){if(r=l!==null?l.baseLanes|n:n,e!==null){for(a=t.child=e.child,i=0;a!==null;)i=i|a.lanes|a.childLanes,a=a.sibling;t.childLanes=i&~r}else t.childLanes=0,t.child=null;return Pb(e,t,r,n)}if(n&536870912)t.memoizedState={baseLanes:0,cachePool:null},e!==null&&hu(t,l!==null?l.cachePool:null),l!==null?Bb(t,l):Eh(),C_(t);else return t.lanes=t.childLanes=536870912,Pb(e,t,l!==null?l.baseLanes|n:n,n)}else l!==null?(hu(t,l.cachePool),Bb(t,l),Br(),t.memoizedState=null):(e!==null&&hu(t,null),Eh(),Br());return tt(e,t,a,n),t.child}function Pb(e,t,n,r){var a=Bm();return a=a===null?null:{parent:Ye._currentValue,pool:a},t.memoizedState={baseLanes:n,cachePool:a},e!==null&&hu(t,null),Eh(),C_(t),e!==null&&is(e,t,r,!0),null}function to(e,t){var n=t.ref;if(n===null)e!==null&&e.ref!==null&&(t.flags|=2097664);else{if(typeof n!="function"&&typeof n!="object")throw Error(A(284));(e===null||e.ref!==n)&&(t.flags|=2097664)}}function Rh(e,t,n,r,a){return Aa(t),n=zm(e,t,n,r,void 0,a),r=Lm(),e!==null&&!Ze?(Um(e,t,a),br(e,t,a)):(ae&&r&&Nm(t),t.flags|=1,tt(e,t,n,a),t.child)}function Fb(e,t,n,r,a,i){return Aa(t),t.updateQueue=null,n=M_(t,r,n,a),N_(e),r=Lm(),e!==null&&!Ze?(Um(e,t,i),br(e,t,i)):(ae&&r&&Nm(t),t.flags|=1,tt(e,t,n,i),t.child)}function Qb(e,t,n,r,a){if(Aa(t),t.stateNode===null){var i=vi,l=n.contextType;typeof l=="object"&&l!==null&&(i=ut(l)),i=new n(r,i),t.memoizedState=i.state!==null&&i.state!==void 0?i.state:null,i.updater=wh,t.stateNode=i,i._reactInternals=t,i=t.stateNode,i.props=r,i.state=t.memoizedState,i.refs={},Km(t),l=n.contextType,i.context=typeof l=="object"&&l!==null?ut(l):vi,i.state=t.memoizedState,l=n.getDerivedStateFromProps,typeof l=="function"&&(Yd(t,n,l,r),i.state=t.memoizedState),typeof n.getDerivedStateFromProps=="function"||typeof i.getSnapshotBeforeUpdate=="function"||typeof i.UNSAFE_componentWillMount!="function"&&typeof i.componentWillMount!="function"||(l=i.state,typeof i.componentWillMount=="function"&&i.componentWillMount(),typeof i.UNSAFE_componentWillMount=="function"&&i.UNSAFE_componentWillMount(),l!==i.state&&wh.enqueueReplaceState(i,i.state,null),ao(t,r,i,a),ro(),i.state=t.memoizedState),typeof i.componentDidMount=="function"&&(t.flags|=4194308),r=!0}else if(e===null){i=t.stateNode;var o=t.memoizedProps,s=wa(n,o);i.props=s;var u=i.context,c=n.contextType;l=vi,typeof c=="object"&&c!==null&&(l=ut(c));var d=n.getDerivedStateFromProps;c=typeof d=="function"||typeof i.getSnapshotBeforeUpdate=="function",o=t.pendingProps!==o,c||typeof i.UNSAFE_componentWillReceiveProps!="function"&&typeof i.componentWillReceiveProps!="function"||(o||u!==l)&&Gb(t,i,r,l),Mr=!1;var f=t.memoizedState;i.state=f,ao(t,r,i,a),ro(),u=t.memoizedState,o||f!==u||Mr?(typeof d=="function"&&(Yd(t,n,d,r),u=t.memoizedState),(s=Mr||qb(t,n,s,r,f,u,l))?(c||typeof i.UNSAFE_componentWillMount!="function"&&typeof i.componentWillMount!="function"||(typeof i.componentWillMount=="function"&&i.componentWillMount(),typeof i.UNSAFE_componentWillMount=="function"&&i.UNSAFE_componentWillMount()),typeof i.componentDidMount=="function"&&(t.flags|=4194308)):(typeof i.componentDidMount=="function"&&(t.flags|=4194308),t.memoizedProps=r,t.memoizedState=u),i.props=r,i.state=u,i.context=l,r=s):(typeof i.componentDidMount=="function"&&(t.flags|=4194308),r=!1)}else{i=t.stateNode,Dh(e,t),l=t.memoizedProps,c=wa(n,l),i.props=c,d=t.pendingProps,f=i.context,u=n.contextType,s=vi,typeof u=="object"&&u!==null&&(s=ut(u)),o=n.getDerivedStateFromProps,(u=typeof o=="function"||typeof i.getSnapshotBeforeUpdate=="function")||typeof i.UNSAFE_componentWillReceiveProps!="function"&&typeof i.componentWillReceiveProps!="function"||(l!==d||f!==s)&&Gb(t,i,r,s),Mr=!1,f=t.memoizedState,i.state=f,ao(t,r,i,a),ro();var h=t.memoizedState;l!==d||f!==h||Mr||e!==null&&e.dependencies!==null&&sc(e.dependencies)?(typeof o=="function"&&(Yd(t,n,o,r),h=t.memoizedState),(c=Mr||qb(t,n,c,r,f,h,s)||e!==null&&e.dependencies!==null&&sc(e.dependencies))?(u||typeof i.UNSAFE_componentWillUpdate!="function"&&typeof i.componentWillUpdate!="function"||(typeof i.componentWillUpdate=="function"&&i.componentWillUpdate(r,h,s),typeof i.UNSAFE_componentWillUpdate=="function"&&i.UNSAFE_componentWillUpdate(r,h,s)),typeof i.componentDidUpdate=="function"&&(t.flags|=4),typeof i.getSnapshotBeforeUpdate=="function"&&(t.flags|=1024)):(typeof i.componentDidUpdate!="function"||l===e.memoizedProps&&f===e.memoizedState||(t.flags|=4),typeof i.getSnapshotBeforeUpdate!="function"||l===e.memoizedProps&&f===e.memoizedState||(t.flags|=1024),t.memoizedProps=r,t.memoizedState=h),i.props=r,i.state=h,i.context=s,r=c):(typeof i.componentDidUpdate!="function"||l===e.memoizedProps&&f===e.memoizedState||(t.flags|=4),typeof i.getSnapshotBeforeUpdate!="function"||l===e.memoizedProps&&f===e.memoizedState||(t.flags|=1024),r=!1)}return i=r,to(e,t),r=(t.flags&128)!==0,i||r?(i=t.stateNode,n=r&&typeof n.getDerivedStateFromError!="function"?null:i.render(),t.flags|=1,e!==null&&r?(t.child=_a(t,e.child,null,a),t.child=_a(t,null,n,a)):tt(e,t,n,a),t.memoizedState=i.state,e=t.child):e=br(e,t,a),e}function Kb(e,t,n,r){return ns(),t.flags|=256,tt(e,t,n,r),t.child}var Vd={dehydrated:null,treeContext:null,retryLane:0};function Xd(e){return{baseLanes:e,cachePool:$_()}}function Pd(e,t,n){return e=e!==null?e.childLanes&~n:0,t&&(e|=en),e}function v2(e,t,n){var r=t.pendingProps,a=!1,i=(t.flags&128)!==0,l;if((l=i)||(l=e!==null&&e.memoizedState===null?!1:(Ve.current&2)!==0),l&&(a=!0,t.flags&=-129),l=(t.flags&32)!==0,t.flags&=-33,e===null){if(ae){if(a?Dr(t):Br(),ae){var o=rt,s;if(s=o){e:{for(s=o,o=Rn;s.nodeType!==8;){if(!o){o=null;break e}if(s=gn(s.nextSibling),s===null){o=null;break e}}o=s}o!==null?(t.memoizedState={dehydrated:o,treeContext:va!==null?{id:or,overflow:sr}:null,retryLane:536870912},s=It(18,null,null,0),s.stateNode=o,s.return=t,t.child=s,ht=t,rt=null,s=!0):s=!1}s||Oa(t)}if(o=t.memoizedState,o!==null&&(o=o.dehydrated,o!==null))return o.data==="$!"?t.lanes=16:t.lanes=536870912,null;ur(t)}return o=r.children,r=r.fallback,a?(Br(),a=t.mode,o=$h({mode:"hidden",children:o},a),r=ba(r,a,n,null),o.return=t,r.return=t,o.sibling=r,t.child=o,a=t.child,a.memoizedState=Xd(n),a.childLanes=Pd(e,l,n),t.memoizedState=Vd,r):(Dr(t),Ch(t,o))}if(s=e.memoizedState,s!==null&&(o=s.dehydrated,o!==null)){if(i)t.flags&256?(Dr(t),t.flags&=-257,t=Fd(e,t,n)):t.memoizedState!==null?(Br(),t.child=e.child,t.flags|=128,t=null):(Br(),a=r.fallback,o=t.mode,r=$h({mode:"visible",children:r.children},o),a=ba(a,o,n,null),a.flags|=2,r.return=t,a.return=t,r.sibling=a,t.child=r,_a(t,e.child,null,n),r=t.child,r.memoizedState=Xd(n),r.childLanes=Pd(e,l,n),t.memoizedState=Vd,t=a);else if(Dr(t),o.data==="$!"){if(l=o.nextSibling&&o.nextSibling.dataset,l)var u=l.dgst;l=u,r=Error(A(419)),r.stack="",r.digest=l,wo({value:r,source:null,stack:null}),t=Fd(e,t,n)}else if(Ze||is(e,t,n,!1),l=(n&e.childLanes)!==0,Ze||l){if(l=be,l!==null){if(r=n&-n,r&42)r=1;else switch(r){case 2:r=1;break;case 8:r=4;break;case 32:r=16;break;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:r=64;break;case 268435456:r=134217728;break;default:r=0}if(r=r&(l.suspendedLanes|n)?0:r,r!==0&&r!==s.retryLane)throw s.retryLane=r,Jr(e,r),mt(l,e,r),h2}o.data==="$?"||Gh(),t=Fd(e,t,n)}else o.data==="$?"?(t.flags|=128,t.child=e.child,t=sz.bind(null,e),o._reactRetry=t,t=null):(e=s.treeContext,rt=gn(o.nextSibling),ht=t,ae=!0,yn=null,Rn=!1,e!==null&&(Qt[Kt++]=or,Qt[Kt++]=sr,Qt[Kt++]=va,or=e.id,sr=e.overflow,va=t),t=Ch(t,r.children),t.flags|=4096);return t}return a?(Br(),a=r.fallback,o=t.mode,s=e.child,u=s.sibling,r=Pr(s,{mode:"hidden",children:r.children}),r.subtreeFlags=s.subtreeFlags&31457280,u!==null?a=Pr(u,a):(a=ba(a,o,n,null),a.flags|=2),a.return=t,r.return=t,r.sibling=a,t.child=r,r=a,a=t.child,o=e.child.memoizedState,o===null?o=Xd(n):(s=o.cachePool,s!==null?(u=Ye._currentValue,s=s.parent!==u?{parent:u,pool:u}:s):s=$_(),o={baseLanes:o.baseLanes|n,cachePool:s}),a.memoizedState=o,a.childLanes=Pd(e,l,n),t.memoizedState=Vd,r):(Dr(t),n=e.child,e=n.sibling,n=Pr(n,{mode:"visible",children:r.children}),n.return=t,n.sibling=null,e!==null&&(l=t.deletions,l===null?(t.deletions=[e],t.flags|=16):l.push(e)),t.child=n,t.memoizedState=null,n)}function Ch(e,t){return t=$h({mode:"visible",children:t},e.mode),t.return=e,e.child=t}function $h(e,t){return U2(e,t,0,null)}function Fd(e,t,n){return _a(t,e.child,null,n),e=Ch(t,t.pendingProps.children),e.flags|=2,t.memoizedState=null,e}function Zb(e,t,n){e.lanes|=t;var r=e.alternate;r!==null&&(r.lanes|=t),Mh(e.return,t,n)}function Qd(e,t,n,r,a){var i=e.memoizedState;i===null?e.memoizedState={isBackwards:t,rendering:null,renderingStartTime:0,last:r,tail:n,tailMode:a}:(i.isBackwards=t,i.rendering=null,i.renderingStartTime=0,i.last=r,i.tail=n,i.tailMode=a)}function g2(e,t,n){var r=t.pendingProps,a=r.revealOrder,i=r.tail;if(tt(e,t,r.children,n),r=Ve.current,r&2)r=r&1|2,t.flags|=128;else{if(e!==null&&e.flags&128)e:for(e=t.child;e!==null;){if(e.tag===13)e.memoizedState!==null&&Zb(e,n,t);else if(e.tag===19)Zb(e,n,t);else if(e.child!==null){e.child.return=e,e=e.child;continue}if(e===t)break e;for(;e.sibling===null;){if(e.return===null||e.return===t)break e;e=e.return}e.sibling.return=e.return,e=e.sibling}r&=1}switch(Te(Ve,r),a){case"forwards":for(n=t.child,a=null;n!==null;)e=n.alternate,e!==null&&rc(e)===null&&(a=n),n=n.sibling;n=a,n===null?(a=t.child,t.child=null):(a=n.sibling,n.sibling=null),Qd(t,!1,a,n,i);break;case"backwards":for(n=null,a=t.child,t.child=null;a!==null;){if(e=a.alternate,e!==null&&rc(e)===null){t.child=a;break}e=a.sibling,a.sibling=n,n=a,a=e}Qd(t,!0,n,null,i);break;case"together":Qd(t,!1,null,null,void 0);break;default:t.memoizedState=null}return t.child}function br(e,t,n){if(e!==null&&(t.dependencies=e.dependencies),ta|=t.lanes,!(n&t.childLanes))if(e!==null){if(is(e,t,n,!1),(n&t.childLanes)===0)return null}else return null;if(e!==null&&t.child!==e.child)throw Error(A(153));if(t.child!==null){for(e=t.child,n=Pr(e,e.pendingProps),t.child=n,n.return=t;e.sibling!==null;)e=e.sibling,n=n.sibling=Pr(e,e.pendingProps),n.return=t;n.sibling=null}return t.child}function Qm(e,t){return e.lanes&t?!0:(e=e.dependencies,!!(e!==null&&sc(e)))}function Qk(e,t,n){switch(t.tag){case 3:Qu(t,t.stateNode.containerInfo),kr(t,Ye,e.memoizedState.cache),ns();break;case 27:case 5:ch(t);break;case 4:Qu(t,t.stateNode.containerInfo);break;case 10:kr(t,t.type,t.memoizedProps.value);break;case 13:var r=t.memoizedState;if(r!==null)return r.dehydrated!==null?(Dr(t),t.flags|=128,null):n&t.child.childLanes?v2(e,t,n):(Dr(t),e=br(e,t,n),e!==null?e.sibling:null);Dr(t);break;case 19:var a=(e.flags&128)!==0;if(r=(n&t.childLanes)!==0,r||(is(e,t,n,!1),r=(n&t.childLanes)!==0),a){if(r)return g2(e,t,n);t.flags|=128}if(a=t.memoizedState,a!==null&&(a.rendering=null,a.tail=null,a.lastEffect=null),Te(Ve,Ve.current),r)break;return null;case 22:case 23:return t.lanes=0,m2(e,t,n);case 24:kr(t,Ye,e.memoizedState.cache)}return br(e,t,n)}function b2(e,t,n){if(e!==null)if(e.memoizedProps!==t.pendingProps)Ze=!0;else{if(!Qm(e,n)&&!(t.flags&128))return Ze=!1,Qk(e,t,n);Ze=!!(e.flags&131072)}else Ze=!1,ae&&t.flags&1048576&&__(t,tc,t.index);switch(t.lanes=0,t.tag){case 16:e:{e=t.pendingProps;var r=t.elementType,a=r._init;if(r=a(r._payload),t.type=r,typeof r=="function")Jm(r)?(e=wa(r,e),t.tag=1,t=Qb(null,t,r,e,n)):(t.tag=0,t=Rh(null,t,r,e,n));else{if(r!=null){if(a=r.$$typeof,a===bm){t.tag=11,t=Vb(null,t,r,e,n);break e}else if(a===Sm){t.tag=14,t=Xb(null,t,r,e,n);break e}}throw t=sh(r)||r,Error(A(306,t,""))}}return t;case 0:return Rh(e,t,t.type,t.pendingProps,n);case 1:return r=t.type,a=wa(r,t.pendingProps),Qb(e,t,r,a,n);case 3:e:{if(Qu(t,t.stateNode.containerInfo),e===null)throw Error(A(387));var i=t.pendingProps;a=t.memoizedState,r=a.element,Dh(e,t),ao(t,i,null,n);var l=t.memoizedState;if(i=l.cache,kr(t,Ye,i),i!==a.cache&&jh(t,[Ye],n,!0),ro(),i=l.element,a.isDehydrated)if(a={element:i,isDehydrated:!1,cache:l.cache},t.updateQueue.baseState=a,t.memoizedState=a,t.flags&256){t=Kb(e,t,i,n);break e}else if(i!==r){r=Wt(Error(A(424)),t),wo(r),t=Kb(e,t,i,n);break e}else for(rt=gn(t.stateNode.containerInfo.firstChild),ht=t,ae=!0,yn=null,Rn=!0,n=R_(t,null,i,n),t.child=n;n;)n.flags=n.flags&-3|4096,n=n.sibling;else{if(ns(),i===r){t=br(e,t,n);break e}tt(e,t,i,n)}t=t.child}return t;case 26:return to(e,t),e===null?(n=h1(t.type,null,t.pendingProps,null))?t.memoizedState=n:ae||(n=t.type,e=t.pendingProps,r=yc(Yr.current).createElement(n),r[st]=t,r[xt]=e,at(r,n,e),Ke(r),t.stateNode=r):t.memoizedState=h1(t.type,e.memoizedProps,t.pendingProps,e.memoizedState),null;case 27:return ch(t),e===null&&ae&&(r=t.stateNode=rT(t.type,t.pendingProps,Yr.current),ht=t,Rn=!0,rt=gn(r.firstChild)),r=t.pendingProps.children,e!==null||ae?tt(e,t,r,n):t.child=_a(t,null,r,n),to(e,t),t.child;case 5:return e===null&&ae&&((a=r=rt)&&(r=_z(r,t.type,t.pendingProps,Rn),r!==null?(t.stateNode=r,ht=t,rt=gn(r.firstChild),Rn=!1,a=!0):a=!1),a||Oa(t)),ch(t),a=t.type,i=t.pendingProps,l=e!==null?e.memoizedProps:null,r=i.children,Fh(a,i)?r=null:l!==null&&Fh(a,l)&&(t.flags|=32),t.memoizedState!==null&&(a=zm(e,t,qk,null,null,n),Mo._currentValue=a),to(e,t),tt(e,t,r,n),t.child;case 6:return e===null&&ae&&((e=n=rt)&&(n=Tz(n,t.pendingProps,Rn),n!==null?(t.stateNode=n,ht=t,rt=null,e=!0):e=!1),e||Oa(t)),null;case 13:return v2(e,t,n);case 4:return Qu(t,t.stateNode.containerInfo),r=t.pendingProps,e===null?t.child=_a(t,null,r,n):tt(e,t,r,n),t.child;case 11:return Vb(e,t,t.type,t.pendingProps,n);case 7:return tt(e,t,t.pendingProps,n),t.child;case 8:return tt(e,t,t.pendingProps.children,n),t.child;case 12:return tt(e,t,t.pendingProps.children,n),t.child;case 10:return r=t.pendingProps,kr(t,t.type,r.value),tt(e,t,r.children,n),t.child;case 9:return a=t.type._context,r=t.pendingProps.children,Aa(t),a=ut(a),r=r(a),t.flags|=1,tt(e,t,r,n),t.child;case 14:return Xb(e,t,t.type,t.pendingProps,n);case 15:return y2(e,t,t.type,t.pendingProps,n);case 19:return g2(e,t,n);case 22:return m2(e,t,n);case 24:return Aa(t),r=ut(Ye),e===null?(a=Bm(),a===null&&(a=be,i=Dm(),a.pooledCache=i,i.refCount++,i!==null&&(a.pooledCacheLanes|=n),a=i),t.memoizedState={parent:r,cache:a},Km(t),kr(t,Ye,a)):(e.lanes&n&&(Dh(e,t),ao(t,null,null,n),ro()),a=e.memoizedState,i=t.memoizedState,a.parent!==r?(a={parent:r,cache:r},t.memoizedState=a,t.lanes===0&&(t.memoizedState=t.updateQueue.baseState=a),kr(t,Ye,r)):(r=i.cache,kr(t,Ye,r),r!==a.cache&&jh(t,[Ye],n,!0))),tt(e,t,t.pendingProps.children,n),t.child;case 29:throw t.pendingProps}throw Error(A(156,t.tag))}var Nh=Gn(null),La=null,cr=null;function kr(e,t,n){Te(Nh,t._currentValue),t._currentValue=n}function pr(e){e._currentValue=Nh.current,We(Nh)}function Mh(e,t,n){for(;e!==null;){var r=e.alternate;if((e.childLanes&t)!==t?(e.childLanes|=t,r!==null&&(r.childLanes|=t)):r!==null&&(r.childLanes&t)!==t&&(r.childLanes|=t),e===n)break;e=e.return}}function jh(e,t,n,r){var a=e.child;for(a!==null&&(a.return=e);a!==null;){var i=a.dependencies;if(i!==null){var l=a.child;i=i.firstContext;e:for(;i!==null;){var o=i;i=a;for(var s=0;s<t.length;s++)if(o.context===t[s]){i.lanes|=n,o=i.alternate,o!==null&&(o.lanes|=n),Mh(i.return,n,e),r||(l=null);break e}i=o.next}}else if(a.tag===18){if(l=a.return,l===null)throw Error(A(341));l.lanes|=n,i=l.alternate,i!==null&&(i.lanes|=n),Mh(l,n,e),l=null}else l=a.child;if(l!==null)l.return=a;else for(l=a;l!==null;){if(l===e){l=null;break}if(a=l.sibling,a!==null){a.return=l.return,l=a;break}l=l.return}a=l}}function is(e,t,n,r){e=null;for(var a=t,i=!1;a!==null;){if(!i){if(a.flags&524288)i=!0;else if(a.flags&262144)break}if(a.tag===10){var l=a.alternate;if(l===null)throw Error(A(387));if(l=l.memoizedProps,l!==null){var o=a.type;Bt(a.pendingProps.value,l.value)||(e!==null?e.push(o):e=[o])}}else if(a===Fu.current){if(l=a.alternate,l===null)throw Error(A(387));l.memoizedState.memoizedState!==a.memoizedState.memoizedState&&(e!==null?e.push(Mo):e=[Mo])}a=a.return}e!==null&&jh(t,e,n,r),t.flags|=262144}function sc(e){for(e=e.firstContext;e!==null;){if(!Bt(e.context._currentValue,e.memoizedValue))return!0;e=e.next}return!1}function Aa(e){La=e,cr=null,e=e.dependencies,e!==null&&(e.firstContext=null)}function ut(e){return S2(La,e)}function Ds(e,t){return La===null&&Aa(e),S2(e,t)}function S2(e,t){var n=t._currentValue;if(t={context:t,memoizedValue:n,next:null},cr===null){if(e===null)throw Error(A(308));cr=t,e.dependencies={lanes:0,firstContext:t},e.flags|=524288}else cr=cr.next=t;return n}var Mr=!1;function Km(e){e.updateQueue={baseState:e.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function Dh(e,t){e=e.updateQueue,t.updateQueue===e&&(t.updateQueue={baseState:e.baseState,firstBaseUpdate:e.firstBaseUpdate,lastBaseUpdate:e.lastBaseUpdate,shared:e.shared,callbacks:null})}function Vr(e){return{lane:e,tag:0,payload:null,callback:null,next:null}}function Xr(e,t,n){var r=e.updateQueue;if(r===null)return null;if(r=r.shared,Re&2){var a=r.pending;return a===null?t.next=t:(t.next=a.next,a.next=t),r.pending=t,t=Iu(e),O_(e,null,n),t}return Tf(e,r,t,n),Iu(e)}function no(e,t,n){if(t=t.updateQueue,t!==null&&(t=t.shared,(n&4194176)!==0)){var r=t.lanes;r&=e.pendingLanes,n|=r,t.lanes=n,ZO(e,n)}}function Kd(e,t){var n=e.updateQueue,r=e.alternate;if(r!==null&&(r=r.updateQueue,n===r)){var a=null,i=null;if(n=n.firstBaseUpdate,n!==null){do{var l={lane:n.lane,tag:n.tag,payload:n.payload,callback:null,next:null};i===null?a=i=l:i=i.next=l,n=n.next}while(n!==null);i===null?a=i=t:i=i.next=t}else a=i=t;n={baseState:r.baseState,firstBaseUpdate:a,lastBaseUpdate:i,shared:r.shared,callbacks:r.callbacks},e.updateQueue=n;return}e=n.lastBaseUpdate,e===null?n.firstBaseUpdate=t:e.next=t,n.lastBaseUpdate=t}var Bh=!1;function ro(){if(Bh){var e=Ci;if(e!==null)throw e}}function ao(e,t,n,r){Bh=!1;var a=e.updateQueue;Mr=!1;var i=a.firstBaseUpdate,l=a.lastBaseUpdate,o=a.shared.pending;if(o!==null){a.shared.pending=null;var s=o,u=s.next;s.next=null,l===null?i=u:l.next=u,l=s;var c=e.alternate;c!==null&&(c=c.updateQueue,o=c.lastBaseUpdate,o!==l&&(o===null?c.firstBaseUpdate=u:o.next=u,c.lastBaseUpdate=s))}if(i!==null){var d=a.baseState;l=0,c=u=s=null,o=i;do{var f=o.lane&-536870913,h=f!==o.lane;if(h?(te&f)===f:(r&f)===f){f!==0&&f===Fi&&(Bh=!0),c!==null&&(c=c.next={lane:0,tag:o.tag,payload:o.payload,callback:null,next:null});e:{var b=e,S=o;f=t;var x=n;switch(S.tag){case 1:if(b=S.payload,typeof b=="function"){d=b.call(x,d,f);break e}d=b;break e;case 3:b.flags=b.flags&-65537|128;case 0:if(b=S.payload,f=typeof b=="function"?b.call(x,d,f):b,f==null)break e;d=Ee({},d,f);break e;case 2:Mr=!0}}f=o.callback,f!==null&&(e.flags|=64,h&&(e.flags|=8192),h=a.callbacks,h===null?a.callbacks=[f]:h.push(f))}else h={lane:f,tag:o.tag,payload:o.payload,callback:o.callback,next:null},c===null?(u=c=h,s=d):c=c.next=h,l|=f;if(o=o.next,o===null){if(o=a.shared.pending,o===null)break;h=o,o=h.next,h.next=null,a.lastBaseUpdate=h,a.shared.pending=null}}while(!0);c===null&&(s=d),a.baseState=s,a.firstBaseUpdate=u,a.lastBaseUpdate=c,i===null&&(a.shared.lanes=0),ta|=l,e.lanes=l,e.memoizedState=d}}function E2(e,t){if(typeof e!="function")throw Error(A(191,e));e.call(t)}function x2(e,t){var n=e.callbacks;if(n!==null)for(e.callbacks=null,e=0;e<n.length;e++)E2(n[e],t)}function ls(e,t){try{var n=t.updateQueue,r=n!==null?n.lastEffect:null;if(r!==null){var a=r.next;n=a;do{if((n.tag&e)===e){r=void 0;var i=n.create,l=n.inst;r=i(),l.destroy=r}n=n.next}while(n!==a)}}catch(o){ye(t,t.return,o)}}function ea(e,t,n){try{var r=t.updateQueue,a=r!==null?r.lastEffect:null;if(a!==null){var i=a.next;r=i;do{if((r.tag&e)===e){var l=r.inst,o=l.destroy;if(o!==void 0){l.destroy=void 0,a=t;var s=n;try{o()}catch(u){ye(a,s,u)}}}r=r.next}while(r!==i)}}catch(u){ye(t,t.return,u)}}function O2(e){var t=e.updateQueue;if(t!==null){var n=e.stateNode;try{x2(t,n)}catch(r){ye(e,e.return,r)}}}function _2(e,t,n){n.props=wa(e.type,e.memoizedProps),n.state=e.memoizedState;try{n.componentWillUnmount()}catch(r){ye(e,t,r)}}function pa(e,t){try{var n=e.ref;if(n!==null){var r=e.stateNode;switch(e.tag){case 26:case 27:case 5:var a=r;break;default:a=r}typeof n=="function"?e.refCleanup=n(a):n.current=a}}catch(i){ye(e,t,i)}}function Ct(e,t){var n=e.ref,r=e.refCleanup;if(n!==null)if(typeof r=="function")try{r()}catch(a){ye(e,t,a)}finally{e.refCleanup=null,e=e.alternate,e!=null&&(e.refCleanup=null)}else if(typeof n=="function")try{n(null)}catch(a){ye(e,t,a)}else n.current=null}function T2(e){var t=e.type,n=e.memoizedProps,r=e.stateNode;try{e:switch(t){case"button":case"input":case"select":case"textarea":n.autoFocus&&r.focus();break e;case"img":n.src?r.src=n.src:n.srcSet&&(r.srcset=n.srcSet)}}catch(a){ye(e,e.return,a)}}function Wb(e,t,n){try{var r=e.stateNode;bz(r,e.type,n,t),r[xt]=t}catch(a){ye(e,e.return,a)}}function w2(e){return e.tag===5||e.tag===3||e.tag===26||e.tag===27||e.tag===4}function Zd(e){e:for(;;){for(;e.sibling===null;){if(e.return===null||w2(e.return))return null;e=e.return}for(e.sibling.return=e.return,e=e.sibling;e.tag!==5&&e.tag!==6&&e.tag!==27&&e.tag!==18;){if(e.flags&2||e.child===null||e.tag===4)continue e;e.child.return=e,e=e.child}if(!(e.flags&2))return e.stateNode}}function kh(e,t,n){var r=e.tag;if(r===5||r===6)e=e.stateNode,t?n.nodeType===8?n.parentNode.insertBefore(e,t):n.insertBefore(e,t):(n.nodeType===8?(t=n.parentNode,t.insertBefore(e,n)):(t=n,t.appendChild(e)),n=n._reactRootContainer,n!=null||t.onclick!==null||(t.onclick=Mf));else if(r!==4&&r!==27&&(e=e.child,e!==null))for(kh(e,t,n),e=e.sibling;e!==null;)kh(e,t,n),e=e.sibling}function uc(e,t,n){var r=e.tag;if(r===5||r===6)e=e.stateNode,t?n.insertBefore(e,t):n.appendChild(e);else if(r!==4&&r!==27&&(e=e.child,e!==null))for(uc(e,t,n),e=e.sibling;e!==null;)uc(e,t,n),e=e.sibling}var nr=!1,Ne=!1,Wd=!1,Jb=typeof WeakSet=="function"?WeakSet:Set,Qe=null,Ib=!1;function Kk(e,t){if(e=e.containerInfo,Xh=bc,e=m_(e),Rm(e)){if("selectionStart"in e)var n={start:e.selectionStart,end:e.selectionEnd};else e:{n=(n=e.ownerDocument)&&n.defaultView||window;var r=n.getSelection&&n.getSelection();if(r&&r.rangeCount!==0){n=r.anchorNode;var a=r.anchorOffset,i=r.focusNode;r=r.focusOffset;try{n.nodeType,i.nodeType}catch{n=null;break e}var l=0,o=-1,s=-1,u=0,c=0,d=e,f=null;t:for(;;){for(var h;d!==n||a!==0&&d.nodeType!==3||(o=l+a),d!==i||r!==0&&d.nodeType!==3||(s=l+r),d.nodeType===3&&(l+=d.nodeValue.length),(h=d.firstChild)!==null;)f=d,d=h;for(;;){if(d===e)break t;if(f===n&&++u===a&&(o=l),f===i&&++c===r&&(s=l),(h=d.nextSibling)!==null)break;d=f,f=d.parentNode}d=h}n=o===-1||s===-1?null:{start:o,end:s}}else n=null}n=n||{start:0,end:0}}else n=null;for(Ph={focusedElem:e,selectionRange:n},bc=!1,Qe=t;Qe!==null;)if(t=Qe,e=t.child,(t.subtreeFlags&1028)!==0&&e!==null)e.return=t,Qe=e;else for(;Qe!==null;){switch(t=Qe,i=t.alternate,e=t.flags,t.tag){case 0:break;case 11:case 15:break;case 1:if(e&1024&&i!==null){e=void 0,n=t,a=i.memoizedProps,i=i.memoizedState,r=n.stateNode;try{var b=wa(n.type,a,n.elementType===n.type);e=r.getSnapshotBeforeUpdate(b,i),r.__reactInternalSnapshotBeforeUpdate=e}catch(S){ye(n,n.return,S)}}break;case 3:if(e&1024){if(e=t.stateNode.containerInfo,n=e.nodeType,n===9)Qh(e);else if(n===1)switch(e.nodeName){case"HEAD":case"HTML":case"BODY":Qh(e);break;default:e.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if(e&1024)throw Error(A(163))}if(e=t.sibling,e!==null){e.return=t.return,Qe=e;break}Qe=t.return}return b=Ib,Ib=!1,b}function A2(e,t,n){var r=n.flags;switch(n.tag){case 0:case 11:case 15:Wn(e,n),r&4&&ls(5,n);break;case 1:if(Wn(e,n),r&4)if(e=n.stateNode,t===null)try{e.componentDidMount()}catch(o){ye(n,n.return,o)}else{var a=wa(n.type,t.memoizedProps);t=t.memoizedState;try{e.componentDidUpdate(a,t,e.__reactInternalSnapshotBeforeUpdate)}catch(o){ye(n,n.return,o)}}r&64&&O2(n),r&512&&pa(n,n.return);break;case 3:if(Wn(e,n),r&64&&(r=n.updateQueue,r!==null)){if(e=null,n.child!==null)switch(n.child.tag){case 27:case 5:e=n.child.stateNode;break;case 1:e=n.child.stateNode}try{x2(r,e)}catch(o){ye(n,n.return,o)}}break;case 26:Wn(e,n),r&512&&pa(n,n.return);break;case 27:case 5:Wn(e,n),t===null&&r&4&&T2(n),r&512&&pa(n,n.return);break;case 12:Wn(e,n);break;case 13:Wn(e,n),r&4&&$2(e,n);break;case 22:if(a=n.memoizedState!==null||nr,!a){t=t!==null&&t.memoizedState!==null||Ne;var i=nr,l=Ne;nr=a,(Ne=t)&&!l?Cr(e,n,(n.subtreeFlags&8772)!==0):Wn(e,n),nr=i,Ne=l}r&512&&(n.memoizedProps.mode==="manual"?pa(n,n.return):Ct(n,n.return));break;default:Wn(e,n)}}function R2(e){var t=e.alternate;t!==null&&(e.alternate=null,R2(t)),e.child=null,e.deletions=null,e.sibling=null,e.tag===5&&(t=e.stateNode,t!==null&&xm(t)),e.stateNode=null,e.return=null,e.dependencies=null,e.memoizedProps=null,e.memoizedState=null,e.pendingProps=null,e.stateNode=null,e.updateQueue=null}var Le=null,wt=!1;function Zn(e,t,n){for(n=n.child;n!==null;)C2(e,t,n),n=n.sibling}function C2(e,t,n){if(Nt&&typeof Nt.onCommitFiberUnmount=="function")try{Nt.onCommitFiberUnmount(Jo,n)}catch{}switch(n.tag){case 26:Ne||Ct(n,t),Zn(e,t,n),n.memoizedState?n.memoizedState.count--:n.stateNode&&(n=n.stateNode,n.parentNode.removeChild(n));break;case 27:Ne||Ct(n,t);var r=Le,a=wt;for(Le=n.stateNode,Zn(e,t,n),n=n.stateNode,t=n.attributes;t.length;)n.removeAttributeNode(t[0]);xm(n),Le=r,wt=a;break;case 5:Ne||Ct(n,t);case 6:a=Le;var i=wt;if(Le=null,Zn(e,t,n),Le=a,wt=i,Le!==null)if(wt)try{e=Le,r=n.stateNode,e.nodeType===8?e.parentNode.removeChild(r):e.removeChild(r)}catch(l){ye(n,t,l)}else try{Le.removeChild(n.stateNode)}catch(l){ye(n,t,l)}break;case 18:Le!==null&&(wt?(t=Le,n=n.stateNode,t.nodeType===8?up(t.parentNode,n):t.nodeType===1&&up(t,n),Bo(t)):up(Le,n.stateNode));break;case 4:r=Le,a=wt,Le=n.stateNode.containerInfo,wt=!0,Zn(e,t,n),Le=r,wt=a;break;case 0:case 11:case 14:case 15:Ne||ea(2,n,t),Ne||ea(4,n,t),Zn(e,t,n);break;case 1:Ne||(Ct(n,t),r=n.stateNode,typeof r.componentWillUnmount=="function"&&_2(n,t,r)),Zn(e,t,n);break;case 21:Zn(e,t,n);break;case 22:Ne||Ct(n,t),Ne=(r=Ne)||n.memoizedState!==null,Zn(e,t,n),Ne=r;break;default:Zn(e,t,n)}}function $2(e,t){if(t.memoizedState===null&&(e=t.alternate,e!==null&&(e=e.memoizedState,e!==null&&(e=e.dehydrated,e!==null))))try{Bo(e)}catch(n){ye(t,t.return,n)}}function Zk(e){switch(e.tag){case 13:case 19:var t=e.stateNode;return t===null&&(t=e.stateNode=new Jb),t;case 22:return e=e.stateNode,t=e._retryCache,t===null&&(t=e._retryCache=new Jb),t;default:throw Error(A(435,e.tag))}}function Jd(e,t){var n=Zk(e);t.forEach(function(r){var a=uz.bind(null,e,r);n.has(r)||(n.add(r),r.then(a,a))})}function Ht(e,t){var n=t.deletions;if(n!==null)for(var r=0;r<n.length;r++){var a=n[r],i=e,l=t,o=l;e:for(;o!==null;){switch(o.tag){case 27:case 5:Le=o.stateNode,wt=!1;break e;case 3:Le=o.stateNode.containerInfo,wt=!0;break e;case 4:Le=o.stateNode.containerInfo,wt=!0;break e}o=o.return}if(Le===null)throw Error(A(160));C2(i,l,a),Le=null,wt=!1,i=a.alternate,i!==null&&(i.return=null),a.return=null}if(t.subtreeFlags&13878)for(t=t.child;t!==null;)N2(t,e),t=t.sibling}var hn=null;function N2(e,t){var n=e.alternate,r=e.flags;switch(e.tag){case 0:case 11:case 14:case 15:Ht(t,e),qt(e),r&4&&(ea(3,e,e.return),ls(3,e),ea(5,e,e.return));break;case 1:Ht(t,e),qt(e),r&512&&(Ne||n===null||Ct(n,n.return)),r&64&&nr&&(e=e.updateQueue,e!==null&&(r=e.callbacks,r!==null&&(n=e.shared.hiddenCallbacks,e.shared.hiddenCallbacks=n===null?r:n.concat(r))));break;case 26:var a=hn;if(Ht(t,e),qt(e),r&512&&(Ne||n===null||Ct(n,n.return)),r&4){var i=n!==null?n.memoizedState:null;if(r=e.memoizedState,n===null)if(r===null)if(e.stateNode===null){e:{r=e.type,n=e.memoizedProps,a=a.ownerDocument||a;t:switch(r){case"title":i=a.getElementsByTagName("title")[0],(!i||i[xo]||i[st]||i.namespaceURI==="http://www.w3.org/2000/svg"||i.hasAttribute("itemprop"))&&(i=a.createElement(r),a.head.insertBefore(i,a.querySelector("head > title"))),at(i,r,n),i[st]=e,Ke(i),r=i;break e;case"link":var l=m1("link","href",a).get(r+(n.href||""));if(l){for(var o=0;o<l.length;o++)if(i=l[o],i.getAttribute("href")===(n.href==null?null:n.href)&&i.getAttribute("rel")===(n.rel==null?null:n.rel)&&i.getAttribute("title")===(n.title==null?null:n.title)&&i.getAttribute("crossorigin")===(n.crossOrigin==null?null:n.crossOrigin)){l.splice(o,1);break t}}i=a.createElement(r),at(i,r,n),a.head.appendChild(i);break;case"meta":if(l=m1("meta","content",a).get(r+(n.content||""))){for(o=0;o<l.length;o++)if(i=l[o],i.getAttribute("content")===(n.content==null?null:""+n.content)&&i.getAttribute("name")===(n.name==null?null:n.name)&&i.getAttribute("property")===(n.property==null?null:n.property)&&i.getAttribute("http-equiv")===(n.httpEquiv==null?null:n.httpEquiv)&&i.getAttribute("charset")===(n.charSet==null?null:n.charSet)){l.splice(o,1);break t}}i=a.createElement(r),at(i,r,n),a.head.appendChild(i);break;default:throw Error(A(468,r))}i[st]=e,Ke(i),r=i}e.stateNode=r}else v1(a,e.type,e.stateNode);else e.stateNode=y1(a,r,e.memoizedProps);else i!==r?(i===null?n.stateNode!==null&&(n=n.stateNode,n.parentNode.removeChild(n)):i.count--,r===null?v1(a,e.type,e.stateNode):y1(a,r,e.memoizedProps)):r===null&&e.stateNode!==null&&Wb(e,e.memoizedProps,n.memoizedProps)}break;case 27:if(r&4&&e.alternate===null){a=e.stateNode,i=e.memoizedProps;try{for(var s=a.firstChild;s;){var u=s.nextSibling,c=s.nodeName;s[xo]||c==="HEAD"||c==="BODY"||c==="SCRIPT"||c==="STYLE"||c==="LINK"&&s.rel.toLowerCase()==="stylesheet"||a.removeChild(s),s=u}for(var d=e.type,f=a.attributes;f.length;)a.removeAttributeNode(f[0]);at(a,d,i),a[st]=e,a[xt]=i}catch(b){ye(e,e.return,b)}}case 5:if(Ht(t,e),qt(e),r&512&&(Ne||n===null||Ct(n,n.return)),e.flags&32){a=e.stateNode;try{Xi(a,"")}catch(b){ye(e,e.return,b)}}r&4&&e.stateNode!=null&&(a=e.memoizedProps,Wb(e,a,n!==null?n.memoizedProps:a)),r&1024&&(Wd=!0);break;case 6:if(Ht(t,e),qt(e),r&4){if(e.stateNode===null)throw Error(A(162));r=e.memoizedProps,n=e.stateNode;try{n.nodeValue=r}catch(b){ye(e,e.return,b)}}break;case 3:if(Su=null,a=hn,hn=mc(t.containerInfo),Ht(t,e),hn=a,qt(e),r&4&&n!==null&&n.memoizedState.isDehydrated)try{Bo(t.containerInfo)}catch(b){ye(e,e.return,b)}Wd&&(Wd=!1,M2(e));break;case 4:r=hn,hn=mc(e.stateNode.containerInfo),Ht(t,e),qt(e),hn=r;break;case 12:Ht(t,e),qt(e);break;case 13:Ht(t,e),qt(e),e.child.flags&8192&&e.memoizedState!==null!=(n!==null&&n.memoizedState!==null)&&(tv=kn()),r&4&&(r=e.updateQueue,r!==null&&(e.updateQueue=null,Jd(e,r)));break;case 22:if(r&512&&(Ne||n===null||Ct(n,n.return)),s=e.memoizedState!==null,u=n!==null&&n.memoizedState!==null,c=nr,d=Ne,nr=c||s,Ne=d||u,Ht(t,e),Ne=d,nr=c,qt(e),t=e.stateNode,t._current=e,t._visibility&=-3,t._visibility|=t._pendingVisibility&2,r&8192&&(t._visibility=s?t._visibility&-2:t._visibility|1,s&&(t=nr||Ne,n===null||u||t||li(e)),e.memoizedProps===null||e.memoizedProps.mode!=="manual"))e:for(n=null,t=e;;){if(t.tag===5||t.tag===26||t.tag===27){if(n===null){u=n=t;try{if(a=u.stateNode,s)i=a.style,typeof i.setProperty=="function"?i.setProperty("display","none","important"):i.display="none";else{l=u.stateNode,o=u.memoizedProps.style;var h=o!=null&&o.hasOwnProperty("display")?o.display:null;l.style.display=h==null||typeof h=="boolean"?"":(""+h).trim()}}catch(b){ye(u,u.return,b)}}}else if(t.tag===6){if(n===null){u=t;try{u.stateNode.nodeValue=s?"":u.memoizedProps}catch(b){ye(u,u.return,b)}}}else if((t.tag!==22&&t.tag!==23||t.memoizedState===null||t===e)&&t.child!==null){t.child.return=t,t=t.child;continue}if(t===e)break e;for(;t.sibling===null;){if(t.return===null||t.return===e)break e;n===t&&(n=null),t=t.return}n===t&&(n=null),t.sibling.return=t.return,t=t.sibling}r&4&&(r=e.updateQueue,r!==null&&(n=r.retryQueue,n!==null&&(r.retryQueue=null,Jd(e,n))));break;case 19:Ht(t,e),qt(e),r&4&&(r=e.updateQueue,r!==null&&(e.updateQueue=null,Jd(e,r)));break;case 21:break;default:Ht(t,e),qt(e)}}function qt(e){var t=e.flags;if(t&2){try{if(e.tag!==27){e:{for(var n=e.return;n!==null;){if(w2(n)){var r=n;break e}n=n.return}throw Error(A(160))}switch(r.tag){case 27:var a=r.stateNode,i=Zd(e);uc(e,i,a);break;case 5:var l=r.stateNode;r.flags&32&&(Xi(l,""),r.flags&=-33);var o=Zd(e);uc(e,o,l);break;case 3:case 4:var s=r.stateNode.containerInfo,u=Zd(e);kh(e,u,s);break;default:throw Error(A(161))}}}catch(c){ye(e,e.return,c)}e.flags&=-3}t&4096&&(e.flags&=-4097)}function M2(e){if(e.subtreeFlags&1024)for(e=e.child;e!==null;){var t=e;M2(t),t.tag===5&&t.flags&1024&&t.stateNode.reset(),e=e.sibling}}function Wn(e,t){if(t.subtreeFlags&8772)for(t=t.child;t!==null;)A2(e,t.alternate,t),t=t.sibling}function li(e){for(e=e.child;e!==null;){var t=e;switch(t.tag){case 0:case 11:case 14:case 15:ea(4,t,t.return),li(t);break;case 1:Ct(t,t.return);var n=t.stateNode;typeof n.componentWillUnmount=="function"&&_2(t,t.return,n),li(t);break;case 26:case 27:case 5:Ct(t,t.return),li(t);break;case 22:Ct(t,t.return),t.memoizedState===null&&li(t);break;default:li(t)}e=e.sibling}}function Cr(e,t,n){for(n=n&&(t.subtreeFlags&8772)!==0,t=t.child;t!==null;){var r=t.alternate,a=e,i=t,l=i.flags;switch(i.tag){case 0:case 11:case 15:Cr(a,i,n),ls(4,i);break;case 1:if(Cr(a,i,n),r=i,a=r.stateNode,typeof a.componentDidMount=="function")try{a.componentDidMount()}catch(u){ye(r,r.return,u)}if(r=i,a=r.updateQueue,a!==null){var o=r.stateNode;try{var s=a.shared.hiddenCallbacks;if(s!==null)for(a.shared.hiddenCallbacks=null,a=0;a<s.length;a++)E2(s[a],o)}catch(u){ye(r,r.return,u)}}n&&l&64&&O2(i),pa(i,i.return);break;case 26:case 27:case 5:Cr(a,i,n),n&&r===null&&l&4&&T2(i),pa(i,i.return);break;case 12:Cr(a,i,n);break;case 13:Cr(a,i,n),n&&l&4&&$2(a,i);break;case 22:i.memoizedState===null&&Cr(a,i,n),pa(i,i.return);break;default:Cr(a,i,n)}t=t.sibling}}function Zm(e,t){var n=null;e!==null&&e.memoizedState!==null&&e.memoizedState.cachePool!==null&&(n=e.memoizedState.cachePool.pool),e=null,t.memoizedState!==null&&t.memoizedState.cachePool!==null&&(e=t.memoizedState.cachePool.pool),e!==n&&(e!=null&&e.refCount++,n!=null&&rs(n))}function Wm(e,t){e=null,t.alternate!==null&&(e=t.alternate.memoizedState.cache),t=t.memoizedState.cache,t!==e&&(t.refCount++,e!=null&&rs(e))}function Rr(e,t,n,r){if(t.subtreeFlags&10256)for(t=t.child;t!==null;)j2(e,t,n,r),t=t.sibling}function j2(e,t,n,r){var a=t.flags;switch(t.tag){case 0:case 11:case 15:Rr(e,t,n,r),a&2048&&ls(9,t);break;case 3:Rr(e,t,n,r),a&2048&&(e=null,t.alternate!==null&&(e=t.alternate.memoizedState.cache),t=t.memoizedState.cache,t!==e&&(t.refCount++,e!=null&&rs(e)));break;case 12:if(a&2048){Rr(e,t,n,r),e=t.stateNode;try{var i=t.memoizedProps,l=i.id,o=i.onPostCommit;typeof o=="function"&&o(l,t.alternate===null?"mount":"update",e.passiveEffectDuration,-0)}catch(s){ye(t,t.return,s)}}else Rr(e,t,n,r);break;case 23:break;case 22:i=t.stateNode,t.memoizedState!==null?i._visibility&4?Rr(e,t,n,r):io(e,t):i._visibility&4?Rr(e,t,n,r):(i._visibility|=4,oi(e,t,n,r,(t.subtreeFlags&10256)!==0)),a&2048&&Zm(t.alternate,t);break;case 24:Rr(e,t,n,r),a&2048&&Wm(t.alternate,t);break;default:Rr(e,t,n,r)}}function oi(e,t,n,r,a){for(a=a&&(t.subtreeFlags&10256)!==0,t=t.child;t!==null;){var i=e,l=t,o=n,s=r,u=l.flags;switch(l.tag){case 0:case 11:case 15:oi(i,l,o,s,a),ls(8,l);break;case 23:break;case 22:var c=l.stateNode;l.memoizedState!==null?c._visibility&4?oi(i,l,o,s,a):io(i,l):(c._visibility|=4,oi(i,l,o,s,a)),a&&u&2048&&Zm(l.alternate,l);break;case 24:oi(i,l,o,s,a),a&&u&2048&&Wm(l.alternate,l);break;default:oi(i,l,o,s,a)}t=t.sibling}}function io(e,t){if(t.subtreeFlags&10256)for(t=t.child;t!==null;){var n=e,r=t,a=r.flags;switch(r.tag){case 22:io(n,r),a&2048&&Zm(r.alternate,r);break;case 24:io(n,r),a&2048&&Wm(r.alternate,r);break;default:io(n,r)}t=t.sibling}}var zl=8192;function ti(e){if(e.subtreeFlags&zl)for(e=e.child;e!==null;)D2(e),e=e.sibling}function D2(e){switch(e.tag){case 26:ti(e),e.flags&zl&&e.memoizedState!==null&&Lz(hn,e.memoizedState,e.memoizedProps);break;case 5:ti(e);break;case 3:case 4:var t=hn;hn=mc(e.stateNode.containerInfo),ti(e),hn=t;break;case 22:e.memoizedState===null&&(t=e.alternate,t!==null&&t.memoizedState!==null?(t=zl,zl=16777216,ti(e),zl=t):ti(e));break;default:ti(e)}}function B2(e){var t=e.alternate;if(t!==null&&(e=t.child,e!==null)){t.child=null;do t=e.sibling,e.sibling=null,e=t;while(e!==null)}}function wl(e){var t=e.deletions;if(e.flags&16){if(t!==null)for(var n=0;n<t.length;n++){var r=t[n];Qe=r,z2(r,e)}B2(e)}if(e.subtreeFlags&10256)for(e=e.child;e!==null;)k2(e),e=e.sibling}function k2(e){switch(e.tag){case 0:case 11:case 15:wl(e),e.flags&2048&&ea(9,e,e.return);break;case 3:wl(e);break;case 12:wl(e);break;case 22:var t=e.stateNode;e.memoizedState!==null&&t._visibility&4&&(e.return===null||e.return.tag!==13)?(t._visibility&=-5,vu(e)):wl(e);break;default:wl(e)}}function vu(e){var t=e.deletions;if(e.flags&16){if(t!==null)for(var n=0;n<t.length;n++){var r=t[n];Qe=r,z2(r,e)}B2(e)}for(e=e.child;e!==null;){switch(t=e,t.tag){case 0:case 11:case 15:ea(8,t,t.return),vu(t);break;case 22:n=t.stateNode,n._visibility&4&&(n._visibility&=-5,vu(t));break;default:vu(t)}e=e.sibling}}function z2(e,t){for(;Qe!==null;){var n=Qe;switch(n.tag){case 0:case 11:case 15:ea(8,n,t);break;case 23:case 22:if(n.memoizedState!==null&&n.memoizedState.cachePool!==null){var r=n.memoizedState.cachePool.pool;r!=null&&r.refCount++}break;case 24:rs(n.memoizedState.cache)}if(r=n.child,r!==null)r.return=n,Qe=r;else e:for(n=e;Qe!==null;){r=Qe;var a=r.sibling,i=r.return;if(R2(r),r===n){Qe=null;break e}if(a!==null){a.return=i,Qe=a;break e}Qe=i}}}function Wk(e,t,n,r){this.tag=e,this.key=n,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=t,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=r,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function It(e,t,n,r){return new Wk(e,t,n,r)}function Jm(e){return e=e.prototype,!(!e||!e.isReactComponent)}function Pr(e,t){var n=e.alternate;return n===null?(n=It(e.tag,t,e.key,e.mode),n.elementType=e.elementType,n.type=e.type,n.stateNode=e.stateNode,n.alternate=e,e.alternate=n):(n.pendingProps=t,n.type=e.type,n.flags=0,n.subtreeFlags=0,n.deletions=null),n.flags=e.flags&31457280,n.childLanes=e.childLanes,n.lanes=e.lanes,n.child=e.child,n.memoizedProps=e.memoizedProps,n.memoizedState=e.memoizedState,n.updateQueue=e.updateQueue,t=e.dependencies,n.dependencies=t===null?null:{lanes:t.lanes,firstContext:t.firstContext},n.sibling=e.sibling,n.index=e.index,n.ref=e.ref,n.refCleanup=e.refCleanup,n}function L2(e,t){e.flags&=31457282;var n=e.alternate;return n===null?(e.childLanes=0,e.lanes=t,e.child=null,e.subtreeFlags=0,e.memoizedProps=null,e.memoizedState=null,e.updateQueue=null,e.dependencies=null,e.stateNode=null):(e.childLanes=n.childLanes,e.lanes=n.lanes,e.child=n.child,e.subtreeFlags=0,e.deletions=null,e.memoizedProps=n.memoizedProps,e.memoizedState=n.memoizedState,e.updateQueue=n.updateQueue,e.type=n.type,t=n.dependencies,e.dependencies=t===null?null:{lanes:t.lanes,firstContext:t.firstContext}),e}function gu(e,t,n,r,a,i){var l=0;if(r=e,typeof e=="function")Jm(e)&&(l=1);else if(typeof e=="string")l=kz(e,n,Bn.current)?26:e==="html"||e==="head"||e==="body"?27:5;else e:switch(e){case ci:return ba(n.children,a,i,t);case UO:l=8,a|=24;break;case ih:return e=It(12,n,t,a|2),e.elementType=ih,e.lanes=i,e;case lh:return e=It(13,n,t,a),e.elementType=lh,e.lanes=i,e;case oh:return e=It(19,n,t,a),e.elementType=oh,e.lanes=i,e;case qO:return U2(n,a,i,t);default:if(typeof e=="object"&&e!==null)switch(e.$$typeof){case x8:case lr:l=10;break e;case HO:l=9;break e;case bm:l=11;break e;case Sm:l=14;break e;case Nr:l=16,r=null;break e}l=29,n=Error(A(130,e===null?"null":typeof e,"")),r=null}return t=It(l,n,t,a),t.elementType=e,t.type=r,t.lanes=i,t}function ba(e,t,n,r){return e=It(7,e,r,t),e.lanes=n,e}function U2(e,t,n,r){e=It(22,e,r,t),e.elementType=qO,e.lanes=n;var a={_visibility:1,_pendingVisibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null,_current:null,detach:function(){var i=a._current;if(i===null)throw Error(A(456));if(!(a._pendingVisibility&2)){var l=Jr(i,2);l!==null&&(a._pendingVisibility|=2,mt(l,i,2))}},attach:function(){var i=a._current;if(i===null)throw Error(A(456));if(a._pendingVisibility&2){var l=Jr(i,2);l!==null&&(a._pendingVisibility&=-3,mt(l,i,2))}}};return e.stateNode=a,e}function Id(e,t,n){return e=It(6,e,null,t),e.lanes=n,e}function ep(e,t,n){return t=It(4,e.children!==null?e.children:[],e.key,t),t.lanes=n,t.stateNode={containerInfo:e.containerInfo,pendingChildren:null,implementation:e.implementation},t}function Jn(e){e.flags|=4}function e1(e,t){if(t.type!=="stylesheet"||t.state.loading&4)e.flags&=-16777217;else if(e.flags|=16777216,!lT(t)){if(t=an.current,t!==null&&((te&4194176)===te?zn!==null:(te&62914560)!==te&&!(te&536870912)||t!==zn))throw Jl=Sh,T_;e.flags|=8192}}function Bs(e,t){t!==null&&(e.flags|=4),e.flags&16384&&(t=e.tag!==22?QO():536870912,e.lanes|=t,Ki|=t)}function Al(e,t){if(!ae)switch(e.tailMode){case"hidden":t=e.tail;for(var n=null;t!==null;)t.alternate!==null&&(n=t),t=t.sibling;n===null?e.tail=null:n.sibling=null;break;case"collapsed":n=e.tail;for(var r=null;n!==null;)n.alternate!==null&&(r=n),n=n.sibling;r===null?t||e.tail===null?e.tail=null:e.tail.sibling=null:r.sibling=null}}function we(e){var t=e.alternate!==null&&e.alternate.child===e.child,n=0,r=0;if(t)for(var a=e.child;a!==null;)n|=a.lanes|a.childLanes,r|=a.subtreeFlags&31457280,r|=a.flags&31457280,a.return=e,a=a.sibling;else for(a=e.child;a!==null;)n|=a.lanes|a.childLanes,r|=a.subtreeFlags,r|=a.flags,a.return=e,a=a.sibling;return e.subtreeFlags|=r,e.childLanes=n,t}function Jk(e,t,n){var r=t.pendingProps;switch(Mm(t),t.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return we(t),null;case 1:return we(t),null;case 3:return n=t.stateNode,r=null,e!==null&&(r=e.memoizedState.cache),t.memoizedState.cache!==r&&(t.flags|=2048),pr(Ye),Yi(),n.pendingContext&&(n.context=n.pendingContext,n.pendingContext=null),(e===null||e.child===null)&&(_l(t)?Jn(t):e===null||e.memoizedState.isDehydrated&&!(t.flags&256)||(t.flags|=1024,yn!==null&&(qh(yn),yn=null))),we(t),null;case 26:return n=t.memoizedState,e===null?(Jn(t),n!==null?(we(t),e1(t,n)):(we(t),t.flags&=-16777217)):n?n!==e.memoizedState?(Jn(t),we(t),e1(t,n)):(we(t),t.flags&=-16777217):(e.memoizedProps!==r&&Jn(t),we(t),t.flags&=-16777217),null;case 27:Ku(t),n=Yr.current;var a=t.type;if(e!==null&&t.stateNode!=null)e.memoizedProps!==r&&Jn(t);else{if(!r){if(t.stateNode===null)throw Error(A(166));return we(t),null}e=Bn.current,_l(t)?$b(t):(e=rT(a,r,n),t.stateNode=e,Jn(t))}return we(t),null;case 5:if(Ku(t),n=t.type,e!==null&&t.stateNode!=null)e.memoizedProps!==r&&Jn(t);else{if(!r){if(t.stateNode===null)throw Error(A(166));return we(t),null}if(e=Bn.current,_l(t))$b(t);else{switch(a=yc(Yr.current),e){case 1:e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case 2:e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;default:switch(n){case"svg":e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case"math":e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;case"script":e=a.createElement("div"),e.innerHTML="<script><\/script>",e=e.removeChild(e.firstChild);break;case"select":e=typeof r.is=="string"?a.createElement("select",{is:r.is}):a.createElement("select"),r.multiple?e.multiple=!0:r.size&&(e.size=r.size);break;default:e=typeof r.is=="string"?a.createElement(n,{is:r.is}):a.createElement(n)}}e[st]=t,e[xt]=r;e:for(a=t.child;a!==null;){if(a.tag===5||a.tag===6)e.appendChild(a.stateNode);else if(a.tag!==4&&a.tag!==27&&a.child!==null){a.child.return=a,a=a.child;continue}if(a===t)break e;for(;a.sibling===null;){if(a.return===null||a.return===t)break e;a=a.return}a.sibling.return=a.return,a=a.sibling}t.stateNode=e;e:switch(at(e,n,r),n){case"button":case"input":case"select":case"textarea":e=!!r.autoFocus;break e;case"img":e=!0;break e;default:e=!1}e&&Jn(t)}}return we(t),t.flags&=-16777217,null;case 6:if(e&&t.stateNode!=null)e.memoizedProps!==r&&Jn(t);else{if(typeof r!="string"&&t.stateNode===null)throw Error(A(166));if(e=Yr.current,_l(t)){if(e=t.stateNode,n=t.memoizedProps,r=null,a=ht,a!==null)switch(a.tag){case 27:case 5:r=a.memoizedProps}e[st]=t,e=!!(e.nodeValue===n||r!==null&&r.suppressHydrationWarning===!0||eT(e.nodeValue,n)),e||Oa(t)}else e=yc(e).createTextNode(r),e[st]=t,t.stateNode=e}return we(t),null;case 13:if(r=t.memoizedState,e===null||e.memoizedState!==null&&e.memoizedState.dehydrated!==null){if(a=_l(t),r!==null&&r.dehydrated!==null){if(e===null){if(!a)throw Error(A(318));if(a=t.memoizedState,a=a!==null?a.dehydrated:null,!a)throw Error(A(317));a[st]=t}else ns(),!(t.flags&128)&&(t.memoizedState=null),t.flags|=4;we(t),a=!1}else yn!==null&&(qh(yn),yn=null),a=!0;if(!a)return t.flags&256?(ur(t),t):(ur(t),null)}if(ur(t),t.flags&128)return t.lanes=n,t;if(n=r!==null,e=e!==null&&e.memoizedState!==null,n){r=t.child,a=null,r.alternate!==null&&r.alternate.memoizedState!==null&&r.alternate.memoizedState.cachePool!==null&&(a=r.alternate.memoizedState.cachePool.pool);var i=null;r.memoizedState!==null&&r.memoizedState.cachePool!==null&&(i=r.memoizedState.cachePool.pool),i!==a&&(r.flags|=2048)}return n!==e&&n&&(t.child.flags|=8192),Bs(t,t.updateQueue),we(t),null;case 4:return Yi(),e===null&&av(t.stateNode.containerInfo),we(t),null;case 10:return pr(t.type),we(t),null;case 19:if(We(Ve),a=t.memoizedState,a===null)return we(t),null;if(r=(t.flags&128)!==0,i=a.rendering,i===null)if(r)Al(a,!1);else{if(Me!==0||e!==null&&e.flags&128)for(e=t.child;e!==null;){if(i=rc(e),i!==null){for(t.flags|=128,Al(a,!1),e=i.updateQueue,t.updateQueue=e,Bs(t,e),t.subtreeFlags=0,e=n,n=t.child;n!==null;)L2(n,e),n=n.sibling;return Te(Ve,Ve.current&1|2),t.child}e=e.sibling}a.tail!==null&&kn()>cc&&(t.flags|=128,r=!0,Al(a,!1),t.lanes=4194304)}else{if(!r)if(e=rc(i),e!==null){if(t.flags|=128,r=!0,e=e.updateQueue,t.updateQueue=e,Bs(t,e),Al(a,!0),a.tail===null&&a.tailMode==="hidden"&&!i.alternate&&!ae)return we(t),null}else 2*kn()-a.renderingStartTime>cc&&n!==536870912&&(t.flags|=128,r=!0,Al(a,!1),t.lanes=4194304);a.isBackwards?(i.sibling=t.child,t.child=i):(e=a.last,e!==null?e.sibling=i:t.child=i,a.last=i)}return a.tail!==null?(t=a.tail,a.rendering=t,a.tail=t.sibling,a.renderingStartTime=kn(),t.sibling=null,e=Ve.current,Te(Ve,r?e&1|2:e&1),t):(we(t),null);case 22:case 23:return ur(t),jm(),r=t.memoizedState!==null,e!==null?e.memoizedState!==null!==r&&(t.flags|=8192):r&&(t.flags|=8192),r?n&536870912&&!(t.flags&128)&&(we(t),t.subtreeFlags&6&&(t.flags|=8192)):we(t),n=t.updateQueue,n!==null&&Bs(t,n.retryQueue),n=null,e!==null&&e.memoizedState!==null&&e.memoizedState.cachePool!==null&&(n=e.memoizedState.cachePool.pool),r=null,t.memoizedState!==null&&t.memoizedState.cachePool!==null&&(r=t.memoizedState.cachePool.pool),r!==n&&(t.flags|=2048),e!==null&&We(ga),null;case 24:return n=null,e!==null&&(n=e.memoizedState.cache),t.memoizedState.cache!==n&&(t.flags|=2048),pr(Ye),we(t),null;case 25:return null}throw Error(A(156,t.tag))}function Ik(e,t){switch(Mm(t),t.tag){case 1:return e=t.flags,e&65536?(t.flags=e&-65537|128,t):null;case 3:return pr(Ye),Yi(),e=t.flags,e&65536&&!(e&128)?(t.flags=e&-65537|128,t):null;case 26:case 27:case 5:return Ku(t),null;case 13:if(ur(t),e=t.memoizedState,e!==null&&e.dehydrated!==null){if(t.alternate===null)throw Error(A(340));ns()}return e=t.flags,e&65536?(t.flags=e&-65537|128,t):null;case 19:return We(Ve),null;case 4:return Yi(),null;case 10:return pr(t.type),null;case 22:case 23:return ur(t),jm(),e!==null&&We(ga),e=t.flags,e&65536?(t.flags=e&-65537|128,t):null;case 24:return pr(Ye),null;case 25:return null;default:return null}}function H2(e,t){switch(Mm(t),t.tag){case 3:pr(Ye),Yi();break;case 26:case 27:case 5:Ku(t);break;case 4:Yi();break;case 13:ur(t);break;case 19:We(Ve);break;case 10:pr(t.type);break;case 22:case 23:ur(t),jm(),e!==null&&We(ga);break;case 24:pr(Ye)}}var ez={getCacheForType:function(e){var t=ut(Ye),n=t.data.get(e);return n===void 0&&(n=e(),t.data.set(e,n)),n}},tz=typeof WeakMap=="function"?WeakMap:Map,Re=0,be=null,Z=null,te=0,ge=0,At=null,rr=!1,sl=!1,Im=!1,Sr=0,Me=0,ta=0,Sa=0,ev=0,en=0,Ki=0,lo=null,Cn=null,zh=!1,tv=0,cc=1/0,fc=null,Fr=null,ks=!1,ca=null,oo=0,Lh=0,Uh=null,so=0,Hh=null;function jt(){if(Re&2&&te!==0)return te&-te;if(P.T!==null){var e=Fi;return e!==0?e:rv()}return JO()}function q2(){en===0&&(en=!(te&536870912)||ae?FO():536870912);var e=an.current;return e!==null&&(e.flags|=32),en}function mt(e,t,n){(e===be&&ge===2||e.cancelPendingCommit!==null)&&(Zi(e,0),ar(e,te,en,!1)),es(e,n),(!(Re&2)||e!==be)&&(e===be&&(!(Re&2)&&(Sa|=n),Me===4&&ar(e,te,en,!1)),Yn(e))}function G2(e,t,n){if(Re&6)throw Error(A(327));var r=!n&&(t&60)===0&&(t&e.expiredLanes)===0||Io(e,t),a=r?az(e,t):tp(e,t,!0),i=r;do{if(a===0){sl&&!r&&ar(e,t,0,!1);break}else if(a===6)ar(e,t,0,!rr);else{if(n=e.current.alternate,i&&!nz(n)){a=tp(e,t,!1),i=!1;continue}if(a===2){if(i=t,e.errorRecoveryDisabledLanes&i)var l=0;else l=e.pendingLanes&-536870913,l=l!==0?l:l&536870912?536870912:0;if(l!==0){t=l;e:{var o=e;a=lo;var s=o.current.memoizedState.isDehydrated;if(s&&(Zi(o,l).flags|=256),l=tp(o,l,!1),l!==2){if(Im&&!s){o.errorRecoveryDisabledLanes|=i,Sa|=i,a=4;break e}i=Cn,Cn=a,i!==null&&qh(i)}a=l}if(i=!1,a!==2)continue}}if(a===1){Zi(e,0),ar(e,t,0,!0);break}e:{switch(r=e,a){case 0:case 1:throw Error(A(345));case 4:if((t&4194176)===t){ar(r,t,en,!rr);break e}break;case 2:Cn=null;break;case 3:case 5:break;default:throw Error(A(329))}if(r.finishedWork=n,r.finishedLanes=t,(t&62914560)===t&&(i=tv+300-kn(),10<i)){if(ar(r,t,en,!rr),Sf(r,0)!==0)break e;r.timeoutHandle=nT(t1.bind(null,r,n,Cn,fc,zh,t,en,Sa,Ki,rr,2,-0,0),i);break e}t1(r,n,Cn,fc,zh,t,en,Sa,Ki,rr,0,-0,0)}}break}while(!0);Yn(e)}function qh(e){Cn===null?Cn=e:Cn.push.apply(Cn,e)}function t1(e,t,n,r,a,i,l,o,s,u,c,d,f){var h=t.subtreeFlags;if((h&8192||(h&16785408)===16785408)&&(No={stylesheets:null,count:0,unsuspend:zz},D2(t),t=Uz(),t!==null)){e.cancelPendingCommit=t(r1.bind(null,e,n,r,a,l,o,s,1,d,f)),ar(e,i,l,!u);return}r1(e,n,r,a,l,o,s,c,d,f)}function nz(e){for(var t=e;;){var n=t.tag;if((n===0||n===11||n===15)&&t.flags&16384&&(n=t.updateQueue,n!==null&&(n=n.stores,n!==null)))for(var r=0;r<n.length;r++){var a=n[r],i=a.getSnapshot;a=a.value;try{if(!Bt(i(),a))return!1}catch{return!1}}if(n=t.child,t.subtreeFlags&16384&&n!==null)n.return=t,t=n;else{if(t===e)break;for(;t.sibling===null;){if(t.return===null||t.return===e)return!0;t=t.return}t.sibling.return=t.return,t=t.sibling}}return!0}function ar(e,t,n,r){t&=~ev,t&=~Sa,e.suspendedLanes|=t,e.pingedLanes&=~t,r&&(e.warmLanes|=t),r=e.expirationTimes;for(var a=t;0<a;){var i=31-Mt(a),l=1<<i;r[i]=-1,a&=~l}n!==0&&KO(e,n,t)}function $f(){return Re&6?!0:(os(0),!1)}function nv(){if(Z!==null){if(ge===0)var e=Z.return;else e=Z,cr=La=null,Hm(e),Ri=null,Ao=0,e=Z;for(;e!==null;)H2(e.alternate,e),e=e.return;Z=null}}function Zi(e,t){e.finishedWork=null,e.finishedLanes=0;var n=e.timeoutHandle;n!==-1&&(e.timeoutHandle=-1,Ez(n)),n=e.cancelPendingCommit,n!==null&&(e.cancelPendingCommit=null,n()),nv(),be=e,Z=n=Pr(e.current,null),te=t,ge=0,At=null,rr=!1,sl=Io(e,t),Im=!1,Ki=en=ev=Sa=ta=Me=0,Cn=lo=null,zh=!1,t&8&&(t|=t&32);var r=e.entangledLanes;if(r!==0)for(e=e.entanglements,r&=t;0<r;){var a=31-Mt(r),i=1<<a;t|=e[a],r&=~i}return Sr=t,_f(),n}function Y2(e,t){Q=null,P.H=Un,t===Wl?(t=jb(),ge=3):t===T_?(t=jb(),ge=4):ge=t===h2?8:t!==null&&typeof t=="object"&&typeof t.then=="function"?6:1,At=t,Z===null&&(Me=1,oc(e,Wt(t,e.current)))}function V2(){var e=P.H;return P.H=Un,e===null?Un:e}function X2(){var e=P.A;return P.A=ez,e}function Gh(){Me=4,rr||(te&4194176)!==te&&an.current!==null||(sl=!0),!(ta&134217727)&&!(Sa&134217727)||be===null||ar(be,te,en,!1)}function tp(e,t,n){var r=Re;Re|=2;var a=V2(),i=X2();(be!==e||te!==t)&&(fc=null,Zi(e,t)),t=!1;var l=Me;e:do try{if(ge!==0&&Z!==null){var o=Z,s=At;switch(ge){case 8:nv(),l=6;break e;case 3:case 2:case 6:an.current===null&&(t=!0);var u=ge;if(ge=0,At=null,Si(e,o,s,u),n&&sl){l=0;break e}break;default:u=ge,ge=0,At=null,Si(e,o,s,u)}}rz(),l=Me;break}catch(c){Y2(e,c)}while(!0);return t&&e.shellSuspendCounter++,cr=La=null,Re=r,P.H=a,P.A=i,Z===null&&(be=null,te=0,_f()),l}function rz(){for(;Z!==null;)P2(Z)}function az(e,t){var n=Re;Re|=2;var r=V2(),a=X2();be!==e||te!==t?(fc=null,cc=kn()+500,Zi(e,t)):sl=Io(e,t);e:do try{if(ge!==0&&Z!==null){t=Z;var i=At;t:switch(ge){case 1:ge=0,At=null,Si(e,t,i,1);break;case 2:if(Mb(i)){ge=0,At=null,n1(t);break}t=function(){ge===2&&be===e&&(ge=7),Yn(e)},i.then(t,t);break e;case 3:ge=7;break e;case 4:ge=5;break e;case 7:Mb(i)?(ge=0,At=null,n1(t)):(ge=0,At=null,Si(e,t,i,7));break;case 5:var l=null;switch(Z.tag){case 26:l=Z.memoizedState;case 5:case 27:var o=Z;if(!l||lT(l)){ge=0,At=null;var s=o.sibling;if(s!==null)Z=s;else{var u=o.return;u!==null?(Z=u,Nf(u)):Z=null}break t}}ge=0,At=null,Si(e,t,i,5);break;case 6:ge=0,At=null,Si(e,t,i,6);break;case 8:nv(),Me=6;break e;default:throw Error(A(462))}}iz();break}catch(c){Y2(e,c)}while(!0);return cr=La=null,P.H=r,P.A=a,Re=n,Z!==null?0:(be=null,te=0,_f(),Me)}function iz(){for(;Z!==null&&!A8();)P2(Z)}function P2(e){var t=b2(e.alternate,e,Sr);e.memoizedProps=e.pendingProps,t===null?Nf(e):Z=t}function n1(e){var t=e,n=t.alternate;switch(t.tag){case 15:case 0:t=Fb(n,t,t.pendingProps,t.type,void 0,te);break;case 11:t=Fb(n,t,t.pendingProps,t.type.render,t.ref,te);break;case 5:Hm(t);default:H2(n,t),t=Z=L2(t,Sr),t=b2(n,t,Sr)}e.memoizedProps=e.pendingProps,t===null?Nf(e):Z=t}function Si(e,t,n,r){cr=La=null,Hm(t),Ri=null,Ao=0;var a=t.return;try{if(Fk(e,a,t,n,te)){Me=1,oc(e,Wt(n,e.current)),Z=null;return}}catch(i){if(a!==null)throw Z=a,i;Me=1,oc(e,Wt(n,e.current)),Z=null;return}t.flags&32768?(ae||r===1?e=!0:sl||te&536870912?e=!1:(rr=e=!0,(r===2||r===3||r===6)&&(r=an.current,r!==null&&r.tag===13&&(r.flags|=16384))),F2(t,e)):Nf(t)}function Nf(e){var t=e;do{if(t.flags&32768){F2(t,rr);return}e=t.return;var n=Jk(t.alternate,t,Sr);if(n!==null){Z=n;return}if(t=t.sibling,t!==null){Z=t;return}Z=t=e}while(t!==null);Me===0&&(Me=5)}function F2(e,t){do{var n=Ik(e.alternate,e);if(n!==null){n.flags&=32767,Z=n;return}if(n=e.return,n!==null&&(n.flags|=32768,n.subtreeFlags=0,n.deletions=null),!t&&(e=e.sibling,e!==null)){Z=e;return}Z=e=n}while(e!==null);Me=6,Z=null}function r1(e,t,n,r,a,i,l,o,s,u){var c=P.T,d=Se.p;try{Se.p=2,P.T=null,lz(e,t,n,r,d,a,i,l,o,s,u)}finally{P.T=c,Se.p=d}}function lz(e,t,n,r,a,i,l,o){do Mi();while(ca!==null);if(Re&6)throw Error(A(327));var s=e.finishedWork;if(r=e.finishedLanes,s===null)return null;if(e.finishedWork=null,e.finishedLanes=0,s===e.current)throw Error(A(177));e.callbackNode=null,e.callbackPriority=0,e.cancelPendingCommit=null;var u=s.lanes|s.childLanes;if(u|=Cm,L8(e,r,u,i,l,o),e===be&&(Z=be=null,te=0),!(s.subtreeFlags&10256)&&!(s.flags&10256)||ks||(ks=!0,Lh=u,Uh=n,cz(Zu,function(){return Mi(),null})),n=(s.flags&15990)!==0,s.subtreeFlags&15990||n?(n=P.T,P.T=null,i=Se.p,Se.p=2,l=Re,Re|=4,Kk(e,s),N2(s,e),$k(Ph,e.containerInfo),bc=!!Xh,Ph=Xh=null,e.current=s,A2(e,s.alternate,s),R8(),Re=l,Se.p=i,P.T=n):e.current=s,ks?(ks=!1,ca=e,oo=r):Q2(e,u),u=e.pendingLanes,u===0&&(Fr=null),j8(s.stateNode),Yn(e),t!==null)for(a=e.onRecoverableError,s=0;s<t.length;s++)u=t[s],a(u.value,{componentStack:u.stack});return oo&3&&Mi(),u=e.pendingLanes,r&4194218&&u&42?e===Hh?so++:(so=0,Hh=e):so=0,os(0),null}function Q2(e,t){(e.pooledCacheLanes&=t)===0&&(t=e.pooledCache,t!=null&&(e.pooledCache=null,rs(t)))}function Mi(){if(ca!==null){var e=ca,t=Lh;Lh=0;var n=WO(oo),r=P.T,a=Se.p;try{if(Se.p=32>n?32:n,P.T=null,ca===null)var i=!1;else{n=Uh,Uh=null;var l=ca,o=oo;if(ca=null,oo=0,Re&6)throw Error(A(331));var s=Re;if(Re|=4,k2(l.current),j2(l,l.current,o,n),Re=s,os(0,!1),Nt&&typeof Nt.onPostCommitFiberRoot=="function")try{Nt.onPostCommitFiberRoot(Jo,l)}catch{}i=!0}return i}finally{Se.p=a,P.T=r,Q2(e,t)}}return!1}function a1(e,t,n){t=Wt(n,t),t=Ah(e.stateNode,t,2),e=Xr(e,t,2),e!==null&&(es(e,2),Yn(e))}function ye(e,t,n){if(e.tag===3)a1(e,e,n);else for(;t!==null;){if(t.tag===3){a1(t,e,n);break}else if(t.tag===1){var r=t.stateNode;if(typeof t.type.getDerivedStateFromError=="function"||typeof r.componentDidCatch=="function"&&(Fr===null||!Fr.has(r))){e=Wt(n,e),n=d2(2),r=Xr(t,n,2),r!==null&&(p2(n,r,t,e),es(r,2),Yn(r));break}}t=t.return}}function np(e,t,n){var r=e.pingCache;if(r===null){r=e.pingCache=new tz;var a=new Set;r.set(t,a)}else a=r.get(t),a===void 0&&(a=new Set,r.set(t,a));a.has(n)||(Im=!0,a.add(n),e=oz.bind(null,e,t,n),t.then(e,e))}function oz(e,t,n){var r=e.pingCache;r!==null&&r.delete(t),e.pingedLanes|=e.suspendedLanes&n,e.warmLanes&=~n,be===e&&(te&n)===n&&(Me===4||Me===3&&(te&62914560)===te&&300>kn()-tv?!(Re&2)&&Zi(e,0):ev|=n,Ki===te&&(Ki=0)),Yn(e)}function K2(e,t){t===0&&(t=QO()),e=Jr(e,t),e!==null&&(es(e,t),Yn(e))}function sz(e){var t=e.memoizedState,n=0;t!==null&&(n=t.retryLane),K2(e,n)}function uz(e,t){var n=0;switch(e.tag){case 13:var r=e.stateNode,a=e.memoizedState;a!==null&&(n=a.retryLane);break;case 19:r=e.stateNode;break;case 22:r=e.stateNode._retryCache;break;default:throw Error(A(314))}r!==null&&r.delete(t),K2(e,n)}function cz(e,t){return Em(e,t)}var dc=null,si=null,Yh=!1,pc=!1,rp=!1,Ea=0;function Yn(e){e!==si&&e.next===null&&(si===null?dc=si=e:si=si.next=e),pc=!0,Yh||(Yh=!0,dz(fz))}function os(e,t){if(!rp&&pc){rp=!0;do for(var n=!1,r=dc;r!==null;){if(e!==0){var a=r.pendingLanes;if(a===0)var i=0;else{var l=r.suspendedLanes,o=r.pingedLanes;i=(1<<31-Mt(42|e)+1)-1,i&=a&~(l&~o),i=i&201326677?i&201326677|1:i?i|2:0}i!==0&&(n=!0,i1(r,i))}else i=te,i=Sf(r,r===be?i:0),!(i&3)||Io(r,i)||(n=!0,i1(r,i));r=r.next}while(n);rp=!1}}function fz(){pc=Yh=!1;var e=0;Ea!==0&&(Sz()&&(e=Ea),Ea=0);for(var t=kn(),n=null,r=dc;r!==null;){var a=r.next,i=Z2(r,t);i===0?(r.next=null,n===null?dc=a:n.next=a,a===null&&(si=n)):(n=r,(e!==0||i&3)&&(pc=!0)),r=a}os(e)}function Z2(e,t){for(var n=e.suspendedLanes,r=e.pingedLanes,a=e.expirationTimes,i=e.pendingLanes&-62914561;0<i;){var l=31-Mt(i),o=1<<l,s=a[l];s===-1?(!(o&n)||o&r)&&(a[l]=z8(o,t)):s<=t&&(e.expiredLanes|=o),i&=~o}if(t=be,n=te,n=Sf(e,e===t?n:0),r=e.callbackNode,n===0||e===t&&ge===2||e.cancelPendingCommit!==null)return r!==null&&r!==null&&jd(r),e.callbackNode=null,e.callbackPriority=0;if(!(n&3)||Io(e,n)){if(t=n&-n,t===e.callbackPriority)return t;switch(r!==null&&jd(r),WO(n)){case 2:case 8:n=XO;break;case 32:n=Zu;break;case 268435456:n=PO;break;default:n=Zu}return r=W2.bind(null,e),n=Em(n,r),e.callbackPriority=t,e.callbackNode=n,t}return r!==null&&r!==null&&jd(r),e.callbackPriority=2,e.callbackNode=null,2}function W2(e,t){var n=e.callbackNode;if(Mi()&&e.callbackNode!==n)return null;var r=te;return r=Sf(e,e===be?r:0),r===0?null:(G2(e,r,t),Z2(e,kn()),e.callbackNode!=null&&e.callbackNode===n?W2.bind(null,e):null)}function i1(e,t){if(Mi())return null;G2(e,t,!0)}function dz(e){xz(function(){Re&6?Em(VO,e):e()})}function rv(){return Ea===0&&(Ea=FO()),Ea}function l1(e){return e==null||typeof e=="symbol"||typeof e=="boolean"?null:typeof e=="function"?e:fu(""+e)}function o1(e,t){var n=t.ownerDocument.createElement("input");return n.name=t.name,n.value=t.value,e.id&&n.setAttribute("form",e.id),t.parentNode.insertBefore(n,t),e=new FormData(e),n.parentNode.removeChild(n),e}function pz(e,t,n,r,a){if(t==="submit"&&n&&n.stateNode===a){var i=l1((a[xt]||null).action),l=r.submitter;l&&(t=(t=l[xt]||null)?l1(t.formAction):l.getAttribute("formAction"),t!==null&&(i=t,l=null));var o=new Ef("action","action",null,r,a);e.push({event:o,listeners:[{instance:null,listener:function(){if(r.defaultPrevented){if(Ea!==0){var s=l?o1(a,l):new FormData(a);Th(n,{pending:!0,data:s,method:a.method,action:i},null,s)}}else typeof i=="function"&&(o.preventDefault(),s=l?o1(a,l):new FormData(a),Th(n,{pending:!0,data:s,method:a.method,action:i},i,s))},currentTarget:a}]})}}for(var ap=0;ap<Rb.length;ap++){var ip=Rb[ap],hz=ip.toLowerCase(),yz=ip[0].toUpperCase()+ip.slice(1);En(hz,"on"+yz)}En(g_,"onAnimationEnd");En(b_,"onAnimationIteration");En(S_,"onAnimationStart");En("dblclick","onDoubleClick");En("focusin","onFocus");En("focusout","onBlur");En(Mk,"onTransitionRun");En(jk,"onTransitionStart");En(Dk,"onTransitionCancel");En(E_,"onTransitionEnd");Vi("onMouseEnter",["mouseout","mouseover"]);Vi("onMouseLeave",["mouseout","mouseover"]);Vi("onPointerEnter",["pointerout","pointerover"]);Vi("onPointerLeave",["pointerout","pointerover"]);ja("onChange","change click focusin focusout input keydown keyup selectionchange".split(" "));ja("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" "));ja("onBeforeInput",["compositionend","keypress","textInput","paste"]);ja("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" "));ja("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" "));ja("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var Co="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),mz=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(Co));function J2(e,t){t=(t&4)!==0;for(var n=0;n<e.length;n++){var r=e[n],a=r.event;r=r.listeners;e:{var i=void 0;if(t)for(var l=r.length-1;0<=l;l--){var o=r[l],s=o.instance,u=o.currentTarget;if(o=o.listener,s!==i&&a.isPropagationStopped())break e;i=o,a.currentTarget=u;try{i(a)}catch(c){lc(c)}a.currentTarget=null,i=s}else for(l=0;l<r.length;l++){if(o=r[l],s=o.instance,u=o.currentTarget,o=o.listener,s!==i&&a.isPropagationStopped())break e;i=o,a.currentTarget=u;try{i(a)}catch(c){lc(c)}a.currentTarget=null,i=s}}}}function J(e,t){var n=t[dh];n===void 0&&(n=t[dh]=new Set);var r=e+"__bubble";n.has(r)||(I2(t,e,2,!1),n.add(r))}function lp(e,t,n){var r=0;t&&(r|=4),I2(n,e,r,t)}var zs="_reactListening"+Math.random().toString(36).slice(2);function av(e){if(!e[zs]){e[zs]=!0,IO.forEach(function(n){n!=="selectionchange"&&(mz.has(n)||lp(n,!1,e),lp(n,!0,e))});var t=e.nodeType===9?e:e.ownerDocument;t===null||t[zs]||(t[zs]=!0,lp("selectionchange",!1,t))}}function I2(e,t,n,r){switch(fT(t)){case 2:var a=Gz;break;case 8:a=Yz;break;default:a=sv}n=a.bind(null,t,n,e),a=void 0,!mh||t!=="touchstart"&&t!=="touchmove"&&t!=="wheel"||(a=!0),r?a!==void 0?e.addEventListener(t,n,{capture:!0,passive:a}):e.addEventListener(t,n,!0):a!==void 0?e.addEventListener(t,n,{passive:a}):e.addEventListener(t,n,!1)}function op(e,t,n,r,a){var i=r;if(!(t&1)&&!(t&2)&&r!==null)e:for(;;){if(r===null)return;var l=r.tag;if(l===3||l===4){var o=r.stateNode.containerInfo;if(o===a||o.nodeType===8&&o.parentNode===a)break;if(l===4)for(l=r.return;l!==null;){var s=l.tag;if((s===3||s===4)&&(s=l.stateNode.containerInfo,s===a||s.nodeType===8&&s.parentNode===a))return;l=l.return}for(;o!==null;){if(l=da(o),l===null)return;if(s=l.tag,s===5||s===6||s===26||s===27){r=i=l;continue e}o=o.parentNode}}r=r.return}o_(function(){var u=i,c=_m(n),d=[];e:{var f=x_.get(e);if(f!==void 0){var h=Ef,b=e;switch(e){case"keypress":if(pu(n)===0)break e;case"keydown":case"keyup":h=uk;break;case"focusin":b="focus",h=Ld;break;case"focusout":b="blur",h=Ld;break;case"beforeblur":case"afterblur":h=Ld;break;case"click":if(n.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":h=mb;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":h=W8;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":h=dk;break;case g_:case b_:case S_:h=ek;break;case E_:h=hk;break;case"scroll":case"scrollend":h=K8;break;case"wheel":h=mk;break;case"copy":case"cut":case"paste":h=nk;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":h=gb;break;case"toggle":case"beforetoggle":h=gk}var S=(t&4)!==0,x=!S&&(e==="scroll"||e==="scrollend"),y=S?f!==null?f+"Capture":null:f;S=[];for(var p=u,v;p!==null;){var E=p;if(v=E.stateNode,E=E.tag,E!==5&&E!==26&&E!==27||v===null||y===null||(E=Oo(p,y),E!=null&&S.push($o(p,E,v))),x)break;p=p.return}0<S.length&&(f=new h(f,b,null,n,c),d.push({event:f,listeners:S}))}}if(!(t&7)){e:{if(f=e==="mouseover"||e==="pointerover",h=e==="mouseout"||e==="pointerout",f&&n!==yh&&(b=n.relatedTarget||n.fromElement)&&(da(b)||b[ll]))break e;if((h||f)&&(f=c.window===c?c:(f=c.ownerDocument)?f.defaultView||f.parentWindow:window,h?(b=n.relatedTarget||n.toElement,h=u,b=b?da(b):null,b!==null&&(x=il(b),S=b.tag,b!==x||S!==5&&S!==27&&S!==6)&&(b=null)):(h=null,b=u),h!==b)){if(S=mb,E="onMouseLeave",y="onMouseEnter",p="mouse",(e==="pointerout"||e==="pointerover")&&(S=gb,E="onPointerLeave",y="onPointerEnter",p="pointer"),x=h==null?f:kl(h),v=b==null?f:kl(b),f=new S(E,p+"leave",h,n,c),f.target=x,f.relatedTarget=v,E=null,da(c)===u&&(S=new S(y,p+"enter",b,n,c),S.target=v,S.relatedTarget=x,E=S),x=E,h&&b)t:{for(S=h,y=b,p=0,v=S;v;v=ni(v))p++;for(v=0,E=y;E;E=ni(E))v++;for(;0<p-v;)S=ni(S),p--;for(;0<v-p;)y=ni(y),v--;for(;p--;){if(S===y||y!==null&&S===y.alternate)break t;S=ni(S),y=ni(y)}S=null}else S=null;h!==null&&s1(d,f,h,S,!1),b!==null&&x!==null&&s1(d,x,b,S,!0)}}e:{if(f=u?kl(u):window,h=f.nodeName&&f.nodeName.toLowerCase(),h==="select"||h==="input"&&f.type==="file")var _=xb;else if(Eb(f))if(p_)_=Rk;else{_=wk;var O=Tk}else h=f.nodeName,!h||h.toLowerCase()!=="input"||f.type!=="checkbox"&&f.type!=="radio"?u&&Om(u.elementType)&&(_=xb):_=Ak;if(_&&(_=_(e,u))){d_(d,_,n,c);break e}O&&O(e,f,u),e==="focusout"&&u&&f.type==="number"&&u.memoizedProps.value!=null&&hh(f,"number",f.value)}switch(O=u?kl(u):window,e){case"focusin":(Eb(O)||O.contentEditable==="true")&&(hi=O,vh=u,Zl=null);break;case"focusout":Zl=vh=hi=null;break;case"mousedown":gh=!0;break;case"contextmenu":case"mouseup":case"dragend":gh=!1,Ab(d,n,c);break;case"selectionchange":if(Nk)break;case"keydown":case"keyup":Ab(d,n,c)}var T;if(Am)e:{switch(e){case"compositionstart":var w="onCompositionStart";break e;case"compositionend":w="onCompositionEnd";break e;case"compositionupdate":w="onCompositionUpdate";break e}w=void 0}else pi?c_(e,n)&&(w="onCompositionEnd"):e==="keydown"&&n.keyCode===229&&(w="onCompositionStart");w&&(u_&&n.locale!=="ko"&&(pi||w!=="onCompositionStart"?w==="onCompositionEnd"&&pi&&(T=s_()):(Hr=c,Tm="value"in Hr?Hr.value:Hr.textContent,pi=!0)),O=hc(u,w),0<O.length&&(w=new vb(w,e,null,n,c),d.push({event:w,listeners:O}),T?w.data=T:(T=f_(n),T!==null&&(w.data=T)))),(T=Sk?Ek(e,n):xk(e,n))&&(w=hc(u,"onBeforeInput"),0<w.length&&(O=new vb("onBeforeInput","beforeinput",null,n,c),d.push({event:O,listeners:w}),O.data=T)),pz(d,e,u,n,c)}J2(d,t)})}function $o(e,t,n){return{instance:e,listener:t,currentTarget:n}}function hc(e,t){for(var n=t+"Capture",r=[];e!==null;){var a=e,i=a.stateNode;a=a.tag,a!==5&&a!==26&&a!==27||i===null||(a=Oo(e,n),a!=null&&r.unshift($o(e,a,i)),a=Oo(e,t),a!=null&&r.push($o(e,a,i))),e=e.return}return r}function ni(e){if(e===null)return null;do e=e.return;while(e&&e.tag!==5&&e.tag!==27);return e||null}function s1(e,t,n,r,a){for(var i=t._reactName,l=[];n!==null&&n!==r;){var o=n,s=o.alternate,u=o.stateNode;if(o=o.tag,s!==null&&s===r)break;o!==5&&o!==26&&o!==27||u===null||(s=u,a?(u=Oo(n,i),u!=null&&l.unshift($o(n,u,s))):a||(u=Oo(n,i),u!=null&&l.push($o(n,u,s)))),n=n.return}l.length!==0&&e.push({event:t,listeners:l})}var vz=/\r\n?/g,gz=/\u0000|\uFFFD/g;function u1(e){return(typeof e=="string"?e:""+e).replace(vz,`
`).replace(gz,"")}function eT(e,t){return t=u1(t),u1(e)===t}function Mf(){}function de(e,t,n,r,a,i){switch(n){case"children":typeof r=="string"?t==="body"||t==="textarea"&&r===""||Xi(e,r):(typeof r=="number"||typeof r=="bigint")&&t!=="body"&&Xi(e,""+r);break;case"className":Cs(e,"class",r);break;case"tabIndex":Cs(e,"tabindex",r);break;case"dir":case"role":case"viewBox":case"width":case"height":Cs(e,n,r);break;case"style":l_(e,r,i);break;case"data":if(t!=="object"){Cs(e,"data",r);break}case"src":case"href":if(r===""&&(t!=="a"||n!=="href")){e.removeAttribute(n);break}if(r==null||typeof r=="function"||typeof r=="symbol"||typeof r=="boolean"){e.removeAttribute(n);break}r=fu(""+r),e.setAttribute(n,r);break;case"action":case"formAction":if(typeof r=="function"){e.setAttribute(n,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof i=="function"&&(n==="formAction"?(t!=="input"&&de(e,t,"name",a.name,a,null),de(e,t,"formEncType",a.formEncType,a,null),de(e,t,"formMethod",a.formMethod,a,null),de(e,t,"formTarget",a.formTarget,a,null)):(de(e,t,"encType",a.encType,a,null),de(e,t,"method",a.method,a,null),de(e,t,"target",a.target,a,null)));if(r==null||typeof r=="symbol"||typeof r=="boolean"){e.removeAttribute(n);break}r=fu(""+r),e.setAttribute(n,r);break;case"onClick":r!=null&&(e.onclick=Mf);break;case"onScroll":r!=null&&J("scroll",e);break;case"onScrollEnd":r!=null&&J("scrollend",e);break;case"dangerouslySetInnerHTML":if(r!=null){if(typeof r!="object"||!("__html"in r))throw Error(A(61));if(n=r.__html,n!=null){if(a.children!=null)throw Error(A(60));e.innerHTML=n}}break;case"multiple":e.multiple=r&&typeof r!="function"&&typeof r!="symbol";break;case"muted":e.muted=r&&typeof r!="function"&&typeof r!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(r==null||typeof r=="function"||typeof r=="boolean"||typeof r=="symbol"){e.removeAttribute("xlink:href");break}n=fu(""+r),e.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",n);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":r!=null&&typeof r!="function"&&typeof r!="symbol"?e.setAttribute(n,""+r):e.removeAttribute(n);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":r&&typeof r!="function"&&typeof r!="symbol"?e.setAttribute(n,""):e.removeAttribute(n);break;case"capture":case"download":r===!0?e.setAttribute(n,""):r!==!1&&r!=null&&typeof r!="function"&&typeof r!="symbol"?e.setAttribute(n,r):e.removeAttribute(n);break;case"cols":case"rows":case"size":case"span":r!=null&&typeof r!="function"&&typeof r!="symbol"&&!isNaN(r)&&1<=r?e.setAttribute(n,r):e.removeAttribute(n);break;case"rowSpan":case"start":r==null||typeof r=="function"||typeof r=="symbol"||isNaN(r)?e.removeAttribute(n):e.setAttribute(n,r);break;case"popover":J("beforetoggle",e),J("toggle",e),cu(e,"popover",r);break;case"xlinkActuate":Kn(e,"http://www.w3.org/1999/xlink","xlink:actuate",r);break;case"xlinkArcrole":Kn(e,"http://www.w3.org/1999/xlink","xlink:arcrole",r);break;case"xlinkRole":Kn(e,"http://www.w3.org/1999/xlink","xlink:role",r);break;case"xlinkShow":Kn(e,"http://www.w3.org/1999/xlink","xlink:show",r);break;case"xlinkTitle":Kn(e,"http://www.w3.org/1999/xlink","xlink:title",r);break;case"xlinkType":Kn(e,"http://www.w3.org/1999/xlink","xlink:type",r);break;case"xmlBase":Kn(e,"http://www.w3.org/XML/1998/namespace","xml:base",r);break;case"xmlLang":Kn(e,"http://www.w3.org/XML/1998/namespace","xml:lang",r);break;case"xmlSpace":Kn(e,"http://www.w3.org/XML/1998/namespace","xml:space",r);break;case"is":cu(e,"is",r);break;case"innerText":case"textContent":break;default:(!(2<n.length)||n[0]!=="o"&&n[0]!=="O"||n[1]!=="n"&&n[1]!=="N")&&(n=F8.get(n)||n,cu(e,n,r))}}function Vh(e,t,n,r,a,i){switch(n){case"style":l_(e,r,i);break;case"dangerouslySetInnerHTML":if(r!=null){if(typeof r!="object"||!("__html"in r))throw Error(A(61));if(n=r.__html,n!=null){if(a.children!=null)throw Error(A(60));e.innerHTML=n}}break;case"children":typeof r=="string"?Xi(e,r):(typeof r=="number"||typeof r=="bigint")&&Xi(e,""+r);break;case"onScroll":r!=null&&J("scroll",e);break;case"onScrollEnd":r!=null&&J("scrollend",e);break;case"onClick":r!=null&&(e.onclick=Mf);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!e_.hasOwnProperty(n))e:{if(n[0]==="o"&&n[1]==="n"&&(a=n.endsWith("Capture"),t=n.slice(2,a?n.length-7:void 0),i=e[xt]||null,i=i!=null?i[n]:null,typeof i=="function"&&e.removeEventListener(t,i,a),typeof r=="function")){typeof i!="function"&&i!==null&&(n in e?e[n]=null:e.hasAttribute(n)&&e.removeAttribute(n)),e.addEventListener(t,r,a);break e}n in e?e[n]=r:r===!0?e.setAttribute(n,""):cu(e,n,r)}}}function at(e,t,n){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":J("error",e),J("load",e);var r=!1,a=!1,i;for(i in n)if(n.hasOwnProperty(i)){var l=n[i];if(l!=null)switch(i){case"src":r=!0;break;case"srcSet":a=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(A(137,t));default:de(e,t,i,l,n,null)}}a&&de(e,t,"srcSet",n.srcSet,n,null),r&&de(e,t,"src",n.src,n,null);return;case"input":J("invalid",e);var o=i=l=a=null,s=null,u=null;for(r in n)if(n.hasOwnProperty(r)){var c=n[r];if(c!=null)switch(r){case"name":a=c;break;case"type":l=c;break;case"checked":s=c;break;case"defaultChecked":u=c;break;case"value":i=c;break;case"defaultValue":o=c;break;case"children":case"dangerouslySetInnerHTML":if(c!=null)throw Error(A(137,t));break;default:de(e,t,r,c,n,null)}}r_(e,i,o,s,u,l,a,!1),Wu(e);return;case"select":J("invalid",e),r=l=i=null;for(a in n)if(n.hasOwnProperty(a)&&(o=n[a],o!=null))switch(a){case"value":i=o;break;case"defaultValue":l=o;break;case"multiple":r=o;default:de(e,t,a,o,n,null)}t=i,n=l,e.multiple=!!r,t!=null?wi(e,!!r,t,!1):n!=null&&wi(e,!!r,n,!0);return;case"textarea":J("invalid",e),i=a=r=null;for(l in n)if(n.hasOwnProperty(l)&&(o=n[l],o!=null))switch(l){case"value":r=o;break;case"defaultValue":a=o;break;case"children":i=o;break;case"dangerouslySetInnerHTML":if(o!=null)throw Error(A(91));break;default:de(e,t,l,o,n,null)}i_(e,r,a,i),Wu(e);return;case"option":for(s in n)if(n.hasOwnProperty(s)&&(r=n[s],r!=null))switch(s){case"selected":e.selected=r&&typeof r!="function"&&typeof r!="symbol";break;default:de(e,t,s,r,n,null)}return;case"dialog":J("cancel",e),J("close",e);break;case"iframe":case"object":J("load",e);break;case"video":case"audio":for(r=0;r<Co.length;r++)J(Co[r],e);break;case"image":J("error",e),J("load",e);break;case"details":J("toggle",e);break;case"embed":case"source":case"link":J("error",e),J("load",e);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(u in n)if(n.hasOwnProperty(u)&&(r=n[u],r!=null))switch(u){case"children":case"dangerouslySetInnerHTML":throw Error(A(137,t));default:de(e,t,u,r,n,null)}return;default:if(Om(t)){for(c in n)n.hasOwnProperty(c)&&(r=n[c],r!==void 0&&Vh(e,t,c,r,n,void 0));return}}for(o in n)n.hasOwnProperty(o)&&(r=n[o],r!=null&&de(e,t,o,r,n,null))}function bz(e,t,n,r){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var a=null,i=null,l=null,o=null,s=null,u=null,c=null;for(h in n){var d=n[h];if(n.hasOwnProperty(h)&&d!=null)switch(h){case"checked":break;case"value":break;case"defaultValue":s=d;default:r.hasOwnProperty(h)||de(e,t,h,null,r,d)}}for(var f in r){var h=r[f];if(d=n[f],r.hasOwnProperty(f)&&(h!=null||d!=null))switch(f){case"type":i=h;break;case"name":a=h;break;case"checked":u=h;break;case"defaultChecked":c=h;break;case"value":l=h;break;case"defaultValue":o=h;break;case"children":case"dangerouslySetInnerHTML":if(h!=null)throw Error(A(137,t));break;default:h!==d&&de(e,t,f,h,r,d)}}ph(e,l,o,s,u,c,i,a);return;case"select":h=l=o=f=null;for(i in n)if(s=n[i],n.hasOwnProperty(i)&&s!=null)switch(i){case"value":break;case"multiple":h=s;default:r.hasOwnProperty(i)||de(e,t,i,null,r,s)}for(a in r)if(i=r[a],s=n[a],r.hasOwnProperty(a)&&(i!=null||s!=null))switch(a){case"value":f=i;break;case"defaultValue":o=i;break;case"multiple":l=i;default:i!==s&&de(e,t,a,i,r,s)}t=o,n=l,r=h,f!=null?wi(e,!!n,f,!1):!!r!=!!n&&(t!=null?wi(e,!!n,t,!0):wi(e,!!n,n?[]:"",!1));return;case"textarea":h=f=null;for(o in n)if(a=n[o],n.hasOwnProperty(o)&&a!=null&&!r.hasOwnProperty(o))switch(o){case"value":break;case"children":break;default:de(e,t,o,null,r,a)}for(l in r)if(a=r[l],i=n[l],r.hasOwnProperty(l)&&(a!=null||i!=null))switch(l){case"value":f=a;break;case"defaultValue":h=a;break;case"children":break;case"dangerouslySetInnerHTML":if(a!=null)throw Error(A(91));break;default:a!==i&&de(e,t,l,a,r,i)}a_(e,f,h);return;case"option":for(var b in n)if(f=n[b],n.hasOwnProperty(b)&&f!=null&&!r.hasOwnProperty(b))switch(b){case"selected":e.selected=!1;break;default:de(e,t,b,null,r,f)}for(s in r)if(f=r[s],h=n[s],r.hasOwnProperty(s)&&f!==h&&(f!=null||h!=null))switch(s){case"selected":e.selected=f&&typeof f!="function"&&typeof f!="symbol";break;default:de(e,t,s,f,r,h)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var S in n)f=n[S],n.hasOwnProperty(S)&&f!=null&&!r.hasOwnProperty(S)&&de(e,t,S,null,r,f);for(u in r)if(f=r[u],h=n[u],r.hasOwnProperty(u)&&f!==h&&(f!=null||h!=null))switch(u){case"children":case"dangerouslySetInnerHTML":if(f!=null)throw Error(A(137,t));break;default:de(e,t,u,f,r,h)}return;default:if(Om(t)){for(var x in n)f=n[x],n.hasOwnProperty(x)&&f!==void 0&&!r.hasOwnProperty(x)&&Vh(e,t,x,void 0,r,f);for(c in r)f=r[c],h=n[c],!r.hasOwnProperty(c)||f===h||f===void 0&&h===void 0||Vh(e,t,c,f,r,h);return}}for(var y in n)f=n[y],n.hasOwnProperty(y)&&f!=null&&!r.hasOwnProperty(y)&&de(e,t,y,null,r,f);for(d in r)f=r[d],h=n[d],!r.hasOwnProperty(d)||f===h||f==null&&h==null||de(e,t,d,f,r,h)}var Xh=null,Ph=null;function yc(e){return e.nodeType===9?e:e.ownerDocument}function c1(e){switch(e){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function tT(e,t){if(e===0)switch(t){case"svg":return 1;case"math":return 2;default:return 0}return e===1&&t==="foreignObject"?0:e}function Fh(e,t){return e==="textarea"||e==="noscript"||typeof t.children=="string"||typeof t.children=="number"||typeof t.children=="bigint"||typeof t.dangerouslySetInnerHTML=="object"&&t.dangerouslySetInnerHTML!==null&&t.dangerouslySetInnerHTML.__html!=null}var sp=null;function Sz(){var e=window.event;return e&&e.type==="popstate"?e===sp?!1:(sp=e,!0):(sp=null,!1)}var nT=typeof setTimeout=="function"?setTimeout:void 0,Ez=typeof clearTimeout=="function"?clearTimeout:void 0,f1=typeof Promise=="function"?Promise:void 0,xz=typeof queueMicrotask=="function"?queueMicrotask:typeof f1<"u"?function(e){return f1.resolve(null).then(e).catch(Oz)}:nT;function Oz(e){setTimeout(function(){throw e})}function up(e,t){var n=t,r=0;do{var a=n.nextSibling;if(e.removeChild(n),a&&a.nodeType===8)if(n=a.data,n==="/$"){if(r===0){e.removeChild(a),Bo(t);return}r--}else n!=="$"&&n!=="$?"&&n!=="$!"||r++;n=a}while(n);Bo(t)}function Qh(e){var t=e.firstChild;for(t&&t.nodeType===10&&(t=t.nextSibling);t;){var n=t;switch(t=t.nextSibling,n.nodeName){case"HTML":case"HEAD":case"BODY":Qh(n),xm(n);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(n.rel.toLowerCase()==="stylesheet")continue}e.removeChild(n)}}function _z(e,t,n,r){for(;e.nodeType===1;){var a=n;if(e.nodeName.toLowerCase()!==t.toLowerCase()){if(!r&&(e.nodeName!=="INPUT"||e.type!=="hidden"))break}else if(r){if(!e[xo])switch(t){case"meta":if(!e.hasAttribute("itemprop"))break;return e;case"link":if(i=e.getAttribute("rel"),i==="stylesheet"&&e.hasAttribute("data-precedence"))break;if(i!==a.rel||e.getAttribute("href")!==(a.href==null?null:a.href)||e.getAttribute("crossorigin")!==(a.crossOrigin==null?null:a.crossOrigin)||e.getAttribute("title")!==(a.title==null?null:a.title))break;return e;case"style":if(e.hasAttribute("data-precedence"))break;return e;case"script":if(i=e.getAttribute("src"),(i!==(a.src==null?null:a.src)||e.getAttribute("type")!==(a.type==null?null:a.type)||e.getAttribute("crossorigin")!==(a.crossOrigin==null?null:a.crossOrigin))&&i&&e.hasAttribute("async")&&!e.hasAttribute("itemprop"))break;return e;default:return e}}else if(t==="input"&&e.type==="hidden"){var i=a.name==null?null:""+a.name;if(a.type==="hidden"&&e.getAttribute("name")===i)return e}else return e;if(e=gn(e.nextSibling),e===null)break}return null}function Tz(e,t,n){if(t==="")return null;for(;e.nodeType!==3;)if((e.nodeType!==1||e.nodeName!=="INPUT"||e.type!=="hidden")&&!n||(e=gn(e.nextSibling),e===null))return null;return e}function gn(e){for(;e!=null;e=e.nextSibling){var t=e.nodeType;if(t===1||t===3)break;if(t===8){if(t=e.data,t==="$"||t==="$!"||t==="$?"||t==="F!"||t==="F")break;if(t==="/$")return null}}return e}function d1(e){e=e.previousSibling;for(var t=0;e;){if(e.nodeType===8){var n=e.data;if(n==="$"||n==="$!"||n==="$?"){if(t===0)return e;t--}else n==="/$"&&t++}e=e.previousSibling}return null}function rT(e,t,n){switch(t=yc(n),e){case"html":if(e=t.documentElement,!e)throw Error(A(452));return e;case"head":if(e=t.head,!e)throw Error(A(453));return e;case"body":if(e=t.body,!e)throw Error(A(454));return e;default:throw Error(A(451))}}var ln=new Map,p1=new Set;function mc(e){return typeof e.getRootNode=="function"?e.getRootNode():e.ownerDocument}var _r=Se.d;Se.d={f:wz,r:Az,D:Rz,C:Cz,L:$z,m:Nz,X:jz,S:Mz,M:Dz};function wz(){var e=_r.f(),t=$f();return e||t}function Az(e){var t=ol(e);t!==null&&t.tag===5&&t.type==="form"?r2(t):_r.r(e)}var ul=typeof document>"u"?null:document;function aT(e,t,n){var r=ul;if(r&&typeof t=="string"&&t){var a=Zt(t);a='link[rel="'+e+'"][href="'+a+'"]',typeof n=="string"&&(a+='[crossorigin="'+n+'"]'),p1.has(a)||(p1.add(a),e={rel:e,crossOrigin:n,href:t},r.querySelector(a)===null&&(t=r.createElement("link"),at(t,"link",e),Ke(t),r.head.appendChild(t)))}}function Rz(e){_r.D(e),aT("dns-prefetch",e,null)}function Cz(e,t){_r.C(e,t),aT("preconnect",e,t)}function $z(e,t,n){_r.L(e,t,n);var r=ul;if(r&&e&&t){var a='link[rel="preload"][as="'+Zt(t)+'"]';t==="image"&&n&&n.imageSrcSet?(a+='[imagesrcset="'+Zt(n.imageSrcSet)+'"]',typeof n.imageSizes=="string"&&(a+='[imagesizes="'+Zt(n.imageSizes)+'"]')):a+='[href="'+Zt(e)+'"]';var i=a;switch(t){case"style":i=Wi(e);break;case"script":i=cl(e)}ln.has(i)||(e=Ee({rel:"preload",href:t==="image"&&n&&n.imageSrcSet?void 0:e,as:t},n),ln.set(i,e),r.querySelector(a)!==null||t==="style"&&r.querySelector(ss(i))||t==="script"&&r.querySelector(us(i))||(t=r.createElement("link"),at(t,"link",e),Ke(t),r.head.appendChild(t)))}}function Nz(e,t){_r.m(e,t);var n=ul;if(n&&e){var r=t&&typeof t.as=="string"?t.as:"script",a='link[rel="modulepreload"][as="'+Zt(r)+'"][href="'+Zt(e)+'"]',i=a;switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":i=cl(e)}if(!ln.has(i)&&(e=Ee({rel:"modulepreload",href:e},t),ln.set(i,e),n.querySelector(a)===null)){switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(n.querySelector(us(i)))return}r=n.createElement("link"),at(r,"link",e),Ke(r),n.head.appendChild(r)}}}function Mz(e,t,n){_r.S(e,t,n);var r=ul;if(r&&e){var a=Ti(r).hoistableStyles,i=Wi(e);t=t||"default";var l=a.get(i);if(!l){var o={loading:0,preload:null};if(l=r.querySelector(ss(i)))o.loading=5;else{e=Ee({rel:"stylesheet",href:e,"data-precedence":t},n),(n=ln.get(i))&&iv(e,n);var s=l=r.createElement("link");Ke(s),at(s,"link",e),s._p=new Promise(function(u,c){s.onload=u,s.onerror=c}),s.addEventListener("load",function(){o.loading|=1}),s.addEventListener("error",function(){o.loading|=2}),o.loading|=4,bu(l,t,r)}l={type:"stylesheet",instance:l,count:1,state:o},a.set(i,l)}}}function jz(e,t){_r.X(e,t);var n=ul;if(n&&e){var r=Ti(n).hoistableScripts,a=cl(e),i=r.get(a);i||(i=n.querySelector(us(a)),i||(e=Ee({src:e,async:!0},t),(t=ln.get(a))&&lv(e,t),i=n.createElement("script"),Ke(i),at(i,"link",e),n.head.appendChild(i)),i={type:"script",instance:i,count:1,state:null},r.set(a,i))}}function Dz(e,t){_r.M(e,t);var n=ul;if(n&&e){var r=Ti(n).hoistableScripts,a=cl(e),i=r.get(a);i||(i=n.querySelector(us(a)),i||(e=Ee({src:e,async:!0,type:"module"},t),(t=ln.get(a))&&lv(e,t),i=n.createElement("script"),Ke(i),at(i,"link",e),n.head.appendChild(i)),i={type:"script",instance:i,count:1,state:null},r.set(a,i))}}function h1(e,t,n,r){var a=(a=Yr.current)?mc(a):null;if(!a)throw Error(A(446));switch(e){case"meta":case"title":return null;case"style":return typeof n.precedence=="string"&&typeof n.href=="string"?(t=Wi(n.href),n=Ti(a).hoistableStyles,r=n.get(t),r||(r={type:"style",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};case"link":if(n.rel==="stylesheet"&&typeof n.href=="string"&&typeof n.precedence=="string"){e=Wi(n.href);var i=Ti(a).hoistableStyles,l=i.get(e);if(l||(a=a.ownerDocument||a,l={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},i.set(e,l),(i=a.querySelector(ss(e)))&&!i._p&&(l.instance=i,l.state.loading=5),ln.has(e)||(n={rel:"preload",as:"style",href:n.href,crossOrigin:n.crossOrigin,integrity:n.integrity,media:n.media,hrefLang:n.hrefLang,referrerPolicy:n.referrerPolicy},ln.set(e,n),i||Bz(a,e,n,l.state))),t&&r===null)throw Error(A(528,""));return l}if(t&&r!==null)throw Error(A(529,""));return null;case"script":return t=n.async,n=n.src,typeof n=="string"&&t&&typeof t!="function"&&typeof t!="symbol"?(t=cl(n),n=Ti(a).hoistableScripts,r=n.get(t),r||(r={type:"script",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};default:throw Error(A(444,e))}}function Wi(e){return'href="'+Zt(e)+'"'}function ss(e){return'link[rel="stylesheet"]['+e+"]"}function iT(e){return Ee({},e,{"data-precedence":e.precedence,precedence:null})}function Bz(e,t,n,r){e.querySelector('link[rel="preload"][as="style"]['+t+"]")?r.loading=1:(t=e.createElement("link"),r.preload=t,t.addEventListener("load",function(){return r.loading|=1}),t.addEventListener("error",function(){return r.loading|=2}),at(t,"link",n),Ke(t),e.head.appendChild(t))}function cl(e){return'[src="'+Zt(e)+'"]'}function us(e){return"script[async]"+e}function y1(e,t,n){if(t.count++,t.instance===null)switch(t.type){case"style":var r=e.querySelector('style[data-href~="'+Zt(n.href)+'"]');if(r)return t.instance=r,Ke(r),r;var a=Ee({},n,{"data-href":n.href,"data-precedence":n.precedence,href:null,precedence:null});return r=(e.ownerDocument||e).createElement("style"),Ke(r),at(r,"style",a),bu(r,n.precedence,e),t.instance=r;case"stylesheet":a=Wi(n.href);var i=e.querySelector(ss(a));if(i)return t.state.loading|=4,t.instance=i,Ke(i),i;r=iT(n),(a=ln.get(a))&&iv(r,a),i=(e.ownerDocument||e).createElement("link"),Ke(i);var l=i;return l._p=new Promise(function(o,s){l.onload=o,l.onerror=s}),at(i,"link",r),t.state.loading|=4,bu(i,n.precedence,e),t.instance=i;case"script":return i=cl(n.src),(a=e.querySelector(us(i)))?(t.instance=a,Ke(a),a):(r=n,(a=ln.get(i))&&(r=Ee({},n),lv(r,a)),e=e.ownerDocument||e,a=e.createElement("script"),Ke(a),at(a,"link",r),e.head.appendChild(a),t.instance=a);case"void":return null;default:throw Error(A(443,t.type))}else t.type==="stylesheet"&&!(t.state.loading&4)&&(r=t.instance,t.state.loading|=4,bu(r,n.precedence,e));return t.instance}function bu(e,t,n){for(var r=n.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),a=r.length?r[r.length-1]:null,i=a,l=0;l<r.length;l++){var o=r[l];if(o.dataset.precedence===t)i=o;else if(i!==a)break}i?i.parentNode.insertBefore(e,i.nextSibling):(t=n.nodeType===9?n.head:n,t.insertBefore(e,t.firstChild))}function iv(e,t){e.crossOrigin==null&&(e.crossOrigin=t.crossOrigin),e.referrerPolicy==null&&(e.referrerPolicy=t.referrerPolicy),e.title==null&&(e.title=t.title)}function lv(e,t){e.crossOrigin==null&&(e.crossOrigin=t.crossOrigin),e.referrerPolicy==null&&(e.referrerPolicy=t.referrerPolicy),e.integrity==null&&(e.integrity=t.integrity)}var Su=null;function m1(e,t,n){if(Su===null){var r=new Map,a=Su=new Map;a.set(n,r)}else a=Su,r=a.get(n),r||(r=new Map,a.set(n,r));if(r.has(e))return r;for(r.set(e,null),n=n.getElementsByTagName(e),a=0;a<n.length;a++){var i=n[a];if(!(i[xo]||i[st]||e==="link"&&i.getAttribute("rel")==="stylesheet")&&i.namespaceURI!=="http://www.w3.org/2000/svg"){var l=i.getAttribute(t)||"";l=e+l;var o=r.get(l);o?o.push(i):r.set(l,[i])}}return r}function v1(e,t,n){e=e.ownerDocument||e,e.head.insertBefore(n,t==="title"?e.querySelector("head > title"):null)}function kz(e,t,n){if(n===1||t.itemProp!=null)return!1;switch(e){case"meta":case"title":return!0;case"style":if(typeof t.precedence!="string"||typeof t.href!="string"||t.href==="")break;return!0;case"link":if(typeof t.rel!="string"||typeof t.href!="string"||t.href===""||t.onLoad||t.onError)break;switch(t.rel){case"stylesheet":return e=t.disabled,typeof t.precedence=="string"&&e==null;default:return!0}case"script":if(t.async&&typeof t.async!="function"&&typeof t.async!="symbol"&&!t.onLoad&&!t.onError&&t.src&&typeof t.src=="string")return!0}return!1}function lT(e){return!(e.type==="stylesheet"&&!(e.state.loading&3))}var No=null;function zz(){}function Lz(e,t,n){if(No===null)throw Error(A(475));var r=No;if(t.type==="stylesheet"&&(typeof n.media!="string"||matchMedia(n.media).matches!==!1)&&!(t.state.loading&4)){if(t.instance===null){var a=Wi(n.href),i=e.querySelector(ss(a));if(i){e=i._p,e!==null&&typeof e=="object"&&typeof e.then=="function"&&(r.count++,r=vc.bind(r),e.then(r,r)),t.state.loading|=4,t.instance=i,Ke(i);return}i=e.ownerDocument||e,n=iT(n),(a=ln.get(a))&&iv(n,a),i=i.createElement("link"),Ke(i);var l=i;l._p=new Promise(function(o,s){l.onload=o,l.onerror=s}),at(i,"link",n),t.instance=i}r.stylesheets===null&&(r.stylesheets=new Map),r.stylesheets.set(t,e),(e=t.state.preload)&&!(t.state.loading&3)&&(r.count++,t=vc.bind(r),e.addEventListener("load",t),e.addEventListener("error",t))}}function Uz(){if(No===null)throw Error(A(475));var e=No;return e.stylesheets&&e.count===0&&Kh(e,e.stylesheets),0<e.count?function(t){var n=setTimeout(function(){if(e.stylesheets&&Kh(e,e.stylesheets),e.unsuspend){var r=e.unsuspend;e.unsuspend=null,r()}},6e4);return e.unsuspend=t,function(){e.unsuspend=null,clearTimeout(n)}}:null}function vc(){if(this.count--,this.count===0){if(this.stylesheets)Kh(this,this.stylesheets);else if(this.unsuspend){var e=this.unsuspend;this.unsuspend=null,e()}}}var gc=null;function Kh(e,t){e.stylesheets=null,e.unsuspend!==null&&(e.count++,gc=new Map,t.forEach(Hz,e),gc=null,vc.call(e))}function Hz(e,t){if(!(t.state.loading&4)){var n=gc.get(e);if(n)var r=n.get(null);else{n=new Map,gc.set(e,n);for(var a=e.querySelectorAll("link[data-precedence],style[data-precedence]"),i=0;i<a.length;i++){var l=a[i];(l.nodeName==="LINK"||l.getAttribute("media")!=="not all")&&(n.set(l.dataset.precedence,l),r=l)}r&&n.set(null,r)}a=t.instance,l=a.getAttribute("data-precedence"),i=n.get(l)||r,i===r&&n.set(null,a),n.set(l,a),this.count++,r=vc.bind(this),a.addEventListener("load",r),a.addEventListener("error",r),i?i.parentNode.insertBefore(a,i.nextSibling):(e=e.nodeType===9?e.head:e,e.insertBefore(a,e.firstChild)),t.state.loading|=4}}var Mo={$$typeof:lr,Provider:null,Consumer:null,_currentValue:ma,_currentValue2:ma,_threadCount:0};function qz(e,t,n,r,a,i,l,o){this.tag=1,this.containerInfo=e,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Dd(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.finishedLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Dd(0),this.hiddenUpdates=Dd(null),this.identifierPrefix=r,this.onUncaughtError=a,this.onCaughtError=i,this.onRecoverableError=l,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=o,this.incompleteTransitions=new Map}function oT(e,t,n,r,a,i,l,o,s,u,c,d){return e=new qz(e,t,n,l,o,s,u,d),t=1,i===!0&&(t|=24),i=It(3,null,null,t),e.current=i,i.stateNode=e,t=Dm(),t.refCount++,e.pooledCache=t,t.refCount++,i.memoizedState={element:r,isDehydrated:n,cache:t},Km(i),e}function sT(e){return e?(e=vi,e):vi}function uT(e,t,n,r,a,i){a=sT(a),r.context===null?r.context=a:r.pendingContext=a,r=Vr(t),r.payload={element:n},i=i===void 0?null:i,i!==null&&(r.callback=i),n=Xr(e,r,t),n!==null&&(mt(n,e,t),no(n,e,t))}function g1(e,t){if(e=e.memoizedState,e!==null&&e.dehydrated!==null){var n=e.retryLane;e.retryLane=n!==0&&n<t?n:t}}function ov(e,t){g1(e,t),(e=e.alternate)&&g1(e,t)}function cT(e){if(e.tag===13){var t=Jr(e,67108864);t!==null&&mt(t,e,67108864),ov(e,67108864)}}var bc=!0;function Gz(e,t,n,r){var a=P.T;P.T=null;var i=Se.p;try{Se.p=2,sv(e,t,n,r)}finally{Se.p=i,P.T=a}}function Yz(e,t,n,r){var a=P.T;P.T=null;var i=Se.p;try{Se.p=8,sv(e,t,n,r)}finally{Se.p=i,P.T=a}}function sv(e,t,n,r){if(bc){var a=Zh(r);if(a===null)op(e,t,r,Sc,n),b1(e,r);else if(Xz(a,e,t,n,r))r.stopPropagation();else if(b1(e,r),t&4&&-1<Vz.indexOf(e)){for(;a!==null;){var i=ol(a);if(i!==null)switch(i.tag){case 3:if(i=i.stateNode,i.current.memoizedState.isDehydrated){var l=sa(i.pendingLanes);if(l!==0){var o=i;for(o.pendingLanes|=2,o.entangledLanes|=2;l;){var s=1<<31-Mt(l);o.entanglements[1]|=s,l&=~s}Yn(i),!(Re&6)&&(cc=kn()+500,os(0))}}break;case 13:o=Jr(i,2),o!==null&&mt(o,i,2),$f(),ov(i,2)}if(i=Zh(r),i===null&&op(e,t,r,Sc,n),i===a)break;a=i}a!==null&&r.stopPropagation()}else op(e,t,r,null,n)}}function Zh(e){return e=_m(e),uv(e)}var Sc=null;function uv(e){if(Sc=null,e=da(e),e!==null){var t=il(e);if(t===null)e=null;else{var n=t.tag;if(n===13){if(e=GO(t),e!==null)return e;e=null}else if(n===3){if(t.stateNode.current.memoizedState.isDehydrated)return t.tag===3?t.stateNode.containerInfo:null;e=null}else t!==e&&(e=null)}}return Sc=e,null}function fT(e){switch(e){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(C8()){case VO:return 2;case XO:return 8;case Zu:case $8:return 32;case PO:return 268435456;default:return 32}default:return 32}}var Wh=!1,Qr=null,Kr=null,Zr=null,jo=new Map,Do=new Map,zr=[],Vz="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function b1(e,t){switch(e){case"focusin":case"focusout":Qr=null;break;case"dragenter":case"dragleave":Kr=null;break;case"mouseover":case"mouseout":Zr=null;break;case"pointerover":case"pointerout":jo.delete(t.pointerId);break;case"gotpointercapture":case"lostpointercapture":Do.delete(t.pointerId)}}function Rl(e,t,n,r,a,i){return e===null||e.nativeEvent!==i?(e={blockedOn:t,domEventName:n,eventSystemFlags:r,nativeEvent:i,targetContainers:[a]},t!==null&&(t=ol(t),t!==null&&cT(t)),e):(e.eventSystemFlags|=r,t=e.targetContainers,a!==null&&t.indexOf(a)===-1&&t.push(a),e)}function Xz(e,t,n,r,a){switch(t){case"focusin":return Qr=Rl(Qr,e,t,n,r,a),!0;case"dragenter":return Kr=Rl(Kr,e,t,n,r,a),!0;case"mouseover":return Zr=Rl(Zr,e,t,n,r,a),!0;case"pointerover":var i=a.pointerId;return jo.set(i,Rl(jo.get(i)||null,e,t,n,r,a)),!0;case"gotpointercapture":return i=a.pointerId,Do.set(i,Rl(Do.get(i)||null,e,t,n,r,a)),!0}return!1}function dT(e){var t=da(e.target);if(t!==null){var n=il(t);if(n!==null){if(t=n.tag,t===13){if(t=GO(n),t!==null){e.blockedOn=t,U8(e.priority,function(){if(n.tag===13){var r=jt(),a=Jr(n,r);a!==null&&mt(a,n,r),ov(n,r)}});return}}else if(t===3&&n.stateNode.current.memoizedState.isDehydrated){e.blockedOn=n.tag===3?n.stateNode.containerInfo:null;return}}}e.blockedOn=null}function Eu(e){if(e.blockedOn!==null)return!1;for(var t=e.targetContainers;0<t.length;){var n=Zh(e.nativeEvent);if(n===null){n=e.nativeEvent;var r=new n.constructor(n.type,n);yh=r,n.target.dispatchEvent(r),yh=null}else return t=ol(n),t!==null&&cT(t),e.blockedOn=n,!1;t.shift()}return!0}function S1(e,t,n){Eu(e)&&n.delete(t)}function Pz(){Wh=!1,Qr!==null&&Eu(Qr)&&(Qr=null),Kr!==null&&Eu(Kr)&&(Kr=null),Zr!==null&&Eu(Zr)&&(Zr=null),jo.forEach(S1),Do.forEach(S1)}function Ls(e,t){e.blockedOn===t&&(e.blockedOn=null,Wh||(Wh=!0,Xe.unstable_scheduleCallback(Xe.unstable_NormalPriority,Pz)))}var Us=null;function E1(e){Us!==e&&(Us=e,Xe.unstable_scheduleCallback(Xe.unstable_NormalPriority,function(){Us===e&&(Us=null);for(var t=0;t<e.length;t+=3){var n=e[t],r=e[t+1],a=e[t+2];if(typeof r!="function"){if(uv(r||n)===null)continue;break}var i=ol(n);i!==null&&(e.splice(t,3),t-=3,Th(i,{pending:!0,data:a,method:n.method,action:r},r,a))}}))}function Bo(e){function t(s){return Ls(s,e)}Qr!==null&&Ls(Qr,e),Kr!==null&&Ls(Kr,e),Zr!==null&&Ls(Zr,e),jo.forEach(t),Do.forEach(t);for(var n=0;n<zr.length;n++){var r=zr[n];r.blockedOn===e&&(r.blockedOn=null)}for(;0<zr.length&&(n=zr[0],n.blockedOn===null);)dT(n),n.blockedOn===null&&zr.shift();if(n=(e.ownerDocument||e).$$reactFormReplay,n!=null)for(r=0;r<n.length;r+=3){var a=n[r],i=n[r+1],l=a[xt]||null;if(typeof i=="function")l||E1(n);else if(l){var o=null;if(i&&i.hasAttribute("formAction")){if(a=i,l=i[xt]||null)o=l.formAction;else if(uv(a)!==null)continue}else o=l.action;typeof o=="function"?n[r+1]=o:(n.splice(r,3),r-=3),E1(n)}}}function cv(e){this._internalRoot=e}jf.prototype.render=cv.prototype.render=function(e){var t=this._internalRoot;if(t===null)throw Error(A(409));var n=t.current,r=jt();uT(n,r,e,t,null,null)};jf.prototype.unmount=cv.prototype.unmount=function(){var e=this._internalRoot;if(e!==null){this._internalRoot=null;var t=e.containerInfo;e.tag===0&&Mi(),uT(e.current,2,null,e,null,null),$f(),t[ll]=null}};function jf(e){this._internalRoot=e}jf.prototype.unstable_scheduleHydration=function(e){if(e){var t=JO();e={blockedOn:null,target:e,priority:t};for(var n=0;n<zr.length&&t!==0&&t<zr[n].priority;n++);zr.splice(n,0,e),n===0&&dT(e)}};var x1=zO.version;if(x1!=="19.0.0")throw Error(A(527,x1,"19.0.0"));Se.findDOMNode=function(e){var t=e._reactInternals;if(t===void 0)throw typeof e.render=="function"?Error(A(188)):(e=Object.keys(e).join(","),Error(A(268,e)));return e=w8(t),e=e!==null?YO(e):null,e=e===null?null:e.stateNode,e};var Fz={bundleType:0,version:"19.0.0",rendererPackageName:"react-dom",currentDispatcherRef:P,findFiberByHostInstance:da,reconcilerVersion:"19.0.0"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var Hs=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!Hs.isDisabled&&Hs.supportsFiber)try{Jo=Hs.inject(Fz),Nt=Hs}catch{}}bf.createRoot=function(e,t){if(!LO(e))throw Error(A(299));var n=!1,r="",a=u2,i=c2,l=f2,o=null;return t!=null&&(t.unstable_strictMode===!0&&(n=!0),t.identifierPrefix!==void 0&&(r=t.identifierPrefix),t.onUncaughtError!==void 0&&(a=t.onUncaughtError),t.onCaughtError!==void 0&&(i=t.onCaughtError),t.onRecoverableError!==void 0&&(l=t.onRecoverableError),t.unstable_transitionCallbacks!==void 0&&(o=t.unstable_transitionCallbacks)),t=oT(e,1,!1,null,null,n,r,a,i,l,o,null),e[ll]=t.current,av(e.nodeType===8?e.parentNode:e),new cv(t)};bf.hydrateRoot=function(e,t,n){if(!LO(e))throw Error(A(299));var r=!1,a="",i=u2,l=c2,o=f2,s=null,u=null;return n!=null&&(n.unstable_strictMode===!0&&(r=!0),n.identifierPrefix!==void 0&&(a=n.identifierPrefix),n.onUncaughtError!==void 0&&(i=n.onUncaughtError),n.onCaughtError!==void 0&&(l=n.onCaughtError),n.onRecoverableError!==void 0&&(o=n.onRecoverableError),n.unstable_transitionCallbacks!==void 0&&(s=n.unstable_transitionCallbacks),n.formState!==void 0&&(u=n.formState)),t=oT(e,1,!0,t,n??null,r,a,i,l,o,s,u),t.context=sT(null),n=t.current,r=jt(),a=Vr(r),a.callback=null,Xr(n,a,r),t.current.lanes=r,es(t,r),Yn(t),e[ll]=t.current,av(e),new jf(t)};bf.version="19.0.0";function pT(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(pT)}catch{}}pT(),DO.exports=bf;var Qz=DO.exports;const Kz=Ra(Qz);Kz.createRoot(document.getElementById("root")).render(g.jsx(ne.StrictMode,{children:g.jsx(g8,{})}));
